@@ -36,14 +36,16 @@ from jax._src import xla_bridge
 from jax._src.interpreters import mlir
 from jax._src.interpreters import pxla
 from jax._src.interpreters import xla
-from jax._src.layout import AutoLayout, DeviceLocalLayout, Layout
+from jax._src.layout import AutoLayout, Layout, Format
 from jax._src.lib import xla_client as xc
-from jax._src.lib import xla_extension as xe
+from jax._src.lib import _jax
+from jax._src.lib import jaxlib_extension_version
 from jax._src.sharding import Sharding
 from jax._src.sharding_impls import (
-    PmapSharding, SingleDeviceSharding, NamedSharding,
-    device_replica_id_map, hashed_index, num_addressable_indices, local_to_global_shape)  # pyformat: disable
-from jax._src.typing import ArrayLike, DLDeviceType
+    PmapSharding, SingleDeviceSharding,
+    device_replica_id_map, hashed_index, num_addressable_indices,
+    local_to_global_shape, _internal_use_concrete_mesh)  # pyformat: disable
+from jax._src.typing import ArrayLike, DLDeviceType, DTypeLike
 from jax._src.util import safe_zip, unzip3, use_cpp_class, use_cpp_method, cache
 import numpy as np
 
@@ -55,7 +57,10 @@ PRNGKeyArray = Any  # TODO(jakevdp): fix cycles and import this.
 
 def _get_device(a: ArrayImpl) -> Device:
   devices = a.sharding._internal_device_list  # pytype: disable=attribute-error
-  assert len(devices) == 1
+  if len(devices) != 1:
+    raise ValueError(
+        "When making an array from single-device arrays the input arrays must "
+        f"have one shard each. An argument array had {len(devices)} shard(s).")
   return devices[0]
 
 
@@ -195,29 +200,37 @@ class ArrayImpl(basearray.Array):
 
     self.aval = aval
     self._sharding = sharding
-    self._arrays = [a._arrays[0] for a in arrays]
     self._committed = committed
     self._npy_value = None
+    arrays = [a._arrays[0] for a in arrays]
 
     # Don't rearrange if skip_checks is enabled because this assumes that the
     # input buffers are already arranged properly. This usually happens when
     # Array's are created as output of a JAX transformation
     # (like pjit, etc).
     if not _skip_checks or config.enable_checks.value:
-      self._check_and_rearrange()
+      arrays = self._check_and_rearrange(arrays, self._sharding, self.aval)
+    self._arrays = arrays
 
-  def _check_and_rearrange(self):
-    device_id_to_buffer = {_get_device(db).id: db for db in self._arrays}
+  def _check_and_rearrange(self, arrays, sharding, aval):
+    device_id_to_buffer = {_get_device(db).id: db for db in arrays}
 
-    addressable_dev = self.sharding.addressable_devices
-    if len(self._arrays) != len(addressable_dev):
+    addressable_dev = sharding.addressable_devices
+    if len(arrays) != len(addressable_dev):
       raise ValueError(
           f"Expected {len(addressable_dev)} per-device arrays "
           "(this is how many devices are addressable by the sharding), but "
-          f"got {len(self._arrays)}")
+          f"got {len(arrays)}")
 
     array_device_ids = set(device_id_to_buffer.keys())
     addressable_device_ids = {d.id for d in addressable_dev}
+    if len(array_device_ids) != len(arrays):
+      buffer_device_ids = [_get_device(db).id for db in arrays]
+      raise ValueError(
+          "When making an array from single-device arrays, the input arrays"
+          " must be from distinct devices, but got device IDs"
+          f" {buffer_device_ids}")
+
     # Calculate a symmetric difference because the device ids between sharding
     # and _arrays should match.
     diff = array_device_ids ^ addressable_device_ids
@@ -235,14 +248,15 @@ class ArrayImpl(basearray.Array):
       raise ValueError(err_msg)
 
     _validate_shape_and_dtype_for_per_device_arrays(
-        self._arrays,
-        sharding=self.sharding,
-        aval=self.aval,
-        expected_shape=self.sharding.shard_shape(self.shape),
+        arrays,
+        sharding=sharding,
+        aval=aval,
+        expected_shape=sharding.shard_shape(aval.shape),
     )
+
     # Rearrange arrays based on the device assignment.
-    addressable_da = self.sharding._addressable_device_assignment
-    self._arrays = [device_id_to_buffer[device.id] for device in addressable_da]
+    addressable_da = sharding._addressable_device_assignment
+    return [device_id_to_buffer[device.id] for device in addressable_da]
 
   @property
   def shape(self) -> Shape:
@@ -330,8 +344,8 @@ class ArrayImpl(basearray.Array):
       return format(self._value, format_spec)
 
   def __getitem__(self, idx):
-    from jax._src.lax import lax
-    from jax._src.numpy import lax_numpy
+    from jax._src.lax import lax  # pytype: disable=import-error
+    from jax._src.numpy import indexing  # pytype: disable=import-error
     self._check_if_deleted()
 
     if isinstance(self.sharding, PmapSharding):
@@ -366,7 +380,7 @@ class ArrayImpl(basearray.Array):
         return ArrayImpl(
             out.aval, sharding, [out], committed=False, _skip_checks=True)
 
-    return lax_numpy._rewriting_take(self, idx)
+    return indexing.rewriting_take(self, idx)
 
   def __iter__(self):
     if self.ndim == 0:
@@ -431,7 +445,7 @@ class ArrayImpl(basearray.Array):
                  max_version: tuple[int, int] | None = None,
                  dl_device: tuple[DLDeviceType, int] | None = None,
                  copy: bool | None = None):
-    from jax._src.dlpack import to_dlpack  # pylint: disable=g-import-not-at-top
+    from jax._src.dlpack import to_dlpack  # pytype: disable=import-error  # pylint: disable=g-import-not-at-top
 
     device_set = self.sharding.device_set
     if len(device_set) > 1:
@@ -451,7 +465,7 @@ class ArrayImpl(basearray.Array):
     if len(self._arrays) != 1:
       raise BufferError("__dlpack__ only supported for unsharded arrays.")
 
-    from jax._src.dlpack import DLDeviceType  # pylint: disable=g-import-not-at-top
+    from jax._src.dlpack import DLDeviceType  # pytype: disable=import-error  # pylint: disable=g-import-not-at-top
 
     if self.platform() == "cpu":
       return DLDeviceType.kDLCPU, 0
@@ -534,17 +548,17 @@ class ArrayImpl(basearray.Array):
     return out
 
   @property
-  def layout(self):
+  def format(self):
     # TODO(yashkatariya): Remove the deleted check from here.
     if self.is_deleted():
-      return Layout(None, self.sharding)
+      return Format(None, self.sharding)
     try:
-      return Layout(DeviceLocalLayout.from_pjrt_layout(self._pjrt_layout),
+      return Format(Layout.from_pjrt_layout(self._pjrt_layout),
                     self.sharding)
-    except xe.XlaRuntimeError as e:
+    except _jax.XlaRuntimeError as e:
       msg, *_ = e.args
       if type(msg) is str and msg.startswith("UNIMPLEMENTED"):
-        return Layout(None, self.sharding)
+        return Format(None, self.sharding)
       else:
         raise
 
@@ -600,8 +614,8 @@ class ArrayImpl(basearray.Array):
     return self
 
   @use_cpp_method()
-  def _single_device_array_to_np_array(self):
-    return np.asarray(self._arrays[0])
+  def _single_device_array_to_np_array_did_copy(self) -> tuple[np.ndarray, bool]:  # type: ignore
+    ...  # pytype: disable=bad-return-type
 
   @use_cpp_method()
   def _copy_single_device_array_to_host_async(self):
@@ -623,13 +637,18 @@ class ArrayImpl(basearray.Array):
     self._check_if_deleted()
 
     if self._npy_value is None:
-      if self.is_fully_replicated:
-        self._npy_value = self._single_device_array_to_np_array()
-        self._npy_value.flags.writeable = False
-        return cast(np.ndarray, self._npy_value)
+      # addressable_device_list can be empty. If it's empty, we will error below
+      if (self.is_fully_replicated and
+          self.sharding._internal_device_list.addressable_device_list):  # type: ignore
+        npy_value, did_copy = self._single_device_array_to_np_array_did_copy()
+        npy_value.flags.writeable = False
+        if did_copy:
+          self._npy_value = npy_value
+        return npy_value
 
       # TODO(yashkatariya): Merge `_process_has_full_value_in_mcjax` with
       # is_fully_addressable.
+      # is_fully_addressable return False if addressable_device_list is empty.
       if (not self.is_fully_addressable and
           not _process_has_full_value_in_mcjax(self.sharding, self.shape)):
         raise RuntimeError(
@@ -645,7 +664,7 @@ class ArrayImpl(basearray.Array):
 
       npy_value = np.empty(self.shape, self.dtype)
       for i, ind in _cached_index_calc(self.sharding, self.shape):
-        npy_value[ind] = self._arrays[i]._single_device_array_to_np_array()
+        npy_value[ind], _ = self._arrays[i]._single_device_array_to_np_array_did_copy()
       self._npy_value = npy_value
       self._npy_value.flags.writeable = False
     return self._npy_value
@@ -665,6 +684,28 @@ def _get_shape_from_index(slc: Index, shape: Shape) -> Shape:
       if isinstance(s, slice)  # If element is int, this dimension is reduced
   )
 
+def _get_and_check_dtype(arrays: Sequence[basearray.Array | np.ndarray],
+                         dtype: DTypeLike | None, fname: str):
+  if arrays:
+    if dtype is None:
+      dtype = arrays[0].dtype
+    else:
+      if arrays[0].dtype != dtype:
+        raise ValueError(
+            f"If `dtype` is provided to `jax.{fname}`, it must match the dtype "
+            f"of the addressable shards. Got dtype={dtype} and shard "
+            f"dtype={arrays[0].dtype}`.")
+  else:
+    if not config.enable_empty_arrays.value:
+      raise ValueError(
+          f"Building an Array with no addressable shards with `jax.{fname}` is "
+          "supported only if `jax.config.enable_empty_arrays` is set to True."
+      )
+    if dtype is None:
+      raise ValueError(
+          "If the Array has no addressable shards, `dtype` must be provided "
+          f"via the `dtype` argument to `jax.{fname}`.")
+  return dtype
 
 # explicitly set to be unhashable.
 setattr(ArrayImpl, "__hash__", None)
@@ -673,8 +714,9 @@ setattr(ArrayImpl, "__array_priority__", 100)
 # TODO(yashkatariya): Remove None from callback input type.
 
 def make_array_from_callback(
-    shape: Shape, sharding: Sharding | Layout,
-    data_callback: Callable[[Index | None], ArrayLike]) -> ArrayImpl:
+    shape: Shape, sharding: Sharding | Format,
+    data_callback: Callable[[Index | None], ArrayLike],
+    dtype: DTypeLike | None = None) -> ArrayImpl:
   # pyformat: disable
   """Returns a ``jax.Array`` via data fetched from ``data_callback``.
 
@@ -690,6 +732,9 @@ def make_array_from_callback(
     data_callback : Callback that takes indices into the global array value as
       input and returns the corresponding data of the global array value.
       The data can be returned as any array-like object, e.g. a ``numpy.ndarray``.
+    dtype: The dtype of the output ``jax.Array``. If not provided, the dtype of
+      the data for the first addressable shard is used. If there are no
+      addressable shards, the ``dtype`` argument must be provided.
 
   Returns:
     A ``jax.Array`` via data fetched from ``data_callback``.
@@ -714,12 +759,12 @@ def make_array_from_callback(
     (4, 2)
   """
   # pyformat: enable
-  dll = sharding.device_local_layout if isinstance(sharding, Layout) else None
+  dll = sharding.layout if isinstance(sharding, Format) else None
   if isinstance(dll, AutoLayout):
     raise TypeError(
-        "`DeviceLocalLayout.AUTO` cannot be used in place of a device-local"
+        "`Layout.AUTO` cannot be used in place of a device-local"
         f" layout when calling `jax.make_array_from_callback`. Got {sharding}")
-  sharding = sharding.sharding if isinstance(sharding, Layout) else sharding  # type: ignore
+  sharding = sharding.sharding if isinstance(sharding, Format) else sharding
   if not isinstance(sharding, Sharding):
     raise TypeError(
         f"sharding should be an instance of `jax.sharding`. Got {sharding} of"
@@ -750,23 +795,28 @@ def make_array_from_callback(
         get_data(device_to_index_map[device]) for device in devices
     ]
 
-  first_value = per_device_values[0]
-  expected_dtype = first_value.dtype
+  dtype = _get_and_check_dtype(
+      per_device_values, dtype, "make_array_from_callback")
   expected_shape = sharding.shard_shape(shape)
-  aval = core.ShapedArray(shape, expected_dtype)
+  aval = core.update_aval_with_sharding(
+      core.ShapedArray(shape, dtype), sharding)
+
   _validate_shape_and_dtype_for_per_device_arrays(
       per_device_values,
       expected_shape=expected_shape,
       aval=aval,
       sharding=sharding,
   )
-  if (isinstance(first_value, ArrayImpl)
-      and first_value._committed
-      and sharding.is_fully_replicated
-      and first_value.is_fully_replicated
-      and first_value.sharding._device_assignment == tuple(devices)
-      and first_value.layout.device_local_layout == dll):
-    return first_value
+  first_value = None
+  if per_device_values:
+    first_value = per_device_values[0]
+    if (isinstance(first_value, ArrayImpl)
+        and first_value._committed
+        and sharding.is_fully_replicated
+        and first_value.is_fully_replicated
+        and first_value.sharding._device_assignment == tuple(devices)
+        and first_value.format.layout == dll):
+      return first_value
 
   if dtypes.issubdtype(aval.dtype, dtypes.extended):
     # TODO(yashkatariya): Can this also use batched_device_put?
@@ -776,7 +826,7 @@ def make_array_from_callback(
     )
 
   if dll is not None:
-    devices = [Layout(dll, SingleDeviceSharding(d)) for d in devices]
+    devices = [Format(dll, SingleDeviceSharding(d)) for d in devices]  # type: ignore
     # pxla.batched_device_put doesn't support Layout... Take the slow route
     arrays = api.device_put(per_device_values, devices)
     return ArrayImpl(aval, sharding, arrays, committed=True)
@@ -968,7 +1018,8 @@ def make_array_from_process_local_data(
 
 
 def make_array_from_single_device_arrays(
-    shape: Shape, sharding: Sharding, arrays: Sequence[basearray.Array]
+    shape: Shape, sharding: Sharding, arrays: Sequence[basearray.Array], *,
+    dtype: DTypeLike | None = None,
 ) -> ArrayImpl:
   r"""Returns a ``jax.Array`` from a sequence of ``jax.Array``\s each on a single device.
       Every device in input ``sharding``\'s mesh must have an array in ``arrays``\s.
@@ -977,10 +1028,12 @@ def make_array_from_single_device_arrays(
     shape : Shape of the output ``jax.Array``. This conveys information already included with
       ``sharding`` and ``arrays`` and serves as a double check.
     sharding: Sharding: A global Sharding instance which describes how the output jax.Array is laid out across devices.
-    arrays: Sequence of ``jax.Array``\s that are each single device addressable. ``len(arrays)``
+    arrays: `list` or `tuple` of ``jax.Array``\s that are each single device addressable. ``len(arrays)``
       must equal ``len(sharding.addressable_devices)`` and the shape of each array must be the same. For multiprocess code,
       each process will call with a different ``arrays`` argument that corresponds to that processes' data.
       These arrays are commonly created via ``jax.device_put``.
+    dtype: The dtype of the output ``jax.Array``. If not provided, the dtype of the first array in
+      ``arrays`` is used. If ``arrays`` is empty, the ``dtype`` argument must be provided.
 
   Returns:
     A global ``jax.Array``, sharded as ``sharding``, with shape equal to ``shape``, and with per-device
@@ -1011,35 +1064,39 @@ def make_array_from_single_device_arrays(
   For cases where you have a local array and want to convert it to a global
   jax.Array, use ``jax.make_array_from_process_local_data``.
   """
+  if isinstance(arrays, Sequence):
+    dtype = _get_and_check_dtype(
+        arrays, dtype, "make_array_from_single_device_arrays")
+
   # All input arrays should be committed. Checking it is expensive on
   # single-controller systems.
-  if any(isinstance(arr, core.Tracer) for arr in arrays):
-    raise ValueError(
-        "jax.make_array_from_single_device_arrays requires a list of concrete"
-        f" arrays as input. got types {set(map(type, arrays))}")
-  aval = core.ShapedArray(shape, arrays[0].dtype, weak_type=False)
+  aval = core.update_aval_with_sharding(
+      core.ShapedArray(shape, dtype, weak_type=False), sharding)
   if dtypes.issubdtype(aval.dtype, dtypes.extended):
     return aval.dtype._rules.make_sharded_array(aval, sharding, arrays,
                                                 committed=True)
+  arrays = list(arrays) if isinstance(arrays, tuple) else arrays
   # TODO(phawkins): ideally the cast() could be checked.
-  return ArrayImpl(aval, sharding, cast(Sequence[ArrayImpl], arrays),
-                   committed=True)
+  try:
+    return ArrayImpl(aval, sharding, cast(Sequence[ArrayImpl], arrays),
+                    committed=True)
+  except TypeError:
+    if not isinstance(arrays, list):
+      raise TypeError("jax.make_array_from_single_device_arrays `arrays` "
+                      "argument must be a list or tuple, but got "
+                      f"{type(arrays)}.")
+    if any(isinstance(arr, core.Tracer) for arr in arrays):
+      raise ValueError(
+          "jax.make_array_from_single_device_arrays requires a list of concrete"
+          f" arrays as input, but got types {set(map(type, arrays))}")
+    raise
 
 xla.canonicalize_dtype_handlers[ArrayImpl] = pxla.identity
 
 def _get_aval_array(self):
-  if config.sharding_in_types.value and isinstance(self.sharding, NamedSharding):
-    return self.aval.update(sharding=NamedSharding(
-        self.sharding.mesh.abstract_mesh,
-        self.sharding.spec._normalized_spec(self.ndim)))
-  else:
-    return self.aval
-
+  return core.update_aval_with_sharding(self.aval, self.sharding)
 core.pytype_aval_mappings[ArrayImpl] = _get_aval_array
 
-# TODO(jakevdp) replace this with true inheritance at the C++ level.
-basearray.Array.register(ArrayImpl)
-
 
 def _array_mlir_constant_handler(val):
   try:
@@ -1056,6 +1113,8 @@ def _array_mlir_constant_handler(val):
 
 mlir.register_constant_handler(ArrayImpl, _array_mlir_constant_handler)
 
+if config.use_simplified_jaxpr_constants.value:
+  core.literalable_types.add(ArrayImpl)
 
 # NOTE(skye): we could refactor to generate _multi_slice parameters directly
 # from the input ShardingSpec, rather than the indices. However, this would
@@ -1092,7 +1151,10 @@ def shard_device_array(x, devices, indices, sharding):
   if sharding.is_fully_replicated:
     shards = [x] * len(devices)
   else:
-    shards = x._multi_slice(start_indices, limit_indices, removed_dims)
+    # TODO(yashkatariya): Maybe this should be set when we call the handler in
+    # InputsHandler.__call__?
+    with _internal_use_concrete_mesh(None):
+      shards = x._multi_slice(start_indices, limit_indices, removed_dims)
   aval = core.shaped_abstractify(x)
   return pxla.batched_device_put(aval, sharding, shards, devices)
 
@@ -1140,7 +1202,7 @@ def _array_shard_arg(xs, shardings, layouts, copy_semantics):
     x._check_if_deleted()
     indices, same_indices = _sharding_indices_and_eq(x.sharding, x.shape, sharding)
     same_layout = (True if layout is None else
-                   x.layout.device_local_layout == layout)
+                   x.format.layout == layout)
 
     if not x.is_fully_addressable:
       if same_indices and same_layout:
@@ -1149,19 +1211,22 @@ def _array_shard_arg(xs, shardings, layouts, copy_semantics):
         raise NotImplementedError(
             "Cannot reshard an input that is not fully addressable")
     else:
-      devices = sharding._addressable_device_assignment
+      devices = sharding._internal_device_list.addressable_device_list
       if same_indices and same_layout:
         # Add a placeholder result that will be filled in later.
         results.append(None)
         # Accumulate arguments to `batched_copy_array_to_devices_with_sharding`.
         batch_xs.append(x)
-        batch_devs.append(list(devices))
+        if jaxlib_extension_version >= 356:
+          batch_devs.append(devices)
+        else:
+          batch_devs.append(list(devices))
         batch_shardings.append(sharding)
         batch_indices.append(i)
         batch_cs.append(cs)
       # Resharding starts here:
       elif not same_layout:
-        results.append(api.device_put(x, Layout(layout, sharding)))
+        results.append(api.device_put(x, Format(layout, sharding)))
       elif dispatch.is_single_device_sharding(x.sharding):
         results.append(shard_device_array(x, devices, indices, sharding))
       else:
@@ -1179,6 +1244,7 @@ pxla.shard_arg_handlers[ArrayImpl] = _array_shard_arg
 
 
 def _array_global_result_handler(global_aval, out_sharding, committed):
+  global_aval = core.update_aval_with_sharding(global_aval, out_sharding)
   if global_aval.dtype == dtypes.float0:
     return lambda _: np.zeros(global_aval.shape, dtypes.float0)
   if dtypes.issubdtype(global_aval.dtype, dtypes.extended):
@@ -1207,16 +1273,20 @@ pxla.local_result_handlers[core.ShapedArray] = _array_local_result_handler
 def _token_shard_arg(xs, shardings, layouts, copy_semantics):
   results = []
   for x, sharding, layout in safe_zip(xs, shardings, layouts):
+    assert layout is None
     x.block_until_ready()
     x = np.array([], dtype=bool)
-    results.append(api.device_put(x, Layout(layout, sharding)))
+    aval = core.get_aval(x)
+    devices = sharding._addressable_device_assignment
+    results.append(pxla.batched_device_put(
+        aval, sharding, [x] * len(devices), devices))
   return results
 pxla.shard_arg_handlers[core.Token] = _token_shard_arg
 
 
 def _token_global_result_handler(global_aval, out_sharding, committed):
   array_handler = _array_global_result_handler(
-      core.token_shaped_array, out_sharding, committed)
+      core.get_token_aval(), out_sharding, committed)
 
   def wrapper(*args, **kwargs):
     out_buf = array_handler(*args, **kwargs)
