@@ -20,32 +20,31 @@ import functools
 import logging
 from typing import Any
 
-import jax
+from jax._src import api
 from jax._src import config
 from jax._src import core
-from jax._src import deprecations
 from jax._src import dispatch
 from jax._src import dtypes
 from jax._src import effects
+from jax._src import ffi
+from jax._src import pickle_util
 from jax._src import sharding_impls
 from jax._src import tree_util
 from jax._src import util
+from jax._src import xla_bridge as xb
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
-from jax._src.lax import lax
-from jax._src.lax.control_flow.loops import map as lax_map
+from jax._src.interpreters import xla
 from jax._src.lib import xla_client as xc
-from jax._src.sharding_impls import SingleDeviceSharding
-from jax._src.typing import DeprecatedArg
+from jax._src.lib.mlir import ir
+from jax._src.lib.mlir.dialects import hlo
+from jax._src.sharding_impls import SdyArray, SdyArrayList, SdyDim, SingleDeviceSharding
+from jax._src.typing import Array, DeprecatedArg
 import numpy as np
 
 logger = logging.getLogger(__name__)
 
-# TODO(dfm): Remove after 6 months.
-# Added Oct 1, 2024
-deprecations.register("jax-callback-vectorized")
-
 # `pure_callback_p` is the main primitive for staging out Python pure callbacks.
 pure_callback_p = core.Primitive("pure_callback")
 pure_callback_p.multiple_results = True
@@ -67,7 +66,7 @@ class _FlatCallback:
   callback_func: Callable[..., Any]
   in_tree: tree_util.PyTreeDef  # (args, kwargs) pytree for `callback_func`.
 
-  def __call__(self, *flat_args: jax.Array) -> Sequence[jax.Array]:
+  def __call__(self, *flat_args: Array) -> Sequence[Array]:
     args, kwargs = tree_util.tree_unflatten(self.in_tree, flat_args)
     return tree_util.tree_leaves(self.callback_func(*args, **kwargs))
 
@@ -77,20 +76,19 @@ def pure_callback_impl(
     result_avals,
     callback: _FlatCallback,
     sharding: SingleDeviceSharding | None,
-    vectorized: bool | DeprecatedArg,
     vmap_method: str | None,
 ):
-  del sharding, vectorized, vmap_method, result_avals
+  del sharding, vmap_method, result_avals
   try:
-    cpu_device, *_ = jax.local_devices(backend="cpu")
+    cpu_device, *_ = xb.local_devices(backend="cpu")
   except RuntimeError as e:
     raise RuntimeError(
         "jax.pure_callback failed to find a local CPU device to place the"
         " inputs on. Make sure \"cpu\" is listed in --jax_platforms or the"
         " JAX_PLATFORMS environment variable."
     ) from e
-  args = jax.device_put(args, cpu_device)
-  with jax.default_device(cpu_device):
+  args = api.device_put(args, cpu_device)
+  with config.default_device(cpu_device):
     try:
       return tree_util.tree_map(np.asarray, callback(*args))
     except BaseException:
@@ -108,10 +106,9 @@ def pure_callback_abstract_eval(
     callback: _FlatCallback,
     result_avals,
     sharding: SingleDeviceSharding | None,
-    vectorized: bool | DeprecatedArg,
     vmap_method: str | None,
 ):
-  del avals, callback, sharding, vectorized, vmap_method
+  del avals, callback, sharding, vmap_method
   return result_avals
 
 
@@ -134,95 +131,8 @@ def pure_callback_transpose_rule(*args, **kwargs):
 ad.primitive_transposes[pure_callback_p] = pure_callback_transpose_rule
 
 
-def callback_batching_rule(
-    prim,
-    args,
-    dims,
-    *,
-    vectorized: bool | None | DeprecatedArg,
-    vmap_method: str | None,
-    result_avals: Sequence[core.ShapedArray],
-    **kwargs: Any,
-):
-  if isinstance(vectorized, DeprecatedArg) and vmap_method is None:
-    deprecations.warn(
-        "jax-callback-vectorized",
-        f"The default behavior of {prim.name} under vmap will soon "
-        "change. Currently, the default behavior is to generate a sequential "
-        "vmap (i.e. a loop), but in the future the default will be to raise "
-        "an error. To keep the current default, set vmap_method='sequential'.",
-        stacklevel=6)
-    vmap_method = "sequential"
-
-  axis_size, = {a.shape[d] for a, d in zip(args, dims)
-                if d is not batching.not_mapped}
-  new_args = [arg if dim is batching.not_mapped else
-              batching.moveaxis(arg, dim, 0) for arg, dim in zip(args, dims)]
-  batched_result_avals = tuple(
-      core.unmapped_aval(axis_size, core.no_axis_name, 0, aval)
-      for aval in result_avals)
-
-  # For FFI calls we must update the layouts. We handle the output layouts
-  # here, but the input layout updates depend on the vmap_method parameter.
-  if vmap_method != "sequential" and kwargs.get("output_layouts") is not None:
-    kwargs["output_layouts"] = tuple(
-        None if layout is None else tuple(n + 1 for n in layout) + (0,)
-        for layout in kwargs["output_layouts"])
-
-  if vmap_method == "legacy_vectorized":
-    # This method is kept to support the behavior that was previously exposed
-    # when using `vectorized=True`.
-    if kwargs.get("input_layouts") is not None:
-      kwargs["input_layouts"] = tuple(
-          layout if d is batching.not_mapped else
-          (None if layout is None else tuple(n + 1 for n in layout) + (0,))
-          for layout, d in zip(kwargs["input_layouts"], dims))
-    outvals = prim.bind(
-        *new_args,
-        vectorized=vectorized,
-        vmap_method=vmap_method,
-        result_avals=batched_result_avals,
-        **kwargs,
-    )
-  elif vmap_method == "expand_dims" or vmap_method == "broadcast_all":
-    size = axis_size if vmap_method == "broadcast_all" else 1
-    bcast_args = [
-        lax.broadcast(x, (size,)) if d is batching.not_mapped else x
-        for x, d in zip(new_args, dims)]
-    if kwargs.get("input_layouts") is not None:
-      kwargs["input_layouts"] = tuple(
-          None if layout is None else tuple(n + 1 for n in layout) + (0,)
-          for layout in kwargs["input_layouts"])
-    outvals = prim.bind(
-      *bcast_args,
-      vectorized=vectorized,
-      vmap_method=vmap_method,
-      result_avals=batched_result_avals,
-      **kwargs,
-    )
-  elif vmap_method == "sequential":
-    is_batched = [d is not batching.not_mapped for d in dims]
-    unbatched_args, batched_args = util.partition_list(is_batched, new_args)
-    def _batch_fun(batched_args):
-      merged_args = util.merge_lists(is_batched, unbatched_args, batched_args)
-      return prim.bind(
-          *merged_args,
-          result_avals=result_avals,
-          vectorized=vectorized,
-          vmap_method=vmap_method,
-          **kwargs,
-      )
-    outvals = lax_map(_batch_fun, batched_args)
-  else:
-    raise NotImplementedError(
-        f"vmap is only supported for the {prim.name} primitive when vmap_method "
-        "is one of 'sequential', 'expand_dims', 'broadcast_all', or "
-        "'legacy_vectorized'.")
-  return tuple(outvals), (0,) * len(outvals)
-
-
 batching.primitive_batchers[pure_callback_p] = functools.partial(
-    callback_batching_rule, pure_callback_p
+    ffi.ffi_batching_rule, pure_callback_p
 )
 
 
@@ -243,13 +153,15 @@ def _callback_op_sharding(
           " computations"
       )
     if config.use_shardy_partitioner.value:
-      assert len(avals_out) == 1
-      op_sharding = sharding_impls.SdyArrayShardingList([
-          sharding_impls.SdyArraySharding(
+      ndim = 0
+      if avals_out and isinstance(avals_out[0], core.ShapedArray):
+        ndim = avals_out[0].ndim
+      op_sharding = SdyArrayList([
+          SdyArray(
               mesh_shape=(),
-              dimension_shardings=[
-                  sharding_impls.SdyDimSharding(axes=[], is_closed=True)
-              ] * avals_out[0].ndim,
+              dim_shardings=[
+                  SdyDim(axes=[], is_open=False)
+              ] * ndim,
               logical_device_ids=())])
     else:
       op_sharding = xc.OpSharding()  # type: ignore[assignment]
@@ -282,10 +194,14 @@ def _callback_op_sharding(
     # program has bulk array semantics, so we run the callback with a MAXIMAL
     # sharding and hence execute it only once on the full logical value).
     if config.use_shardy_partitioner.value:
-      op_sharding = sharding_impls.SdyArrayShardingList([
-          sharding_impls.SdyArraySharding(
+      # For shardy, we need to have the same number of shardy annotations as the
+      # number of result ops. If there are no result ops, we need 1 shardy
+      # annotation.
+      num_sdy_shardings = max(1, len(avals_out))
+      op_sharding = SdyArrayList(num_sdy_shardings * [
+          SdyArray(
               mesh_shape=(),
-              dimension_shardings=[],
+              dim_shardings=[],
               logical_device_ids=(device_index,))])
     else:
       op_sharding = xc.OpSharding()  # type: ignore[assignment]
@@ -313,7 +229,7 @@ def pure_callback_lowering(
 
   op_sharding = _callback_op_sharding(
       ctx.module_context.axis_context, sharding, ctx.avals_out)
-  result, _, _ = mlir.emit_python_callback(
+  result, _, _ = emit_python_callback(
       ctx,
       _callback,
       None,
@@ -326,7 +242,9 @@ def pure_callback_lowering(
   return result
 
 
-mlir.register_lowering(pure_callback_p, pure_callback_lowering)
+# TODO(phawkins): On TPU, these have an embedded channel ID that should be
+# unique for each callback. Caching defeats this.
+mlir.register_lowering(pure_callback_p, pure_callback_lowering, cacheable=False)
 
 def _check_shape_dtype(shape_dtype):
   dt = np.dtype(shape_dtype.dtype)
@@ -359,20 +277,28 @@ def pure_callback(
   `jit`-decorated function has no data dependence on its value. Pure callbacks
   may also be reordered if data-dependence allows.
 
+  .. warning::
+
+     In the context of JAX transformations, Python exceptions should be
+     considered side-effects: this means that intentionally raising an error
+     within a `pure_callback` breaks the API contract, and the behavior of
+     the resulting program is undefined.
+
   When `vmap`-ed the behavior will depend on the value of the ``vmap_method``.
 
   * Calling :func:`~jax.vmap` on a callback without an explicit ``vmap_method``
-    is deprecated and it will eventually raise ``NotImplementedError``.
+    raises a ``NotImplementedError``.
   * ``vmap_method="sequential"`` uses :func:`~jax.lax.map` to loop over
     the batched arguments, calling ``callback`` once for each batch element.
+  * ``vmap_method="sequential_unrolled"`` is like ``sequential``, but the loop
+    is unrolled.
   * ``vmap_method="expand_dims"`` calls ``callback`` with new axes of size ``1``
     added as the leading dimension unbatched inputs.
   * ``vmap_method="broadcast_all"`` behaves like ``expand_dims``, but the
     inputs are tiled to the expected batched shape.
 
-  If necessary, the legacy behavior provided by the deprecated
-  ``vectorized=True`` argument can be recovered using
-  ``vmap_method="legacy_vectorized"``.
+  If necessary, the legacy behavior provided by the removed ``vectorized=True``
+  argument can be recovered using ``vmap_method="legacy_vectorized"``.
 
   The current default behavior is to use ``vmap_method="sequential"`` when
   not specified, but this behavior is deprecated, and in the future, the
@@ -439,22 +365,15 @@ def pure_callback(
     (4,) (4,)
     Array([1., 2., 3., 4.], dtype=float32)
 
-  .. _External Callbacks: https://jax.readthedocs.io/en/latest/notebooks/external_callbacks.html
+  .. _External Callbacks: https://docs.jax.dev/en/latest/external-callbacks.html
   """
-  if not isinstance(vectorized, DeprecatedArg) and not vectorized is None:
-    deprecations.warn(
-        "jax-callback-vectorized",
-        "The vectorized argument of jax.pure_callback is deprecated and setting "
-        "it will soon raise an error. To avoid an error in the future, and to "
-        "suppress this warning, please use the vmap_method argument instead.",
-        stacklevel=2)
-    if vmap_method is not None:
-      raise ValueError(
-          "the vectorized and vmap_method arguments of jax.pure_callback cannot "
-          "be used together. Please use the vmap_method argument.")
-    vmap_method = "legacy_vectorized" if vectorized else "sequential"
-  allowed_vmap_methods = ["sequential", "expand_dims", "broadcast_all",
-                          "legacy_vectorized", None]
+  # TODO(danfm): Remove this check 3 months after v0.6.0 is released.
+  if not isinstance(vectorized, DeprecatedArg):
+    raise ValueError(
+        "The 'vectorized' argument of jax.pure_callback was removed in JAX "
+        "v0.6.0. Use 'vmap_method' instead.")
+  allowed_vmap_methods = ["sequential", "sequential_unrolled", "expand_dims",
+                          "broadcast_all", "legacy_vectorized", None]
   if vmap_method not in allowed_vmap_methods:
     raise ValueError(
         f"vmap_method must be on of the allowed methods {allowed_vmap_methods}, "
@@ -470,7 +389,6 @@ def pure_callback(
       callback=_FlatCallback(callback, in_tree),
       result_avals=tuple(flat_result_avals),
       sharding=sharding,
-      vectorized=vectorized,
       vmap_method=vmap_method,
   )
   return tree_util.tree_unflatten(out_tree, out_flat)
@@ -507,15 +425,15 @@ def io_callback_impl(
 ):
   del result_avals, sharding, ordered
   try:
-    cpu_device, *_ = jax.local_devices(backend="cpu")
+    cpu_device, *_ = xb.local_devices(backend="cpu")
   except RuntimeError as e:
     raise RuntimeError(
         "jax.io_callback failed to find a local CPU device to place the"
         " inputs on. Make sure \"cpu\" is listed in --jax_platforms or the"
         " JAX_PLATFORMS environment variable."
     ) from e
-  args = jax.device_put(args, cpu_device)
-  with jax.default_device(cpu_device):
+  args = api.device_put(args, cpu_device)
+  with config.default_device(cpu_device):
     try:
       return tree_util.tree_map(np.asarray, callback(*args))
     except BaseException:
@@ -555,6 +473,7 @@ ad.primitive_transposes[io_callback_p] = io_callback_transpose_rule
 def io_callback_batching_rule(
     args, dims, callback, result_avals, sharding, ordered
 ):
+  from jax._src.lax.control_flow.loops import map as lax_map  # pytype: disable=import-error
   if ordered:
     raise ValueError("Cannot `vmap` ordered IO callback.")
   is_batched = [d is not batching.not_mapped for d in dims]
@@ -586,7 +505,7 @@ def io_callback_lowering(ctx, *args, callback, sharding, ordered, **params):
       ctx.module_context.axis_context, sharding, ctx.avals_out)
   if ordered:
     token = ctx.tokens_in.get(_OrderedIOEffect)
-    result, token, _ = mlir.emit_python_callback(
+    result, token, _ = emit_python_callback(
         ctx,
         _callback,
         token,
@@ -598,7 +517,7 @@ def io_callback_lowering(ctx, *args, callback, sharding, ordered, **params):
     )
     ctx.set_tokens_out(mlir.TokenSet({_OrderedIOEffect: token}))
   else:
-    result, _, _ = mlir.emit_python_callback(
+    result, _, _ = emit_python_callback(
         ctx,
         _callback,
         None,
@@ -610,9 +529,9 @@ def io_callback_lowering(ctx, *args, callback, sharding, ordered, **params):
     )
   return result
 
-
-mlir.register_lowering(io_callback_p, io_callback_lowering)
-
+# TODO(phawkins): On TPU, these have an embedded channel ID that should be
+# unique for each callback. Caching defeats this.
+mlir.register_lowering(io_callback_p, io_callback_lowering, cacheable=False)
 
 def io_callback(
     callback: Callable[..., Any],
@@ -648,7 +567,7 @@ def io_callback(
     - :func:`jax.debug.callback`: callback designed for general-purpose debugging.
     - :func:`jax.debug.print`: callback designed for printing.
 
-  .. _External Callbacks: https://jax.readthedocs.io/en/latest/notebooks/external_callbacks.html
+  .. _External Callbacks: https://docs.jax.dev/en/latest/notebooks/external_callbacks.html
   """
   flat_args, in_tree = tree_util.tree_flatten((args, kwargs))
   tree_util.tree_map(_check_shape_dtype, result_shape_dtypes)
@@ -663,3 +582,297 @@ def io_callback(
       ordered=ordered,
   )
   return tree_util.tree_unflatten(out_tree, out_flat)
+
+
+def is_empty_shape(s: core.Shape) -> bool:
+  return any(d == 0 for d in s)
+
+
+def send_to_host(
+    channel: int,
+    token: hlo.TokenType,
+    operand: Any,
+    name: str,
+    *,
+    sharding: SdyArrayList | xc.OpSharding | None = None,
+) -> ir.Value:
+  channel_handle = hlo.ChannelHandle.get(channel, mlir.SEND_TO_HOST_TYPE)
+  send_op = hlo.SendOp([operand], token, channel_handle,
+                        is_host_transfer=ir.BoolAttr.get(True))
+  send_op.attributes["mhlo.frontend_attributes"] = ir.DictAttr.get(
+      dict(
+          _xla_host_transfer_handler_name=ir.StringAttr.get(str(name)),
+          _xla_host_transfer_rendezvous=ir.StringAttr.get(str(name))))
+  if sharding is not None:
+    if config.use_shardy_partitioner.value:
+      # `SendOp`'s return type is a StableHLO `TokenType`. However JAX passed
+      # in the maximal sharding of the array type. Since a token has no rank,
+      # we need to create an equivalent sharding with no dimensions. If there
+      # are multiple shardings, just grab the first one since all these
+      # shardings should be the same.
+      assert isinstance(sharding, SdyArrayList)
+      assert len(sharding.shardings) >= 1
+      sharding = SdyArrayList([
+          SdyArray(
+              mesh_shape=(), dim_shardings=[],
+              logical_device_ids=sharding.shardings[0].logical_device_ids)])
+    mlir.set_sharding(send_op, sharding)
+  return send_op.result
+
+
+def receive_from_host(
+    channel: int,
+    token: hlo.TokenType,
+    out_aval: core.ShapedArray,
+    name: str,
+    *,
+    sharding: SdyArrayList | xc.OpSharding | None = None,
+) -> tuple[ir.Value, ir.Value]:
+  channel_handle = hlo.ChannelHandle.get(channel, mlir.RECV_FROM_HOST_TYPE)
+  recv_op = hlo.RecvOp([mlir.aval_to_ir_type(out_aval),
+                        hlo.TokenType.get()], token, channel_handle,
+                        is_host_transfer=ir.BoolAttr.get(True))
+  recv_op.attributes["mhlo.frontend_attributes"] = ir.DictAttr.get(
+      dict(
+          _xla_host_transfer_handler_name=ir.StringAttr.get(str(name)),
+          _xla_host_transfer_rendezvous=ir.StringAttr.get(str(name))))
+  if sharding is not None:
+    if config.use_shardy_partitioner.value:
+      assert isinstance(sharding, SdyArrayList)
+      assert len(sharding.shardings) >= 1
+       # `RecvOp`'s last argument is a `TokenType`. Since Shardy requires the
+      # number of shardings to match the number of results, but JAX only sees
+      # the array result, we need to add an equivalent sharding for the token.
+      # Note that even if a function returns N results, we will end up with N
+      # `RecvOp`s, so we only need to get the first sharding. All shardings are
+      # the same anyways, operating on the same single device ID.
+      sharding = SdyArrayList([
+          sharding.shardings[0],
+          SdyArray(
+              mesh_shape=(), dim_shardings=[],
+              logical_device_ids=sharding.shardings[0].logical_device_ids)])
+    mlir.set_sharding(recv_op, sharding)
+  # Token should be at the end of the results
+  result, token = recv_op.results
+  return token, result
+
+
+
+def _aval_to_xla_shape(aval: core.AbstractValue) -> xc.Shape:
+  try:
+    return _xla_shape_handlers[type(aval)](aval)
+  except KeyError as err:
+    raise TypeError(f"No xla_shape_handler for type: {type(aval)}") from err
+
+_xla_shape_handlers: dict[type[core.AbstractValue],
+                         Callable[[Any], xc.Shape]] = {}
+
+def _make_array_shape(aval: core.ShapedArray) -> xc.Shape:
+  aval = core.physical_aval(aval)
+  dtype = np.dtype('bool') if aval.dtype == dtypes.float0 else aval.dtype
+  return xc.Shape.array_shape(dtype, aval.shape)
+_xla_shape_handlers[core.ShapedArray] = _make_array_shape
+
+_xla_shape_handlers[core.AbstractToken] = lambda _: xc.Shape.token_shape()
+
+
+def _emit_tpu_python_callback(
+    backend: xb.XlaBackend,
+    ctx: mlir.LoweringRuleContext,
+    callback,
+    token: Any | None,
+    operands: Sequence[ir.Value],
+    operand_avals: Sequence[core.ShapedArray],
+    operand_shapes: Sequence[xc.Shape],
+    result_avals: Sequence[core.ShapedArray],
+    result_shapes: Sequence[xc.Shape],
+    *,
+    sharding: SdyArrayList | xc.OpSharding | None = None,
+) -> tuple[Sequence[ir.Value], Any]:
+  token = token or hlo.create_token()
+  _wrapped_callback = callback
+
+  send_channels = []
+  if not operand_avals:
+    # If there are no operands to the callback, we need to insert a dummy send
+    # op or the callback will never be triggered!
+    # TODO(sharadmv,chky): Enable this fix in the runtime as opposed to in
+    # MLIR builder.
+    callback_without_args = _wrapped_callback
+    def _wrapped_callback(*args):  # pylint: disable=function-redefined
+      del args
+      return callback_without_args()
+    send_channel = ctx.module_context.new_channel()
+    dummy_send_aval = core.ShapedArray((1,), np.float32)
+    dummy_send_val = mlir.ir_constant(np.zeros(1, np.float32))
+    operand_shapes = [*operand_shapes, _aval_to_xla_shape(dummy_send_aval)]
+    token = send_to_host(send_channel, token, dummy_send_val, callback.__name__,
+                         sharding=sharding)
+    send_channels.append(send_channel)
+  else:
+    for operand in operands:
+      channel = ctx.module_context.new_channel()
+      token = send_to_host(channel, token, operand, callback.__name__,
+                           sharding=sharding)
+      send_channels.append(channel)
+
+  recv_channels = []
+  outputs = []
+  for result_aval in result_avals:
+    channel = ctx.module_context.new_channel()
+    assert isinstance(result_aval, core.ShapedArray)
+    token, out = receive_from_host(channel, token, result_aval,
+                                   callback.__name__, sharding=sharding)
+    outputs.append(out)
+    recv_channels.append(channel)
+  ifrt_callback = backend.make_python_callback_from_host_send_and_recv(
+      _wrapped_callback, operand_shapes, result_shapes, send_channels,
+      recv_channels, pickle_util.dumps)
+  ctx.module_context.add_host_callback(ifrt_callback)
+  return outputs, token
+
+
+def emit_python_callback(
+    ctx: mlir.LoweringRuleContext,
+    callback,
+    token: Any | None,
+    operands: Sequence[ir.Value],
+    operand_avals: Sequence[core.ShapedArray],
+    result_avals: Sequence[core.ShapedArray],
+    *,
+    has_side_effect: bool,
+    partitioned: bool = False,
+    sharding: SdyArrayList | xc.OpSharding | None = None,
+) -> tuple[Sequence[mlir.IrValues], Any, Any]:
+  """Emits MLIR that calls back to a provided Python function.
+
+  Args:
+    ctx: The lowering context.
+    callback: The Python callback function.
+    token: The token to use for the callback.
+    operands: The operands to the callback.
+    operand_avals: The abstract values of the operands.
+    result_avals: The abstract values of the results.
+    has_side_effect: Whether the callback has side effects.
+    partitioned: If True, then `callback` is called on local shards only. If
+      False, then `callback` is called on all shards.
+    sharding: The sharding of the callback.
+
+  Returns:
+    A tuple of MLIR result values, a new token (if any), and the host callback
+    object.
+  """
+  if len(ctx.module_context.platforms) > 1:
+    raise NotImplementedError("multi-platform lowering for python_callback")
+  platform = ctx.module_context.platforms[0]
+  if platform not in {"cpu", "cuda", "rocm", "tpu"}:
+    raise ValueError(
+        f"`EmitPythonCallback` not supported on {platform} backend.")
+  if partitioned:
+    if platform not in {"cpu", "cuda", "rocm"}:
+      raise NotImplementedError(
+          f"Partitioned callback not implemented on {platform} backend.")
+    if result_avals:
+      raise ValueError("Partitioned callback not supported with return values.")
+  backend = ctx.module_context.get_backend()
+  result_shapes = [_aval_to_xla_shape(aval) for aval in result_avals]
+  operand_shapes = [_aval_to_xla_shape(aval) for aval in operand_avals]
+
+  # First we apply checks to ensure output shapes and dtypes match the expected
+  # ones.
+  def _wrapped_callback(*args):
+    out_vals = callback(*args)
+    if len(out_vals) != len(result_avals):
+      raise RuntimeError(
+          "Mismatched number of outputs from callback. "
+          "Expected: {}, Actual: {}".format(len(result_avals), len(out_vals)))
+    # Handle Python literals, and custom arrays, e.g., tf.Tensor.
+    out_vals = tuple(xla.canonicalize_dtype(np.asarray(a)) for a in out_vals)
+    for i, (out_val, out_aval) in enumerate(zip(out_vals, result_avals)):
+      if out_val.shape != out_aval.shape:
+        raise RuntimeError(
+            f"Incorrect output shape for return value #{i}: "
+            f"Expected: {out_aval.shape}, Actual: {out_val.shape}")
+      if out_val.dtype != out_aval.dtype:
+        raise RuntimeError(
+            f"Incorrect output dtype for return value #{i}: "
+            f"Expected: {out_aval.dtype}, Actual: {out_val.dtype}")
+
+    if platform == "tpu":
+      # On TPU we cannot receive empty arrays. So, we return from the wrapped
+      # callback only the non-empty results, and we will create empty constants
+      # in the receiving computation.
+      # TODO(b/238239458): fix TPU Recv to work with empty arrays.
+      non_empty_out_vals = tuple(
+          out_val
+          for out_val, result_aval in zip(out_vals, result_avals)
+          if not is_empty_shape(result_aval.shape))
+      return non_empty_out_vals
+    else:
+      return out_vals
+
+  if platform == "tpu":
+    non_empty_result_avals, non_empty_result_shapes = util.unzip2([
+        (aval, shape)
+        for aval, shape in zip(result_avals, result_shapes)
+        if not is_empty_shape(aval.shape)])
+    non_empty_outputs, token = _emit_tpu_python_callback(
+        backend, ctx, _wrapped_callback,  token,
+        operands, operand_avals, operand_shapes,
+        non_empty_result_avals, non_empty_result_shapes,
+        sharding=sharding)
+    non_empty_outputs_iter = iter(non_empty_outputs)
+    outputs = [
+        mlir.ir_constant(np.zeros(result_aval.shape, dtype=result_aval.dtype))
+        if is_empty_shape(result_aval.shape) else next(non_empty_outputs_iter)
+        for result_aval in result_avals]
+    return outputs, token, None
+
+  device = "gpu" if platform in {"cuda", "rocm"} else "cpu"
+  partition = "_partitioned" if partitioned else ""
+  call_target_name = f"xla_ffi{partition}_python_{device}_callback"
+  if token:
+    callback_without_token = _wrapped_callback
+    def _wrapped_callback(token, *args):  # type: ignore  # pylint: disable=function-redefined
+      return (token, *callback_without_token(*args))
+    operands = [token, *operands]
+    if (
+        config.use_shardy_partitioner.value
+        and sharding is not None
+        and len(ctx.avals_out) > 0
+        and isinstance(sharding, SdyArrayList)
+    ):
+      # Add a sharding annotation for the token if we have at least one
+      # output. Otherwise, the single shardy annotation required of all ops
+      # (even those without any results) can annotate the token.
+      sharding = SdyArrayList([
+          SdyArray(
+              mesh_shape=(),
+              dim_shardings=[],
+              logical_device_ids=()),
+          *sharding.shardings])
+    ctx = dataclasses.replace(
+        ctx,
+        avals_in=[core.abstract_token, *ctx.avals_in],
+        avals_out=[core.abstract_token, *ctx.avals_out],
+    )
+
+  # TODO(dsuo): Remove this line once we deprecate the XLA custom call
+  # handler.
+  ifrt_callback = _wrapped_callback
+  ctx.module_context.add_host_callback(ifrt_callback)
+  index = np.uint64(len(ctx.module_context.host_callbacks) - 1)
+  result = ffi.build_ffi_lowering_function(  # type: ignore
+      call_target_name,
+      has_side_effect=has_side_effect,
+  )(ctx, *operands, index=np.uint64(index))
+
+  if sharding is not None:
+    mlir.set_sharding(result, sharding)
+
+  results = result.results  # type: ignore
+
+  if token:
+    token, *results = results  # type: ignore
+
+  return results, token, ifrt_callback  # type: ignore
