@@ -12,7 +12,10 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import datetime
 import os
+import re
+import warnings
 from jax import version
 from jax._src import config
 from jax._src import hardware_utils
@@ -70,7 +73,21 @@ def cloud_tpu_init() -> None:
 
   # Exit early if we're not running on a Cloud TPU VM or libtpu isn't installed.
   libtpu_path = get_tpu_library_path()
-  num_tpu_chips = hardware_utils.num_available_tpu_chips_and_device_id()[0]
+  num_tpu_chips, tpu_id = hardware_utils.num_available_tpu_chips_and_device_id()
+  if num_tpu_chips == 0:
+    os.environ['TPU_SKIP_MDS_QUERY'] = '1'
+  if (
+      tpu_id is not None
+      and tpu_id >= hardware_utils.TpuVersion.v5e
+      and not hardware_utils.transparent_hugepages_enabled()
+  ):
+    warnings.warn(
+        'Transparent hugepages are not enabled. TPU runtime startup and'
+        ' shutdown time should be significantly improved on TPU v5e and newer.'
+        ' If not already set, you may need to enable transparent hugepages in'
+        ' your VM image (sudo sh -c "echo always >'
+        ' /sys/kernel/mm/transparent_hugepage/enabled")'
+    )
   if (libtpu_path is None or num_tpu_chips == 0) and not jax_force_tpu_init():
     return
 
@@ -98,3 +115,22 @@ def cloud_tpu_init() -> None:
       'jax_pjrt_client_create_options',
       f'ml_framework_name:JAX;ml_framework_version:{version.__version__}'
       )
+
+
+def is_cloud_tpu_older_than(year: int, month: int, day: int):
+  # We import locally because the functions above must run before the runtime
+  # modules are imported.
+  from jax._src import xla_bridge  # pytype: disable=import-error
+  date = datetime.date(year, month, day)
+  if not running_in_cloud_tpu_vm:
+    return False
+  # The format of Cloud TPU platform_version is like:
+  # PJRT C API
+  # TFRT TPU v2
+  # Built on Oct 30 2023 03:04:42 (1698660263) cl/577737722
+  platform_version = xla_bridge.get_backend().platform_version.split('\n')[-1]
+  results = re.findall(r'\(.*?\)', platform_version)
+  if len(results) != 1:
+    return True
+  build_date = date.fromtimestamp(int(results[0][1:-1]))
+  return build_date < date
