@@ -22,19 +22,22 @@ from __future__ import annotations
 from functools import partial
 import inspect
 from typing import Any
+from collections.abc import Callable
 import weakref
 
 import numpy as np
-import jax
-from jax import tree_util
+
+from jax._src import api
 from jax._src import api_util
 from jax._src import config
 from jax._src import core
 from jax._src import custom_api_util
 from jax._src import dispatch
+from jax._src import errors
 from jax._src import linear_util as lu
 from jax._src import mesh as mesh_lib
 from jax._src import sharding_impls
+from jax._src import tree_util
 from jax._src import xla_bridge as xb
 from jax._src.custom_partitioning_sharding_rule import sdy_sharding_rule_to_mlir, SdyShardingRule, str_to_sdy_sharding_rule
 from jax._src.interpreters import mlir
@@ -42,7 +45,7 @@ from jax._src.interpreters import partial_eval as pe
 from jax._src.lib import xla_client as xc
 from jax._src.lib.mlir import ir
 from jax._src.lib.mlir.dialects import hlo
-from jax.errors import UnexpectedTracerError
+from jax._src.sharding import Sharding
 
 
 def _resolve_kwargs(fun, args, kwargs):
@@ -93,7 +96,7 @@ def _to_jax_shape(s):
 
 
 def _to_jax_sharded_shape(s, sharding):
-  return jax.ShapeDtypeStruct(
+  return api.ShapeDtypeStruct(
       s.dimensions(), s.numpy_dtype(), sharding=sharding
   )
 
@@ -140,7 +143,7 @@ def _custom_partitioning_propagate_user_sharding(user_sharding, shape,
 
 
 def _to_hlo_sharding(sharding, num_dimensions):
-  if not isinstance(sharding, jax.sharding.Sharding):
+  if not isinstance(sharding, Sharding):
     raise ValueError("Custom Partitioning rules must return Sharding.")
   return sharding._to_xla_hlo_sharding(num_dimensions)
 
@@ -178,10 +181,11 @@ def _custom_partitioning_partition(arg_shapes, arg_shardings, result_shape,
       _to_jax_shape(sharding.tile(s))
       for sharding, s in zip(result_shardings, result_shapes)
   ]
-  closed_jaxpr = jax.make_jaxpr(lower_fn, axis_env=list(mesh.shape.items()))(
-      *tiled_args
+  closed_jaxpr = api.make_jaxpr(lower_fn, axis_env=list(mesh.shape.items()))(
+      *info.in_tree.unflatten(tiled_args)
   )
-  if closed_jaxpr.out_avals != tiled_results:
+  if ([(o.shape, o.dtype) for o in closed_jaxpr.out_avals] !=
+      [(t.shape, t.dtype) for t in tiled_results]):
     raise ValueError(
         "Mismatch in result shapes. %s vs %s"
         % (repr(closed_jaxpr.out_avals), repr(tiled_results))
@@ -250,7 +254,7 @@ custom_partitioning_p.def_impl(_custom_partitioning_impl)
 
 def _check_for_tracers(x):
   if any(isinstance(leaf, core.Tracer) for leaf in tree_util.tree_leaves(x)):
-    raise UnexpectedTracerError(
+    raise errors.UnexpectedTracerError(
         "Found a JAX Tracer object passed as an argument to a"
         "custom_partitioning function in a position indicated as static by"
         "static_argnums. "
@@ -288,17 +292,12 @@ class custom_partitioning:
     f.def_partition(partition, propagate_user_sharding,
                     infer_sharding_from_operands=infer_sharding_from_operands,
                     sharding_rule='i j -> 'i j')
-    When config.use_shardy_partitioner.value is True, the sharding_rule is
-    used; otherwise, propagate_user_sharding and infer_sharding_from_operands
-    are used.
-    Instead of using an Einsum-like notation string, sharding_rule can also be
-    a SdyShardingRule object, such as sharding_rule=SdyShardingRule(("i", "j"), ("i", "j")).
 
   The args to ``def_partition`` are as follows:
 
   * ``propagate_user_sharding``: Callable which takes the sharding of a user (in the dag)
-    and returns a suggestion for a new `NamedSharding`. The default
-    implementation is just to return the suggested sharding.
+    and returns a suggestion for a new `NamedSharding`. The default value is None.
+    A trivial implementation is just to return the input sharding.
   * ``partition``: Callable which takes the SPMD suggested partition shapes and
     partition specs and returns the mesh, a per-shard lowering function, and the final
     input and output sharding specs (the SPMD partitioner will repartition the
@@ -309,10 +308,16 @@ class custom_partitioning:
   * ``decode_shardings``: When set to True, convert input ``GSPMDSharding``s to
     ``NamedSharding`` if possible. This may not be possible if the user does not
     provide a contextual mesh.
-  * ``sharding_rule``: Either an SdyShardingRule object or an Einsum-like
-    notation string that describes the sharding rule. We borrow the idea from
-    the einops.rearrange string , to use a space separator between factors and
-    allow multiple letters factor names.
+  * ``sharding_rule``: an SdyShardingRule object, an Einsum-like notation string
+    that describes the sharding rule, or a Callable that produces either of
+    these. We borrow the idea from the einops.rearrange string , to use a space
+    separator between factors and allow multiple letters factor names. See
+    `jax-shardy-guide <https://colab.sandbox.google.com/github/openxla/shardy/blob/main/docs/getting_started_jax.ipynb>`_
+    for more details and examples on how to use this.
+
+  When config.use_shardy_partitioner.value is True, `sharding_rule` is used;
+  otherwise, `propagate_user_sharding` and `infer_sharding_from_operands` are
+  used.
 
   Positional arguments can be specified as static using static_argnums. JAX uses
   :code:`inspect.signature(fun)` to resolve these positional arguments.
@@ -451,48 +456,46 @@ class custom_partitioning:
 
   __getattr__: Any = custom_api_util.forward_attr
 
-  def def_partition(self, partition, infer_sharding_from_operands,
+  def def_partition(self, partition, infer_sharding_from_operands=None,
                     propagate_user_sharding=None, decode_shardings=True,
                     sharding_rule=None):
-    if config.use_shardy_partitioner.value:
-      infer_sharding_from_operands = None
-      propagate_user_sharding = None
-    else:
-      sharding_rule = None
     self.partition = partition
     self.propagate_user_sharding = propagate_user_sharding
     self.infer_sharding_from_operands = infer_sharding_from_operands
     self.decode_shardings = decode_shardings
-    self.sharding_rule = None if sharding_rule is None \
-      else sharding_rule if isinstance(sharding_rule, SdyShardingRule) \
-          else str_to_sdy_sharding_rule(sharding_rule)
+    if (sharding_rule is None or isinstance(sharding_rule, Callable) or
+        isinstance(sharding_rule, SdyShardingRule)):
+      self.sharding_rule = sharding_rule
+    else:
+      self.sharding_rule = str_to_sdy_sharding_rule(sharding_rule)
     return partition
 
   def __call__(self, *args, **kwargs):
     args = _resolve_kwargs(self.fun, args, kwargs)
+    debug = api_util.debug_info("custom_partitioning", self.fun,
+                                args, kwargs,
+                                static_argnums=self.static_argnums)
     if self.static_argnums:
       static_argnums = set(self.static_argnums)
       args = tuple(x if i in static_argnums else x for i, x in enumerate(args))
       dyn_argnums = [i for i in range(len(args)) if i not in static_argnums]
       f_, dyn_args = api_util.argnums_partial(
-          lu.wrap_init(self.fun),
+          lu.wrap_init(self.fun, debug_info=debug),
           dyn_argnums,
           args,
           require_static_args_hashable=False,
       )
-      static_args = [args[i] for i in self.static_argnums]
+      static_args = tuple(args[i] for i in self.static_argnums)
       _check_for_tracers(static_args)
     else:
-      static_args = []
-      f_, dyn_args = lu.wrap_init(self.fun), args
+      static_args = ()
+      f_, dyn_args = lu.wrap_init(self.fun, debug_info=debug), args
     args_flat, in_tree = tree_util.tree_flatten(dyn_args)
     flat_fun, out_tree = api_util.flatten_fun_nokwargs(f_, in_tree)
     in_avals = [core.get_aval(x) for x in args_flat]
-    debug = pe.tracing_debug_info(self.fun, in_tree, out_tree, False,
-                          "custom_partitioning")
     mesh = mesh_lib.thread_resources.env.physical_mesh
     with core.extend_axis_env_nd(mesh.shape.items()):
-      jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(flat_fun, in_avals, debug)
+      jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(flat_fun, in_avals)
     assert not len(consts)
     closed_call = core.ClosedJaxpr(pe.convert_constvars_jaxpr(jaxpr), ())
 
@@ -500,6 +503,14 @@ class custom_partitioning:
     infer_sharding_from_operands = None
     sharding_rule = None
     if config.use_shardy_partitioner.value:
+      if (self.sharding_rule is None and
+          (self.propagate_user_sharding is not None or
+            self.infer_sharding_from_operands is not None)):
+        raise NotImplementedError(
+            "Shardy is used, but sharding propagation callbacks instead of "
+            "sharding_rule are provided. Need to provide sharding_rule to "
+            "migrate to Shardy."
+        )
       sharding_rule = self.sharding_rule
     else:
       propagate_user_sharding = self.propagate_user_sharding
@@ -557,11 +568,11 @@ def _custom_partitioning_lowering_rule(ctx: mlir.LoweringRuleContext, *values,
       return hlo_sharding
     if mesh.empty or not decode_shardings:
       assert devices is not None
-      return sharding_impls._op_sharding_to_pos_sharding(hlo_sharding, devices)
+      return sharding_impls.GSPMDSharding(devices, hlo_sharding)
     pspec = sharding_impls.parse_flatten_op_sharding(
-        hlo_sharding, mesh)[0].get_partition_spec()
-    pspec = jax.sharding.PartitionSpec(*pspec, *((None,) * (ndim - len(pspec))))
-    return jax.sharding.NamedSharding(mesh, pspec)
+        hlo_sharding, mesh)[0]
+    pspec = sharding_impls.PartitionSpec(*pspec, *((None,) * (ndim - len(pspec))))
+    return sharding_impls.NamedSharding(mesh, pspec)
 
   sharding_callback_info = _ShardingCallbackInfo(propagate_user_sharding,
       partition, to_mesh_pspec_sharding, in_tree, out_tree,
@@ -585,7 +596,16 @@ def _custom_partitioning_lowering_rule(ctx: mlir.LoweringRuleContext, *values,
       result_layouts=None)
   if sharding_rule is not None:
     value_types = [mlir.aval_to_ir_type(s) for s in call.in_avals]
-    out.attributes['sdy.sharding_rule'] = sdy_sharding_rule_to_mlir(sharding_rule, value_types, result_types)
+    if callable(sharding_rule):
+      sharding_rule = sharding_rule(*static_args, mesh, value_types, result_types)
+      if isinstance(sharding_rule, str):
+        sharding_rule = str_to_sdy_sharding_rule(sharding_rule)
+      elif not isinstance(sharding_rule, SdyShardingRule):
+          raise ValueError("sharding_rule callable must produce either an "
+                           "SdyShardingRule object or an Einsum-like notation "
+                           "string.")
+    out.attributes['sdy.sharding_rule'] = sdy_sharding_rule_to_mlir(
+      sharding_rule, value_types, result_types)
   return out.results
 
 mlir.register_lowering(custom_partitioning_p,
