@@ -14,8 +14,9 @@
 
 from __future__ import annotations
 
+from collections.abc import Callable
 import functools
-from typing import Any, Callable
+from typing import Any
 
 from jax._src import ad_util
 from jax._src import api_util
@@ -30,7 +31,8 @@ from jax._src.interpreters import mlir
 from jax._src.interpreters import partial_eval as pe
 from jax._src.interpreters import xla
 from jax._src.tree_util import (tree_flatten, tree_leaves, tree_map,
-                                tree_structure, treedef_tuple, tree_unflatten)
+                                tree_structure, treedef_tuple, tree_unflatten,
+                                PyTreeDef)
 
 
 source_info_util.register_exclusion(__file__)
@@ -89,11 +91,21 @@ class custom_transpose:
     # TODO(frostig,mattjj): could, and should, we avoid flattening
     # self.fun at this point?
 
-    flat_fun, out_tree2 = flatten_fun_nokwargs(lu.wrap_init(self.fun), in_tree)
+    flat_fun, out_tree2 = flatten_fun_nokwargs(
+        lu.wrap_init(
+            self.fun,
+            debug_info=api_util.debug_info("custom_transpose fun", self.fun,
+                                           (res_arg, lin_arg), {})),
+        in_tree)
     out_types_flat, out_tree = tree_flatten(out_types)
+    transpose_wrapped = lu.wrap_init(
+        self.transpose,
+        debug_info=api_util.debug_info("custom_transpose transpose_fun",
+                                       self.transpose,
+                                       (res_arg, out_types), {})                             )
     out_flat = custom_transpose_p.bind(flat_fun, *args_flat,
-                                       transpose=self.transpose,
-                                       out_types=out_types_flat,
+                                       transpose=transpose_wrapped,
+                                       out_types=tuple(out_types_flat),
                                        lin_tree=lin_tree,
                                        res_tree=res_tree,
                                        out_tree=out_tree)
@@ -124,19 +136,20 @@ def is_treedef_prefix(entire, prefix):
 def rule_name(rule):
   return getattr(rule, '__name__', '<unnamed transpose rule>')
 
-def check_transpose_rule_trees(rule, lin_tree, rule_out_tree):
+def check_transpose_rule_trees(rule: lu.WrappedFun,
+                               lin_tree: PyTreeDef,
+                               rule_out_tree: PyTreeDef):
   if not is_treedef_prefix(lin_tree, rule_out_tree):
-    if hasattr(rule, '_transpose_type_error'):
-      raise rule._transpose_type_error(lin_tree, rule_out_tree)
-    else:
-      raise TypeError(
-          'structure of custom transpose rule\'s output does not prefix-match '
-          'structure of primal function\'s linear inputs under '
-          f'custom transpose rule ({rule_name(rule)}).\n'
-          f'Transpose rule output: {rule_out_tree}\n'
-          f'Linear primal inputs: {lin_tree}')
+    rule_name = rule.debug_info.func_src_info if rule.debug_info else "<unknown>"
+    raise TypeError(
+        'structure of custom transpose rule\'s output does not prefix-match '
+        'structure of primal function\'s linear inputs under '
+        f'custom transpose rule ({rule_name}).\n'
+        f'Transpose rule output: {rule_out_tree}\n'
+        f'Linear primal inputs: {lin_tree}')
 
-def make_transpose_from_thunk(thunk, lin_tree):
+def make_transpose_from_thunk(thunk: Callable,
+                              lin_tree: PyTreeDef) -> lu.WrappedFun:
   transpose_jaxpr, transpose_consts = thunk()
   transpose_jaxpr = core.ClosedJaxpr(
       pe.convert_constvars_jaxpr(transpose_jaxpr), ())
@@ -144,7 +157,7 @@ def make_transpose_from_thunk(thunk, lin_tree):
     args_flat = tree_leaves((res_arg, ct_out))
     ct_ins = core.jaxpr_as_fun(transpose_jaxpr)(*transpose_consts, *args_flat)
     return tree_unflatten(lin_tree, ct_ins)
-  return transpose
+  return lu.wrap_init(transpose, debug_info=transpose_jaxpr.jaxpr.debug_info)
 
 
 ### custom_transpose primitive and rules
@@ -154,29 +167,29 @@ class CustomTransposePrimitive(core.Primitive):
   map_primitive = False
   multiple_results = True
 
-  def bind(self, call, *args, **params):
-    # TODO(frostig,mattjj): This doesn't handle closures yet, which is
-    # a bit involved. Closures are complicated by us binding `call`
-    # twice in the JVP rule for custom transpose. The `env_trace_todo`
-    # output by `process_env_traces` due to one of those two bindings
-    # should be passable to the other, and need to be passed onward
-    # since the second bind is deferred by partial eval (since it
-    # typically receives unknowns)
-    top_trace = core.find_top_trace(args)
-    tracers = map(top_trace.full_raise, args)
-    outs = top_trace.process_custom_transpose(self, call, tracers, **params)
-    return outs
+  def bind(self, *args, **params):
+    return self._true_bind(*args, **params)
+
+  def bind_with_trace(self, trace, call_args, params):
+    call, tracers = call_args[0], call_args[1:]
+    return trace.process_custom_transpose(self, call, tracers, **params)
 
   # TODO(frostig,mattjj): consider keeping `call` as a named parameter
   # instead of following this "call primitive" convention.
   def get_bind_params(self, params):
-    assert 'call_jaxpr' in params
-    assert 'transpose_jaxpr_thunk' in params
-    new_params = dict(params)
-    new_params['transpose'] = make_transpose_from_thunk(
-        new_params.pop('transpose_jaxpr_thunk'),
-        new_params['lin_tree'])
-    call = lu.wrap_init(core.jaxpr_as_fun(new_params.pop('call_jaxpr')))
+    if 'call_jaxpr' in params:
+      assert 'transpose_jaxpr_thunk' in params
+      new_params: dict[str, Any] = dict(params)
+      new_params['transpose'] = make_transpose_from_thunk(
+          new_params.pop('transpose_jaxpr_thunk'),
+          new_params['lin_tree'])
+      call_jaxpr: core.ClosedJaxpr = new_params.pop('call_jaxpr')
+      call = lu.wrap_init(core.jaxpr_as_fun(call_jaxpr),
+                          debug_info=call_jaxpr.jaxpr.debug_info)
+    else:
+      assert 'transpose' in params
+      new_params: dict[str, Any] = dict(params)
+      call = new_params.pop("call")
     return [call], new_params
 
 
@@ -188,7 +201,7 @@ def custom_transpose_typecheck(_, *in_atoms, out_types, **params):
 
 def custom_transpose_transpose_rule(
     cts, *args, out_types, res_tree, lin_tree, out_tree, **params):
-
+  transpose: lu.WrappedFun
   if 'transpose_jaxpr_thunk' in params:
     assert 'call_jaxpr' in params
     transpose = make_transpose_from_thunk(
@@ -204,18 +217,24 @@ def custom_transpose_transpose_rule(
   # Consider passing this information to the custom transpose rule?
 
   res_arg, lin_arg = tree_unflatten(call_in_tree, args)
-  del lin_arg
   assert all(not ad.is_undefined_primal(x) for x in tree_leaves(res_arg))
 
   cts = [ad_util.zeros_like_aval(ct.aval) if type(ct) is ad_util.Zero else ct
          for ct in cts]
   ct_out = tree_unflatten(out_tree, cts)
-  ct_lin = transpose(res_arg, ct_out)
+  ct_lin = transpose.call_wrapped(res_arg, ct_out)
   check_transpose_rule_trees(transpose, lin_tree, tree_structure(ct_lin))
-  ct_lin_flat, _ = tree_flatten(
-      tree_broadcast(lin_tree, ct_lin, is_leaf=lambda x: x is None),
-      is_leaf=lambda x: x is None)
-  return [None] * len(tree_leaves(res_arg)) + ct_lin_flat
+  ct_lin = tree_broadcast(lin_tree, ct_lin, is_leaf=lambda x: x is None)
+
+  # When the transpose returns None, we treat that as a Zero, except when the
+  # input is also None. In that case, the cotangent corresponding to that input
+  # should be dropped.
+  zero = object()
+  ct_lin = tree_map(lambda l, ct: zero if ct is None and l is not None else ct,
+                    lin_arg, ct_lin, is_leaf=ad.is_undefined_primal)
+
+  ct_lin_flat, _ = tree_flatten(ct_lin)
+  return [None] * res_tree.num_leaves + [None if ct is zero else ct for ct in ct_lin_flat]
 
 
 def custom_transpose_lowering(*args, call_jaxpr, **params):
