@@ -17,7 +17,6 @@ from __future__ import annotations
 
 import atexit
 from collections.abc import Sequence
-import contextlib
 import dataclasses
 import enum
 from functools import partial
@@ -25,9 +24,8 @@ import itertools
 import logging
 import threading
 import time
-from typing import Any, NamedTuple
+from typing import Any
 
-import jax
 from jax._src import api
 from jax._src import array
 from jax._src import basearray
@@ -35,24 +33,28 @@ from jax._src import config
 from jax._src import core
 from jax._src import dtypes
 from jax._src import lib
-from jax._src import source_info_util
+from jax._src import pjit
 from jax._src import traceback_util
 from jax._src import util
+
+from jax._src import xla_bridge
 from jax._src.abstract_arrays import array_types
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
 from jax._src.interpreters import pxla
 from jax._src.interpreters import xla
-from jax._src.layout import DeviceLocalLayout, Layout
+from jax._src.api_util import InternalFloatingPointError
+from jax._src.layout import Layout, Format
 from jax._src.lib import xla_client as xc
 from jax._src.mesh import AbstractMesh, Mesh
-from jax._src.monitoring import record_event_duration_secs, record_event_time_span
+from jax._src.monitoring import record_scalar, record_event_duration_secs, record_event_time_span
 from jax._src.partition_spec import PartitionSpec
 from jax._src.sharding import Sharding
-from jax._src.sharding_impls import ( NamedSharding,
-    SingleDeviceSharding, TransferToMemoryKind,
+from jax._src.sharding_impls import (
+    NamedSharding, SingleDeviceSharding, TransferToMemoryKind, GSPMDSharding,
     is_single_device_sharding)
+from jax._src.stages import SourceInfo
 import numpy as np
 
 
@@ -99,6 +101,7 @@ def xla_primitive_callable(prim: core.Primitive, **params):
       return prim.bind(*args, **params)
   prim_fun.__name__ = prim.name
   prim_fun.__qualname__ = prim.name
+  prim_fun._apply_primitive = True
   return api.jit(prim_fun)
 
 
@@ -132,11 +135,12 @@ class RuntimeTokenSet(threading.local):
       # TODO(yueshengys): This might still be buggy in a multi-process SPMD
       # scenario. Revise the logic later. A distributed shutdown barrier inside
       # the XLA program may be needed.
-      return jax.device_put(tok, jax.sharding.PositionalSharding(devices))
+      return api.device_put(
+          tok, NamedSharding(Mesh(devices, 'x'), PartitionSpec('x')))
 
     # We only use replicated sharding for the first time when the token for the
     # order effect hasn't been created.
-    s = jax.sharding.GSPMDSharding.get_replicated(devices)
+    s = GSPMDSharding.get_replicated(devices)
     sharded_tok = core.Token(pxla.shard_args([s], [None], [None], [tok])[0])
     self.current_tokens[eff] = sharded_tok
     return sharded_tok
@@ -168,22 +172,40 @@ def wait_for_tokens():
   runtime_tokens.block_until_ready()
 
 
-@contextlib.contextmanager
-def log_elapsed_time(fmt: str, fun_name: str, event: str | None = None):
-  if _on_exit:
-    yield
-  else:
-    log_priority = logging.WARNING if config.log_compiles.value else logging.DEBUG
-    start_time = time.time()
-    yield
+class LogElapsedTimeContextManager:
+  __slots__ = ['fmt', 'fun_name', 'event', 'start_time']
+
+  def __init__(self, fmt: str, fun_name: str, event: str | None = None):
+    self.fmt = fmt
+    self.fun_name = fun_name
+    self.event = event
+
+  def __enter__(self):
+    self.start_time = time.time()
+    if self.event is not None:
+      record_scalar(
+          self.event, self.start_time, fun_name=self.fun_name
+      )
+
+  def __exit__(self, exc_type, exc_value, traceback):
+    if _on_exit:
+      return
+
     end_time = time.time()
-    elapsed_time = end_time - start_time
+    elapsed_time = end_time - self.start_time
+    log_priority = logging.WARNING if config.log_compiles.value else logging.DEBUG
     if logger.isEnabledFor(log_priority):
-      logger.log(log_priority, fmt.format(
-          fun_name=fun_name, elapsed_time=elapsed_time))
-    if event is not None:
-      record_event_duration_secs(event, elapsed_time)
-      record_event_time_span(event, start_time, end_time)
+      logger.log(log_priority, self.fmt.format(
+          fun_name=self.fun_name, elapsed_time=elapsed_time))
+    if self.event is not None:
+      record_event_duration_secs(
+          self.event, elapsed_time, fun_name=self.fun_name
+      )
+      record_event_time_span(
+          self.event, self.start_time, end_time, fun_name=self.fun_name
+      )
+
+log_elapsed_time = LogElapsedTimeContextManager
 
 
 def should_tuple_args(num_args: int, platform: str) -> bool:
@@ -221,16 +243,10 @@ def jaxpr_has_prim_requiring_devices(jaxpr: core.Jaxpr) -> bool:
   return False
 
 
-class SourceInfo(NamedTuple):
-  source_info: source_info_util.SourceInfo
-  eqn_name: str
-
-
 @util.weakref_lru_cache
 def get_intermediate_shardings(
     jaxpr: core.Jaxpr) -> Sequence[tuple[Sharding, SourceInfo]]:
-  from jax._src import pjit
-  from jax.experimental import shard_map
+  from jax._src import shard_map  # pytype: disable=import-error
 
   out = []
   for eqn in jaxpr.eqns:
@@ -240,19 +256,17 @@ def get_intermediate_shardings(
         continue
       source_info = SourceInfo(eqn.source_info, eqn.primitive.name)
       out.append((s, source_info))
-    elif eqn.primitive is pjit.pjit_p:
+    elif eqn.primitive is pjit.jit_p:
       source_info = SourceInfo(eqn.source_info, eqn.primitive.name)
       out.extend((i, source_info) for i in eqn.params['in_shardings'])
       out.extend((o, source_info) for o in eqn.params['out_shardings'])
     elif eqn.primitive is shard_map.shard_map_p:
-      if isinstance(eqn.params['mesh'], AbstractMesh):
+      mesh = eqn.params['mesh']
+      if isinstance(mesh, AbstractMesh):
         continue
       source_info = SourceInfo(eqn.source_info, eqn.primitive.name)
-      def _names_to_pspec(names):
-        ndmin = max(names) + 1 if names else 0
-        return PartitionSpec(*(names.get(i) for i in range(ndmin)))
-      out.extend((NamedSharding(eqn.params['mesh'], _names_to_pspec(names)), source_info)
-                 for names in [*eqn.params['in_names'], *eqn.params['out_names']])
+      out.extend((NamedSharding(mesh, spec), source_info)
+                 for spec in [*eqn.params['in_specs'], *eqn.params['out_specs']])
     elif eqn.primitive is device_put_p:
       source_info = SourceInfo(eqn.source_info, eqn.primitive.name)
       out.extend((s, source_info) for s in eqn.params['devices']
@@ -320,9 +334,9 @@ def check_special(name: str, bufs: Sequence[basearray.Array]) -> None:
 def _check_special(name: str, dtype: np.dtype, buf: basearray.Array) -> None:
   if dtypes.issubdtype(dtype, np.inexact):
     if config.debug_nans.value and np.any(np.isnan(np.asarray(buf))):
-      raise FloatingPointError(f"invalid value (nan) encountered in {name}")
+      raise InternalFloatingPointError(name, "nan")
     if config.debug_infs.value and np.any(np.isinf(np.asarray(buf))):
-      raise FloatingPointError(f"invalid value (inf) encountered in {name}")
+      raise InternalFloatingPointError(name, "inf")
 
 class CopySemantics(enum.Enum):
   ALIAS = enum.auto()
@@ -342,16 +356,6 @@ def _different_device_order_reshard(x, target_sharding, copy: CopySemantics):
     return api.jit(_identity_fn, out_shardings=target_sharding,
                    donate_argnums=donate_argnums)(x)
 
-  if inp_sharding.device_set != target_sharding.device_set:
-    inp_ids = [d.id for d in inp_sharding._device_assignment]
-    inp_plat = inp_sharding._device_assignment[0].platform.upper()
-    target_ids = [d.id for d in target_sharding._device_assignment]
-    target_plat = target_sharding._device_assignment[0].platform.upper()
-    raise ValueError("Input and target sharding should have the same set of "
-                     f"devices. Got input's device set ids: {inp_ids} on "
-                     f"platform {inp_plat} and target sharding's device set "
-                     f"ids: {target_ids} on platform {target_plat}")
-
   if inp_sharding.is_fully_replicated:
     permute_order = None
   else:
@@ -364,11 +368,48 @@ def _different_device_order_reshard(x, target_sharding, copy: CopySemantics):
       new_mesh, inp_sharding.spec, memory_kind=target_sharding.memory_kind,
       _logical_device_ids=(None if permute_order is None else
                             tuple(permute_order.tolist())))
-  new_x = array.make_array_from_single_device_arrays(x.shape, new_s, x._arrays)
+  new_x = _reorder_shards(x, new_s, CopySemantics.ALIAS)
   return api.jit(_identity_fn, out_shardings=target_sharding,
                 donate_argnums=donate_argnums)(new_x)
 
 
+def _reorder_shards(x, new_s, copy_semantics: CopySemantics):
+  """Reorders array shards to match the order indicated by the new sharding."""
+  xc_copy_semantics = pxla.to_xc_copy_semantics([copy_semantics])[0]
+  return xc.reorder_shards(x, new_s, xc_copy_semantics)  # type: ignore
+
+
+@util.cache()
+def _is_supported_cross_host_transfer(ndim, src_sharding, dst_sharding):
+  """Returns True if src->dst is a supported cross-host transfer."""
+  if xla_bridge.process_count() == 1:
+    return False
+  if (src_sharding._internal_device_list.device_kind !=
+      dst_sharding._internal_device_list.device_kind):
+    return False
+  if (src_sharding._to_xla_hlo_sharding(ndim) !=
+      dst_sharding._to_xla_hlo_sharding(ndim)):
+    return False
+  # This check excludes the case where the source and destination shardings
+  # have the same process index sets but there are shards that require
+  # cross-host transfers. This case is supportable but expensive to check for.
+  different_process_inds = (
+      src_sharding._internal_device_list.process_indices !=
+      dst_sharding._internal_device_list.process_indices)
+  backend = xla_bridge.get_backend()
+  # If a cross-host device transfer is requested but the backend does not
+  # support it, then the user must set the flags to enable DCN-based transfers.
+  if (different_process_inds and
+      not getattr(backend, 'supports_cross_host_transfers', False) and
+      not xla_bridge.CROSS_HOST_TRANSFER_SOCKET_ADDRESS.value):
+    raise ValueError(
+        f"The backend ({backend.platform}, {backend.platform_version}) does "
+        "not support cross-host device transfers via ICI/NCCL. Please set "
+        "jax_cross_host_transfer_socket_address and (optionally) "
+        "jax_cross_host_transport_addresses flags to enable DCN-based cross "
+        "host device transfers.")
+  return different_process_inds
+
 @dataclasses.dataclass(frozen=True)
 class _DeferredShardArg:
   """Deferred call to `pxla.shard_args`.
@@ -390,7 +431,7 @@ class _DeferredShardArg:
 
 
 def _device_put_sharding_impl(x, aval, device, copy):
-  from jax.experimental import multihost_utils
+  from jax.experimental import multihost_utils  # pytype: disable=import-error
 
   if isinstance(device, Sharding):
     s = device
@@ -399,7 +440,8 @@ def _device_put_sharding_impl(x, aval, device, copy):
       return x
 
     if (not s.is_fully_addressable and
-        isinstance(x, array.ArrayImpl) and not x.is_fully_addressable):
+        isinstance(x, array.ArrayImpl) and not x.is_fully_addressable and
+        s.device_set == x.sharding.device_set):
       assert isinstance(s, Sharding)
       return _different_device_order_reshard(x, s, copy)
 
@@ -410,14 +452,47 @@ def _device_put_sharding_impl(x, aval, device, copy):
       assert isinstance(s, Sharding)
       return _different_device_order_reshard(x, s, copy)
 
+    if (isinstance(x, array.ArrayImpl) and x._committed
+        and _is_supported_cross_host_transfer(x.ndim, x.sharding, s)):
+      return xc.batched_copy_array_to_devices_with_sharding(
+          [x], [s._internal_device_list], [s],  # pytype: disable=attribute-error
+          pxla.to_xc_copy_semantics([copy]))[0]
+
     if not s.is_fully_addressable:
+      # If both the source and target shardings are not fully addressable and
+      # one of the above conditions has not been met, then assume that the user
+      # is attempting a different device order reshard.
+      if (isinstance(x, array.ArrayImpl) and not x.is_fully_addressable
+          and s.device_set != x.sharding.device_set):
+        inp_ids = [d.id for d in x.sharding._device_assignment]
+        inp_plat = x.sharding._device_assignment[0].platform.upper()
+        target_ids = [d.id for d in s._device_assignment]
+        target_plat = s._device_assignment[0].platform.upper()
+        raise ValueError(
+            "For a cross-host reshard in multi-controller JAX, input and target"
+            " sharding should have the same set of devices. Got input's device"
+            f" set ids: {inp_ids} on platform {inp_plat} and target sharding's"
+            f" device set ids: {target_ids} on platform {target_plat}.\n\n"
+            "There is experimental support for cross-host transfers with "
+            "different device sets, when input/output shardings have the same "
+            "indices and layouts, in the TFRT TPU runtime only.")
+
       if ((isinstance(x, array.ArrayImpl) and not x._committed) or
-          type(x) in array_types):
-        multihost_utils.assert_equal(
-            x, fail_message=(
-                f"{type(x)} passed to device_put is not the same on each"
-                " process. Make sure you are passing the same value of"
-                f" {type(x)} on each process."))
+          type(x) in array_types or type(x) in dtypes.python_scalar_dtypes):
+        # If all hosts participate in the sharding, assert that the input is the
+        # same on all hosts. If some hosts have no addressable devices in the
+        # sharding, bypass the check, since we can't easily distinguish between
+        # these two cases: (1) the sharding contains the same subset of global
+        # devices on all hosts (and hosts with no addressable devices in the
+        # sharding do not transfer data) or (2) the sharding contains a
+        # different subset of devices on each host. For (1), the input should be
+        # the same on all hosts, but for (2) it need not be.
+        if xla_bridge.process_count() == len(s._internal_device_list.process_indices):  # pytype: disable=attribute-error
+          multihost_utils.assert_equal(
+              x, fail_message=(
+                  f"{type(x)} passed to device_put is not the same on each"
+                  " process. Make sure you are passing the same value of"
+                  f" {type(x)} on each process."))
         return _DeferredShardArg(x, s, aval, True, copy)
       # TODO(yashkatariya,mattjj): Link to a doc about McJAX and jax.Array.
       raise ValueError(
@@ -439,6 +514,9 @@ def _device_put_sharding_impl(x, aval, device, copy):
         return _DeferredShardArg(x, x.sharding, aval, x.committed, copy)
     elif is_single_device_sharding(x.sharding):
       device = x.sharding._device_assignment[0] if device is None else device
+      if copy == CopySemantics.COPY:
+        return xc.batched_device_put(aval, SingleDeviceSharding(device), [x],
+                                     [device], True, True)
       return pxla.batched_device_put(aval, SingleDeviceSharding(device), [x],
                                      [device])
 
@@ -448,8 +526,8 @@ def _device_put_sharding_impl(x, aval, device, copy):
 
 
 def _device_put_impl(
-    x, *, device: Device | Sharding | Layout | None,
-    src: Device | Sharding | Layout | None, copy: CopySemantics):
+    x, *, device: Device | Sharding | Format | None,
+    src: Device | Sharding | Format | None, copy: CopySemantics):
   if (isinstance(device, TransferToMemoryKind) or
       isinstance(src, TransferToMemoryKind)):
     raise ValueError(
@@ -463,18 +541,18 @@ def _device_put_impl(
     raise TypeError(
         f"Argument '{x}' of type {type(x)} is not a valid JAX type") from err
 
-  if isinstance(device, Layout):
+  if isinstance(device, Format):
     l = device
-    dll = l.device_local_layout
-    x_dll = x.layout.device_local_layout if hasattr(x, 'layout') else None
+    dll = l.layout
+    x_dll = x.format.layout if hasattr(x, 'format') else None
     if dll is None and l.sharding is None:
       return _device_put_sharding_impl(x, aval, l.sharding, copy)
     if (not isinstance(l.sharding, Sharding) or
-        not isinstance(dll, (DeviceLocalLayout, type(None)))):
+        not isinstance(dll, (Layout, type(None)))):
       raise ValueError(
-          "sharding and device_local_layout in `Layout` instance should be"
+          "sharding and layout in `Layout` instance should be"
           f" concrete. Got layout: {l} for input {aval.str_short()}")
-    if (getattr(x, 'layout', None) == l and getattr(x, '_committed', False) and
+    if (getattr(x, 'format', None) == l and getattr(x, '_committed', False) and
         copy == CopySemantics.ALIAS):
       return x
     if x_dll is None and dll is None:
@@ -488,8 +566,8 @@ def _device_put_impl(
 
 def _batched_device_put_impl(
     *xs,
-    devices: Sequence[Device | Sharding | Layout | None],
-    srcs: Sequence[Device | Sharding | Layout | None],
+    devices: Sequence[Device | Sharding | Format | None],
+    srcs: Sequence[Device | Sharding | Format | None],
     copy_semantics: Sequence[CopySemantics]):
   ys = []
   dsa_indices, dsa_xs, dsa_shardings, dsa_copy_semantics = [], [], [], []
@@ -505,7 +583,7 @@ def _batched_device_put_impl(
   if dsa_xs:
     # Batch shard_arg calls. Helps improve efficiency for backends that support
     # efficient batch transfer.
-    # device_put handles `Layout` via a different path, so just pass `None` as
+    # device_put handles `Format` via a different path, so just pass `None` as
     # the layout here.
     shard_arg_results = pxla.shard_args(dsa_shardings, [None] * len(dsa_xs),
                                         dsa_copy_semantics, dsa_xs)
@@ -544,7 +622,7 @@ def _device_put_transpose(cts, *_, devices, srcs, copy_semantics):
         assert cp == CopySemantics.COPY
         new_copy_semantics.append(CopySemantics.COPY)
     ys = device_put_p.bind(*args, devices=srcs, srcs=devices,
-                           copy_semantics=new_copy_semantics)
+                           copy_semantics=tuple(new_copy_semantics))
     for i, y in zip(indices, ys):
       results[i] = y
   return results
