@@ -22,7 +22,8 @@ from typing import Any
 from jax._src import clusters
 from jax._src import config
 from jax._src import xla_bridge
-from jax._src.lib import xla_extension
+from jax._src.lib import _jax
+from jax._src.lib import jaxlib_extension_version
 
 logger = logging.getLogger(__name__)
 
@@ -37,10 +38,11 @@ _CHECK_PROXY_ENVS = config.bool_flag(
 class State:
   process_id: int = 0
   num_processes: int = 1
-  service: Any | None = None
-  client: Any | None = None
+  service: _jax.DistributedRuntimeService | Any | None = None
+  client: _jax.DistributedRuntimeClient | Any | None = None
   preemption_sync_manager: Any | None = None
   coordinator_address: str | None = None
+  slice_index: int | None = None
 
   def initialize(self,
                  coordinator_address: str | None = None,
@@ -50,10 +52,8 @@ class State:
                  cluster_detection_method: str | None = None,
                  initialization_timeout: int = 300,
                  coordinator_bind_address: str | None = None,
-                 service_heartbeat_interval_seconds: int = 10,
-                 service_max_missing_heartbeats: int = 10,
-                 client_heartbeat_interval_seconds: int = 10,
-                 client_max_missing_heartbeats: int = 10):
+                 heartbeat_timeout_seconds: int = 100,
+                 slice_index: int | None = None):
     coordinator_address = (coordinator_address or
                            os.environ.get('JAX_COORDINATOR_ADDRESS'))
     if isinstance(local_device_ids, int):
@@ -124,59 +124,76 @@ class State:
       )
       logger.warning(warning)
 
+
+    heartbeat_kwargs = {}
+    if jaxlib_extension_version >= 361:
+      # In jaxlib version 361, the heartbeat_timeout argument replaced the old
+      # heartbeat_interval and max_missing_heartbeats arguments.
+      heartbeat_kwargs['heartbeat_timeout'] = heartbeat_timeout_seconds
+    else:
+      heartbeat_kwargs['heartbeat_interval'] = heartbeat_timeout_seconds
+      heartbeat_kwargs['max_missing_heartbeats'] = 1
+
     if process_id == 0:
       if self.service is not None:
         raise RuntimeError('distributed.initialize should only be called once.')
       logger.info(
           'Starting JAX distributed service on %s', coordinator_bind_address
       )
-      self.service = xla_extension.get_distributed_runtime_service(
-          coordinator_bind_address, num_processes,
-          heartbeat_interval=service_heartbeat_interval_seconds,
-          max_missing_heartbeats=service_max_missing_heartbeats)
+      self.service = _jax.get_distributed_runtime_service(
+          coordinator_bind_address, num_processes, **heartbeat_kwargs) # type: ignore
 
     self.num_processes = num_processes
 
     if self.client is not None:
       raise RuntimeError('distributed.initialize should only be called once.')
 
-    self.client = xla_extension.get_distributed_runtime_client(
+    self.client = _jax.get_distributed_runtime_client(
         coordinator_address, process_id, init_timeout=initialization_timeout,
-        heartbeat_interval=client_heartbeat_interval_seconds,
-        max_missing_heartbeats=client_max_missing_heartbeats, use_compression=True)
+        use_compression=True, **heartbeat_kwargs) # type: ignore
     logger.info('Connecting to JAX distributed service on %s', coordinator_address)
     self.client.connect()
 
     self.initialize_preemption_sync_manager()
 
+    if slice_index is None and 'JAX_SLICE_INDEX' in os.environ:
+      slice_index = int(os.environ.get('JAX_SLICE_INDEX'))  # type: ignore
+    self.slice_index = slice_index
+
   def shutdown(self):
+    if self.preemption_sync_manager:
+      # It's important to shut down the preemption sync manager before the
+      # client because the preemption sync manager depends on the client.
+      # TODO: Delete hasattr check once 0.6.1 is the minimum jaxlib version
+      if hasattr(self.preemption_sync_manager, "shutdown"):
+        self.preemption_sync_manager.shutdown()
+      self.preemption_sync_manager = None
     if self.client:
       self.client.shutdown()
       self.client = None
     if self.service:
       self.service.shutdown()
       self.service = None
-    if self.preemption_sync_manager:
-      self.preemption_sync_manager = None
 
   def initialize_preemption_sync_manager(self):
     if self.preemption_sync_manager is not None:
       raise RuntimeError(
           'Preemption sync manager should only be initialized once.')
     self.preemption_sync_manager = (
-        xla_extension.create_preemption_sync_manager())
+        _jax.create_preemption_sync_manager())
     self.preemption_sync_manager.initialize(self.client)
 
 global_state = State()
 
-
 def initialize(coordinator_address: str | None = None,
                num_processes: int | None = None,
                process_id: int | None = None,
                local_device_ids: int | Sequence[int] | None = None,
                cluster_detection_method: str | None = None,
                initialization_timeout: int = 300,
-               coordinator_bind_address: str | None = None):
+               heartbeat_timeout_seconds: int = 100,
+               coordinator_bind_address: str | None = None,
+               slice_index: int | None = None):
   """Initializes the JAX distributed system.
 
   Calling :func:`~jax.distributed.initialize` prepares JAX for execution on
@@ -232,11 +249,16 @@ def initialize(coordinator_address: str | None = None,
     initialization_timeout: Time period (in seconds) for which connection will
       be retried. If the initialization takes more than the timeout specified,
       the initialization will error. Defaults to 300 secs i.e. 5 mins.
+    heartbeat_timeout_seconds: The time (in seconds) after which a process is
+      considered dead if it hasn't successfully sent any heartbeats. Defaults
+      to 100 seconds.
     coordinator_bind_address: the address and port to which the coordinator service
       on process `0` should bind. If this is not specified, the default is to bind to
       all available addresses on the same port as ``coordinator_address``. On systems
       that have multiple network interfaces per node it may be insufficient to only
       have the coordinator service listen on one address/interface.
+    slice_index: The slice index assigned to this process' local devices. If any process sets ``slice_index``,
+      then all processes must do so. If ``None`` the slice indices will be chosen automatically.
 
   Raises:
     RuntimeError: If :func:`~jax.distributed.initialize` is called more than once
@@ -258,12 +280,19 @@ def initialize(coordinator_address: str | None = None,
   """
   if xla_bridge.backends_are_initialized():
     raise RuntimeError("jax.distributed.initialize() must be called before "
-                        "any JAX computations are executed.")
+                        "any JAX calls that might initialise the XLA backend. "
+                        "This includes any computation, but also calls to jax.devices, jax.device_put, and others.")
   global_state.initialize(coordinator_address, num_processes, process_id,
                           local_device_ids, cluster_detection_method,
-                          initialization_timeout, coordinator_bind_address)
+                          initialization_timeout, coordinator_bind_address,
+                          heartbeat_timeout_seconds=heartbeat_timeout_seconds,
+                          slice_index=slice_index)
 
 
+def is_initialized() -> bool:
+  """Check if the JAX distributed system is initialized."""
+  return global_state.client is not None
+
 def shutdown():
   """Shuts down the distributed system.
 
