@@ -11,9 +11,7 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-"""JAX APIs for exporting JAX functions for interoperation.
-
-"""
+"""JAX APIs for exporting JAX functions for interoperation."""
 
 from __future__ import annotations
 
@@ -27,34 +25,41 @@ import json
 import re
 from typing import Any, Protocol, TypeVar, Union, cast
 
-from absl import logging
+import logging
 import numpy as np
 
-import jax
-from jax import sharding
-
 from jax._src import ad_util
+from jax._src import api
 from jax._src import config
 from jax._src import core
+from jax._src import custom_derivatives
 from jax._src import dispatch
 from jax._src import dtypes
 from jax._src import effects
+from jax._src import mesh as mesh_lib
 from jax._src.interpreters import mlir
 from jax._src.interpreters import pxla
+from jax._src.lib import ifrt_version
 from jax._src.lib import xla_client
-from jax._src.lib.mlir import ir
+from jax._src.lib import _jax
+from jax._src.lib import jaxlib_extension_version
+from jax._src.lib.mlir import ir, passmanager
 from jax._src.lib.mlir.dialects import hlo
-from jax._src.lib.mlir.dialects import func as func_dialect
+from jax._src.lib.mlir.dialects import func as func_dialect, sdy
 from jax._src import pjit
+from jax._src import sharding
 from jax._src import sharding_impls
 from jax._src import source_info_util
 from jax._src import stages
 from jax._src import tree_util
+from jax._src import typing
 from jax._src import util
 from jax._src import xla_bridge as xb
 
 from jax._src.export import shape_poly
 
+logger = logging.getLogger(__name__)
+
 map = util.safe_map
 zip = util.safe_zip
 
@@ -65,9 +70,9 @@ LoweringSharding = Union[sharding.Sharding, pxla.UnspecifiedValue]
 HloSharding = xla_client.HloSharding
 
 # The minimum and maximum supported calling convention version.
-# See https://jax.readthedocs.io/en/latest/export/export.html#export-calling-convention-version
+# See https://docs.jax.dev/en/latest/export/export.html#export-calling-convention-version
 minimum_supported_calling_convention_version = 9
-maximum_supported_calling_convention_version = 9
+maximum_supported_calling_convention_version = 10
 
 
 class DisabledSafetyCheck:
@@ -151,16 +156,16 @@ class Exported:
     platforms: a tuple containing the platforms for which the function should
         be exported. The set of platforms in JAX is open-ended; users can
         add platforms. JAX built-in platforms are: 'tpu', 'cpu', 'cuda', 'rocm'.
-        See https://jax.readthedocs.io/en/latest/export/export.html#cross-platform-and-multi-platform-export.
+        See https://docs.jax.dev/en/latest/export/export.html#cross-platform-and-multi-platform-export.
     ordered_effects: the ordered effects present in the serialized module.
-        This is present from serialization version 9. See https://jax.readthedocs.io/en/latest/export/export.html#module-calling-convention
+        This is present from serialization version 9. See https://docs.jax.dev/en/latest/export/export.html#module-calling-convention
         for the calling convention in presence of ordered effects.
     unordered_effects: the unordered effects present in the serialized module.
         This is present from serialization version 9.
     mlir_module_serialized: the serialized lowered VHLO module.
     calling_convention_version: a version number for the calling
         convention of the exported module.
-        See more versioning details at https://jax.readthedocs.io/en/latest/export/export.html#calling-convention-versions.
+        See more versioning details at https://docs.jax.dev/en/latest/export/export.html#calling-convention-versions.
     module_kept_var_idx: the sorted indices of the arguments among `in_avals` that
         must be passed to the module. The other arguments have been dropped
         because they are not used.
@@ -179,7 +184,7 @@ class Exported:
         for each primal output. It returns a tuple with the cotangents
         corresponding to the flattened primal inputs.
 
-  See a [description of the calling convention for the `mlir_module`](https://jax.readthedocs.io/en/latest/export/export.html#module-calling-convention).
+  See a [description of the calling convention for the `mlir_module`](https://docs.jax.dev/en/latest/export/export.html#module-calling-convention).
   """
   fun_name: str
   in_tree: tree_util.PyTreeDef
@@ -213,7 +218,7 @@ class Exported:
 
   def in_shardings_jax(
     self,
-    mesh: sharding.Mesh) -> Sequence[sharding.Sharding | None]:
+    mesh: mesh_lib.Mesh) -> Sequence[sharding.Sharding | None]:
     """Creates Shardings corresponding to self.in_shardings_hlo.
 
     The Exported object stores `in_shardings_hlo` as HloShardings, which are
@@ -223,7 +228,9 @@ class Exported:
 
     Example usage:
 
-      >>> from jax import export
+      >>> from jax._src.lib import ifrt_version
+      >>> if ifrt_version < 12: jax.config.update("jax_use_shardy_partitioner", False)
+      >>> from jax import export, sharding
       >>> # Prepare the exported object:
       >>> exp_mesh = sharding.Mesh(jax.devices(), ("a",))
       >>> exp = export.export(jax.jit(lambda x: jax.numpy.add(x, x),
@@ -248,17 +255,17 @@ class Exported:
        Shard(device=CpuDevice(id=0), index=(slice(7, 8, None),), replica_id=0, data=[14])]
 
     """
-    return tuple(_hlo_sharding_to_xla_compatible_sharding(s, mesh)
+    return tuple(_hlo_sharding_to_named_sharding(s, mesh)
                  for s in self.in_shardings_hlo)
 
   def out_shardings_jax(
       self,
-      mesh: sharding.Mesh) -> Sequence[sharding.Sharding | None]:
+      mesh: mesh_lib.Mesh) -> Sequence[sharding.Sharding | None]:
     """Creates Shardings corresponding to `self.out_shardings_hlo`.
 
     See documentation for in_shardings_jax.
     """
-    return tuple(_hlo_sharding_to_xla_compatible_sharding(s, mesh)
+    return tuple(_hlo_sharding_to_named_sharding(s, mesh)
                  for s in self.out_shardings_hlo)
 
   def has_vjp(self) -> bool:
@@ -304,7 +311,7 @@ class Exported:
 
     The invocation supports reverse-mode AD, and all the features supported
     by exporting: shape polymorphism, multi-platform, device polymorphism.
-    See the examples in the [JAX export documentation](https://jax.readthedocs.io/en/latest/export/export.html).
+    See the examples in the [JAX export documentation](https://docs.jax.dev/en/latest/export/export.html).
     """
     return call_exported(self)(*args, **kwargs)
 
@@ -510,13 +517,13 @@ def default_export_platform() -> str:
   One of: `tpu`, `cpu`, `cuda`, `rocm`.
   """
   # Canonicalize to turn 'gpu' into 'cuda' or 'rocm'
-  return xb.canonicalize_platform(jax.default_backend())
+  return xb.canonicalize_platform(xb.default_backend())
 
 default_lowering_platform = default_export_platform
 
 def shape_and_dtype_jax_array(a) -> tuple[Sequence[int | None], DType]:
   """Returns the shape and dtype of a jax.Array or a j"""
-  if isinstance(a, jax.ShapeDtypeStruct):
+  if isinstance(a, api.ShapeDtypeStruct):
     return a.shape, a.dtype
   aval = core.get_aval(a)
   return aval.shape, aval.dtype
@@ -527,6 +534,7 @@ def export(
     *,
     platforms: Sequence[str] | None = None,
     disabled_checks: Sequence[DisabledSafetyCheck] = (),
+    _override_lowering_rules: Sequence[tuple[Any, Any]] | None = None
     ) -> Callable[..., Exported]:
   """Exports a JAX function for persistent serialization.
 
@@ -538,7 +546,14 @@ def export(
         the exported code takes an argument specifying the platform.
         If None, then use the default JAX backend.
         The calling convention for multiple platforms is explained at
-        https://jax.readthedocs.io/en/latest/export/export.html#module-calling-convention.
+        https://docs.jax.dev/en/latest/export/export.html#module-calling-convention.
+    _override_lowering_rules: an optional sequence of custom lowering rules
+        for some JAX primitives. Each element of the sequence is a pair
+        of a JAX primitive and a lowering function. Defining lowering rules
+        is an advanced feature using JAX internal APIs, which are subject
+        to change. Furthermore, the responsibility for the stability of the
+        MLIR emitted through these custom lowering rules, rests with the user
+        of these rules.
     disabled_checks: the safety checks to disable. See documentation for
         of `jax.export.DisabledSafetyCheck`.
 
@@ -566,7 +581,8 @@ def export(
       Array([0.09983342, 0.19866933, 0.29552022, 0.38941833], dtype=float32)
   """
   return _export_internal(fun_jit, platforms=platforms,
-                          disabled_checks=disabled_checks)
+                          disabled_checks=disabled_checks,
+                          override_lowering_rules=_override_lowering_rules)
 
 
 # TODO(necula): remove this once we improve the integration with jax2tf.
@@ -575,13 +591,14 @@ def _export_internal(
     *,
     platforms: Sequence[str] | None = None,
     disabled_checks: Sequence[DisabledSafetyCheck] = (),
-    _device_assignment_for_internal_jax2tf_use_only = None,
+    _device_assignment_for_internal_jax2tf_use_only=None,
+    override_lowering_rules=None,
     ) -> Callable[..., Exported]:
   """Exports native serialization for a JAX function.
 
   Note: this function exists only for internal usage by jax2tf. Use
     `jax.export` instead.
-    See https://jax.readthedocs.io/en/latest/export/export.html
+    See https://docs.jax.dev/en/latest/export/export.html
 
   See docstring of `export` for more details.
   """
@@ -602,6 +619,7 @@ def _export_internal(
     lowered = traced.lower(
         lowering_platforms=actual_lowering_platforms,
         _private_parameters=mlir.LoweringParameters(
+            override_lowering_rules=override_lowering_rules,
             for_export=True,
             export_ignore_forward_compatibility=config.export_ignore_forward_compatibility.value))
     return _export_lowered(
@@ -656,6 +674,12 @@ def _export_lowered(
     # For pmap
     module_kept_var_idx = tuple(range(len(args_avals_flat)))
   shape_poly_state = lowering.compile_args["shape_poly_state"]
+
+  # Make a copy of mlir module as we should not mutate it
+  # because it may be cached
+  context = mlir.make_ir_context()
+  with context, ir.Location.unknown(context):
+    mlir_module = ir.Module.parse(mlir.module_to_bytecode(mlir_module))
   if (not all(core.is_constant_shape(a.shape) for a in args_avals_flat)
       or lowering.compile_args.get("ordered_effects", [])):
     mlir_module = _wrap_main_func(
@@ -669,6 +693,12 @@ def _export_lowered(
     mlir_module_attrs["jax.uses_shape_polymorphism"] = (
         mlir.ir.BoolAttr.get(shape_poly_state.uses_dim_vars))
 
+  # Shardy was used during lowering if we can find the Shardy mesh in the
+  # module. Note that the mesh should have been lifted by the
+  # `sdy-lift-inlined-meshes` pass in mlir.py.
+  shardy_enabled = has_sdy_mesh(ir.SymbolTable(mlir_module.operation),
+                                mlir_module)
+
   mlir_module_serialized = _module_to_bytecode(mlir_module)
 
   # Figure out the result types and shapes
@@ -681,19 +711,19 @@ def _export_lowered(
     out_avals_flat = lowered.compile_args["out_avals"]  # type: ignore
 
   # Log and then check the module.
-  if logging.vlog_is_on(3):
-    logmsg = (f"fun_name={fun_name} version={version} "
-              f"lowering_platforms={lowering._platforms} "  # type: ignore[unused-ignore,attribute-error]
-              f"disabled_checks={disabled_checks}")
-    logging.info("Exported JAX function: %s\n", logmsg)
-    logging.info(mlir.dump_module_message(mlir_module, "export"))
-    logging.info(
-        "Size of mlir_module_serialized: %d byte",
-        len(mlir_module_serialized),
-    )
+  logmsg = (f"fun_name={fun_name} version={version} "
+            f"lowering_platforms={lowering._platforms} "  # type: ignore[unused-ignore,attribute-error]
+            f"disabled_checks={disabled_checks}")
+  logger.debug("Exported JAX function: %s\n", logmsg)
+  logger.debug(mlir.dump_module_message(mlir_module, "export"))
+  logger.debug(
+      "Size of mlir_module_serialized: %d byte",
+      len(mlir_module_serialized),
+  )
 
   _check_module(mlir_module,
-                disabled_checks=disabled_checks)
+                disabled_checks=disabled_checks,
+                shardy_enabled=shardy_enabled)
 
   ordered_effects = tuple(lowering.compile_args["ordered_effects"])
   unordered_effects = tuple(lowering.compile_args["unordered_effects"])
@@ -715,24 +745,51 @@ def _export_lowered(
     export_sharding(s, aval)
     for s, aval in zip(lowering.compile_args["out_shardings"], out_avals_flat))
 
-  device_assignment = lowering.compile_args["device_assignment"]
+  device_assignment = lowering._device_list  # type: ignore
   if _device_assignment_for_internal_jax2tf_use_only is not None:
     _device_assignment_for_internal_jax2tf_use_only[0] = device_assignment
+
+  cur_mesh = cur_arg = cur_k_path = None
+  # lowered.args_info is a tree of the args, but we need the out avals too to
+  # get the key paths for.
+  out_avals_tree = tree_util.tree_unflatten(lowered.out_tree, out_avals_flat)
+  if config.use_shardy_partitioner.value:
+    for sharding, (k_path, arg) in zip(
+        itertools.chain.from_iterable([
+            all_in_shardings, lowering.compile_args["out_shardings"]]),
+        itertools.chain.from_iterable([
+            tree_util.tree_flatten_with_path(lowered.args_info)[0],
+            tree_util.tree_flatten_with_path(out_avals_tree)[0]])):
+      if isinstance(sharding, sharding_impls.NamedSharding):
+        if cur_mesh is None:
+          cur_mesh, cur_arg, cur_k_path = sharding.mesh, arg, k_path
+        elif cur_mesh.shape_tuple != sharding.mesh.shape_tuple:
+          raise ValueError(
+              "Mesh for all inputs/outputs should be equal. Got one mesh "
+              f"{cur_mesh} on an array {cur_arg._aval} at "  # type: ignore[union-attr]
+              f"{shape_poly.args_kwargs_path_to_str(cur_k_path)} and another mesh: "  # type: ignore[arg-type]
+              f"{sharding.mesh}' on a tensor {arg._aval} at "
+              f"{shape_poly.args_kwargs_path_to_str(k_path)}")
+    if cur_mesh and isinstance(cur_mesh, mesh_lib.Mesh):
+      cur_mesh = cur_mesh.abstract_mesh
+
   def _get_exported_vjp(exp_primal: Exported) -> Exported:
     # Turn the primal jaxpr into a function, in preparation for exporting
     # the VJP. Note that jaxpr_as_fun produces a function with flat arguments
     assert(jaxpr is not None)  # None only when the lowered was created outside JAX
     fun_jax = core.jaxpr_as_fun(jaxpr)
 
-    fun_vjp_jax, vjp_in_avals = _get_vjp_fun(fun_jax,
-                                             in_tree=exp_primal.in_tree,
-                                             in_avals=exp_primal.in_avals,
-                                             in_shardings_hlo=exp_primal.in_shardings_hlo,
-                                             out_avals=exp_primal.out_avals,
-                                             out_shardings_hlo=exp_primal.out_shardings_hlo,
-                                             device_assignment=device_assignment,
-                                             apply_jit=True,
-                                             flat_primal_fun=True)
+    fun_vjp_jax, vjp_in_avals = _get_vjp_fun(
+        fun_jax,
+        in_tree=exp_primal.in_tree,
+        in_avals=exp_primal.in_avals,
+        in_shardings_hlo=exp_primal.in_shardings_hlo,
+        out_avals=exp_primal.out_avals,
+        out_shardings_hlo=exp_primal.out_shardings_hlo,
+        device_assignment=device_assignment,
+        apply_jit=True,
+        flat_primal_fun=True,
+        mesh=cur_mesh)  # type: ignore[arg-type]
     return export(fun_vjp_jax,  # type: ignore[arg-type]
                   platforms=exp_primal.platforms,
                   disabled_checks=exp_primal.disabled_safety_checks)(*vjp_in_avals)
@@ -781,8 +838,15 @@ def _module_to_bytecode(module: ir.Module) -> bytes:
   else:
     target_version = hlo.get_version_from_compatibility_requirement(
       hlo.StablehloCompatibilityRequirement.WEEK_4)
-  module_serialized = xla_client._xla.mlir.serialize_portable_artifact(  # type: ignore
-      mlir_str, target_version)
+  if jaxlib_extension_version >= 357 and ifrt_version >= 14:
+    module_serialized = xla_client._xla.mlir.serialize_portable_artifact(  # type: ignore
+        mlir_str, target_version, xb.get_backend().serialize_with_sdy)
+  else:
+    # pylint: disable=no-value-for-parameter
+    module_serialized = xla_client._xla.mlir.serialize_portable_artifact(  # type: ignore
+        mlir_str, target_version)
+    # pylint: disable=no-value-for-parameter
+
   return module_serialized
 
 
@@ -797,10 +861,10 @@ def _wrap_main_func(
 ) -> ir.Module:
   """Wraps the lowered module with a new "main" handling dimension arguments.
 
-  See calling convention documentation https://jax.readthedocs.io/en/latest/export/export.html#module-calling-convention.
+  See calling convention documentation https://docs.jax.dev/en/latest/export/export.html#module-calling-convention.
 
   Args:
-    module: the HLO module as obtained from lowering.
+    module: a copy of HLO module as obtained from lowering.
     args_avals_flat: the avals for all the arguments of the lowered function,
       which correspond to the array arguments of the `module`.
     args_kwargs_tree: the PyTreeDef corresponding to `(args, kwargs)`, for error
@@ -814,10 +878,9 @@ def _wrap_main_func(
   Returns the wrapped module, without dimension and token arguments.
   """
   dim_vars = shape_poly.all_dim_vars(args_avals_flat)
-  context = mlir.make_ir_context()
+  context = module.context
+  wrapped_module = module
   with context, ir.Location.unknown(context):
-    # Make a copy, do not mutate because it may be cached
-    wrapped_module = ir.Module.parse(mlir.module_to_bytecode(module))
     symbol_table = ir.SymbolTable(wrapped_module.operation)
     orig_main = symbol_table["main"]
     orig_main.attributes["sym_visibility"] = ir.StringAttr.get("private")
@@ -907,7 +970,8 @@ def _wrap_main_func(
           ))
       ctx = mlir.LoweringRuleContext(
         module_context=module_context,
-        name_stack=source_info_util.new_name_stack(), primitive=None,
+        name_stack=source_info_util.new_name_stack(), traceback=None,
+        primitive=None,
         avals_in=args_avals_flat, avals_out=None,
         tokens_in=mlir.TokenSet(), tokens_out=None)
       # We compute dim_values from the array arguments.
@@ -951,6 +1015,9 @@ def _wrap_main_func(
                                  orig_main_args)
       func_dialect.ReturnOp([call.results[idx] for idx in new_main_result_indices])
     symbol_table.set_symbol_name(new_main_op, "main")
+    pipeline = passmanager.PassManager.parse(
+        'builtin.module(symbol-dce)')
+    pipeline.run(wrapped_module.operation)
 
   return wrapped_module
 
@@ -1035,6 +1102,8 @@ _GPU_FFI_KERNELS = [
     "hipsolver_gesvd_ffi", "hipsolver_gesvdj_ffi",
     # tridiagonal on GPU
     "cusolver_sytrd_ffi",
+    # tridiagonal_solve on GPU
+    "cusparse_gtsv2_ffi",
 ]
 # These are the JAX custom call target names that are guaranteed to be stable.
 # Their backwards compatibility is tested by back_compat_test.py.
@@ -1042,33 +1111,21 @@ _CUSTOM_CALL_TARGETS_GUARANTEED_STABLE = {
     *_CPU_FFI_KERNELS,
     *_GPU_FFI_KERNELS,
     "Sharding", "SPMDFullToShardShape", "SPMDShardToFullShape",
-    "cu_threefry2x32", "cu_threefry2x32_ffi",
+    "annotate_device_placement",
+    "cu_threefry2x32_ffi",
     # Triton IR does not guarantee stability.
     # "__gpu$xla.gpu.triton",
-    # cholesky on CPU
-    "lapack_spotrf", "lapack_dpotrf", "lapack_cpotrf", "lapack_zpotrf",
     # eigh on TPU
     "Eigh",
-    # eig on CPU
-    "lapack_sgeev", "lapack_dgeev", "lapack_cgeev", "lapack_zgeev",
-    # svd on CPU
-    "lapack_sgesdd", "lapack_dgesdd", "lapack_cgesdd", "lapack_zgesdd",
     # qr and svd on TPU
     "Qr", "ProductOfElementaryHouseholderReflectors",
-    # triangular_solve on CPU
-    "blas_strsm", "blas_dtrsm", "blas_ctrsm", "blas_ztrsm",
-    # schur on CPU
-    "lapack_sgees", "lapack_dgees", "lapack_cgees", "lapack_zgees",
-    # tridiagonal on CPU
-    "lapack_ssytrd", "lapack_dsytrd", "lapack_chetrd", "lapack_zhetrd",
-    # hessenberg on CPU
-    "lapack_sgehrd", "lapack_dgehrd", "lapack_cgehrd", "lapack_zgehrd",
     # lu on TPU
     "LuDecomposition",
     # ApproxTopK on TPU
     "ApproxTopK", "stablehlo.dynamic_approx_top_k",
     "tf.call_tf_function",  # From jax2tf.call_tf(func, call_tf_graph=True)
     "tpu_custom_call",  # Pallas/TPU kernels
+    "mosaic_gpu",  # Pallas Mosaic GPU kernels
     # TODO(burmako): maintain backwards compatibility for these, until they
     # are upstreamed to StableHLO.
     # See https://github.com/openxla/stablehlo/issues/8.
@@ -1078,10 +1135,11 @@ _CUSTOM_CALL_TARGETS_GUARANTEED_STABLE = {
     "shape_assertion",  # Used by shape_poly to evaluate assertions
 }
 
-check_sharding_pattern = re.compile(r"^({replicated}|{unknown shard_as.*}|"")$")
+check_sharding_pattern = re.compile(r"^({replicated}|{unknown shard_as.*}|.*\[({}, )*{}\]"")$")
 
 def _check_module(mod: ir.Module, *,
-                  disabled_checks: Sequence[DisabledSafetyCheck]) -> bool:
+                  disabled_checks: Sequence[DisabledSafetyCheck],
+                  shardy_enabled: bool) -> bool:
   """Run a number of checks on the module.
 
   Args:
@@ -1103,13 +1161,15 @@ def _check_module(mod: ir.Module, *,
   module_uses_non_replicated_sharding = False
   def check_sharding(op: ir.Operation, loc: ir.Location):
     try:
-      sharding = op.attributes["mhlo.sharding"]
+      sharding = (op.attributes["sharding"] if shardy_enabled else
+                  op.attributes["mhlo.sharding"])
     except KeyError:
       pass
     else:
       nonlocal module_uses_non_replicated_sharding
       try:
-        sharding_value = ir.StringAttr(sharding).value
+        sharding_value = (str(sharding) if shardy_enabled else
+                          ir.StringAttr(sharding).value)
       except UnicodeDecodeError:
         # The mhlo.sharding attribute may be in pretty-printed format, or
         # as an encoding of an HloSharding protobuf in some rare situations.
@@ -1130,6 +1190,8 @@ def _check_module(mod: ir.Module, *,
         disallowed_custom_call_ops.append(f"{op} at {op.location}")
       if call_target_name_attr == sharding_attr:
         check_sharding(op, op.location)
+    elif op_name == "sdy.sharding_constraint":
+      check_sharding(op, op.location)
 
   def walk_operations(op):
     check_op(op)
@@ -1143,7 +1205,7 @@ def _check_module(mod: ir.Module, *,
     disallowed_custom_call_ops_str = "\n".join(disallowed_custom_call_ops)
     msg = ("Cannot serialize code with custom calls whose targets have no "
            "compatibility guarantees. "
-           "See https://jax.readthedocs.io/en/latest/export/export.html#compatibility-guarantees-for-custom-calls. "
+           "See https://docs.jax.dev/en/latest/export/export.html#compatibility-guarantees-for-custom-calls. "
            "Examples are:\n"
            f"{disallowed_custom_call_ops_str}.\n")
     raise ValueError(msg)
@@ -1163,32 +1225,38 @@ def expand_in_shardings(in_shardings: Sequence[LoweringSharding],
     all_in_shardings[idx] = in_s
   return tuple(all_in_shardings)
 
-def _hlo_sharding_to_xla_compatible_sharding(
-    hlo_sharding: HloSharding | None,
-    mesh: sharding.Mesh) -> sharding.Sharding | None:
-  if hlo_sharding is None:
-    return None
-  return sharding_impls._gspmd_to_named_sharding_via_mesh(
-      _hlo_sharding_to_gspmd_sharding(hlo_sharding, tuple(mesh.devices.flat)),  # type: ignore[arg-type]
-      mesh)
 
 def _hlo_sharding_to_gspmd_sharding(
     hlo_sharding: HloSharding | None,
-    device_assignment: Sequence[jax.Device]) -> sharding.GSPMDSharding | None:
+    device_assignment: Sequence[_jax.Device]
+    ) -> sharding_impls.GSPMDSharding | None:
   if hlo_sharding is None:
     return None
-  return sharding.GSPMDSharding(device_assignment, hlo_sharding)
+  return sharding_impls.GSPMDSharding(device_assignment, hlo_sharding)
 
-def _get_vjp_fun(primal_fun: Callable, *,
-                 in_tree: tree_util.PyTreeDef,
-                 in_avals: Sequence[core.AbstractValue],
-                 out_avals: Sequence[core.AbstractValue],
-                 in_shardings_hlo: tuple[HloSharding | None, ...],
-                 out_shardings_hlo: tuple[HloSharding | None, ...],
-                 device_assignment: Sequence[sharding_impls.Device] | None,
-                 apply_jit: bool,
-                 flat_primal_fun: bool = False,
-                 ) -> tuple[Callable, Sequence[core.AbstractValue]]:
+
+def _hlo_sharding_to_named_sharding(
+    hlo_sharding: HloSharding | None,
+    mesh: mesh_lib.Mesh | mesh_lib.AbstractMesh):
+  if hlo_sharding is None:
+    return None
+  return sharding_impls.cached_named_sharding(
+      mesh, sharding_impls.parse_flatten_op_sharding(hlo_sharding, mesh)[0])
+
+
+def _get_vjp_fun(
+    primal_fun: Callable,
+    *,
+    in_tree: tree_util.PyTreeDef,
+    in_avals: Sequence[core.AbstractValue],
+    out_avals: Sequence[core.AbstractValue],
+    in_shardings_hlo: tuple[HloSharding | None, ...],
+    out_shardings_hlo: tuple[HloSharding | None, ...],
+    device_assignment: Sequence[sharding_impls.Device] | None,
+    apply_jit: bool,
+    flat_primal_fun: bool = False,
+    mesh: mesh_lib.AbstractMesh | None = None,
+) -> tuple[Callable, Sequence[core.AbstractValue]]:
   # Since jax.vjp does not handle kwargs, it is easier to do all the work
   # here with flattened functions.
   # apply_jit=False is only used for backwards compatibility with the graph
@@ -1205,7 +1273,7 @@ def _get_vjp_fun(primal_fun: Callable, *,
 
     args_flat_jax, out_cts_flat_jax = util.split_list(args_and_out_cts_flat_jax,
                                                       [len(in_avals)])
-    _, pullback_jax = jax.vjp(primal_fun if flat_primal_fun else flattened_primal_fun_jax,
+    _, pullback_jax = api.vjp(primal_fun if flat_primal_fun else flattened_primal_fun_jax,
                               *args_flat_jax)
     return pullback_jax(out_cts_flat_jax)
 
@@ -1214,27 +1282,35 @@ def _get_vjp_fun(primal_fun: Callable, *,
                       map(lambda a: a.to_tangent_aval(), out_avals)))
 
   if apply_jit:
-    assert device_assignment is not None
-    vjp_in_shardings = tuple(
-        _hlo_sharding_to_gspmd_sharding(s, device_assignment)
-        for s in itertools.chain(in_shardings_hlo, out_shardings_hlo))
-    vjp_out_shardings = tuple(
-        _hlo_sharding_to_gspmd_sharding(s, device_assignment)
-        for s in in_shardings_hlo)
+    if mesh:
+      vjp_in_shardings = tuple(
+          _hlo_sharding_to_named_sharding(s, mesh)
+          for s in itertools.chain(in_shardings_hlo, out_shardings_hlo))
+      vjp_out_shardings = tuple(_hlo_sharding_to_named_sharding(s, mesh)
+                                for s in in_shardings_hlo)
+    else:
+      assert device_assignment is not None
+      vjp_in_shardings = tuple(
+          _hlo_sharding_to_gspmd_sharding(s, device_assignment)
+          for s in itertools.chain(in_shardings_hlo, out_shardings_hlo))
+      vjp_out_shardings = tuple(
+          _hlo_sharding_to_gspmd_sharding(s, device_assignment)
+          for s in in_shardings_hlo)
     return pjit.pjit(fun_vjp_jax,
                      in_shardings=vjp_in_shardings,
                      out_shardings=vjp_out_shardings), vjp_in_avals
   else:
     return fun_vjp_jax, vjp_in_avals
 
+
 ### Calling the exported function
 
-def call(exported: Exported) -> Callable[..., jax.Array]:
+def call(exported: Exported) -> Callable[..., typing.Array]:
   if not isinstance(exported, Exported):
     raise ValueError(
       "The exported argument must be an export.Exported. "
       f"Found {exported}.")
-  @jax.custom_vjp
+  @custom_derivatives.custom_vjp
   def f_flat(*args_flat):
     return call_exported_p.bind(*args_flat, exported=exported)
 
@@ -1290,11 +1366,12 @@ call_exported = call
 call_exported_p = core.Primitive("call_exported")
 call_exported_p.multiple_results = True
 
+
 @util.cache()
 def _call_exported_abstract_eval(
     *in_avals: core.AbstractValue,
     exported: Exported
-    ) -> tuple[tuple[core.AbstractValue, ...], set[effects.Effect]]:
+) -> tuple[tuple[core.AbstractValue, ...], set[effects.Effect]]:
   exported_dim_vars = shape_poly.all_dim_vars(exported.in_avals)
   assert len(in_avals) == len(exported.in_avals)  # since the pytrees have the same structure
   # Check that the expected shapes match the actual ones
@@ -1342,7 +1419,7 @@ def _call_exported_abstract_eval(
   # it would be ambiguous whether we should continue tracing with a result
   # of type `f32[c]` or `f32[d]`.
   shape_constraints.check_statically(synthetic_eval)
-  exported_dim_values = [synthetic_eval.evaluate(solution[var])
+  exported_dim_values = [synthetic_eval.evaluate(solution[var])  # type: ignore[arg-type]
                          for var in exported_dim_vars]
   out_avals = tuple(
       core.ShapedArray(core.evaluate_shape(out_aval.shape, exported_dim_vars,
@@ -1359,12 +1436,67 @@ def _call_exported_impl(*args, exported: Exported):
 
 call_exported_p.def_impl(_call_exported_impl)
 
+
+def get_mesh_from_symbol(symtab: ir.SymbolTable) -> mesh_lib.AbstractMesh:
+  if "mesh" not in symtab:
+    return mesh_lib.empty_abstract_mesh
+  mesh_attr = sdy.MeshAttr(symtab["mesh"].mesh)
+  axes = [sdy.MeshAxisAttr(a) for a in mesh_attr.axes]
+  if not axes:
+    return mesh_lib.empty_abstract_mesh
+  axes_sizes = tuple(a.size for a in axes)
+  axes_names = tuple(a.name for a in axes)
+  return mesh_lib.AbstractMesh(axes_sizes, axes_names)
+
+def has_sdy_meshes_in_frontend_attributes(submodule: ir.Module) -> bool:
+  if "mhlo.frontend_attributes" not in submodule.operation.attributes:
+    return False
+  frontend_attributes = submodule.operation.attributes[
+      "mhlo.frontend_attributes"
+  ]
+  return "xla.sdy.meshes" in frontend_attributes
+
+def has_sdy_mesh(symtab: ir.SymbolTable, submodule: ir.Module) -> bool:
+  for mesh_name in ("mesh", "empty_mesh", "maximal_mesh_0"):
+    if mesh_name in symtab:
+      return isinstance(symtab[mesh_name], sdy.MeshOp)
+  return has_sdy_meshes_in_frontend_attributes(submodule)
+
 def _call_exported_lowering(ctx: mlir.LoweringRuleContext, *args,
                             exported: Exported):
   if exported.uses_global_constants:
     ctx.module_context.shape_poly_state.uses_dim_vars = True
   submodule = ir.Module.parse(exported.mlir_module())
 
+  symtab = ir.SymbolTable(submodule.operation)
+  shardy_enabled = has_sdy_mesh(symtab, submodule)
+  if shardy_enabled:
+    if not config.use_shardy_partitioner.value:
+      raise ValueError(
+          "The function was exported with shardy enabled but you are calling "
+          "it with Shardy disabled. Please enable Shardy using "
+          "`--jax_use_shardy_partitioner=True`.")
+    # TODO(b/422690222): remove this pass once we don't need to support 6m
+    # old exported modules.
+    if has_sdy_meshes_in_frontend_attributes(submodule):
+      if jaxlib_extension_version < 356:
+        submodule = ir.Module.parse(
+            _jax.sdy.sdy_round_trip_import_shardings(
+                mlir.module_to_bytecode(submodule)))  # type: ignore[module-attr]
+      else:
+        with submodule.context:
+          pipeline = passmanager.PassManager.parse(
+              'builtin.module(xla-sdy-round-trip-import-shardy-attrs)')
+          pipeline.run(submodule.operation)
+
+  with submodule.context:
+    pipeline = passmanager.PassManager.parse(
+        'builtin.module(sdy-lift-inlined-meshes)')
+    pipeline.run(submodule.operation)
+  mesh = None
+  if shardy_enabled:
+    mesh = get_mesh_from_symbol(symtab)
+
   axis_context = ctx.module_context.axis_context
   if isinstance(axis_context, sharding_impls.ShardingContext):
     num_devices = axis_context.num_devices
@@ -1374,28 +1506,29 @@ def _call_exported_lowering(ctx: mlir.LoweringRuleContext, *args,
     num_devices = axis_context.axis_env.nreps
   else:
     raise NotImplementedError(type(axis_context))
-  if num_devices != exported.nr_devices:
-    # In some special cases we allow running with a different number of devices
-    # than the function was exported for.
-    err_msg = ""
-    if exported.nr_devices != 1:
-      err_msg = "the function was exported for more than 1 device."
-    elif (_check_module(submodule, disabled_checks=()) or
-          any(s is not None and not s.is_replicated()
-              for s in exported.in_shardings_hlo + exported.out_shardings_hlo)):
-      err_msg = "the function contains non-replicated sharding annotations."
-    if err_msg:
-      raise ValueError(
+  if num_devices != exported.nr_devices and exported.nr_devices != 1:
+    raise ValueError(
         f"Function {exported.fun_name} was exported for "
         f"{exported.nr_devices} devices and is called in a context with "
-        f"{num_devices} devices. This is disallowed because: {err_msg}"
-      )
+        f"{num_devices} devices, which is not allowed."
+    )
 
   # Apply in_shardings
-  args = tuple(
-    wrap_with_sharding(ctx, x, x_aval, x_sharding)
-    for x, x_aval, x_sharding in zip(args, ctx.avals_in, exported.in_shardings_hlo))
-  symtab = ir.SymbolTable(submodule.operation)
+  if mesh:
+    # A mesh only exists if Shardy is enabled.
+    args = tuple(
+        wrap_with_sharding(
+            ctx, x, x_aval,
+            _hlo_sharding_to_named_sharding(x_sharding, mesh), use_shardy=True)  # type: ignore[arg-type]
+        for x, x_aval, x_sharding in zip(args, ctx.avals_in, exported.in_shardings_hlo))
+  else:
+    # Since there is no mesh - either due to shardy being disabled or the loaded
+    # function being lowered for GSPMD (so no shardy mesh) - need to create a
+    # GSPMD sharding from the HLO sharding (can't use shardy lowering).
+    args = tuple(
+        wrap_with_sharding(ctx, x, x_aval, x_sharding, use_shardy=False)
+        for x, x_aval, x_sharding in zip(args, ctx.avals_in, exported.in_shardings_hlo))
+
   # The called function may have been exported with polymorphic shapes and called
   # now with more refined shapes. We insert hlo.ConvertOp to ensure the module
   # is valid.
@@ -1481,19 +1614,37 @@ def _call_exported_lowering(ctx: mlir.LoweringRuleContext, *args,
       for out, out_aval, refined_out_aval in zip(call.results[len(ordered_effects):],
                                                  exported.out_avals, ctx.avals_out))
   # Apply out_shardings
-  results = tuple(
-    wrap_with_sharding(ctx, x, x_aval, x_sharding)
-    for x, x_aval, x_sharding in zip(results, ctx.avals_out, exported.out_shardings_hlo)
-  )
+  if mesh:
+    # A mesh only exists if Shardy is enabled.
+    results = tuple(
+        wrap_with_sharding(
+            ctx, x, x_aval, _hlo_sharding_to_named_sharding(x_sharding, mesh),
+            use_shardy=True)  # type: ignore[arg-type]
+        for x, x_aval, x_sharding in zip(results, ctx.avals_out, exported.out_shardings_hlo))
+  else:
+    # Since there is no mesh - either due to shardy being disabled or the loaded
+    # function being lowered for GSPMD (so no shardy mesh) - need to create a
+    # GSPMD sharding from the HLO sharding (can't use shardy lowering).
+    results = tuple(
+        wrap_with_sharding(ctx, x, x_aval, x_sharding, use_shardy=False)
+        for x, x_aval, x_sharding in zip(results, ctx.avals_out, exported.out_shardings_hlo))
   return results
 
 mlir.register_lowering(call_exported_p, _call_exported_lowering)
 
-def wrap_with_sharding(ctx: mlir.LoweringRuleContext,
-                       x: ir.Value,
-                       x_aval: core.AbstractValue,
-                       x_sharding: HloSharding | None) -> ir.Value:
+
+def wrap_with_sharding(
+    ctx: mlir.LoweringRuleContext,
+    x: ir.Value,
+    x_aval: core.AbstractValue,
+    x_sharding: sharding_impls.NamedSharding | sharding_impls.GSPMDSharding | HloSharding | None,
+    use_shardy: bool,
+) -> ir.Value:
   if x_sharding is None:
     return x
-  return mlir.wrap_with_sharding_op(
-    ctx, x, x_aval, x_sharding.to_proto())
+  if use_shardy:
+    x_sharding = x_sharding._to_sdy_sharding(x_aval.ndim)  # type: ignore
+  else:
+    x_sharding = x_sharding.to_proto()  # type: ignore
+  return mlir.wrap_with_sharding_op(ctx, x, x_aval, x_sharding,  # type: ignore[arg-type]
+                                    allow_shardy_lowering=use_shardy)
