@@ -20,14 +20,13 @@ import functools
 import operator
 from typing import Any, Generic, TypeVar
 
-from jax import lax
-from jax.api_util import flatten_fun_nokwargs
+from jax._src import api_util
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
 from jax._src.interpreters import partial_eval as pe
-from jax.tree_util import (tree_flatten, tree_structure, tree_unflatten,
-                           treedef_tuple, tree_map, tree_leaves, PyTreeDef)
+from jax._src.tree_util import (tree_flatten, tree_structure, tree_unflatten,
+                                treedef_tuple, tree_map, tree_leaves, PyTreeDef)
 
 from jax._src import ad_util
 from jax._src import core
@@ -35,6 +34,7 @@ from jax._src import dispatch
 from jax._src import dtypes
 from jax._src import linear_util as lu
 from jax._src import source_info_util
+from jax._src.lax import lax
 from jax._src.state.types import (ReadEffect, AbstractRef, StateEffect)
 from jax._src.state import discharge as state_discharge
 from jax._src.state import primitives as state_primitives
@@ -67,15 +67,18 @@ discharge_state = state_discharge.discharge_state
 
 for_p = core.Primitive('for')
 for_p.multiple_results = True
+for_p.skip_canonicalization = True
 
 ### Tracing utilities
 
-def _trace_to_jaxpr_with_refs(f, state_tree: PyTreeDef,
-                              state_avals: Sequence[core.AbstractValue]
+def _trace_to_jaxpr_with_refs(f: Callable, state_tree: PyTreeDef,
+                              state_avals: Sequence[core.AbstractValue],
+                              debug_info: core.DebugInfo,
                               ) -> tuple[core.Jaxpr, list[Any], PyTreeDef]:
-  f, out_tree_thunk = flatten_fun_nokwargs(
-      lu.wrap_init(f), treedef_tuple((tree_structure(0), state_tree)))
-  jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(
+  f, out_tree_thunk = api_util.flatten_fun_nokwargs(
+      lu.wrap_init(f, debug_info=debug_info),
+      treedef_tuple((tree_structure(0), state_tree)))
+  jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(
       f, state_avals)
   return jaxpr, consts, out_tree_thunk()
 
@@ -129,12 +132,13 @@ def for_loop(nsteps: int | Sequence[int],
           rest_steps, functools.partial(body, i), vals, unroll=unroll)
       tree_map(lambda ref, val: ref_set(ref, (), val), refs, vals)
     return for_loop(outer_step, wrapped_body, init_state, unroll=unroll)
+  dbg = api_util.debug_info("for_loop", body, (0, init_state), {})
   nsteps, = nsteps
   flat_state, state_tree = tree_flatten(init_state)
   state_avals = map(state_utils.val_to_ref_aval, flat_state)
   idx_aval = core.ShapedArray((), dtypes.canonicalize_dtype(np.int64))
   jaxpr, consts, out_tree = _trace_to_jaxpr_with_refs(
-      body, state_tree, [idx_aval, *state_avals])
+      body, state_tree, [idx_aval, *state_avals], dbg)
   if out_tree != tree_structure(None):
     raise Exception("`body` should not return anything.")
   jaxpr = state_utils.hoist_consts_to_refs(jaxpr, index=1)
@@ -161,7 +165,7 @@ def scan(f: Callable[[Carry, X], tuple[Carry, Y]],
     raise TypeError("scan: f argument should be a callable.")
   if unroll < 1:
     raise ValueError("`unroll` must be a positive integer.")
-  xs_flat, xs_tree = tree_flatten(xs)
+  xs_flat, _ = tree_flatten(xs)
 
   try:
     lengths = [x.shape[0] for x in xs_flat]
@@ -192,26 +196,32 @@ def scan(f: Callable[[Carry, X], tuple[Carry, Y]],
   x_dtypes = [dtypes.canonicalize_dtype(x.dtype) for x in xs_flat]
   x_avals = tuple(map(core.ShapedArray, x_shapes, x_dtypes))
 
-  def _create_jaxpr(init):
-    init_flat = tree_leaves(init)
-    _, in_tree = tree_flatten((init, xs))
+  init_flat = tree_leaves(init)
+  _, in_tree = tree_flatten((init, xs))
+  dbg = api_util.debug_info("scan", f, (init, xs), {})
+  carry_avals = tuple(map(core.get_aval, init_flat))
+  jaxpr, consts, out_tree = _initial_style_jaxpr(
+      f, in_tree, carry_avals + x_avals, dbg)
+  del carry_avals
 
-    carry_avals = tuple(map(core.get_aval, init_flat))
-    jaxpr, _, out_tree = _initial_style_jaxpr(
-        f, in_tree, carry_avals + x_avals, "scan")
-    return jaxpr, out_tree
-  jaxpr, out_tree = _create_jaxpr(init)
   _, ys_avals = tree_unflatten(out_tree, jaxpr.out_avals)
   ys = tree_map(lambda aval: lax.full([length, *aval.shape], 0, aval.dtype),
                 ys_avals)
   def for_body(i, refs):
     carry_refs, xs_refs, ys_refs = refs
-    carry = tree_map(lambda x: x[()], carry_refs)
-    x = tree_map(lambda x: x[i], xs_refs)
-    carry, y = f(carry, x)
-    tree_map(lambda c_ref, c: ref_set(c_ref, (), c), carry_refs, carry)
-    tree_map(lambda y_ref, y: ref_set(y_ref, (i,), y), ys_refs, y)
+    flat_carry_refs = tree_leaves(carry_refs)
+    flat_xs_refs = tree_leaves(xs_refs)
+    flat_ys_refs = tree_leaves(ys_refs)
+    flat_outs = core.eval_jaxpr(jaxpr.jaxpr, jaxpr.consts,
+                                * consts,
+                                * map(lambda x: x[()], flat_carry_refs),
+                                * map(lambda x: x[i], flat_xs_refs))
+    flat_carry_out, flat_ys_out = split_list(flat_outs, [len(flat_carry_refs)])
+    map(lambda c_ref, c: ref_set(c_ref, (), c), flat_carry_refs, flat_carry_out)
+    map(lambda y_ref, y: ref_set(y_ref, (i,), y), flat_ys_refs, flat_ys_out)
+
   assert isinstance(length, int)
+  api_util.save_wrapped_fun_debug_info(for_body, dbg)
   init, _, ys = for_loop(length, for_body, (init, xs, ys), reverse=reverse,
                          unroll=unroll)
   return init, ys
@@ -267,7 +277,7 @@ def _for_impl_unrolled(body, nsteps, unroll, *args):
       state = body(i, state)
       i = i + 1
     return i, state
-  _, state = lax.while_loop(cond, while_body, (i, state))
+  _, state = loops.while_loop(cond, while_body, (i, state))
   return state
 
 mlir.register_lowering(for_p, mlir.lower_fun(_for_impl, multiple_results=True))
@@ -291,7 +301,8 @@ def _for_vmap(axis_data, args, dims, *,
     batched = map(operator.or_, batched, out_batched)
   else:
     raise Exception("Invalid fixpoint")
-  args = [batching.broadcast(x, axis_data.size, 0) if now_bat and not was_bat
+  args = [batching.broadcast(x, axis_data.size, 0, axis_data.explicit_mesh_axis)
+          if now_bat and not was_bat
           else batching.moveaxis(x, d, 0) if now_bat else x
           for x, d, was_bat, now_bat in zip(args, dims, init_batched, batched)]
   batched_jaxpr_, _ = batching.batch_jaxpr(
@@ -475,7 +486,7 @@ def _for_partial_eval(trace: pe.JaxprTrace, *tracers: pe.JaxprTracer,
   # Since not all inputs are used in jaxpr_unknown, we filter the input tracers
   # down using the output of `dce_jaxpr`.
   used_and_known = map(operator.and_, used_refs, map(operator.not_, in_unknowns))
-  tracers = [trace.instantiate_const(t) if u_and_k else t for t, u_and_k  # type: ignore
+  tracers = [trace.instantiate_const(t) if u_and_k else t for t, u_and_k
              in zip(tracers, used_and_known)]
   _, known_used = partition_list(used_refs, used_and_known)
   _, used_tracers = partition_list(used_refs, tracers)
@@ -492,7 +503,7 @@ def _for_partial_eval(trace: pe.JaxprTrace, *tracers: pe.JaxprTracer,
 
   assert len(unknown_inputs) == len(res_ref_unknown_outputs)
   assert len(unknown_inputs) == len(jaxpr_unknown.invars) - 1
-  eqn = pe.new_eqn_recipe(unknown_inputs, res_ref_unknown_outputs,
+  eqn = pe.new_eqn_recipe(trace, unknown_inputs, res_ref_unknown_outputs,
                           for_p, dict(jaxpr=jaxpr_unknown, nsteps=nsteps,
                                       reverse=reverse,
                                       which_linear=which_linear_unknown,
@@ -571,7 +582,6 @@ def _for_partial_eval_custom(saveable, in_unknowns, in_inst, eqn):
   newvar = core.gensym()
   resvars = map(newvar, res_avals)
 
-  @lu.wrap_init
   def known(*known_vals):
     empty_res = map(ad_util.zeros_like_aval, res_avals)
     jaxpr_known_args = [*known_vals, *empty_res]
@@ -579,12 +589,13 @@ def _for_partial_eval_custom(saveable, in_unknowns, in_inst, eqn):
     return for_p.bind(*jaxpr_known_args, jaxpr=jaxpr_known, nsteps=nsteps,
                       reverse=reverse, which_linear=jaxpr_known_which_linear,
                       unroll=unroll)
-  call_jaxpr_, _, call_jaxpr_consts, () = pe.trace_to_jaxpr_dynamic(
-      known, [v.aval for v in known_invars])
+  call_jaxpr_, _, call_jaxpr_consts = pe.trace_to_jaxpr_dynamic(
+      lu.wrap_init(known, debug_info=jaxpr.debug_info),
+      [v.aval for v in known_invars])
   call_jaxpr = core.ClosedJaxpr(call_jaxpr_, call_jaxpr_consts)
   eqn_known = pe.new_jaxpr_eqn(known_invars, [*known_outvars, *resvars],
                                core.closed_call_p, dict(call_jaxpr=call_jaxpr),
-                               call_jaxpr.effects, eqn.source_info)
+                               call_jaxpr.effects, eqn.source_info, eqn.ctx)
 
   jaxpr_staged = _convert_inputs_to_reads(nsteps, len(res_avals),
                                           jaxpr_staged_resin_,
@@ -595,20 +606,20 @@ def _for_partial_eval_custom(saveable, in_unknowns, in_inst, eqn):
                                    which_linear=which_linear_unknown,
                                    unroll=unroll)
 
-  @lu.wrap_init
   def staged(*res_and_refs):
     out_flat = for_p.bind(*res_and_refs, **params_staged)
     _, ans = split_list(out_flat, [num_res])
     _, ans = partition_list(out_inst, ans)
     return ans
-  call_jaxpr_, _, call_jaxpr_consts, () = pe.trace_to_jaxpr_dynamic(
-      staged, [v.aval for v in [*resvars, *eqn.invars]])
+  call_jaxpr_, _, call_jaxpr_consts = pe.trace_to_jaxpr_dynamic(
+      lu.wrap_init(staged, debug_info=jaxpr_staged.debug_info),
+      [v.aval for v in [*resvars, *eqn.invars]])
   assert len(jaxpr_staged.invars) - 1 == len(call_jaxpr_.invars)
   call_jaxpr = core.ClosedJaxpr(call_jaxpr_, call_jaxpr_consts)
   _, outvars = partition_list(out_inst, eqn.outvars)
   eqn_staged = pe.new_jaxpr_eqn([*resvars, *eqn.invars], outvars,
                                core.closed_call_p, dict(call_jaxpr=call_jaxpr),
-                               call_jaxpr.effects, eqn.source_info)
+                               call_jaxpr.effects, eqn.source_info, eqn.ctx)
   new_vars = [*new_inst, *resvars]
   return eqn_known, eqn_staged, in_unknowns, out_inst, new_vars
 
@@ -620,7 +631,7 @@ def _convert_outputs_to_writes(
   assert not jaxpr.constvars, "Jaxpr shouldn't have constvars."
 
   in_avals = [v.aval for v in jaxpr.invars]  # [i, *orig_ref_avals]
-  @lu.wrap_init
+
   def eval_jaxpr(i, *refs):
     # We split the refs into the original input refs and the dummy residual
     # refs.
@@ -639,8 +650,9 @@ def _convert_outputs_to_writes(
       AbstractRef(core.ShapedArray((nsteps, *v.aval.shape),  # pytype: disable=attribute-error
                   v.aval.dtype))  # pytype: disable=attribute-error
       for v, loop_invar in zip(jaxpr.outvars, loop_invar_res)]
-  jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(
-      eval_jaxpr, [*in_avals, *res_ref_avals])
+  jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(
+      lu.wrap_init(eval_jaxpr, debug_info=jaxpr.debug_info),
+      [*in_avals, *res_ref_avals])
   assert not consts
   return jaxpr, [core.ShapedArray(a.shape, a.dtype) for a in res_ref_avals]  # pytype: disable=attribute-error
 
@@ -649,7 +661,6 @@ def _convert_inputs_to_reads(
     loop_invar_res: Sequence[bool]) -> core.Jaxpr:
   assert not jaxpr.constvars, "Jaxpr should not have constvars"
 
-  @lu.wrap_init
   def eval_jaxpr(i, *refs):
     residual_refs, orig_refs = split_list(refs, [num_res])
     residual_vals = [r[()] if loop_invar else r[i] for r, loop_invar
@@ -665,8 +676,9 @@ def _convert_inputs_to_reads(
                   aval.dtype))  # pytype: disable=attribute-error
       for aval, loop_invar in zip(res_val_avals, loop_invar_res)]
 
-  jaxpr, _, (), () = pe.trace_to_jaxpr_dynamic(
-      eval_jaxpr, [i_aval, *res_ref_avals, *orig_ref_avals])
+  jaxpr, _, () = pe.trace_to_jaxpr_dynamic(
+      lu.wrap_init(eval_jaxpr, debug_info=jaxpr.debug_info),
+      [i_aval, *res_ref_avals, *orig_ref_avals])
   return jaxpr
 
 def transpose_jaxpr(jaxpr: core.Jaxpr, which_linear: list[bool]) -> core.Jaxpr:
@@ -691,8 +703,9 @@ def transpose_jaxpr(jaxpr: core.Jaxpr, which_linear: list[bool]) -> core.Jaxpr:
     ct_args = [x for x, u in zip(args, used_ct) if u]
     ad.backward_pass(tangent_jaxpr, False, (), (*primals_args, *ct_args), ())
     return []
-  jaxpr_trans, _, _, () = pe.trace_to_jaxpr_dynamic(
-      lu.wrap_init(trans), [v.aval for v in jaxpr.invars])
+  jaxpr_trans, _, _ = pe.trace_to_jaxpr_dynamic(
+      lu.wrap_init(trans, debug_info=jaxpr.debug_info),
+      [v.aval for v in jaxpr.invars])
   return jaxpr_trans
 
 def _for_transpose(in_cts, *args, jaxpr, nsteps, reverse, which_linear, unroll):
@@ -745,8 +758,9 @@ def discharged_for_loop(nsteps, body, init_state, *, reverse: bool = False):
   flat_state, state_tree = tree_flatten(init_state)
   state_avals = map(state_utils.val_to_ref_aval, flat_state)
   idx_aval = core.ShapedArray((), dtypes.canonicalize_dtype(np.int64))
+  debug = api_util.debug_info("discharged_for_loop", body, (0, init_state), {})
   jaxpr, consts, out_tree = _trace_to_jaxpr_with_refs(
-      body, state_tree, [idx_aval, *state_avals])
+      body, state_tree, [idx_aval, *state_avals], debug)
   if out_tree != tree_structure(None):
     raise Exception("`body` should not return anything.")
   discharged_jaxpr, discharged_consts = discharge_state(jaxpr, consts)
