@@ -19,7 +19,8 @@ For example,
 
    from jax._src import linear_util as lu
 
-   wf = lu.wrap_init(f)  # Produce a WrappedFun for applying transformations on `f`
+   # Produce a WrappedFun for applying transformations on `f`
+   wf = lu.wrap_init(f, debug_info=api_util.debug_info("test", f, (), {}))
 
 A `WrappedFun` object represents a function `f`, together with a sequence of
 nested transformations that are to be applied to the positional and keyword
@@ -63,15 +64,20 @@ data must be immutable, because it will be stored in function memoization tables
 """
 from __future__ import annotations
 
-from collections.abc import Callable
+from collections.abc import Callable, Sequence
 from functools import partial
+import re
+import time
 from typing import Any, NamedTuple
+from collections.abc import Hashable
+import warnings
 import weakref
 
 from jax._src import config
 from jax._src import core
 from jax._src import traceback_util
-from jax._src.util import curry, cache_clearing_funs
+from jax._src.tree_util import KeyPath, generate_key_paths, keystr
+from jax._src.util import HashableFunction, curry, fun_name, register_cache
 
 
 traceback_util.register_exclusion(__file__)
@@ -141,22 +147,36 @@ class WrappedFun:
 
   Args:
     f: the function to be transformed.
-    transforms: a list of `(gen, gen_static_args)` tuples representing
+    f_transformed: transformed function.
+    transforms: a tuple of `(gen, gen_static_args)` tuples representing
       transformations to apply to `f.` Here `gen` is a generator function and
       `gen_static_args` is a tuple of static arguments for the generator. See
       description at the start of this module for the expected behavior of the
       generator.
     stores: a list of out_store for the auxiliary output of the `transforms`.
-    params: extra parameters to pass as keyword arguments to `f`, along with the
-      transformed keyword arguments.
+    params: a tuple of `(name, param)` tuples representing extra parameters to
+      pass as keyword arguments to `f`, along with the transformed keyword
+      arguments.
+    in_type: optional input type
+    debug_info: debugging info about the function being wrapped.
   """
   __slots__ = ("f", "f_transformed", "transforms", "stores", "params", "in_type", "debug_info")
 
+  f: Callable
+  f_transformed: Callable
+  transforms: tuple[tuple[Callable, tuple[Hashable, ...]], ...]
+  stores: tuple[Store | EqualStore | None, ...]
+  params: tuple[tuple[str, Any], ...]
+  in_type: core.InputType | None
+  debug_info: DebugInfo
+
   def __init__(self, f: Callable,
                f_transformed: Callable,
-               transforms,
-               stores: tuple[Store | EqualStore | None, ...], params, in_type,
-               debug_info: TracingDebugInfo | None):
+               transforms: tuple[tuple[Callable, tuple[Hashable, ...]], ...],
+               stores: tuple[Store | EqualStore | None, ...],
+               params: tuple[tuple[str, Hashable], ...],
+               in_type: core.InputType | None,
+               debug_info: DebugInfo):
     self.f = f
     self.f_transformed = f_transformed
     self.transforms = transforms
@@ -167,7 +187,7 @@ class WrappedFun:
 
   @property
   def __name__(self):
-    return getattr(self.f, '__name__', '<unnamed wrapped function>')
+    return fun_name(self.f, "<unnamed wrapped function>")
 
   def wrap(self, gen, gen_static_args,
            out_store: Store | EqualStore | None) -> WrappedFun:
@@ -175,11 +195,11 @@ class WrappedFun:
     if out_store is None:
       return WrappedFun(self.f, partial(gen, self.f_transformed, *gen_static_args),
                         ((gen, gen_static_args),) + self.transforms,
-                        (out_store,) + self.stores, self.params, None, None)
+                        (out_store,) + self.stores, self.params, None, self.debug_info)
     else:
       return WrappedFun(self.f, partial(gen, self.f_transformed, out_store, *gen_static_args),
                         ((gen, gen_static_args),) + self.transforms,
-                        (out_store,) + self.stores, self.params, None, None)
+                        (out_store,) + self.stores, self.params, None, self.debug_info)
 
   def populate_stores(self, stores):
     """Copy the values from the `stores` into `self.stores`."""
@@ -247,38 +267,143 @@ def transformation_with_aux2(
   out_thunk = lambda: out_store.val
   return fun.wrap(gen, gen_static_args, out_store), out_thunk
 
-def fun_name(f):
-  try:
-    return f.__name__
-  except:
-    return str(f)
 
-class TracingDebugInfo(NamedTuple):
-  # Packages up trace/staging-time debug info about a func and its parameters,
-  # formed just before staging to a jaxpr and read in trace-time error messages.
+class DebugInfo(NamedTuple):
+  """Debugging info about a func, its arguments, and results."""
   traced_for: str             # e.g. 'jit', 'scan', etc
-  func_src_info: str | None   # e.g. f'{fun.__name__} at {filename}:{lineno}'
-  arg_names: tuple[str | None, ...]  # e.g. ('args[0]', ... )
-  # e.g. ('[0]', '[1]', ...)
-  result_paths_thunk: Callable[[], tuple[str, ...]] | None
 
-  @classmethod
-  def from_jaxpr(cls, jaxpr: core.ClosedJaxpr) -> TracingDebugInfo | None:
-    jaxpr_dbg = jaxpr.jaxpr._debug_info
-    if jaxpr_dbg is None: return None
-    return TracingDebugInfo(jaxpr_dbg.traced_for,
-                            jaxpr_dbg.func_src_info,
-                            jaxpr_dbg.arg_names,
-                            lambda: jaxpr_dbg.result_paths)
+  func_src_info: str
+  """e.g. f'{fun.__name__} at {filename}:{lineno}' or {fun.__name__} if we have
+  no source location information. The first word is always the function name,
+  which may be '<unknown>'.
+  """
+
+  arg_names: tuple[str, ...]
+  """The paths of the flattened non-static argnames,
+  e.g. `('x', 'dict_arg["a"]', ... )`.
+  Uses the empty string for the args that do not correspond to
+  user-named arguments, e.g., tangent args in `jax.jvp`, or for arguments that
+  we are not yet tracking properly.
+  At the moment, `arg_names` accuracy is best-effort.
+  Use `safe_arg_names` to detect and handle an unexpected
+  number of elements in `arg_names`.
+  """
+
+  result_paths: tuple[str, ...] | Callable[[], tuple[str, ...]] | None
+  """The paths to the flattened results, e.g., `('result[0]', result[1])` for a
+  function that returns a tuple of arrays, or `(result,)` for a function that
+  returns a single array.
+  The result paths are not available while we are tracing the function,
+  instead we keep a thunk. It is possible for the result paths to be `None`
+  only when we first create a `DebugInfo`, before we put it in `lu.WrappedFun`
+  and before we start tracing.
+  Inside a `lu.WrappedFun` it can be only a thunk or a tuple of strings.
+  Once we are done tracing, we use
+  `self.resolve_result_paths()` to execute the thunk and replace the
+  actual result paths.
+  At the moment, `result_paths` accuracy is best-effort.
+  Use `safe_result_paths` to detect and handle an unexpected
+  number of elements in `result_paths`.
+  """
+
+  def resolve_result_paths(self) -> DebugInfo:
+    """Return a debug info with resolved result paths."""
+    assert self.result_paths is not None
+    if callable(self.result_paths):
+      return self._replace(result_paths=tuple(self.result_paths()))
+    return self
+
+  @property
+  def func_name(self) -> str:
+    return self.func_src_info.split(" ")[0]
+
+  def replace_func_name(self, name: str) -> DebugInfo:
+    func_src_comps = self.func_src_info.split(" ")
+    func_src_comps[0] = name
+    return self._replace(func_src_info=" ".join(func_src_comps))
+
+  @property
+  def func_filename(self) -> str | None:
+    m = _re_func_src_info.match(self.func_src_info)
+    if not m: return None
+    return m.group(3)
+
+  @property
+  def func_lineno(self) -> int | None:
+    m = _re_func_src_info.match(self.func_src_info)
+    if not m or m.group(4) is None: return None
+    return int(m.group(4))
+
+  def safe_arg_names(self, expected: int) -> tuple[str, ...]:
+    """Get the arg_names with a safety check."""
+    if len(self.arg_names) == expected:
+      return self.arg_names
+    else:
+      # TODO(necula): this should not happen
+      return ("",) * expected
+
+  def filter_arg_names(self, keep: Sequence[bool]) -> tuple[str, ...]:
+    """Keep only the arg_names for which `keep` is True."""
+    return tuple(v for v, b in zip(self.safe_arg_names(len(keep)), keep) if b)
+
+  def safe_result_paths(self, expected: int) -> tuple[str, ...]:
+    """Get the result paths with a safety check."""
+    assert self.result_paths is not None and not callable(self.result_paths), self
+    if self.result_paths is not None and len(self.result_paths) == expected:
+      return self.result_paths
+    else:
+      # TODO(necula): this should not happen
+      return ("",) * expected
+
+  def filter_result_paths(self, keep: Sequence[bool]) -> tuple[str, ...]:
+    """Keep only the result_paths for which `keep` is True."""
+    assert self.result_paths is not None and not callable(self.result_paths), self
+    return tuple(v for v, b in zip(self.safe_result_paths(len(keep)), keep) if b)
+
+_re_func_src_info = re.compile(r"([^ ]+)( at (.+):(\d+))?$")
+
+def _missing_debug_info(for_what: str) -> DebugInfo:
+  warnings.warn(
+      f"{for_what} is missing a DebugInfo object. "
+      "This behavior is deprecated, use api_util.debug_info() to "
+      "construct a proper DebugInfo object and propagate it to this function. "
+      "See https://github.com/jax-ml/jax/issues/26480 for more details.",
+      DeprecationWarning, stacklevel=2)
+  return DebugInfo("missing_debug_info", "<missing_debug_info>", (), ())
 
 def wrap_init(f: Callable, params=None, *,
-              debug_info: TracingDebugInfo | None = None) -> WrappedFun:
+              debug_info: DebugInfo) -> WrappedFun:
   """Wraps function `f` as a `WrappedFun`, suitable for transformation."""
   params_dict = {} if params is None else params
   params = () if params is None else tuple(sorted(params.items()))
-  return WrappedFun(f, partial(f, **params_dict), (), (), params, None, debug_info)
+  fun = WrappedFun(f, partial(f, **params_dict), (), (), params, None, debug_info)
+  if debug_info.result_paths is None:
+    fun, result_paths_thunk = _get_result_paths_thunk(fun)
+    debug_info = debug_info._replace(
+        result_paths=HashableFunction(result_paths_thunk, closure=()))
+  fun = WrappedFun(fun.f, fun.f_transformed, fun.transforms, fun.stores,
+                   fun.params, fun.in_type, debug_info)
+  return fun
 
 
+# We replace <flat index 0> with 0
+_re_clean_keystr_arg_names = re.compile(r"<flat index ([^>]+)>")
+def _clean_keystr_arg_names(k: KeyPath) -> str:
+  res = keystr(k)
+  return _re_clean_keystr_arg_names.sub(r"\1", res)
+
+@transformation_with_aux2
+def _get_result_paths_thunk(_fun: Callable, _store: Store, *args, **kwargs):
+  ans = _fun(*args, **kwargs)
+  result_paths = tuple(f"result{_clean_keystr_arg_names(path)}" for path, _ in generate_key_paths(ans))
+  if _store:
+    # In some instances a lu.WrappedFun is called multiple times, e.g.,
+    # the bwd function in a custom_vjp
+    assert _store.val == result_paths, (_store, result_paths)
+  else:
+    _store.store(result_paths)
+  return ans
+
 def annotate(f: WrappedFun, in_type: core.InputType | None) -> WrappedFun:
   assert f.in_type is None
   if in_type is None:
@@ -315,16 +440,9 @@ def _check_input_type(in_type: core.InputType) -> None:
           provided[d.val] = True
   assert all(provided)
 
-def add_debug_info(f: WrappedFun, debug_info: TracingDebugInfo | None
-                   ) -> WrappedFun:
-  """Produce a new WrappedFun with debug_info attached."""
-  assert f.debug_info is None
-  if debug_info is None:
-    return f
-  return WrappedFun(f.f, f.f_transformed, f.transforms, f.stores, f.params, f.in_type, debug_info)
 
-
-def cache(call: Callable, *, explain: Callable | None = None):
+def cache(call: Callable, *,
+          explain: Callable[[WrappedFun, bool, dict, tuple, float], None] | None = None):
   """Memoization decorator for functions taking a WrappedFun as first argument.
 
   Args:
@@ -332,6 +450,10 @@ def cache(call: Callable, *, explain: Callable | None = None):
       underlying transforms and params on the WrappedFun are used as part of the
       memoization cache key.
 
+    explain: a function that is invoked upon cache misses to log an explanation
+      of the miss.
+      Invoked with `(fun, is_cache_first_use, cache, key, elapsed_sec)`.
+
   Returns:
      A memoized version of ``call``.
   """
@@ -339,16 +461,17 @@ def cache(call: Callable, *, explain: Callable | None = None):
 
   def memoized_fun(fun: WrappedFun, *args):
     cache = fun_caches.setdefault(fun.f, new_cache := {})  # type: ignore
-    key = (fun.transforms, fun.params, fun.in_type, args, config.enable_x64.value,
-           config.default_device.value, config.trace_context())
+    key = (fun.transforms, fun.params, fun.in_type, args, config.trace_context())
     result = cache.get(key, None)
     if result is not None:
       ans, stores = result
       fun.populate_stores(stores)
     else:
+      if do_explain := explain and config.explain_cache_misses.value:
+        start = time.time()
       ans = call(fun, *args)
-      if explain and config.explain_cache_misses.value:
-        explain(fun.f, cache is new_cache, cache, key)
+      if do_explain:
+        explain(fun, cache is new_cache, cache, key, time.time() - start)  # type: ignore
       cache[key] = (ans, fun.stores)
 
     return ans
@@ -358,7 +481,7 @@ def cache(call: Callable, *, explain: Callable | None = None):
 
   memoized_fun.cache_clear = fun_caches.clear  # type: ignore
   memoized_fun.evict_function = _evict_function  # type: ignore
-  cache_clearing_funs.add(memoized_fun.cache_clear)
+  register_cache(memoized_fun, str(call))
   return memoized_fun
 
 @transformation2
