@@ -28,9 +28,10 @@ import numpy as np
 
 from jax._src import config as jax_config
 from jax._src import xla_bridge as xb
-from jax._src import util
+from jax._src.util import safe_zip, cache, tuple_delete
 from jax._src.lib import xla_client as xc
 
+zip, unsafe_zip = safe_zip, zip
 
 MeshAxisName = Any
 ResourceAxisName = Hashable
@@ -73,7 +74,7 @@ class ResourceEnv(NamedTuple):
     return f"ResourceEnv(mesh=Mesh({mesh_repr}))"
 
 
-@util.cache(max_size=128, trace_context_in_key=False)
+@cache(max_size=128, trace_context_in_key=False)
 def _get_local_mesh(global_mesh: Mesh, process_index: int) -> Mesh:
   if global_mesh.empty:
     return global_mesh
@@ -82,7 +83,7 @@ def _get_local_mesh(global_mesh: Mesh, process_index: int) -> Mesh:
   subcube_indices = []
   # We take the smallest slice of each dimension that doesn't skip any local device.
   for axis in range(global_mesh.devices.ndim):
-    other_axes = util.tuple_delete(tuple(range(global_mesh.devices.ndim)), axis)
+    other_axes = tuple_delete(tuple(range(global_mesh.devices.ndim)), axis)
     # NOTE: This re-reduces over many axes multiple times, so we could definitely
     #       optimize it, but I hope it won't be a bottleneck anytime soon.
     local_slices = is_local_device.any(other_axes, keepdims=False)
@@ -102,42 +103,128 @@ def _get_local_mesh(global_mesh: Mesh, process_index: int) -> Mesh:
   return Mesh(global_mesh.devices[subcube_indices_tuple], global_mesh.axis_names)
 
 
-class AxisTypes(enum.Enum):
-  Hidden = enum.auto()
-  Visible = enum.auto()
-  Collective = enum.auto()
+class AxisType(enum.Enum):
+  Auto = enum.auto()
+  Explicit = enum.auto()
+  Manual = enum.auto()
 
   def __repr__(self):
     return self.name
 
-def axis_names_to_types(axis_types) -> dict[str, AxisTypes]:
-  return {n: t for t, names in axis_types.items()
-          for n in ((names,) if not isinstance(names, tuple) else names)}
+def _normalize_axis_types(axis_names, axis_types, name):
+  axis_types = ((AxisType.Auto,) * len(axis_names)
+                if axis_types is None else axis_types)
+  if not isinstance(axis_types, tuple):
+    axis_types = (axis_types,)
 
-def axis_types_to_names(name_to_type: dict[str, AxisTypes]):
-  d = collections.defaultdict(list)
-  for n, t in name_to_type.items():
-    d[t].append(n)
-  return {t: ns[0] if len(ns) == 1 else tuple(ns) for t, ns in d.items()}
+  if not all(isinstance(a, AxisType) for a in axis_types):
+    raise TypeError(
+        f"axis_types passed to {name} must be of type `jax.sharding.AxisType`."
+        f" Got {axis_types} of type {tuple(type(a) for a in axis_types)}")
+  if len(axis_names) != len(axis_types):
+    raise ValueError(
+        "Number of axis names should match the number of axis_types. Got"
+        f" axis_names={axis_names} and axis_types={axis_types}")
+  return axis_types
 
+def all_axis_types_match(axis_types, ty: AxisType) -> bool:
+  if not axis_types:
+    return False
+  return all(t == ty for t in axis_types)
+
+def any_axis_types_match(axis_types, ty: AxisType) -> bool:
+  if not axis_types:
+    return False
+  return any(t == ty for t in axis_types)
+
+
+class BaseMesh:
+  axis_names: tuple[MeshAxisName, ...]
+  shape_tuple: tuple[tuple[str, int], ...]
+  _axis_types: tuple[AxisType, ...]
+
+  @property
+  def axis_types(self) -> tuple[AxisType, ...]:
+    return self._axis_types
+
+  @functools.cached_property
+  def _are_all_axes_manual(self) -> bool:
+    return all_axis_types_match(self._axis_types, AxisType.Manual)
+
+  @functools.cached_property
+  def _are_all_axes_auto(self) -> bool:
+    return all_axis_types_match(self._axis_types, AxisType.Auto)
+
+  @functools.cached_property
+  def _are_all_axes_explicit(self) -> bool:
+    return all_axis_types_match(self._axis_types, AxisType.Explicit)
+
+  @functools.cached_property
+  def _are_all_axes_auto_or_manual(self) -> bool:
+    if not self._axis_types:
+      return False
+    return all(t == AxisType.Auto or t == AxisType.Manual
+               for t in self._axis_types)
+
+  @functools.cached_property
+  def _any_axis_manual(self) -> bool:
+    return any_axis_types_match(self._axis_types, AxisType.Manual)
+
+  @functools.cached_property
+  def _any_axis_auto(self) -> bool:
+    return any_axis_types_match(self._axis_types, AxisType.Auto)
+
+  @functools.cached_property
+  def _any_axis_explicit(self) -> bool:
+    return any_axis_types_match(self._axis_types, AxisType.Explicit)
+
+  @functools.cached_property
+  def _any_axis_auto_or_manual(self) -> bool:
+    if not self._axis_types:
+      return False
+    return any(t == AxisType.Auto or t == AxisType.Manual
+               for t in self._axis_types)
+
+  @functools.cached_property
+  def auto_axes(self):
+    return tuple(n for n, t in safe_zip(self.axis_names, self._axis_types)
+                 if t == AxisType.Auto)
+
+  @functools.cached_property
+  def explicit_axes(self):
+    return tuple(n for n, t in safe_zip(self.axis_names, self._axis_types)
+                 if t == AxisType.Explicit)
+
+  @functools.cached_property
+  def manual_axes(self):
+    return tuple(n for n, t in safe_zip(self.axis_names, self._axis_types)
+                 if t == AxisType.Manual)
+
+  @functools.cached_property
+  def _axis_types_dict(self):
+    if not self.axis_names:
+      return {}
+    d = collections.defaultdict(list)
+    for n, t in safe_zip(self.axis_names, self._axis_types):
+      d[t].append(n)
+    return {t: tuple(n) for t, n in d.items()}
+
+  @functools.cached_property
+  def _name_to_type(self):
+    return dict(safe_zip(self.axis_names, self._axis_types))
+
+
+def _unpicke_mesh(devices, axis_names, axis_types):
+  return Mesh(devices, axis_names, axis_types)
 
 _mesh_object_dict = {}  # type: ignore
 
-MeshAxisType = dict[AxisTypes, str | tuple[str, ...]]
 
-class Mesh(contextlib.ContextDecorator):
+class Mesh(BaseMesh, contextlib.ContextDecorator):
   """Declare the hardware resources available in the scope of this manager.
 
-  In particular, all ``axis_names`` become valid resource names inside the
-  managed block and can be used e.g. in the ``in_axis_resources`` argument of
-  :py:func:`jax.experimental.pjit.pjit`. Also see JAX's multi-process programming
-  model (https://jax.readthedocs.io/en/latest/multi_process.html)
-  and the Distributed arrays and automatic parallelization tutorial
-  (https://jax.readthedocs.io/en/latest/notebooks/Distributed_arrays_and_automatic_parallelization.html)
-
-  If you are compiling in multiple threads, make sure that the
-  ``with Mesh`` context manager is inside the function that the threads will
-  execute.
+  See `Distributed arrays and automatic parallelization`_ and
+  `Explicit Sharding`_ tutorials.
 
   Args:
     devices: A NumPy ndarray object containing JAX device objects (as
@@ -145,44 +232,33 @@ class Mesh(contextlib.ContextDecorator):
     axis_names: A sequence of resource axis names to be assigned to the
       dimensions of the ``devices`` argument. Its length should match the
       rank of ``devices``.
+    axis_types: and optional tuple of :class:`jax.sharding.AxisType` entries corresponding to
+      the ``axis_names``. See `Explicit Sharding`_ for more information.
 
   Examples:
 
-    >>> from jax.experimental.pjit import pjit
     >>> from jax.sharding import Mesh
-    >>> from jax.sharding import PartitionSpec as P
+    >>> from jax.sharding import PartitionSpec as P, NamedSharding
     >>> import numpy as np
     ...
-    >>> inp = np.arange(16).reshape((8, 2))
-    >>> devices = np.array(jax.devices()).reshape(4, 2)
-    ...
     >>> # Declare a 2D mesh with axes `x` and `y`.
-    >>> global_mesh = Mesh(devices, ('x', 'y'))
-    >>> # Use the mesh object directly as a context manager.
-    >>> with global_mesh:
-    ...   out = pjit(lambda x: x, in_shardings=None, out_shardings=None)(inp)
+    >>> devices = np.array(jax.devices()).reshape(4, 2)
+    >>> mesh = Mesh(devices, ('x', 'y'))
+    >>> inp = np.arange(16).reshape(8, 2)
+    >>> arr = jax.device_put(inp, NamedSharding(mesh, P('x', 'y')))
+    >>> out = jax.jit(lambda x: x * 2)(arr)
+    >>> assert out.sharding == NamedSharding(mesh, P('x', 'y'))
 
-    >>> # Initialize the Mesh and use the mesh as the context manager.
-    >>> with Mesh(devices, ('x', 'y')) as global_mesh:
-    ...   out = pjit(lambda x: x, in_shardings=None, out_shardings=None)(inp)
-
-    >>> # Also you can use it as `with ... as ...`.
-    >>> global_mesh = Mesh(devices, ('x', 'y'))
-    >>> with global_mesh as m:
-    ...   out = pjit(lambda x: x, in_shardings=None, out_shardings=None)(inp)
-
-    >>> # You can also use it as `with Mesh(...)`.
-    >>> with Mesh(devices, ('x', 'y')):
-    ...   out = pjit(lambda x: x, in_shardings=None, out_shardings=None)(inp)
+  .. _Distributed arrays and automatic parallelization: https://docs.jax.dev/en/latest/notebooks/Distributed_arrays_and_automatic_parallelization.html
+  .. _Explicit Sharding:  https://docs.jax.dev/en/latest/notebooks/explicit-sharding.html
   """
 
   devices: np.ndarray
   axis_names: tuple[MeshAxisName, ...]
-  axis_types: MeshAxisType
 
   def __new__(cls, devices: np.ndarray | Sequence[xc.Device],
-              axis_names: str | Sequence[MeshAxisName], *,
-              axis_types: MeshAxisType | None = None):
+              axis_names: str | Sequence[MeshAxisName],
+              axis_types: tuple[AxisType, ...] | None = None):
     if not isinstance(devices, np.ndarray):
       devices = np.array(devices)
     if isinstance(axis_names, str):
@@ -198,16 +274,9 @@ class Mesh(contextlib.ContextDecorator):
           f"devices.ndim == {devices.ndim} and "
           f"len(axis_names) == {len(axis_names)}.")
 
-    axis_types = ({AxisTypes.Hidden: axis_names} if axis_types is None else
-                  axis_types)
-    axis_types_tuple = tuple(axis_types.items())
-    if len(axis_names_to_types(axis_types).keys()) != len(axis_names):
-      raise ValueError(
-          "Number of axis names in axis_types should match the number of"
-          f" axis_names. Got axis_names={axis_names} and"
-          f" axis_types={axis_types}")
+    axis_types = _normalize_axis_types(axis_names, axis_types, 'Mesh')
 
-    key = (axis_names, devices.shape, tuple(devices.flat), axis_types_tuple)
+    key = (axis_names, devices.shape, tuple(devices.flat), axis_types)
     val = _mesh_object_dict.get(key, None)
     if val is not None:
       return val
@@ -216,32 +285,31 @@ class Mesh(contextlib.ContextDecorator):
     self.devices = devices.copy()
     self.devices.flags.writeable = False
     self.axis_names = axis_names
-    self.axis_types = axis_types
-    self._axis_types_tuple = axis_types_tuple
+    self._axis_types = axis_types
+    self._size = math.prod(self.shape.values()) if self.devices.ndim else 0
     _mesh_object_dict[key] = self
     return self
 
   def __reduce__(self):
-    return (type(self), (self.devices, self.axis_names),
-            {'axis_types': self.axis_types})
+    return (_unpicke_mesh, (self.devices, self.axis_names, self._axis_types))
 
   def __eq__(self, other):
-    if not isinstance(other, Mesh):
-      return False
     # This is a performance optimization. Comparing thousands of devices
     # can be expensive.
-    if id(self) == id(other):
+    if self is other:
       return True
+    if not isinstance(other, Mesh):
+      return False
     return (self.axis_names == other.axis_names and
             self.devices.shape == other.devices.shape and
-            self._axis_types_tuple == other._axis_types_tuple and
+            self._axis_types == other._axis_types and
             self._internal_device_list == other._internal_device_list)
 
   def __hash__(self):
     if not hasattr(self, '_hash'):
       self._hash = hash(
           (self.axis_names, self._internal_device_list, self.devices.shape,
-           self._axis_types_tuple))
+           self._axis_types))
     return self._hash
 
   def __setattr__(self, name, value):
@@ -274,29 +342,25 @@ class Mesh(contextlib.ContextDecorator):
               if not t.physical_mesh.empty))
     return False
 
-  @property
+  @functools.cached_property
   def shape(self):
     return collections.OrderedDict(
         (name, size)
-        for name, size in util.safe_zip(self.axis_names, self.devices.shape))
+        for name, size in safe_zip(self.axis_names, self.devices.shape))
 
   @functools.cached_property
   def shape_tuple(self):
     return tuple(
         (name, size)
-        for name, size in util.safe_zip(self.axis_names, self.devices.shape))
+        for name, size in safe_zip(self.axis_names, self.devices.shape))
 
   @property
   def axis_sizes(self) -> tuple[int, ...]:
     return self.devices.shape
 
-  @functools.cached_property
-  def _name_to_type(self):
-    return axis_names_to_types(self.axis_types)
-
   @property
   def size(self):
-    return math.prod(self.shape.values()) if self.devices.ndim else 0
+    return self._size
 
   @property
   def empty(self):
@@ -336,14 +400,16 @@ class Mesh(contextlib.ContextDecorator):
 
   def __str__(self):
     mesh_str = ", ".join(f"'{k}': {v}" for k, v in self.shape.items())
-    return f"Mesh({mesh_str})"
+    atr = f", axis_types={self._axis_types}"
+    return f"Mesh({mesh_str}{atr})"
 
   @functools.cached_property
   def _repr(self):
     if self.empty:
-      return "Mesh(device_ids=[], axis_names=())"
-    atr = f", axis_types={self.axis_types}"
-    return f"Mesh(device_ids={self.device_ids!r}, axis_names={self.axis_names!r}{atr})"
+      return "Mesh(axis_sizes=(), axis_names=())"
+    atr = f", axis_types={self._axis_types}"
+    return (f"Mesh(axis_sizes={self.device_ids.shape}, "
+            f"axis_names={self.axis_names!r}{atr})")
 
   def __repr__(self):
     return self._repr
@@ -355,23 +421,8 @@ class Mesh(contextlib.ContextDecorator):
 
   @functools.cached_property
   def abstract_mesh(self):
-    return AbstractMesh(self.shape_tuple, axis_types=self.axis_types)
-
-  @functools.cached_property
-  def _are_all_axes_collective(self) -> bool:
-    return all(t == AxisTypes.Collective for t in self.axis_types.keys())
-
-  @functools.cached_property
-  def _are_all_axes_hidden(self) -> bool:
-    return all(t == AxisTypes.Hidden for t in self.axis_types.keys())
-
-  @functools.cached_property
-  def _any_axis_collective(self) -> bool:
-    return any(t == AxisTypes.Collective for t in self.axis_types.keys())
-
-  @functools.cached_property
-  def _any_axis_hidden(self) -> bool:
-    return any(t == AxisTypes.Hidden for t in self.axis_types.keys())
+    return AbstractMesh(self.axis_sizes, self.axis_names,
+                        axis_types=self._axis_types)
 
 
 EMPTY_ENV = ResourceEnv(Mesh(np.empty((), dtype=object), ()))
@@ -385,7 +436,7 @@ class _ThreadResourcesLocalState(threading.local):
 thread_resources = _ThreadResourcesLocalState()
 
 
-class AbstractMesh:
+class AbstractMesh(BaseMesh):
   """AbstractMesh contains only axis names and axis sizes.
 
   It does not contain concrete devices compared to `jax.sharding.Mesh`. You
@@ -394,60 +445,59 @@ class AbstractMesh:
   your mesh shape and axis names stay the same but the devices change.
   See the description of https://github.com/jax-ml/jax/pull/23022 for more
   details.
+
+  Args:
+    axis_sizes: A tuple of integers specifying the size of each resource axis.
+    axis_names: A tuple of resource axis names to be assigned to the
+      dimensions of the ``devices`` argument. Its length should match the
+      rank of ``devices``.
+    axis_types: and optional tuple of :class:`jax.sharding.AxisType` entries corresponding to
+      the ``axis_names``. See `Explicit Sharding`_ for more information.
+
+  .. _Explicit Sharding:  https://docs.jax.dev/en/latest/notebooks/explicit-sharding.html
   """
 
-  def __init__(self, shape_tuple: tuple[tuple[str, int], ...], *,
-               axis_types: MeshAxisType | None = None):
-    self.shape_tuple = shape_tuple
-    if self.shape_tuple:
-      self._axis_names, self._axis_sizes = list(zip(*self.shape_tuple))
-    else:
-      self._axis_names, self._axis_sizes = (), ()
-    self.axis_types = ({AxisTypes.Hidden: self._axis_names}
-                       if axis_types is None else axis_types)
-    self._axis_types_tuple = tuple(self.axis_types.items())
-    if len(self._name_to_type.keys()) != len(self._axis_names):
-      raise ValueError(
-          "Number of axis names in axis_types should match the number of"
-          f" axis_names in shape_tuple. Got axis_names={self._axis_names} and"
-          f" axis_types={self.axis_types}")
+  def __init__(self, axis_sizes: tuple[int, ...], axis_names: tuple[str, ...],
+               axis_types: AxisType | tuple[AxisType, ...] | None = None):
+    self.axis_sizes = axis_sizes
+    self.axis_names = axis_names
+    self._size = math.prod(self.axis_sizes) if self.axis_sizes else 0
+    self._axis_types = _normalize_axis_types(
+        self.axis_names, axis_types, 'AbstractMesh')
+    self._hash = hash((self.axis_sizes, self.axis_names, self._axis_types))
 
   def __hash__(self):
-    return hash((self.shape_tuple, self._axis_types_tuple))
+    return self._hash
 
   def __eq__(self, other):
+    if self is other:
+      return True
     if not isinstance(other, AbstractMesh):
       return False
-    if id(self) == id(other):
-      return True
-    return (self.shape_tuple == other.shape_tuple and
-            self._axis_types_tuple == other._axis_types_tuple)
+    return (self.axis_sizes == other.axis_sizes and
+            self.axis_names == other.axis_names and
+            self._axis_types == other._axis_types)
 
   def __repr__(self):
-    mesh_repr = ", ".join(f"'{n}': {v}" for n, v in self.shape_tuple)
-    atr = f", axis_types={self.axis_types}"
+    mesh_repr = (", ".join(f"'{n}': {v}" for n, v in self.shape_tuple)
+                 if self.shape_tuple else "()")
+    atr = f", axis_types={self._axis_types}"
     return f"AbstractMesh({mesh_repr}{atr})"
 
   @property
-  def axis_names(self):
-    return self._axis_names
-
-  @property
-  def axis_sizes(self) -> tuple[int, ...]:
-    return self._axis_sizes
-
-  @functools.cached_property
-  def _name_to_type(self):
-    return axis_names_to_types(self.axis_types)
-
-  @functools.cached_property
   def size(self):
-    return math.prod(self._axis_sizes) if self._axis_sizes else 0
+    return self._size
 
   @functools.cached_property
   def shape(self):
     return collections.OrderedDict(self.shape_tuple)
 
+  @functools.cached_property
+  def shape_tuple(self):
+    return tuple(
+        (name, size)
+        for name, size in safe_zip(self.axis_names, self.axis_sizes))
+
   @property
   def _internal_device_list(self):
     return None
@@ -456,32 +506,15 @@ class AbstractMesh:
   def empty(self):
     return self.size == 0
 
-  def update_axis_types(self, new_axis_types) -> AbstractMesh:
-    # dict(self._name_to_type) will copy it.
-    updated_name_to_type = dict(self._name_to_type)
-    updated_name_to_type.update(axis_names_to_types(new_axis_types))
-    new_axis_types = axis_types_to_names(updated_name_to_type)
-    return AbstractMesh(self.shape_tuple, axis_types=new_axis_types)
-
   @property
   def abstract_mesh(self):
     return self
 
-  @functools.cached_property
-  def _are_all_axes_collective(self) -> bool:
-    return all(t == AxisTypes.Collective for t in self.axis_types.keys())
-
-  @functools.cached_property
-  def _are_all_axes_hidden(self) -> bool:
-    return all(t == AxisTypes.Hidden for t in self.axis_types.keys())
-
-  @functools.cached_property
-  def _any_axis_collective(self) -> bool:
-    return any(t == AxisTypes.Collective for t in self.axis_types.keys())
-
-  @functools.cached_property
-  def _any_axis_hidden(self) -> bool:
-    return any(t == AxisTypes.Hidden for t in self.axis_types.keys())
+  def update_axis_types(self, name_to_type: dict[MeshAxisName, AxisType]):
+    new_axis_types = tuple(name_to_type[n] if n in name_to_type else a
+                           for n, a in zip(self.axis_names, self._axis_types))
+    return AbstractMesh(self.axis_sizes, self.axis_names,
+                        axis_types=new_axis_types)
 
   @property
   def devices(self):
@@ -511,8 +544,8 @@ class AbstractMesh:
 
   @staticmethod
   def _extremely_unsafe_enter_tracing_context(mesh: AbstractMesh):
-    jax_config.abstract_mesh_context_manager.set_local(mesh)
-    return
+    prev = jax_config.abstract_mesh_context_manager.swap_local(mesh)
+    return prev
 
 
 # Create this indirection because pytype fails to recognize a property if a
@@ -520,33 +553,29 @@ class AbstractMesh:
 def _raise_value_error(name):
   raise ValueError(f"AbstractMesh does not implement {name}")
 
+empty_abstract_mesh = AbstractMesh((), ())
 
-@contextlib.contextmanager
-def set_abstract_mesh(mesh: AbstractMesh):
-  prev_val = jax_config.abstract_mesh_context_manager.swap_local(mesh)
-  try:
-    yield
-  finally:
-    jax_config.abstract_mesh_context_manager.set_local(prev_val)
+class UseAbstractMeshContextManager:
+  __slots__ = ['mesh', 'prev']
 
-def get_abstract_mesh():
-  return jax_config.abstract_mesh_context_manager.value
+  def __init__(self, mesh: AbstractMesh):
+    if not isinstance(mesh, AbstractMesh):
+      raise ValueError(
+          "Expected mesh of type `jax.sharding.AbstractMesh`. Got type:"
+          f" {type(mesh)}")
+    self.mesh = mesh
 
+  def __enter__(self):
+    self.prev = jax_config.abstract_mesh_context_manager.swap_local(self.mesh)
 
-@contextlib.contextmanager
-def set_concrete_mesh(mesh: Mesh):
-  prev_val = jax_config.device_context.swap_local(mesh)
-  try:
-    yield
-  finally:
-    jax_config.device_context.set_local(prev_val)
+  def __exit__(self, exc_type, exc_value, traceback):
+    jax_config.abstract_mesh_context_manager.set_local(self.prev)
 
-def get_concrete_mesh():
+use_abstract_mesh = UseAbstractMeshContextManager
+
+def get_abstract_mesh() -> AbstractMesh:
+  val = jax_config.abstract_mesh_context_manager.value
+  return empty_abstract_mesh if val is None else val
+
+def get_concrete_mesh() -> Mesh | None:
   return jax_config.device_context.value
-
-
-@contextlib.contextmanager
-def use_mesh(mesh: Mesh):
-  with (set_abstract_mesh(mesh.abstract_mesh),
-        jax_config.sharding_in_types(True), set_concrete_mesh(mesh)):
-    yield
