@@ -1,4 +1,4 @@
-# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
+# Copyright 2021 The JAX Authors. All Rights Reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -19,21 +19,19 @@ from __future__ import annotations
 import collections
 from collections.abc import Callable, Generator, MutableMapping, Sequence
 import itertools
-import logging
 import math
 from typing import Any
 
 from jax._src import xla_bridge as xb
 import numpy as np
 
-logger = logging.getLogger(__name__)
-
 _TPU_V2 = 'TPU v2'
 _TPU_V3 = 'TPU v3'
 _TPU_V4 = 'TPU v4'
 _TPU_V5_LITE = "TPU v5 lite"
 _TPU_V5E = "TPU v5e"
 _TPU_V5P = "TPU v5p"
+_TPU_V6_LITE = "TPU v6 lite"
 
 # Maps physical topology -> mesh shape -> transpose to use for jekbradbury's
 # famous contiguous mesh trick.
@@ -79,18 +77,12 @@ def _tpu_v2_v3_create_device_mesh(
     **unused_kwargs,
 ) -> np.ndarray:
   if len(devices) == 8:
-    logger.info(
-        'Reordering mesh to physical ring order on single-tray TPU v2/v3.'
-    )
     device_mesh = np.asarray(devices)
     device_mesh = device_mesh[np.array(_TRAY_RING_ORDER)]
     device_mesh = device_mesh.reshape(mesh_shape)
     return device_mesh
   elif mesh_shape[-1] == 8:
     device_mesh = np.asarray(devices).reshape(mesh_shape)
-    logger.info(
-        'Reordering mesh to physical ring order on each TPU v2/v3 tray.'
-    )
     perm = np.array(_TRAY_RING_ORDER)
     device_mesh = device_mesh[..., perm]
     return device_mesh
@@ -172,7 +164,7 @@ def _v5p_create_device_mesh(
       devices,
       key=lambda d: tuple(reversed(getattr(d, "coords", (0, 0, 0)))))
 
-  if bound_x == bound_y == 2 and bound_z == 2:
+  if bound_x == bound_y == bound_z == 2 and len(devices) == 8:
     device_mesh = np.asarray(sequential_devices)
     device_mesh = device_mesh[np.array(_V5P_2x2x2_ORDER)]
     device_mesh = device_mesh.reshape(mesh_shape)
@@ -190,6 +182,7 @@ device_kind_handler_dict: dict[
     _TPU_V3: _tpu_v2_v3_create_device_mesh,
     _TPU_V5_LITE: _v5e_create_device_mesh,
     _TPU_V5P: _v5p_create_device_mesh,
+    _TPU_V6_LITE: _v5e_create_device_mesh,
 }
 
 
@@ -597,10 +590,10 @@ def _generate_logical_mesh(
           zip(logical_indices, physical_indices, range(len(logical_indices)))
       )
   )
-  logical_mesh = np.transpose(logical_mesh, transpose_axes)  # type: ignore  # numpy 2.2
+  logical_mesh = np.transpose(logical_mesh, transpose_axes)
 
   # Reshape to add the trivial dimensions back.
-  logical_mesh = np.reshape(logical_mesh, logical_mesh_shape)  # type: ignore  # numpy 2.2
+  logical_mesh = np.reshape(logical_mesh, logical_mesh_shape)
 
   return logical_mesh
 
