@@ -22,37 +22,57 @@ import operator
 import math
 import numpy as np
 from typing import Any, Literal
+import warnings
 
-import jax
-import jax.numpy as jnp
-from jax import custom_jvp
-from jax import lax
+from jax._src import api
 from jax._src import config
 from jax._src import core
+from jax._src import custom_derivatives
 from jax._src import deprecations
 from jax._src import dtypes
+from jax._src import lax
+from jax._src import numpy as jnp
 from jax._src import util
 from jax._src.core import AxisName
-from jax._src.sharding_impls import NamedSharding, PartitionSpec as P
 from jax._src.cudnn.fused_attention_stablehlo import (
     dot_product_attention as cudnn_dot_product_attention, MaskType)
+from jax._src.cudnn.scaled_matmul_stablehlo import (
+    scaled_matmul_wrapper as cudnn_scaled_matmul,
+    scaled_dot_general_wrapper as cudnn_scaled_dot_general,
+    BlockScaleConfig)
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
+from jax._src.numpy import einsum as jnp_einsum
 from jax._src.numpy import util as numpy_util
-from jax._src.typing import Array, ArrayLike, DType
+from jax._src.numpy.reductions import Axis
+from jax._src.sharding_impls import NamedSharding, PartitionSpec as P
+from jax._src.typing import Array, ArrayLike, DType, DTypeLike
 from jax._src.ops.special import logsumexp as _logsumexp
 
 
-class Unspecified:
-  def __repr__(self):
-    return "_UNSPECIFIED"
-_UNSPECIFIED = Unspecified()
-
-
 # activations
+@api.jit
+def identity(x: ArrayLike) -> Array:
+  r"""Identity activation function.
 
-@custom_jvp
-@jax.jit
+  Returns the argument unmodified.
+
+  Args:
+    x : input array
+
+  Returns:
+    The argument `x` unmodified.
+
+  Examples:
+    >>> jax.nn.identity(jax.numpy.array([-2., -1., -0.5, 0, 0.5, 1., 2.]))
+    Array([-2. , -1. , -0.5, 0. , 0.5, 1. , 2. ], dtype=float32)
+
+  """
+  numpy_util.check_arraylike("identity", x)
+  return jnp.asarray(x)
+
+@custom_derivatives.custom_jvp
+@api.jit
 def relu(x: ArrayLike) -> Array:
   r"""Rectified linear unit activation function.
 
@@ -88,7 +108,7 @@ def relu(x: ArrayLike) -> Array:
 # For behavior at 0, see https://dl.acm.org/doi/10.5555/3540261.3540297
 relu.defjvps(lambda g, ans, x: lax.select(x > 0, g, lax.full_like(g, 0)))
 
-@jax.jit
+@api.jit
 def squareplus(x: ArrayLike, b: ArrayLike = 4) -> Array:
   r"""Squareplus activation function.
 
@@ -110,7 +130,7 @@ def squareplus(x: ArrayLike, b: ArrayLike = 4) -> Array:
   y = x + jnp.sqrt(jnp.square(x) + b)
   return y / 2
 
-@jax.jit
+@api.jit
 def softplus(x: ArrayLike) -> Array:
   r"""Softplus activation function.
 
@@ -124,7 +144,7 @@ def softplus(x: ArrayLike) -> Array:
   """
   return jnp.logaddexp(x, 0)
 
-@jax.jit
+@api.jit
 def sparse_plus(x: ArrayLike) -> Array:
   r"""Sparse plus function.
 
@@ -150,7 +170,7 @@ def sparse_plus(x: ArrayLike) -> Array:
   x = jnp.asarray(x)
   return jnp.where(x <= -1.0, 0.0, jnp.where(x >= 1.0, x, (x + 1.0)**2/4))
 
-@jax.jit
+@api.jit
 def soft_sign(x: ArrayLike) -> Array:
   r"""Soft-sign activation function.
 
@@ -166,7 +186,7 @@ def soft_sign(x: ArrayLike) -> Array:
   x_arr = jnp.asarray(x)
   return x_arr / (jnp.abs(x_arr) + 1)
 
-@partial(jax.jit, inline=True)
+@partial(api.jit, inline=True)
 def sigmoid(x: ArrayLike) -> Array:
   r"""Sigmoid activation function.
 
@@ -187,7 +207,7 @@ def sigmoid(x: ArrayLike) -> Array:
   """
   return lax.logistic(x)
 
-@jax.jit
+@api.jit
 def sparse_sigmoid(x: ArrayLike) -> Array:
   r"""Sparse sigmoid activation function.
 
@@ -219,7 +239,7 @@ def sparse_sigmoid(x: ArrayLike) -> Array:
   """
   return 0.5 * jnp.clip(x + 1.0, 0.0, 2.0)
 
-@jax.jit
+@api.jit
 def silu(x: ArrayLike) -> Array:
   r"""SiLU (aka swish) activation function.
 
@@ -245,7 +265,7 @@ def silu(x: ArrayLike) -> Array:
 
 swish = silu
 
-@jax.jit
+@api.jit
 def mish(x: ArrayLike) -> Array:
   r"""Mish activation function.
 
@@ -268,7 +288,7 @@ def mish(x: ArrayLike) -> Array:
   x_arr = jnp.asarray(x)
   return x_arr * jnp.tanh(softplus(x_arr))
 
-@jax.jit
+@api.jit
 def log_sigmoid(x: ArrayLike) -> Array:
   r"""Log-sigmoid activation function.
 
@@ -290,7 +310,7 @@ def log_sigmoid(x: ArrayLike) -> Array:
   x_arr = jnp.asarray(x)
   return -softplus(-x_arr)
 
-@jax.jit
+@api.jit
 def elu(x: ArrayLike, alpha: ArrayLike = 1.0) -> Array:
   r"""Exponential linear unit activation function.
 
@@ -318,7 +338,7 @@ def elu(x: ArrayLike, alpha: ArrayLike = 1.0) -> Array:
                    x_arr,
                    alpha * jnp.expm1(jnp.where(x_arr > 0, 0., x_arr)))
 
-@jax.jit
+@api.jit
 def leaky_relu(x: ArrayLike, negative_slope: ArrayLike = 1e-2) -> Array:
   r"""Leaky rectified linear unit activation function.
 
@@ -346,7 +366,7 @@ def leaky_relu(x: ArrayLike, negative_slope: ArrayLike = 1e-2) -> Array:
   x_arr = jnp.asarray(x)
   return jnp.where(x_arr >= 0, x_arr, negative_slope * x_arr)
 
-@jax.jit
+@api.jit
 def hard_tanh(x: ArrayLike) -> Array:
   r"""Hard :math:`\mathrm{tanh}` activation function.
 
@@ -369,7 +389,7 @@ def hard_tanh(x: ArrayLike) -> Array:
   x_arr = jnp.asarray(x)
   return jnp.where(x_arr > 1, 1, jnp.where(x_arr < -1, -1, x_arr))
 
-@jax.jit
+@api.jit
 def celu(x: ArrayLike, alpha: ArrayLike = 1.0) -> Array:
   r"""Continuously-differentiable exponential linear unit activation.
 
@@ -394,7 +414,7 @@ def celu(x: ArrayLike, alpha: ArrayLike = 1.0) -> Array:
   """
   return jnp.maximum(x, 0.0) + alpha * jnp.expm1(jnp.minimum(x, 0.0) / alpha)
 
-@jax.jit
+@api.jit
 def selu(x: ArrayLike) -> Array:
   r"""Scaled exponential linear unit activation.
 
@@ -427,7 +447,7 @@ def selu(x: ArrayLike) -> Array:
   return scale * elu(x, alpha)
 
 # TODO(phawkins): this jit was found to change numerics in a test. Debug this.
-# @partial(jax.jit, static_argnames=("approximate",))
+# @partial(api.jit, static_argnames=("approximate",))
 def gelu(x: ArrayLike, approximate: bool = True) -> Array:
   r"""Gaussian error linear unit activation function.
 
@@ -462,7 +482,7 @@ def gelu(x: ArrayLike, approximate: bool = True) -> Array:
         0.5 * x_arr * (lax.erfc(-x_arr * sqrt_half)), dtype=x_arr.dtype
     )
 
-@partial(jax.jit, static_argnames=("axis",))
+@partial(api.jit, static_argnames=("axis",))
 def glu(x: ArrayLike, axis: int = -1) -> Array:
   r"""Gated linear unit activation function.
 
@@ -498,11 +518,10 @@ def glu(x: ArrayLike, axis: int = -1) -> Array:
 logsumexp = _logsumexp
 
 
-@partial(jax.jit, static_argnames=("axis",))
+@partial(api.jit, static_argnames=("axis",))
 def log_softmax(x: ArrayLike,
-                axis: int | tuple[int, ...] | None = -1,
-                where: ArrayLike | None = None,
-                initial: Unspecified = _UNSPECIFIED) -> Array:
+                axis: Axis = -1,
+                where: ArrayLike | None = None) -> Array:
   r"""Log-Softmax function.
 
   Computes the logarithm of the :code:`softmax` function, which rescales
@@ -516,7 +535,8 @@ def log_softmax(x: ArrayLike,
     x : input array
     axis: the axis or axes along which the :code:`log_softmax` should be
       computed. Either an integer or a tuple of integers.
-    where: Elements to include in the :code:`log_softmax`.
+    where: Elements to include in the :code:`log_softmax`. The output for any
+      masked-out element is minus infinity.
 
   Returns:
     An array.
@@ -528,29 +548,24 @@ def log_softmax(x: ArrayLike,
   See also:
     :func:`softmax`
   """
-  # TODO(jakevdp): remove the initial argument after JAX v0.4.40.
-  if initial is not _UNSPECIFIED:
-    raise TypeError("The initial argument to jax.nn.log_softmax was removed in JAX v0.4.36.")
-  del initial
   numpy_util.check_arraylike("log_softmax", x)
   x_arr = jnp.asarray(x)
-  x_max = jnp.max(x_arr, axis, where=where, initial=-jnp.inf, keepdims=True)
-  x_safe = x_arr if where is None else jnp.where(where, x_arr, -jnp.inf)
+  x_max = jnp.max(x_arr, axis, where=where, initial=-np.inf, keepdims=True)
+  x_safe = x_arr if where is None else jnp.where(where, x_arr, -np.inf)
   shifted = x_safe - lax.stop_gradient(x_max)
   shifted_logsumexp = jnp.log(
       jnp.sum(jnp.exp(shifted), axis, where=where, keepdims=True))
   result = shifted - shifted_logsumexp
   if where is not None:
-    return jnp.where(where, result, -jnp.inf)
+    return jnp.where(where, result, -np.inf)
   return result
 
 
 # TODO(phawkins): this jit was found to change numerics in a test. Debug this.
-# @partial(jax.jit, static_argnames=("axis",))
+# @partial(api.jit, static_argnames=("axis",))
 def softmax(x: ArrayLike,
-            axis: int | tuple[int, ...] | None = -1,
-            where: ArrayLike | None = None,
-            initial: Unspecified = _UNSPECIFIED) -> Array:
+            axis: Axis = -1,
+            where: ArrayLike | None = None) -> Array:
   r"""Softmax function.
 
   Computes the function which rescales elements to the range :math:`[0, 1]`
@@ -564,7 +579,8 @@ def softmax(x: ArrayLike,
     axis: the axis or axes along which the softmax should be computed. The
       softmax output summed across these dimensions should sum to :math:`1`.
       Either an integer or a tuple of integers.
-    where: Elements to include in the :code:`softmax`.
+    where: Elements to include in the :code:`softmax`. The output for any
+      masked-out element is zero.
 
   Returns:
     An array.
@@ -576,10 +592,6 @@ def softmax(x: ArrayLike,
   See also:
     :func:`log_softmax`
   """
-  # TODO(jakevdp): remove the initial argument after JAX v0.4.40.
-  if initial is not _UNSPECIFIED:
-    raise TypeError("The initial argument to jax.nn.softmax was removed in JAX v0.4.36.")
-  del initial
   if config.softmax_custom_jvp.value:
     # mypy is confused by the `functools.partial` application in the definition
     # of `_softmax` and incorrectly concludes that `_softmax` returns
@@ -589,12 +601,12 @@ def softmax(x: ArrayLike,
     return _softmax_deprecated(x, axis, where)
 
 # TODO(mattjj): replace softmax with _softmax when deprecation flag is removed
-@partial(jax.custom_jvp, nondiff_argnums=(1,))
+@partial(custom_derivatives.custom_jvp, nondiff_argnums=(1,))
 def _softmax(
     x: ArrayLike,
-    axis: int | tuple[int, ...] | None = -1,
+    axis: Axis = -1,
     where: ArrayLike | None = None,
-    initial: ArrayLike | None = -jnp.inf) -> Array:
+    initial: ArrayLike | None = -np.inf) -> Array:
   x_max = jnp.max(x, axis, where=where, initial=initial, keepdims=True)
   x_safe = x if where is None else jnp.where(where, x, initial)
   unnormalized = jnp.exp(x_safe - x_max)
@@ -611,9 +623,9 @@ def _softmax_jvp(axis, primals, tangents):
 
 def _softmax_deprecated(
     x: ArrayLike,
-    axis: int | tuple[int, ...] | None = -1,
+    axis: Axis = -1,
     where: ArrayLike | None = None,
-    initial: ArrayLike | None = -jnp.inf) -> Array:
+    initial: ArrayLike | None = -np.inf) -> Array:
   x_max = jnp.max(x, axis, where=where, initial=initial, keepdims=True)
   x_safe = x if where is None else jnp.where(where, x, initial)
   unnormalized = jnp.exp(x_safe - lax.stop_gradient(x_max))
@@ -623,14 +635,40 @@ def _softmax_deprecated(
   return result
 
 
-@partial(jax.jit, static_argnames=("axis",))
+@partial(api.jit, static_argnames=("axis",))
 def standardize(x: ArrayLike,
-              axis: int | tuple[int, ...] | None = -1,
-              mean: ArrayLike | None = None,
-              variance: ArrayLike | None = None,
-              epsilon: ArrayLike = 1e-5,
-              where: ArrayLike | None = None) -> Array:
-  r"""Normalizes an array by subtracting ``mean`` and dividing by :math:`\sqrt{\mathrm{variance}}`."""
+                axis: Axis = -1,
+                mean: ArrayLike | None = None,
+                variance: ArrayLike | None = None,
+                epsilon: ArrayLike = 1e-5,
+                where: ArrayLike | None = None) -> Array:
+  r"""Standardizes input to zero mean and unit variance.
+
+  The standardization is given by:
+
+  .. math::
+
+     x_{std} = \frac{x - \langle x\rangle}{\sqrt{\langle(x - \langle x\rangle)^2\rangle + \epsilon}}
+
+  where :math:`\langle x\rangle` indicates the mean of :math:`x`, and :math:`\epsilon` is
+  a small correction factor introduced to avoid division by zero.
+
+  Args:
+    x: input array to be standardized.
+    axis: integer or tuple of integers representing the axes along which
+      to standardize. Defaults to the last axis (``-1``).
+    mean: optionally specify the mean used for standardization. If not specified,
+      then ``x.mean(axis, where=where)`` will be used.
+    variance: optionally specify the variance used for standardization. If not
+      specified, then ``x.var(axis, where=where)`` will be used.
+    epsilon: correction factor added to variance to avoid division by zero; defaults
+      to ``1E-5``.
+    where: optional boolean mask specifying which elements to use when computing
+      the mean and variance.
+
+  Returns:
+    An array of the same shape as ``x`` containing the standardized input.
+  """
   numpy_util.check_arraylike("standardize", x)
   numpy_util.check_arraylike_or_none("standardize", mean, variance, where)
   if mean is None:
@@ -645,7 +683,7 @@ def standardize(x: ArrayLike,
   return jnp.subtract(x, jnp.asarray(mean)) * lax.rsqrt(jnp.asarray(variance) + epsilon)
 
 # TODO(slebedev): Change the type of `x` to `ArrayLike`.
-@partial(jax.jit, static_argnames=("num_classes", "dtype", "axis"))
+@partial(api.jit, static_argnames=("num_classes", "dtype", "axis"))
 def _one_hot(x: Array, num_classes: int, *,
              dtype: Any, axis: int | AxisName) -> Array:
   num_classes = core.concrete_dim_or_error(
@@ -653,9 +691,9 @@ def _one_hot(x: Array, num_classes: int, *,
       "The error arose in jax.nn.one_hot argument `num_classes`.")
   dtype = dtypes.canonicalize_dtype(dtype)
   try:
-    output_pos_axis = util.canonicalize_axis(axis, x.ndim + 1)
+    output_pos_axis = util.canonicalize_axis(axis, x.ndim + 1)  # type: ignore[arg-type]
   except TypeError:
-    axis_size = lax.psum(1, axis)
+    axis_size = lax.axis_size(axis)
     if num_classes != axis_size:
       raise ValueError(f"Expected num_classes to match the size of axis {axis}, "
                        f"but {num_classes} != {axis_size}") from None
@@ -665,18 +703,15 @@ def _one_hot(x: Array, num_classes: int, *,
   lhs = lax.expand_dims(x, (axis,))
   rhs_shape = [1] * x.ndim
   rhs_shape.insert(output_pos_axis, num_classes)
-  if config.sharding_in_types.value:
-    # TODO(yashkatariya): Maybe expose `out_sharding` on `one_hot` too?
-    rhs_sharding = NamedSharding(x.sharding.mesh, P(*[None] * len(rhs_shape)))  # pytype: disable=attribute-error
-  else:
-    rhs_sharding = None
+  # TODO(yashkatariya): Maybe expose `out_sharding` on `one_hot` too?
+  rhs_sharding = NamedSharding(x.aval.sharding.mesh, P(*[None] * len(rhs_shape)))  # pytype: disable=attribute-error
   rhs = lax.broadcasted_iota(x.dtype, rhs_shape, output_pos_axis,
-                             sharding=rhs_sharding)
+                             out_sharding=rhs_sharding)
   return (lhs == rhs).astype(dtype)
 
 # TODO(slebedev): Change the type of `x` to `ArrayLike`.
 def one_hot(x: Any, num_classes: int, *,
-            dtype: Any = jnp.float_, axis: int | AxisName = -1) -> Array:
+            dtype: Any = dtypes.float_, axis: int | AxisName = -1) -> Array:
   """One-hot encodes the given indices.
 
   Each index in the input ``x`` is encoded as a vector of zeros of length
@@ -704,7 +739,7 @@ def one_hot(x: Any, num_classes: int, *,
       num_classes,
       "The error arose in jax.nn.one_hot argument `num_classes`.")
   x_arr = jnp.asarray(x)
-  if not jnp.isdtype(x_arr.dtype, "integral"):
+  if not dtypes.isdtype(x_arr.dtype, "integral"):
     # Deprecated 2024-12-18
     deprecations.warn(
       'jax-nn-one-hot-float-input',
@@ -713,8 +748,8 @@ def one_hot(x: Any, num_classes: int, *,
   return _one_hot(x_arr, num_classes, dtype=dtype, axis=axis)
 
 
-@jax.custom_jvp
-@jax.jit
+@custom_derivatives.custom_jvp
+@api.jit
 def relu6(x: ArrayLike) -> Array:
   r"""Rectified Linear Unit 6 activation function.
 
@@ -746,7 +781,7 @@ def relu6(x: ArrayLike) -> Array:
 relu6.defjvps(lambda g, ans, x:
               lax.select((x > 0) & (x < 6), g, lax.full_like(g, 0)))
 
-@jax.jit
+@api.jit
 def hard_sigmoid(x: ArrayLike) -> Array:
   r"""Hard Sigmoid activation function.
 
@@ -766,7 +801,7 @@ def hard_sigmoid(x: ArrayLike) -> Array:
   """
   return relu6(x + 3.) / 6.
 
-@jax.jit
+@api.jit
 def hard_silu(x: ArrayLike) -> Array:
   r"""Hard SiLU (swish) activation function
 
@@ -794,11 +829,11 @@ def hard_silu(x: ArrayLike) -> Array:
 hard_swish = hard_silu
 
 def _get_large_negative(dtype):
-  dtype_max = jnp.finfo(dtype).max
+  dtype_max = dtypes.finfo(dtype).max
   return jnp.asarray(-0.7 * dtype_max, dtype=dtype)
 
 def _get_causal_mask(T, S):
-  mask = jnp.tril(jnp.ones((T, S), dtype=jnp.bool_))
+  mask = jnp.tril(jnp.ones((T, S), dtype=bool))
   return mask[None, None, :, :]
 
 def _get_window_mask(T: int, S: int, local_window_size: tuple[int, int]):
@@ -831,9 +866,9 @@ def _apply_masks(logits, mask, is_causal, q_seqlen, kv_seqlen,
   if mask is None and not is_causal and q_seqlen is None and kv_seqlen is None:
     return logits
 
-  combined_mask = jnp.ones_like(logits, dtype=jnp.bool_)
+  combined_mask = jnp.ones_like(logits, dtype=bool)
   if mask is not None:
-    assert mask.dtype == jnp.bool_
+    assert mask.dtype == np.dtype(bool)
     combined_mask = jnp.logical_and(combined_mask, mask)
 
   T, S = logits.shape[2], logits.shape[3]
@@ -856,9 +891,38 @@ def _apply_masks(logits, mask, is_causal, q_seqlen, kv_seqlen,
 
 def _dot_product_attention_core(query, key, value, bias, mask, is_causal,
                                 scale, q_seqlen, kv_seqlen, local_window_size):
-  logits_dtype = jnp.promote_types(query.dtype, jnp.float32)
-  logits = jnp.einsum('BTNH,BSNH->BNTS', query, key,
-                      preferred_element_type=logits_dtype)
+  logits_dtype = jnp.promote_types(query.dtype, np.float32)
+
+  # If the query and logits dtypes are different, then the default precision
+  # can use inconsistent types in the backwards pass
+  # (see https://github.com/jax-ml/jax/issues/24047).
+  if query.dtype == dtypes.bfloat16:
+    precision = lax.DotAlgorithmPreset.BF16_BF16_F32
+  elif query.dtype == np.float16:
+    precision = lax.DotAlgorithmPreset.F16_F16_F32
+  # TODO(sbodenstein): Implement this fix for all dtypes.
+  else:
+    precision = None
+
+  # Explicit precision will fail on platforms that don't support it. For example,
+  # some GPUs do not support BF16_BF16_F32, and TPU does not support F16_F16_F32.
+  # Use the default precision as a fallback in these cases.
+  try:
+    logits = jnp_einsum.einsum(
+        "BTNH,BSNH->BNTS",
+        query,
+        key,
+        precision=precision,
+        preferred_element_type=logits_dtype,
+    )
+  except:  # pylint: disable=bare-except
+    logits = jnp_einsum.einsum(
+        "BTNH,BSNH->BNTS",
+        query,
+        key,
+        precision=None,
+        preferred_element_type=logits_dtype,
+    )
 
   logits *= jnp.array(scale, dtype=logits.dtype)
 
@@ -869,10 +933,10 @@ def _dot_product_attention_core(query, key, value, bias, mask, is_causal,
                                local_window_size)
 
   # Softmax and it is always carried out in fp32.
-  padded_logits = padded_logits.astype(jnp.float32)
-  probs = jax.nn.softmax(padded_logits, axis=-1).astype(key.dtype)
+  padded_logits = padded_logits.astype(np.float32)
+  probs = softmax(padded_logits, axis=-1).astype(key.dtype)
 
-  encoded = jnp.einsum('BNTS,BSNH->BTNH', probs, value)
+  encoded = jnp_einsum.einsum('BNTS,BSNH->BTNH', probs, value)
   if q_seqlen is not None and kv_seqlen is not None:
     mask = _get_padding_mask_encoded(encoded.shape[1], q_seqlen)
     encoded *= mask.astype(encoded.dtype)
@@ -906,7 +970,7 @@ def _dot_product_attention_xla(
     return t
   bias = _reshape_to_grouped(bias)
   mask = _reshape_to_grouped(mask)
-  vmapped_fn = jax.vmap(
+  vmapped_fn = api.vmap(
       _dot_product_attention_core,
       in_axes=(3, None, None, 2, 2, None, None, None, None, None),
       out_axes=3,
@@ -930,7 +994,7 @@ def bias_bwd_rule(query_head_num, res, g):
 # around a cuDNN issue where bias gradients are only supported when the batch
 # size is 1 and the number of heads matches the query.
 # TODO(kaixih@nvidia): Remove this workaround once cuDNN resolves the issue.
-@partial(jax.custom_vjp, nondiff_argnums=(1,))
+@partial(custom_derivatives.custom_vjp, nondiff_argnums=(1,))
 def check_valid_bias_batch(x, query_head_num):
   output, _ = bias_fwd_rule(x, query_head_num)
   return output
@@ -1043,7 +1107,7 @@ def dot_product_attention(
       token's local window. If set, this specifies the (left_window_size,
       right_window_size) for each token. E.g., if local_window_size == (3, 2)
       and the sequence is [0, 1, 2, 3, 4, 5, c, 7, 8, 9], token `c` can attend
-      to [3, 4, 5, c, 7, 8]. If a single int is given, it will be intepreted as
+      to [3, 4, 5, c, 7, 8]. If a single int is given, it will be interpreted as
       a symmetric window (window_size, window_size).
     implementation: A string to control which implementation backend to use.
       Supported strings are `xla`, `cudnn` (cuDNN flash attention). It defaults
@@ -1089,11 +1153,11 @@ def dot_product_attention(
   B, S, K, H = key_arr.shape
   _check_shape_and_dtype(value_arr, [B, S, K, H], key_arr.dtype, 'value')
   _check_shape_and_dtype(query_arr, [B, -1, -1, H], key_arr.dtype, 'query')
-  _check_shape_and_dtype(mask, [-1] * 4, jnp.bool_, 'mask')
+  _check_shape_and_dtype(mask, [-1] * 4, np.dtype(bool), 'mask')
   _check_shape_and_dtype(bias, [-1] * 4, None, 'bias')
-  _check_shape_and_dtype(query_seq_lengths, [B], jnp.int32,
+  _check_shape_and_dtype(query_seq_lengths, [B], np.dtype('int32'),
                          'query_seq_lengths')
-  _check_shape_and_dtype(key_value_seq_lengths, [B], jnp.int32,
+  _check_shape_and_dtype(key_value_seq_lengths, [B], np.dtype('int32'),
                          'key_value_seq_lengths')
   if query_arr.shape[-2] % K != 0:
     raise ValueError(f"The number of query heads must be a multiple of "
@@ -1119,9 +1183,9 @@ def dot_product_attention(
       if use_padding:
         if query_seq_lengths is None:
           T = query_arr.shape[1]
-          query_seq_lengths = jnp.full((B,), T, dtype=jnp.int32)
+          query_seq_lengths = jnp.full((B,), T, dtype=np.int32)
         if key_value_seq_lengths is None:
-          key_value_seq_lengths = jnp.full((B,), S, dtype=jnp.int32)
+          key_value_seq_lengths = jnp.full((B,), S, dtype=np.int32)
 
       mask_type = MaskType.NO_MASK
       if use_padding and is_causal:
@@ -1159,3 +1223,227 @@ def dot_product_attention(
       raise ValueError(f"Unsupported implementation option: {implementation}")
 
   return jnp.reshape(out, output_shape)
+
+def scaled_matmul(
+    lhs: Array,
+    rhs: Array,
+    lhs_scales: Array,
+    rhs_scales: Array,
+    preferred_element_type: DTypeLike = np.float32,
+) -> Array:
+    r"""Scaled matrix multiplication function.
+
+    Performs block-scaled matmul of `a` and `b` using `a_scales` and `b_scales`.
+    The last dim is the contracting dim, and block size is inferred.
+
+    Mathematically, this operation is equivalent to::
+
+      a_block_size = a.shape[-1] // a_scales.shape[-1]
+      b_block_size = b.shape[-1] // b_scales.shape[-1]
+      a_scaled = a * jnp.repeat(a_scales, a_block_size, axis=-1)
+      b_scaled = b * jnp.repeat(b_scales, b_block_size, axis=-1)
+      jnp.einsum('BMK,BNK->BMN', a_scaled, b_scaled)
+
+    Args:
+      lhs (Array): Operand a, shape (B, M, K).
+      rhs (Array): Operand b, shape (B, N, K).
+      lhs_scales (Array): Shape (B, M, K_a), where `K % K_a == 0`.
+      rhs_scales (Array): Shape (B, N, K_b), where `K % K_b == 0`.
+      preferred_element_type (DTypeLike, optional): Defaults to `jnp.float32`.
+
+    Returns:
+      Array of shape (B, M, N).
+
+    Notes:
+      - We currently do not support user-defined `precision` for customizing the
+        compute data type. It is fixed to `jnp.float32`.
+      - Block size is inferred as `K // K_a` for `a` and `K // K_b` for `b`.
+      - To use cuDNN with Nvidia Blackwell GPUs, inputs must match::
+
+          # mxfp8
+          a, b: jnp.float8_e4m3fn | jnp.float8_e5m2
+          a_scales, b_scales: jnp.float8_e8m0fnu
+          block_size: 32
+          # nvfp4
+          a, b: jnp.float4_e2m1fn
+          a_scales, b_scales: jnp.float8_e4m3fn
+          block_size: 16
+
+    Examples:
+
+      Basic case:
+
+      >>> a = jnp.array([1, 2, 3]).reshape((1, 1, 3))
+      >>> b = jnp.array([4, 5, 6]).reshape((1, 1, 3))
+      >>> a_scales = jnp.array([0.5]).reshape((1, 1, 1))
+      >>> b_scales = jnp.array([0.5]).reshape((1, 1, 1))
+      >>> scaled_matmul(a, b, a_scales, b_scales)  # doctest: +SKIP
+      Array([[[8.]]], dtype=float32)
+
+      Using fused cuDNN call on Blackwell GPUs:
+
+      >>> dtype = jnp.float8_e4m3fn
+      >>> a = jax.random.normal(jax.random.PRNGKey(1), (3, 128, 64), dtype=dtype)
+      >>> b = jax.random.normal(jax.random.PRNGKey(2), (3, 128, 64), dtype=dtype)
+      >>> a_scales = jnp.ones((3, 128, 4), dtype=jnp.float8_e8m0fnu)
+      >>> b_scales = jnp.ones((3, 128, 4), dtype=jnp.float8_e8m0fnu)
+      >>> scaled_matmul(a, b, a_scales, b_scales)  # doctest: +SKIP
+    """
+    a, b, a_scales, b_scales = lhs, rhs, lhs_scales, rhs_scales
+    if not all(x.ndim == 3 for x in (a, b, a_scales, b_scales)):
+        raise ValueError(
+            "scaled_matmul requires all inputs to be 3-dimensional arrays"
+        )
+
+    B_a, M_a, K_a = a.shape
+    B_b, N_b, K_b = b.shape
+    if K_a != K_b or B_a != B_b:
+        raise ValueError(
+            "scaled_matmul requires inputs a and b to have matching batch (B) "
+            f"and contract (K) dimensions, but got shapes {a.shape} and "
+            f"{b.shape}"
+        )
+
+    B_as, M_as, K_as = a_scales.shape
+    B_bs, N_bs, K_bs = b_scales.shape
+    if K_as != K_bs or B_as != B_bs:
+        raise ValueError(
+            "scaled_matmul requires scales to have matching batch (B) and "
+            f"contract (K) dimensions, but got shapes {a_scales.shape} and "
+            f"{b_scales.shape}"
+        )
+
+    if M_as != M_a or N_bs != N_b:
+        raise ValueError(
+            "scaled_matmul requires scales to match non-contract dimensions of "
+            f"inputs, but got shapes a: {a.shape}, b: {b.shape}, a_scales: "
+            f"{a_scales.shape}, b_scales: {b_scales.shape}"
+        )
+
+    preferred_element_type = dtypes.canonicalize_dtype(
+        np.dtype(preferred_element_type)
+    )
+    out = cudnn_scaled_matmul(
+        a,
+        b,
+        a_scales,
+        b_scales,
+        preferred_element_type=preferred_element_type,
+    )
+    return out
+
+def get_scaled_dot_general_config(mode: Literal['nvfp4', 'mxfp8'],
+                                  global_scale: Array | None = None):
+    r"""Get quantization configs for scaled_dot_general.
+
+    Create quantization configs for the `jax.nn.scaled_dot_general`.
+
+    See Also:
+      - :func:`jax.nn.scaled_dot_general`: Scaled dot general function.
+    """
+
+    if mode == 'nvfp4':
+        one = jnp.ones((1,), dtype=np.float32)
+        return BlockScaleConfig(
+            mode='nvfp4',
+            block_size=16,
+            data_type=dtypes.float4_e2m1fn,
+            scale_type=dtypes.float8_e4m3fn,
+            global_scale=one if global_scale is None else global_scale,
+            infer_only=False
+        )
+    elif mode == 'mxfp8':
+        return BlockScaleConfig(
+            mode='mxfp8',
+            block_size=32,
+            data_type=dtypes.float8_e4m3fn,
+            scale_type=dtypes.float8_e8m0fnu,
+            global_scale=None,
+            infer_only=False
+        )
+    else:
+        raise ValueError(f"Unsupported mode: {mode}")
+
+def scaled_dot_general(
+    lhs, rhs,
+    dimension_numbers,
+    preferred_element_type=np.float32,
+    configs: list[BlockScaleConfig] | None = None,
+    implementation: Literal['cudnn'] | None = None,
+  ):
+  r"""Scaled dot general operation.
+
+  Performs a generalized dot product with block-scaled quantization on the
+  lhs and rhs inputs. This operation extends `lax.dot_general` to support
+  user-defined scaling configurations.
+
+  Essentially, the operation follows::
+
+      a, a_scales = quantize(lhs, configs[0])
+      b, b_scales = quantize(rhs, configs[1])
+      c = jax.nn.scaled_matmul(a, b, a_scales, b_scales)
+
+  Args:
+    lhs (ArrayLike): Input array.
+    rhs (ArrayLike): Input array.
+    dimension_numbers (DotDimensionNumbers): A tuple of two tuples specifying
+      the contraction and batch dimensions:
+      `((lhs_contracting_dims, rhs_contracting_dims), (lhs_batch_dims, rhs_batch_dims))`.
+    preferred_element_type (DTypeLike, optional): Output data type of the dot
+      product. Defaults to `jnp.float32`. Other valid types include
+      `jnp.bfloat16` and `jnp.float16`.
+    configs (list of BlockScaleConfig, optional): Scaling configurations for
+      lhs, rhs, and gradients. Users can obtain valid configurations via
+      `jax.nn.get_scaled_dot_general_config`. Currently, `nvfp4` and `mxfp8`
+      are supported. If `None`, falls back to `lax.dot_general`.
+    implementation: str
+      (Deprecated) Backend selector, now ignored. The system chooses the backend
+      automatically. Scheduled for removal in future releases.
+
+  Returns:
+    Array: The resulting tensor, with batch dimensions first, followed by
+    non-contracting/non-batch dimensions of lhs, and then those of rhs.
+
+  See Also:
+    - :func:`jax.nn.scaled_matmul`: Scaled matmul function.
+    - :func:`jax.lax.dot_general`: General dot product operator.
+
+  Notes:
+    - Unlike `nn.scaled_matmul`, which assumes quantized low-precision
+      inputs with explicit scaling factors, this operator takes high-precision
+      inputs, applies quantization internally, and handles the backward pass.
+
+  Examples:
+
+    Creating config for mxfp8:
+
+    >>> configs = [jax.nn.get_scaled_dot_general_config('mxfp8')] * 3
+
+    Creating config for nvfp4:
+
+    >>> global_scale = jnp.array([0.5], jnp.float32)
+    >>> configs = [jax.nn.get_scaled_dot_general_config('nvfp4', global_scale)] * 3
+
+    Using scaled_dot_general with the configs:
+
+    >>> import functools
+    >>> scaled_dot_general_fn = functools.partial(jax.nn.scaled_dot_general, configs=configs)
+    >>> lhs = jax.random.normal(jax.random.PRNGKey(1), (3, 128, 64))
+    >>> rhs = jax.random.normal(jax.random.PRNGKey(2), (3, 128, 64))
+    >>> out = scaled_dot_general_fn(lhs, rhs, (((2,), (2,)), ((0,), (0,))))  # doctest: +SKIP
+  """
+  if implementation is not None:
+    warnings.warn("Backend selector, now ignored. The system chooses the "
+                  "backend automatically.", DeprecationWarning)
+
+  if configs is None:
+    return lax.dot_general(lhs, rhs, dimension_numbers,
+                           preferred_element_type=preferred_element_type)
+
+  out = cudnn_scaled_dot_general(
+      lhs, rhs, dimension_numbers,
+      preferred_element_type=preferred_element_type,
+      configs=configs
+  )
+
+  return out
