@@ -22,15 +22,17 @@ from __future__ import annotations
 from collections.abc import Sequence
 import math
 import typing
-from typing import Any, Literal, Protocol
+from typing import Any, Literal, Protocol, TypeAlias
 
 import numpy as np
 
-import jax.numpy as jnp
-from jax import lax
-from jax import random
 from jax._src import core
 from jax._src import dtypes
+from jax._src import numpy as jnp
+from jax._src import random
+from jax._src.named_sharding import NamedSharding
+from jax._src.partition_spec import PartitionSpec
+from jax._src.sharding_impls import canonicalize_sharding
 from jax._src.typing import Array, ArrayLike
 from jax._src.util import set_module
 
@@ -42,20 +44,23 @@ DTypeLikeFloat = Any
 DTypeLikeComplex = Any
 DTypeLikeInexact = Any  # DTypeLikeFloat | DTypeLikeComplex
 RealNumeric = Any  # Scalar jnp array or float
+OutShardingType: TypeAlias = NamedSharding | PartitionSpec | None
 
 @export
 @typing.runtime_checkable
 class Initializer(Protocol):
-  @staticmethod
-  def __call__(key: Array,
+  def __call__(self,
+               key: Array,
                shape: core.Shape,
-               dtype: DTypeLikeInexact = jnp.float_) -> Array:
+               dtype: DTypeLikeInexact = dtypes.float_,
+               out_sharding: OutShardingType = None) -> Array:
     raise NotImplementedError
 
 @export
 def zeros(key: Array,
           shape: core.Shape,
-          dtype: DTypeLikeInexact = jnp.float_) -> Array:
+          dtype: DTypeLikeInexact = dtypes.float_,
+          out_sharding: OutShardingType = None) -> Array:
   """An initializer that returns a constant array full of zeros.
 
   The ``key`` argument is ignored.
@@ -65,12 +70,14 @@ def zeros(key: Array,
   Array([[0., 0., 0.],
          [0., 0., 0.]], dtype=float32)
   """
-  return jnp.zeros(shape, dtypes.canonicalize_dtype(dtype))
+  return jnp.zeros(shape, dtypes.canonicalize_dtype(dtype),
+                   out_sharding=out_sharding)
 
 @export
 def ones(key: Array,
          shape: core.Shape,
-         dtype: DTypeLikeInexact = jnp.float_) -> Array:
+         dtype: DTypeLikeInexact = dtypes.float_,
+         out_sharding: OutShardingType = None) -> Array:
   """An initializer that returns a constant array full of ones.
 
   The ``key`` argument is ignored.
@@ -81,12 +88,12 @@ def ones(key: Array,
          [1., 1.],
          [1., 1.]], dtype=float32)
   """
-  return jnp.ones(shape, dtypes.canonicalize_dtype(dtype))
+  return jnp.ones(shape, dtypes.canonicalize_dtype(dtype),
+                  out_sharding=out_sharding)
 
 @export
 def constant(value: ArrayLike,
-             dtype: DTypeLikeInexact = jnp.float_
-             ) -> Initializer:
+             dtype: DTypeLikeInexact = dtypes.float_) -> Initializer:
   """Builds an initializer that returns arrays full of a constant ``value``.
 
   Args:
@@ -101,14 +108,16 @@ def constant(value: ArrayLike,
   """
   def init(key: Array,
            shape: core.Shape,
-           dtype: DTypeLikeInexact = dtype) -> Array:
+           dtype: DTypeLikeInexact = dtype,
+           out_sharding: OutShardingType = None) -> Array:
     dtype = dtypes.canonicalize_dtype(dtype)
-    return jnp.full(shape, value, dtype=dtype)
+    out_sharding = canonicalize_sharding(out_sharding, 'nn.initializers.constant')
+    return jnp.full(shape, value, dtype=dtype, device=out_sharding)
   return init
 
 @export
 def uniform(scale: RealNumeric = 1e-2,
-            dtype: DTypeLikeInexact = jnp.float_) -> Initializer:
+            dtype: DTypeLikeInexact = dtypes.float_) -> Initializer:
   """Builds an initializer that returns real uniformly-distributed random arrays.
 
   Args:
@@ -127,14 +136,16 @@ def uniform(scale: RealNumeric = 1e-2,
   """
   def init(key: Array,
            shape: core.Shape,
-           dtype: DTypeLikeInexact = dtype) -> Array:
+           dtype: DTypeLikeInexact = dtype,
+           out_sharding: OutShardingType = None) -> Array:
     dtype = dtypes.canonicalize_dtype(dtype)
-    return random.uniform(key, shape, dtype) * jnp.array(scale, dtype)
+    return random.uniform(key, shape, dtype,
+                          out_sharding=out_sharding) * jnp.array(scale, dtype)
   return init
 
 @export
 def normal(stddev: RealNumeric = 1e-2,
-           dtype: DTypeLikeInexact = jnp.float_) -> Initializer:
+           dtype: DTypeLikeInexact = dtypes.float_) -> Initializer:
   """Builds an initializer that returns real normally-distributed random arrays.
 
   Args:
@@ -153,14 +164,16 @@ def normal(stddev: RealNumeric = 1e-2,
   """
   def init(key: Array,
            shape: core.Shape,
-           dtype: DTypeLikeInexact = dtype) -> Array:
+           dtype: DTypeLikeInexact = dtype,
+           out_sharding: OutShardingType = None) -> Array:
     dtype = dtypes.canonicalize_dtype(dtype)
-    return random.normal(key, shape, dtype) * jnp.array(stddev, dtype)
+    return random.normal(key, shape, dtype,
+                         out_sharding=out_sharding) * jnp.array(stddev, dtype)
   return init
 
 @export
 def truncated_normal(stddev: RealNumeric = 1e-2,
-                     dtype: DTypeLikeInexact = jnp.float_,
+                     dtype: DTypeLikeInexact = dtypes.float_,
                      lower: RealNumeric = -2.0,
                      upper: RealNumeric = 2.0) -> Initializer:
   r"""Builds an initializer that returns truncated-normal random arrays.
@@ -190,10 +203,12 @@ def truncated_normal(stddev: RealNumeric = 1e-2,
 
   def init(key: Array,
            shape: core.Shape,
-           dtype: DTypeLikeInexact = dtype) -> Array:
+           dtype: DTypeLikeInexact = dtype,
+           out_sharding: OutShardingType = None) -> Array:
     dtype = dtypes.canonicalize_dtype(dtype)
     return random.truncated_normal(
-        key, lower, upper, shape, dtype) * jnp.array(stddev, dtype)
+        key, lower, upper, shape, dtype,
+        out_sharding=out_sharding) * jnp.array(stddev, dtype)
   return init
 
 @export
@@ -240,7 +255,7 @@ def _complex_uniform(key: Array,
   real_dtype = np.array(0, dtype).real.dtype
   dtype = dtypes.to_complex_dtype(real_dtype)
   r = jnp.sqrt(2 * random.uniform(key_r, shape, real_dtype)).astype(dtype)
-  theta = 2 * jnp.pi * random.uniform(key_theta, shape, real_dtype).astype(dtype)
+  theta = 2 * np.pi * random.uniform(key_theta, shape, real_dtype).astype(dtype)
   return r * jnp.exp(1j * theta)
 
 def _complex_truncated_normal(key: Array, upper: ArrayLike,
@@ -257,19 +272,19 @@ def _complex_truncated_normal(key: Array, upper: ArrayLike,
   t = ((1 - jnp.exp(jnp.array(-(upper ** 2), dtype)))
        * random.uniform(key_r, shape, real_dtype).astype(dtype))
   r = jnp.sqrt(-jnp.log(1 - t))
-  theta = 2 * jnp.pi * random.uniform(key_theta, shape, real_dtype).astype(dtype)
+  theta = 2 * np.pi * random.uniform(key_theta, shape, real_dtype).astype(dtype)
   return r * jnp.exp(1j * theta)
 
 @export
 def variance_scaling(
   scale: RealNumeric,
-  mode: Literal["fan_in"] | Literal["fan_out"] | Literal["fan_avg"],
+  mode: Literal["fan_in"] | Literal["fan_out"] | Literal["fan_avg"] | Literal["fan_geo_avg"],
   distribution: (Literal["truncated_normal"] | Literal["normal"] |
                       Literal["uniform"]),
   in_axis: int | Sequence[int] = -2,
   out_axis: int | Sequence[int] = -1,
-  batch_axis: Sequence[int] = (),
-  dtype: DTypeLikeInexact = jnp.float_
+  batch_axis: int | Sequence[int] = (),
+  dtype: DTypeLikeInexact = dtypes.float_
 ) -> Initializer:
   r"""
   Initializer that adapts its scale to the shape of the weights tensor.
@@ -277,11 +292,12 @@ def variance_scaling(
   With ``distribution="truncated_normal"`` or ``distribution="normal"``, samples
   are drawn from a (truncated) normal distribution with a mean of zero
   and a standard deviation (after truncation, if applicable) of
-  :math:`\sqrt{\frac{scale}{n}}`, where `n` is:
+  :math:`\sqrt{\frac{scale}{n}}`, where `n` is, for each ``mode``:
 
-  * the number of input units in the weights tensor, if ``mode="fan_in"``,
-  * the number of output units, if ``mode="fan_out"``, or
-  * the average of the numbers of input and output units, if ``mode="fan_avg"``.
+  * ``"fan_in"``: the number of inputs
+  * ``"fan_out"``: the number of outputs
+  * ``"fan_avg"``: the arithmetic average of the numbers of inputs and outputs
+  * ``"fan_geo_avg"``: the geometric average of the numbers of inputs and outputs
 
   This initializer can be configured with ``in_axis``, ``out_axis``, and
   ``batch_axis`` to work with general convolutional or dense layers; axes that
@@ -301,7 +317,7 @@ def variance_scaling(
 
   Args:
     scale: scaling factor (positive float).
-    mode: one of ``"fan_in"``, ``"fan_out"``, and ``"fan_avg"``.
+    mode: one of ``"fan_in"``, ``"fan_out"``, ``"fan_avg"``, and ``"fan_geo_avg"``.
     distribution: random distribution to use. One of ``"truncated_normal"``,
       ``"normal"`` and ``"uniform"``.
     in_axis: axis or sequence of axes of the input dimension in the weights
@@ -315,32 +331,37 @@ def variance_scaling(
 
   def init(key: Array,
            shape: core.Shape,
-           dtype: DTypeLikeInexact = dtype) -> Array:
+           dtype: DTypeLikeInexact = dtype,
+           out_sharding: OutShardingType = None) -> Array:
     shape = core.canonicalize_shape(shape)
     dtype = dtypes.canonicalize_dtype(dtype)
     fan_in, fan_out = _compute_fans(shape, in_axis, out_axis, batch_axis)
     if mode == "fan_in": denominator = fan_in
     elif mode == "fan_out": denominator = fan_out
     elif mode == "fan_avg": denominator = (fan_in + fan_out) / 2
+    elif mode == "fan_geo_avg": denominator = (fan_in * fan_out) ** 0.5
     else:
       raise ValueError(
         f"invalid mode for variance scaling initializer: {mode}")
     variance = jnp.array(scale / denominator, dtype=dtype)
 
     if distribution == "truncated_normal":
-      if jnp.issubdtype(dtype, jnp.floating):
+      if dtypes.issubdtype(dtype, np.floating):
         # constant is stddev of standard normal truncated to (-2, 2)
         stddev = jnp.sqrt(variance) / jnp.array(.87962566103423978, dtype)
-        return random.truncated_normal(key, -2, 2, shape, dtype) * stddev
+        return random.truncated_normal(key, -2, 2, shape, dtype,
+                                       out_sharding=out_sharding) * stddev
       else:
         # constant is stddev of complex standard normal truncated to 2
         stddev = jnp.sqrt(variance) / jnp.array(.95311164380491208, dtype)
         return _complex_truncated_normal(key, 2, shape, dtype) * stddev
     elif distribution == "normal":
-      return random.normal(key, shape, dtype) * jnp.sqrt(variance)
+      return random.normal(key, shape, dtype,
+                           out_sharding=out_sharding) * jnp.sqrt(variance)
     elif distribution == "uniform":
-      if jnp.issubdtype(dtype, jnp.floating):
-        return random.uniform(key, shape, dtype, -1) * jnp.sqrt(3 * variance)
+      if dtypes.issubdtype(dtype, np.floating):
+        return random.uniform(key, shape, dtype, -1,
+                              out_sharding=out_sharding) * jnp.sqrt(3 * variance)
       else:
         return _complex_uniform(key, shape, dtype) * jnp.sqrt(variance)
     else:
@@ -351,8 +372,8 @@ def variance_scaling(
 @export
 def glorot_uniform(in_axis: int | Sequence[int] = -2,
                    out_axis: int | Sequence[int] = -1,
-                   batch_axis: Sequence[int] = (),
-                   dtype: DTypeLikeInexact = jnp.float_) -> Initializer:
+                   batch_axis: int | Sequence[int] = (),
+                   dtype: DTypeLikeInexact = dtypes.float_) -> Initializer:
   """Builds a Glorot uniform initializer (aka Xavier uniform initializer).
 
   A `Glorot uniform initializer`_ is a specialization of
@@ -389,8 +410,8 @@ xavier_uniform = glorot_uniform
 @export
 def glorot_normal(in_axis: int | Sequence[int] = -2,
                   out_axis: int | Sequence[int] = -1,
-                  batch_axis: Sequence[int] = (),
-                  dtype: DTypeLikeInexact = jnp.float_) -> Initializer:
+                  batch_axis: int | Sequence[int] = (),
+                  dtype: DTypeLikeInexact = dtypes.float_) -> Initializer:
   """Builds a Glorot normal initializer (aka Xavier normal initializer).
 
   A `Glorot normal initializer`_ is a specialization of
@@ -427,8 +448,8 @@ xavier_normal = glorot_normal
 @export
 def lecun_uniform(in_axis: int | Sequence[int] = -2,
                   out_axis: int | Sequence[int] = -1,
-                  batch_axis: Sequence[int] = (),
-                  dtype: DTypeLikeInexact = jnp.float_) -> Initializer:
+                  batch_axis: int | Sequence[int] = (),
+                  dtype: DTypeLikeInexact = dtypes.float_) -> Initializer:
   """Builds a Lecun uniform initializer.
 
   A `Lecun uniform initializer`_ is a specialization of
@@ -463,8 +484,8 @@ def lecun_uniform(in_axis: int | Sequence[int] = -2,
 @export
 def lecun_normal(in_axis: int | Sequence[int] = -2,
                  out_axis: int | Sequence[int] = -1,
-                 batch_axis: Sequence[int] = (),
-                 dtype: DTypeLikeInexact = jnp.float_) -> Initializer:
+                 batch_axis: int | Sequence[int] = (),
+                 dtype: DTypeLikeInexact = dtypes.float_) -> Initializer:
   """Builds a Lecun normal initializer.
 
   A `Lecun normal initializer`_ is a specialization of
@@ -499,8 +520,8 @@ def lecun_normal(in_axis: int | Sequence[int] = -2,
 @export
 def he_uniform(in_axis: int | Sequence[int] = -2,
                out_axis: int | Sequence[int] = -1,
-               batch_axis: Sequence[int] = (),
-               dtype: DTypeLikeInexact = jnp.float_) -> Initializer:
+               batch_axis: int | Sequence[int] = (),
+               dtype: DTypeLikeInexact = dtypes.float_) -> Initializer:
   """Builds a He uniform initializer (aka Kaiming uniform initializer).
 
   A `He uniform initializer`_ is a specialization of
@@ -537,8 +558,8 @@ kaiming_uniform = he_uniform
 @export
 def he_normal(in_axis: int | Sequence[int] = -2,
               out_axis: int | Sequence[int] = -1,
-              batch_axis: Sequence[int] = (),
-              dtype: DTypeLikeInexact = jnp.float_) -> Initializer:
+              batch_axis: int | Sequence[int] = (),
+              dtype: DTypeLikeInexact = dtypes.float_) -> Initializer:
   """Builds a He normal initializer (aka Kaiming normal initializer).
 
   A `He normal initializer`_ is a specialization of
@@ -575,7 +596,7 @@ kaiming_normal = he_normal
 @export
 def orthogonal(scale: RealNumeric = 1.0,
                column_axis: int = -1,
-               dtype: DTypeLikeInexact = jnp.float_) -> Initializer:
+               dtype: DTypeLikeInexact = dtypes.float_) -> Initializer:
   """
   Builds an initializer that returns uniformly distributed orthogonal matrices.
 
@@ -600,17 +621,15 @@ def orthogonal(scale: RealNumeric = 1.0,
   """
   def init(key: Array,
            shape: core.Shape,
-           dtype: DTypeLikeInexact = dtype) -> Array:
+           dtype: DTypeLikeInexact = dtype,
+           out_sharding: OutShardingType = None) -> Array:
+    if out_sharding is not None:
+      raise NotImplementedError
     dtype = dtypes.canonicalize_dtype(dtype)
     if len(shape) < 2:
       raise ValueError("orthogonal initializer requires at least a 2D shape")
     n_rows, n_cols = math.prod(shape) // shape[column_axis], shape[column_axis]
-    matrix_shape = (n_cols, n_rows) if n_rows < n_cols else (n_rows, n_cols)
-    A = random.normal(key, matrix_shape, dtype)
-    Q, R = jnp.linalg.qr(A)
-    diag_sign = lax.broadcast_to_rank(jnp.sign(jnp.diag(R)), rank=Q.ndim)
-    Q *= diag_sign # needed for a uniform distribution
-    if n_rows < n_cols: Q = Q.T
+    Q = random.orthogonal(key, n_rows, (), dtype, n_cols)
     Q = jnp.reshape(Q, tuple(np.delete(shape, column_axis)) + (shape[column_axis],))
     Q = jnp.moveaxis(Q, -1, column_axis)
     return jnp.array(scale, dtype) * Q
@@ -620,7 +639,7 @@ def orthogonal(scale: RealNumeric = 1.0,
 def delta_orthogonal(
   scale: RealNumeric = 1.0,
   column_axis: int = -1,
-  dtype: DTypeLikeInexact = jnp.float_) -> Initializer:
+  dtype: DTypeLikeInexact = dtypes.float_) -> Initializer:
   """
   Builds an initializer for delta orthogonal kernels.
 
@@ -655,7 +674,10 @@ def delta_orthogonal(
   """
   def init(key: Array,
            shape: core.Shape,
-           dtype: DTypeLikeInexact = dtype) -> Array:
+           dtype: DTypeLikeInexact = dtype,
+           out_sharding: OutShardingType = None) -> Array:
+    if out_sharding is not None:
+      raise NotImplementedError
     dtype = dtypes.canonicalize_dtype(dtype)
     if len(shape) not in [3, 4, 5]:
       raise ValueError("Delta orthogonal initializer requires a 3D, 4D or 5D "
