@@ -18,22 +18,25 @@ from collections.abc import Sequence
 from functools import partial
 import itertools
 import math
-
-import numpy as np
 import operator
 from typing import Literal, NamedTuple, overload
 
-import jax
-from jax import jit, custom_jvp
-from jax import lax
+import numpy as np
 
+from jax._src import api
+from jax._src import core
+from jax._src import config
 from jax._src import deprecations
-from jax._src.lax import lax as lax_internal
-from jax._src.lax.lax import PrecisionLike
+from jax._src.custom_derivatives import custom_jvp
+from jax._src.lax import lax
 from jax._src.lax import linalg as lax_linalg
+from jax._src.numpy import array_creation
+from jax._src.numpy import einsum
+from jax._src.numpy import indexing
 from jax._src.numpy import lax_numpy as jnp
-from jax._src.numpy import reductions, ufuncs
-from jax._src.numpy.util import promote_dtypes_inexact, check_arraylike
+from jax._src.sharding_impls import NamedSharding, PartitionSpec as P
+from jax._src.numpy import reductions, tensor_contractions, ufuncs
+from jax._src.numpy.util import promote_dtypes_inexact, ensure_arraylike
 from jax._src.util import canonicalize_axis, set_module
 from jax._src.typing import ArrayLike, Array, DTypeLike, DeprecatedArg
 
@@ -42,24 +45,24 @@ export = set_module('jax.numpy.linalg')
 
 
 class EighResult(NamedTuple):
-  eigenvalues: jax.Array
-  eigenvectors: jax.Array
+  eigenvalues: Array
+  eigenvectors: Array
 
 
 class QRResult(NamedTuple):
-  Q: jax.Array
-  R: jax.Array
+  Q: Array
+  R: Array
 
 
 class SlogdetResult(NamedTuple):
-  sign: jax.Array
-  logabsdet: jax.Array
+  sign: Array
+  logabsdet: Array
 
 
 class SVDResult(NamedTuple):
-  U: jax.Array
-  S: jax.Array
-  Vh: jax.Array
+  U: Array
+  S: Array
+  Vh: Array
 
 
 def _H(x: ArrayLike) -> Array:
@@ -70,8 +73,8 @@ def _symmetrize(x: Array) -> Array: return (x + _H(x)) / 2
 
 
 @export
-@partial(jit, static_argnames=['upper'])
-def cholesky(a: ArrayLike, *, upper: bool = False) -> Array:
+@partial(api.jit, static_argnames=['upper', 'symmetrize_input'])
+def cholesky(a: ArrayLike, *, upper: bool = False, symmetrize_input: bool = True) -> Array:
   """Compute the Cholesky decomposition of a matrix.
 
   JAX implementation of :func:`numpy.linalg.cholesky`.
@@ -96,6 +99,10 @@ def cholesky(a: ArrayLike, *, upper: bool = False) -> Array:
       Must have shape ``(..., N, N)``.
     upper: if True, compute the upper Cholesky decomposition `U`. if False
       (default), compute the lower Cholesky decomposition `L`.
+    symmetrize_input: if True (default) then input is symmetrized, which leads
+      to better behavior under automatic differentiation. Note that when this
+      is set to True, both the upper and lower triangles of the input will
+      be used in computing the decomposition.
 
   Returns:
     array of shape ``(..., N, N)`` representing the Cholesky decomposition
@@ -131,9 +138,9 @@ def cholesky(a: ArrayLike, *, upper: bool = False) -> Array:
     >>> jnp.allclose(x, L @ L.T)
     Array(True, dtype=bool)
   """
-  check_arraylike("jnp.linalg.cholesky", a)
-  a, = promote_dtypes_inexact(jnp.asarray(a))
-  L = lax_linalg.cholesky(a)
+  a = ensure_arraylike("jnp.linalg.cholesky", a)
+  a, = promote_dtypes_inexact(a)
+  L = lax_linalg.cholesky(a, symmetrize_input=symmetrize_input)
   return L.mT.conj() if upper else L
 
 
@@ -196,7 +203,7 @@ def svd(
 
 @export
 @partial(
-    jit,
+    api.jit,
     static_argnames=(
         "full_matrices",
         "compute_uv",
@@ -280,8 +287,8 @@ def svd(
     >>> jnp.allclose(x_reconstructed, x)
     Array(True, dtype=bool)
   """
-  check_arraylike("jnp.linalg.svd", a)
-  a, = promote_dtypes_inexact(jnp.asarray(a))
+  a = ensure_arraylike("jnp.linalg.svd", a)
+  a, = promote_dtypes_inexact(a)
   if hermitian:
     w, v = lax_linalg.eigh(a, subset_by_index=subset_by_index)
     s = lax.abs(v)
@@ -292,7 +299,7 @@ def svd(
       s = lax.rev(s, dimensions=[s.ndim - 1])
       idxs = lax.rev(idxs, dimensions=[s.ndim - 1])
       sign = lax.rev(sign, dimensions=[s.ndim - 1])
-      u = jnp.take_along_axis(w, idxs[..., None, :], axis=-1)
+      u = indexing.take_along_axis(w, idxs[..., None, :], axis=-1)
       vh = _H(u * sign[..., None, :].astype(u.dtype))
       return SVDResult(u, s, vh)
     else:
@@ -316,7 +323,7 @@ def svd(
 
 
 @export
-@partial(jit, static_argnames=('n',))
+@partial(api.jit, static_argnames=('n',))
 def matrix_power(a: ArrayLike, n: int) -> Array:
   """Raise a square matrix to an integer power.
 
@@ -361,8 +368,7 @@ def matrix_power(a: ArrayLike, n: int) -> Array:
     Array([[ 5.5 , -2.5 ],
            [-3.75,  1.75]], dtype=float32)
   """
-  check_arraylike("jnp.linalg.matrix_power", a)
-  arr, = promote_dtypes_inexact(jnp.asarray(a))
+  arr = ensure_arraylike("jnp.linalg.matrix_power", a)
 
   if arr.ndim < 2:
     raise TypeError("{}-dimensional array given. Array must be at least "
@@ -398,7 +404,7 @@ def matrix_power(a: ArrayLike, n: int) -> Array:
 
 
 @export
-@jit
+@api.jit
 def matrix_rank(
   M: ArrayLike, rtol: ArrayLike | None = None, *,
   tol: ArrayLike | DeprecatedArg | None = DeprecatedArg()) -> Array:
@@ -437,7 +443,7 @@ def matrix_rank(
     >>> jnp.linalg.matrix_rank(b)
     Array(1, dtype=int32)
   """
-  check_arraylike("jnp.linalg.matrix_rank", M)
+  M = ensure_arraylike("jnp.linalg.matrix_rank", M)
   # TODO(micky774): deprecated 2024-5-14, remove after deprecation expires.
   if not isinstance(tol, DeprecatedArg):
     rtol = tol
@@ -448,9 +454,9 @@ def matrix_rank(
        "Please use rtol instead."),
       stacklevel=2
     )
-  M, = promote_dtypes_inexact(jnp.asarray(M))
+  M, = promote_dtypes_inexact(M)
   if M.ndim < 2:
-    return (M != 0).any().astype(jnp.int32)
+    return (M != 0).any().astype(np.int32)
   S = svd(M, full_matrices=False, compute_uv=False)
   if rtol is None:
     rtol = S.max(-1) * np.max(M.shape[-2:]).astype(S.dtype) * jnp.finfo(S.dtype).eps
@@ -476,7 +482,7 @@ def _slogdet_lu(a: Array) -> tuple[Array, Array]:
                   jnp.array(0, dtype=dtype),
                   sign * jnp.array(-2 * (parity % 2) + 1, dtype=dtype))
   logdet = jnp.where(
-      is_zero, jnp.array(-jnp.inf, dtype=dtype),
+      is_zero, jnp.array(-np.inf, dtype=dtype),
       reductions.sum(ufuncs.log(ufuncs.abs(diag)).astype(dtype), axis=-1))
   return sign, ufuncs.real(logdet)
 
@@ -485,7 +491,7 @@ def _slogdet_qr(a: Array) -> tuple[Array, Array]:
   # Implementation of slogdet using QR decomposition. One reason we might prefer
   # QR decomposition is that it is more amenable to a fast batched
   # implementation on TPU because of the lack of row pivoting.
-  if jnp.issubdtype(lax.dtype(a), jnp.complexfloating):
+  if jnp.issubdtype(lax.dtype(a), np.complexfloating):
     raise NotImplementedError("slogdet method='qr' not implemented for complex "
                               "inputs")
   n = a.shape[-1]
@@ -503,7 +509,7 @@ def _slogdet_qr(a: Array) -> tuple[Array, Array]:
 
 
 @export
-@partial(jit, static_argnames=('method',))
+@partial(api.jit, static_argnames=('method',))
 def slogdet(a: ArrayLike, *, method: str | None = None) -> SlogdetResult:
   """
   Compute the sign and (natural) logarithm of the determinant of an array.
@@ -535,9 +541,9 @@ def slogdet(a: ArrayLike, *, method: str | None = None) -> SlogdetResult:
     >>> jnp.exp(logabsdet)  # Absolute value of determinant
     Array(2., dtype=float32)
   """
-  check_arraylike("jnp.linalg.slogdet", a)
-  a, = promote_dtypes_inexact(jnp.asarray(a))
-  a_shape = jnp.shape(a)
+  a = ensure_arraylike("jnp.linalg.slogdet", a)
+  a, = promote_dtypes_inexact(a)
+  a_shape = np.shape(a)
   if len(a_shape) < 2 or a_shape[-1] != a_shape[-2]:
     raise ValueError(f"Argument to slogdet() must have shape [..., n, n], got {a_shape}")
   if method is None or method == "lu":
@@ -553,11 +559,11 @@ def _slogdet_jvp(primals, tangents):
   g, = tangents
   sign, ans = slogdet(x)
   ans_dot = jnp.trace(solve(x, g), axis1=-1, axis2=-2)
-  if jnp.issubdtype(jnp._dtype(x), jnp.complexfloating):
+  if jnp.issubdtype(jnp._dtype(x), np.complexfloating):
     sign_dot = (ans_dot - ufuncs.real(ans_dot).astype(ans_dot.dtype)) * sign
     ans_dot = ufuncs.real(ans_dot)
   else:
-    sign_dot = jnp.zeros_like(sign)
+    sign_dot = array_creation.zeros_like(sign)
   return (sign, ans), (sign_dot, ans_dot)
 
 _slogdet_lu.defjvp(_slogdet_jvp)
@@ -605,10 +611,11 @@ def _cofactor_solve(a: ArrayLike, b: ArrayLike) -> tuple[Array, Array]:
   Returns:
     det(a) and cofactor(a)^T*b, aka adjugate(a)*b
   """
-  a, = promote_dtypes_inexact(jnp.asarray(a))
-  b, = promote_dtypes_inexact(jnp.asarray(b))
-  a_shape = jnp.shape(a)
-  b_shape = jnp.shape(b)
+  a, b = ensure_arraylike("jnp.linalg._cofactor_solve", a, b)
+  a, = promote_dtypes_inexact(a)
+  b, = promote_dtypes_inexact(b)
+  a_shape = np.shape(a)
+  b_shape = np.shape(b)
   a_ndims = len(a_shape)
   if not (a_ndims >= 2 and a_shape[-1] == a_shape[-2]
     and b_shape[-2:] == a_shape[-2:]):
@@ -636,20 +643,20 @@ def _cofactor_solve(a: ArrayLike, b: ArrayLike) -> tuple[Array, Array]:
   partial_det = reductions.cumprod(diag, axis=-1) * sign[..., None]
   lu = lu.at[..., -1, -1].set(1.0 / partial_det[..., -2])
   permutation = jnp.broadcast_to(permutation, (*batch_dims, a_shape[-1]))
-  iotas = jnp.ix_(*(lax.iota(jnp.int32, b) for b in (*batch_dims, 1)))
+  iotas = jnp.ix_(*(lax.iota(np.int32, b) for b in (*batch_dims, 1)))
   # filter out any matrices that are not full rank
-  d = jnp.ones(x.shape[:-1], x.dtype)
+  d = array_creation.ones(x.shape[:-1], x.dtype)
   d = lax_linalg.triangular_solve(lu, d, left_side=True, lower=False)
   d = reductions.any(ufuncs.logical_or(ufuncs.isnan(d), ufuncs.isinf(d)), axis=-1)
   d = jnp.tile(d[..., None, None], d.ndim*(1,) + x.shape[-2:])
-  x = jnp.where(d, jnp.zeros_like(x), x)  # first filter
+  x = jnp.where(d, array_creation.zeros_like(x), x)  # first filter
   x = x[iotas[:-1] + (permutation, slice(None))]
   x = lax_linalg.triangular_solve(lu, x, left_side=True, lower=True,
                                   unit_diagonal=True)
   x = jnp.concatenate((x[..., :-1, :] * partial_det[..., -1, None, None],
                       x[..., -1:, :]), axis=-2)
   x = lax_linalg.triangular_solve(lu, x, left_side=True, lower=False)
-  x = jnp.where(d, jnp.zeros_like(x), x)  # second filter
+  x = jnp.where(d, array_creation.zeros_like(x), x)  # second filter
 
   return partial_det[..., -1], x
 
@@ -683,7 +690,7 @@ def _det_jvp(primals, tangents):
 
 
 @export
-@jit
+@api.jit
 def det(a: ArrayLike) -> Array:
   """
   Compute the determinant of an array.
@@ -705,9 +712,9 @@ def det(a: ArrayLike) -> Array:
     >>> jnp.linalg.det(a)
     Array(-2., dtype=float32)
   """
-  check_arraylike("jnp.linalg.det", a)
-  a, = promote_dtypes_inexact(jnp.asarray(a))
-  a_shape = jnp.shape(a)
+  a = ensure_arraylike("jnp.linalg.det", a)
+  a, = promote_dtypes_inexact(a)
+  a_shape = np.shape(a)
   if len(a_shape) >= 2 and a_shape[-1] == 2 and a_shape[-2] == 2:
     return _det_2x2(a)
   elif len(a_shape) >= 2 and a_shape[-1] == 3 and a_shape[-2] == 3:
@@ -759,14 +766,14 @@ def eig(a: ArrayLike) -> tuple[Array, Array]:
     Array([[ 0.70710677+0.j, -0.70710677+0.j],
            [ 0.70710677+0.j,  0.70710677+0.j]], dtype=complex64)
   """
-  check_arraylike("jnp.linalg.eig", a)
-  a, = promote_dtypes_inexact(jnp.asarray(a))
+  a = ensure_arraylike("jnp.linalg.eig", a)
+  a, = promote_dtypes_inexact(a)
   w, v = lax_linalg.eig(a, compute_left_eigenvectors=False)
   return w, v
 
 
 @export
-@jit
+@api.jit
 def eigvals(a: ArrayLike) -> Array:
   """
   Compute the eigenvalues of a general matrix.
@@ -797,14 +804,14 @@ def eigvals(a: ArrayLike) -> Array:
     ...  w
     Array([ 3.+0.j, -1.+0.j], dtype=complex64)
   """
-  check_arraylike("jnp.linalg.eigvals", a)
-  a, = promote_dtypes_inexact(jnp.asarray(a))
+  a = ensure_arraylike("jnp.linalg.eigvals", a)
+  a, = promote_dtypes_inexact(a)
   return lax_linalg.eig(a, compute_left_eigenvectors=False,
                         compute_right_eigenvectors=False)[0]
 
 
 @export
-@partial(jit, static_argnames=('UPLO', 'symmetrize_input'))
+@partial(api.jit, static_argnames=('UPLO', 'symmetrize_input'))
 def eigh(a: ArrayLike, UPLO: str | None = None,
          symmetrize_input: bool = True) -> EighResult:
   """
@@ -818,7 +825,9 @@ def eigh(a: ArrayLike, UPLO: str | None = None,
     UPLO: specifies whether the calculation is done with the lower triangular
       part of ``a`` (``'L'``, default) or the upper triangular part (``'U'``).
     symmetrize_input: if True (default) then input is symmetrized, which leads
-      to better behavior under automatic differentiation.
+      to better behavior under automatic differentiation. Note that when this
+      is set to True, both the upper and lower triangles of the input will
+      be used in computing the decomposition.
 
   Returns:
     A namedtuple ``(eigenvalues, eigenvectors)`` where
@@ -845,7 +854,7 @@ def eigh(a: ArrayLike, UPLO: str | None = None,
     Array([[-0.707+0.j   , -0.707+0.j   ],
            [ 0.   +0.707j,  0.   -0.707j]], dtype=complex64)
   """
-  check_arraylike("jnp.linalg.eigh", a)
+  a = ensure_arraylike("jnp.linalg.eigh", a)
   if UPLO is None or UPLO == "L":
     lower = True
   elif UPLO == "U":
@@ -854,14 +863,15 @@ def eigh(a: ArrayLike, UPLO: str | None = None,
     msg = f"UPLO must be one of None, 'L', or 'U', got {UPLO}"
     raise ValueError(msg)
 
-  a, = promote_dtypes_inexact(jnp.asarray(a))
+  a, = promote_dtypes_inexact(a)
   v, w = lax_linalg.eigh(a, lower=lower, symmetrize_input=symmetrize_input)
   return EighResult(w, v)
 
 
 @export
-@partial(jit, static_argnames=('UPLO',))
-def eigvalsh(a: ArrayLike, UPLO: str | None = 'L') -> Array:
+@partial(api.jit, static_argnames=('UPLO', 'symmetrize_input'))
+def eigvalsh(a: ArrayLike, UPLO: str | None = 'L', *,
+             symmetrize_input: bool = True) -> Array:
   """
   Compute the eigenvalues of a Hermitian matrix.
 
@@ -872,6 +882,10 @@ def eigvalsh(a: ArrayLike, UPLO: str | None = 'L') -> Array:
       or symmetric (if real) matrix.
     UPLO: specifies whether the calculation is done with the lower triangular
       part of ``a`` (``'L'``, default) or the upper triangular part (``'U'``).
+    symmetrize_input: if True (default) then input is symmetrized, which leads
+      to better behavior under automatic differentiation. Note that when this
+      is set to True, both the upper and lower triangles of the input will
+      be used in computing the decomposition.
 
   Returns:
     An array of shape ``(..., M)`` containing the eigenvalues, sorted in
@@ -889,9 +903,9 @@ def eigvalsh(a: ArrayLike, UPLO: str | None = 'L') -> Array:
     >>> w
     Array([-1.,  3.], dtype=float32)
   """
-  check_arraylike("jnp.linalg.eigvalsh", a)
-  a, = promote_dtypes_inexact(jnp.asarray(a))
-  w, _ = eigh(a, UPLO)
+  a = ensure_arraylike("jnp.linalg.eigvalsh", a)
+  a, = promote_dtypes_inexact(a)
+  w, _ = eigh(a, UPLO, symmetrize_input=symmetrize_input)
   return w
 
 
@@ -956,15 +970,15 @@ def pinv(a: ArrayLike, rtol: ArrayLike | None = None,
 
 
 @partial(custom_jvp, nondiff_argnums=(1, 2))
-@partial(jit, static_argnames=('hermitian'))
+@partial(api.jit, static_argnames=('hermitian'))
 def _pinv(a: ArrayLike, rtol: ArrayLike | None = None, hermitian: bool = False) -> Array:
   # Uses same algorithm as
   # https://github.com/numpy/numpy/blob/v1.17.0/numpy/linalg/linalg.py#L1890-L1979
-  check_arraylike("jnp.linalg.pinv", a)
-  arr, = promote_dtypes_inexact(jnp.asarray(a))
+  a = ensure_arraylike("jnp.linalg.pinv", a)
+  arr, = promote_dtypes_inexact(a)
   m, n = arr.shape[-2:]
   if m == 0 or n == 0:
-    return jnp.empty(arr.shape[:-2] + (n, m), arr.dtype)
+    return array_creation.empty(arr.shape[:-2] + (n, m), arr.dtype)
   arr = ufuncs.conj(arr)
   if rtol is None:
     max_rows_cols = max(arr.shape[-2:])
@@ -973,16 +987,16 @@ def _pinv(a: ArrayLike, rtol: ArrayLike | None = None, hermitian: bool = False)
   u, s, vh = svd(arr, full_matrices=False, hermitian=hermitian)
   # Singular values less than or equal to ``rtol * largest_singular_value``
   # are set to zero.
-  rtol = lax.expand_dims(rtol[..., jnp.newaxis], range(s.ndim - rtol.ndim - 1))
+  rtol = lax.expand_dims(rtol[..., np.newaxis], range(s.ndim - rtol.ndim - 1))
   cutoff = rtol * s[..., 0:1]
-  s = jnp.where(s > cutoff, s, jnp.inf).astype(u.dtype)
-  res = jnp.matmul(vh.mT, ufuncs.divide(u.mT, s[..., jnp.newaxis]),
-                   precision=lax.Precision.HIGHEST)
+  s = jnp.where(s > cutoff, s, np.inf).astype(u.dtype)
+  res = tensor_contractions.matmul(vh.mT, ufuncs.divide(u.mT, s[..., np.newaxis]),
+                                   precision=lax.Precision.HIGHEST)
   return lax.convert_element_type(res, arr.dtype)
 
 
 @_pinv.defjvp
-@jax.default_matmul_precision("float32")
+@config.default_matmul_precision("float32")
 def _pinv_jvp(rtol, hermitian, primals, tangents):
   # The Differentiation of Pseudo-Inverses and Nonlinear Least Squares Problems
   # Whose Variables Separate. Author(s): G. H. Golub and V. Pereyra. SIAM
@@ -1011,7 +1025,7 @@ def _pinv_jvp(rtol, hermitian, primals, tangents):
 
 
 @export
-@jit
+@api.jit
 def inv(a: ArrayLike) -> Array:
   """Return the inverse of a square matrix
 
@@ -1062,8 +1076,7 @@ def inv(a: ArrayLike) -> Array:
     >>> jnp.linalg.solve(a, b)
      Array([ 0.  ,  1.25, -0.5 ], dtype=float32)
   """
-  check_arraylike("jnp.linalg.inv", a)
-  arr = jnp.asarray(a)
+  arr = ensure_arraylike("jnp.linalg.inv", a)
   if arr.ndim < 2 or arr.shape[-1] != arr.shape[-2]:
     raise ValueError(
       f"Argument to inv must have shape [..., n, n], got {arr.shape}.")
@@ -1072,7 +1085,7 @@ def inv(a: ArrayLike) -> Array:
 
 
 @export
-@partial(jit, static_argnames=('ord', 'axis', 'keepdims'))
+@partial(api.jit, static_argnames=('ord', 'axis', 'keepdims'))
 def norm(x: ArrayLike, ord: int | str | None = None,
          axis: None | tuple[int, ...] | int = None,
          keepdims: bool = False) -> Array:
@@ -1085,7 +1098,8 @@ def norm(x: ArrayLike, ord: int | str | None = None,
     ord: specify the kind of norm to take. Default is Frobenius norm for matrices,
       and the 2-norm for vectors. For other options, see Notes below.
     axis: integer or sequence of integers specifying the axes over which the norm
-      will be computed. Defaults to all axes of ``x``.
+      will be computed. For a single axis, compute a vector norm. For two axes,
+      compute a matrix norm. Defaults to all axes of ``x``.
     keepdims: if True, the output array will have the same number of dimensions as
       the input, with the size of reduced axes replaced by ``1`` (default: False).
 
@@ -1113,6 +1127,9 @@ def norm(x: ArrayLike, ord: int | str | None = None,
     - ``ord=2`` computes the 2-norm, i.e. the largest singular value
     - ``ord=-2`` computes the smallest singular value
 
+    In the special case of ``ord=None`` and ``axis=None``, this function accepts an
+    array of any dimension and computes the vector 2-norm of the flattened array.
+
   Examples:
     Vector norms:
 
@@ -1140,9 +1157,9 @@ def norm(x: ArrayLike, ord: int | str | None = None,
     >>> jnp.linalg.norm(x, axis=1)
     Array([3.7416575, 9.486833 ], dtype=float32)
   """
-  check_arraylike("jnp.linalg.norm", x)
-  x, = promote_dtypes_inexact(jnp.asarray(x))
-  x_shape = jnp.shape(x)
+  x = ensure_arraylike("jnp.linalg.norm", x)
+  x, = promote_dtypes_inexact(x)
+  x_shape = np.shape(x)
   ndim = len(x_shape)
 
   if axis is None:
@@ -1169,40 +1186,39 @@ def norm(x: ArrayLike, ord: int | str | None = None,
       if not keepdims and col_axis > row_axis:
         col_axis -= 1
       return reductions.amax(reductions.sum(ufuncs.abs(x), axis=row_axis, keepdims=keepdims),
-                             axis=col_axis, keepdims=keepdims)
+                             axis=col_axis, keepdims=keepdims, initial=0)
     elif ord == -1:
       if not keepdims and col_axis > row_axis:
         col_axis -= 1
       return reductions.amin(reductions.sum(ufuncs.abs(x), axis=row_axis, keepdims=keepdims),
                              axis=col_axis, keepdims=keepdims)
-    elif ord == jnp.inf:
+    elif ord == np.inf:
       if not keepdims and row_axis > col_axis:
         row_axis -= 1
       return reductions.amax(reductions.sum(ufuncs.abs(x), axis=col_axis, keepdims=keepdims),
-                     axis=row_axis, keepdims=keepdims)
-    elif ord == -jnp.inf:
+                     axis=row_axis, keepdims=keepdims, initial=0)
+    elif ord == -np.inf:
       if not keepdims and row_axis > col_axis:
         row_axis -= 1
       return reductions.amin(reductions.sum(ufuncs.abs(x), axis=col_axis, keepdims=keepdims),
                      axis=row_axis, keepdims=keepdims)
     elif ord in ('nuc', 2, -2):
       x = jnp.moveaxis(x, axis, (-2, -1))
+      s = svd(x, compute_uv=False)
       if ord == 2:
-        reducer = reductions.amax
+        y = reductions.amax(s, axis=-1, initial=0)
       elif ord == -2:
-        reducer = reductions.amin
+        y = reductions.amin(s, axis=-1)
       else:
-        # `sum` takes an extra dtype= argument, unlike `amax` and `amin`.
-        reducer = reductions.sum  # type: ignore[assignment]
-      y = reducer(svd(x, compute_uv=False), axis=-1)
+        y = reductions.sum(s, axis=-1)
       if keepdims:
         y = jnp.expand_dims(y, axis)
       return y
     else:
       raise ValueError(f"Invalid order '{ord}' for matrix norm.")
   else:
-    raise ValueError(
-        f"Invalid axis values ({axis}) for jnp.linalg.norm.")
+    raise ValueError(f"Improper number of axes for norm: {axis=}. Pass one axis to"
+                     " compute a vector-norm, or two axes to compute a matrix-norm.")
 
 @overload
 def qr(a: ArrayLike, mode: Literal["r"]) -> Array: ...
@@ -1210,7 +1226,7 @@ def qr(a: ArrayLike, mode: Literal["r"]) -> Array: ...
 def qr(a: ArrayLike, mode: str = "reduced") -> Array | QRResult: ...
 
 @export
-@partial(jit, static_argnames=('mode',))
+@partial(api.jit, static_argnames=('mode',))
 def qr(a: ArrayLike, mode: str = "reduced") -> Array | QRResult:
   """Compute the QR decomposition of an array
 
@@ -1276,8 +1292,8 @@ def qr(a: ArrayLike, mode: str = "reduced") -> Array | QRResult:
     >>> jnp.allclose(Q @ R, a)
     Array(True, dtype=bool)
   """
-  check_arraylike("jnp.linalg.qr", a)
-  a, = promote_dtypes_inexact(jnp.asarray(a))
+  a = ensure_arraylike("jnp.linalg.qr", a)
+  a, = promote_dtypes_inexact(a)
   if mode == "raw":
     a, taus = lax_linalg.geqrf(a)
     return QRResult(a.mT, taus)
@@ -1294,23 +1310,27 @@ def qr(a: ArrayLike, mode: str = "reduced") -> Array | QRResult:
 
 
 @export
-@jit
+@api.jit
 def solve(a: ArrayLike, b: ArrayLike) -> Array:
-  """Solve a linear system of equations
+  """Solve a linear system of equations.
 
   JAX implementation of :func:`numpy.linalg.solve`.
 
   This solves a (batched) linear system of equations ``a @ x = b``
   for ``x`` given ``a`` and ``b``.
 
+  If ``a`` is singular, this will return ``nan`` or ``inf`` values.
+
   Args:
     a: array of shape ``(..., N, N)``.
     b: array of shape ``(N,)`` (for 1-dimensional right-hand-side) or
       ``(..., N, M)`` (for batched 2-dimensional right-hand-side).
 
   Returns:
-    An array containing the result of the linear solve. The result has shape ``(..., N)``
-    if ``b`` is of shape ``(N,)``, and has shape ``(..., N, M)`` otherwise.
+    An array containing the result of the linear solve if ``a`` is non-singular.
+    The result has shape ``(..., N)`` if ``b`` is of shape ``(N,)``, and has
+    shape ``(..., N, M)`` otherwise.
+    If ``a`` is singular, the result contains ``nan`` or ``inf`` values.
 
   See also:
     - :func:`jax.scipy.linalg.solve`: SciPy-style API for solving linear systems.
@@ -1332,8 +1352,8 @@ def solve(a: ArrayLike, b: ArrayLike) -> Array:
     >>> jnp.allclose(A @ x, b)
     Array(True, dtype=bool)
   """
-  check_arraylike("jnp.linalg.solve", a, b)
-  a, b = promote_dtypes_inexact(jnp.asarray(a), jnp.asarray(b))
+  a, b = ensure_arraylike("jnp.linalg.solve", a, b)
+  a, b = promote_dtypes_inexact(a, b)
 
   if a.ndim < 2:
     raise ValueError(
@@ -1348,6 +1368,7 @@ def solve(a: ArrayLike, b: ArrayLike) -> Array:
       " To recover this behavior, use solve(a, b[..., None]).squeeze(-1).")
 
   signature = "(m,m),(m)->(m)" if b.ndim == 1 else "(m,m),(m,n)->(m,n)"
+  a, b = core.standard_insert_pvary(a, b)
   return jnp.vectorize(lax_linalg._solve, signature=signature)(a, b)
 
 
@@ -1370,9 +1391,9 @@ def _lstsq(a: ArrayLike, b: ArrayLike, rcond: float | None, *,
   m, n = a.shape
   dtype = a.dtype
   if a.size == 0:
-    s = jnp.empty(0, dtype=a.dtype)
+    s = array_creation.empty(0, dtype=a.dtype)
     rank = jnp.array(0, dtype=int)
-    x = jnp.empty((n, *b.shape[1:]), dtype=a.dtype)
+    x = array_creation.empty((n, *b.shape[1:]), dtype=a.dtype)
   else:
     if rcond is None:
       rcond = float(jnp.finfo(dtype).eps) * max(n, m)
@@ -1382,21 +1403,21 @@ def _lstsq(a: ArrayLike, b: ArrayLike, rcond: float | None, *,
     mask = s >= jnp.array(rcond, dtype=s.dtype) * s[0]
     rank = mask.sum()
     safe_s = jnp.where(mask, s, 1).astype(a.dtype)
-    s_inv = jnp.where(mask, 1 / safe_s, 0)[:, jnp.newaxis]
-    uTb = jnp.matmul(u.conj().T, b, precision=lax.Precision.HIGHEST)
-    x = jnp.matmul(vt.conj().T, s_inv * uTb, precision=lax.Precision.HIGHEST)
+    s_inv = jnp.where(mask, 1 / safe_s, 0)[:, np.newaxis]
+    uTb = tensor_contractions.matmul(u.conj().T, b, precision=lax.Precision.HIGHEST)
+    x = tensor_contractions.matmul(vt.conj().T, s_inv * uTb, precision=lax.Precision.HIGHEST)
   # Numpy returns empty residuals in some cases. To allow compilation, we
   # default to returning full residuals in all cases.
   if numpy_resid and (rank < n or m <= n):
     resid = jnp.asarray([])
   else:
-    b_estimate = jnp.matmul(a, x, precision=lax.Precision.HIGHEST)
+    b_estimate = tensor_contractions.matmul(a, x, precision=lax.Precision.HIGHEST)
     resid = norm(b - b_estimate, axis=0) ** 2
   if b_orig_ndim == 1:
     x = x.ravel()
   return x, resid, rank, s
 
-_jit_lstsq = jit(partial(_lstsq, numpy_resid=False))
+_jit_lstsq = api.jit(partial(_lstsq, numpy_resid=False))
 
 
 @export
@@ -1434,7 +1455,7 @@ def lstsq(a: ArrayLike, b: ArrayLike, rcond: float | None = None, *,
     ...   print(x)
     [-4.   4.5]
   """
-  check_arraylike("jnp.linalg.lstsq", a, b)
+  a, b = ensure_arraylike("jnp.linalg.lstsq", a, b)
   if numpy_resid:
     return _lstsq(a, b, rcond, numpy_resid=True)
   return _jit_lstsq(a, b, rcond)
@@ -1476,8 +1497,7 @@ def cross(x1: ArrayLike, x2: ArrayLike, /, *, axis=-1):
            [ 0.,  0.,  1.],
            [ 0., -1.,  0.]], dtype=float32)
   """
-  check_arraylike("jnp.linalg.outer", x1, x2)
-  x1, x2 = jnp.asarray(x1), jnp.asarray(x2)
+  x1, x2 = ensure_arraylike("jnp.linalg.outer", x1, x2)
   if x1.shape[axis] != 3 or x2.shape[axis] != 3:
     raise ValueError(
         "Both input arrays must be (arrays of) 3-dimensional vectors, "
@@ -1510,8 +1530,7 @@ def outer(x1: ArrayLike, x2: ArrayLike, /) -> Array:
            [ 8, 10, 12],
            [12, 15, 18]], dtype=int32)
   """
-  check_arraylike("jnp.linalg.outer", x1, x2)
-  x1, x2 = jnp.asarray(x1), jnp.asarray(x2)
+  x1, x2 = ensure_arraylike("jnp.linalg.outer", x1, x2)
   if x1.ndim != 1 or x2.ndim != 1:
     raise ValueError(f"Input arrays must be one-dimensional, but they are {x1.ndim=} {x2.ndim=}")
   return x1[:, None] * x2[None, :]
@@ -1544,7 +1563,7 @@ def matrix_norm(x: ArrayLike, /, *, keepdims: bool = False, ord: str | int = 'fr
     >>> jnp.linalg.matrix_norm(x)
     Array(16.881943, dtype=float32)
   """
-  check_arraylike('jnp.linalg.matrix_norm', x)
+  x = ensure_arraylike('jnp.linalg.matrix_norm', x)
   return norm(x, ord=ord, keepdims=keepdims, axis=(-2, -1))
 
 
@@ -1596,12 +1615,11 @@ def matrix_transpose(x: ArrayLike, /) -> Array:
            [[5, 7],
             [6, 8]]], dtype=int32)
   """
-  check_arraylike('jnp.linalg.matrix_transpose', x)
-  x_arr = jnp.asarray(x)
+  x_arr = ensure_arraylike('jnp.linalg.matrix_transpose', x)
   ndim = x_arr.ndim
   if ndim < 2:
-    raise ValueError(f"matrix_transpose requres at least 2 dimensions; got {ndim=}")
-  return jax.lax.transpose(x_arr, (*range(ndim - 2), ndim - 1, ndim - 2))
+    raise ValueError(f"matrix_transpose requires at least 2 dimensions; got {ndim=}")
+  return lax.transpose(x_arr, (*range(ndim - 2), ndim - 1, ndim - 2))
 
 
 @export
@@ -1640,13 +1658,13 @@ def vector_norm(x: ArrayLike, /, *, axis: int | tuple[int, ...] | None = None, k
     >>> jnp.linalg.vector_norm(x, axis=1)
     Array([3.7416575, 9.486833 ], dtype=float32)
   """
-  check_arraylike('jnp.linalg.vector_norm', x)
+  x = ensure_arraylike('jnp.linalg.vector_norm', x)
   if ord is None or ord == 2:
     return ufuncs.sqrt(reductions.sum(ufuncs.real(x * ufuncs.conj(x)), axis=axis,
                                       keepdims=keepdims))
-  elif ord == jnp.inf:
-    return reductions.amax(ufuncs.abs(x), axis=axis, keepdims=keepdims)
-  elif ord == -jnp.inf:
+  elif ord == np.inf:
+    return reductions.amax(ufuncs.abs(x), axis=axis, keepdims=keepdims, initial=0)
+  elif ord == -np.inf:
     return reductions.amin(ufuncs.abs(x), axis=axis, keepdims=keepdims)
   elif ord == 0:
     return reductions.sum(x != 0, dtype=jnp.finfo(lax.dtype(x)).dtype,
@@ -1666,14 +1684,14 @@ def vector_norm(x: ArrayLike, /, *, axis: int | tuple[int, ...] | None = None, k
     raise ValueError(msg)
   else:
     abs_x = ufuncs.abs(x)
-    ord_arr = lax_internal._const(abs_x, ord)
-    ord_inv = lax_internal._const(abs_x, 1. / ord_arr)
+    ord_arr = lax._const(abs_x, ord)
+    ord_inv = lax._const(abs_x, 1. / ord_arr)
     out = reductions.sum(abs_x ** ord_arr, axis=axis, keepdims=keepdims)
     return ufuncs.power(out, ord_inv)
 
 @export
 def vecdot(x1: ArrayLike, x2: ArrayLike, /, *, axis: int = -1,
-           precision: PrecisionLike = None,
+           precision: lax.PrecisionLike = None,
            preferred_element_type: DTypeLike | None = None) -> Array:
   """Compute the (batched) vector conjugate dot product of two arrays.
 
@@ -1717,14 +1735,14 @@ def vecdot(x1: ArrayLike, x2: ArrayLike, /, *, axis: int = -1,
     >>> jnp.linalg.vecdot(x1, x2, axis=-1)
     Array([20, 47], dtype=int32)
   """
-  check_arraylike('jnp.linalg.vecdot', x1, x2)
-  return jnp.vecdot(x1, x2, axis=axis, precision=precision,
-                    preferred_element_type=preferred_element_type)
+  x1, x2 = ensure_arraylike('jnp.linalg.vecdot', x1, x2)
+  return tensor_contractions.vecdot(x1, x2, axis=axis, precision=precision,
+                                    preferred_element_type=preferred_element_type)
 
 
 @export
 def matmul(x1: ArrayLike, x2: ArrayLike, /, *,
-           precision: PrecisionLike = None,
+           precision: lax.PrecisionLike = None,
            preferred_element_type: DTypeLike | None = None) -> Array:
   """Perform a matrix multiplication.
 
@@ -1778,16 +1796,17 @@ def matmul(x1: ArrayLike, x2: ArrayLike, /, *,
     Array([[22, 28],
            [49, 64]], dtype=int32)
   """
-  check_arraylike('jnp.linalg.matmul', x1, x2)
-  return jnp.matmul(x1, x2, precision=precision,
-                    preferred_element_type=preferred_element_type)
+  x1, x2 = ensure_arraylike('jnp.linalg.matmul', x1, x2)
+  return tensor_contractions.matmul(x1, x2, precision=precision,
+                                    preferred_element_type=preferred_element_type)
 
 
 @export
 def tensordot(x1: ArrayLike, x2: ArrayLike, /, *,
               axes: int | tuple[Sequence[int], Sequence[int]] = 2,
-              precision: PrecisionLike = None,
-              preferred_element_type: DTypeLike | None = None) -> Array:
+              precision: lax.PrecisionLike = None,
+              preferred_element_type: DTypeLike | None = None,
+              out_sharding: NamedSharding | P | None = None) -> Array:
   """Compute the tensor dot product of two N-dimensional arrays.
 
   JAX implementation of :func:`numpy.linalg.tensordot`.
@@ -1860,9 +1879,10 @@ def tensordot(x1: ArrayLike, x2: ArrayLike, /, *,
     Array([[1, 2, 3],
            [2, 4, 6]], dtype=int32)
   """
-  check_arraylike('jnp.linalg.tensordot', x1, x2)
-  return jnp.tensordot(x1, x2, axes=axes, precision=precision,
-                       preferred_element_type=preferred_element_type)
+  x1, x2 = ensure_arraylike('jnp.linalg.tensordot', x1, x2)
+  return tensor_contractions.tensordot(
+      x1, x2, axes=axes, precision=precision,
+      preferred_element_type=preferred_element_type, out_sharding=out_sharding)
 
 
 @export
@@ -1886,7 +1906,7 @@ def svdvals(x: ArrayLike, /) -> Array:
     >>> jnp.linalg.svdvals(x)
     Array([9.508031 , 0.7728694], dtype=float32)
   """
-  check_arraylike('jnp.linalg.svdvals', x)
+  x = ensure_arraylike('jnp.linalg.svdvals', x)
   return svd(x, compute_uv=False, hermitian=False)
 
 
@@ -1927,7 +1947,7 @@ def diagonal(x: ArrayLike, /, *, offset: int = 0) -> Array:
     Array([[ 0,  5, 10],
            [12, 17, 22]], dtype=int32)
   """
-  check_arraylike('jnp.linalg.diagonal', x)
+  x = ensure_arraylike('jnp.linalg.diagonal', x)
   return jnp.diagonal(x, offset=offset, axis1=-2, axis2=-1)
 
 
@@ -1960,8 +1980,7 @@ def tensorinv(a: ArrayLike, ind: int = 2) -> Array:
     >>> jnp.allclose(xinv_x, jnp.eye(4), atol=1E-4)
     Array(True, dtype=bool)
   """
-  check_arraylike("tensorinv", a)
-  arr = jnp.asarray(a)
+  arr = ensure_arraylike("tensorinv", a)
   ind = operator.index(ind)
   if ind <= 0:
     raise ValueError(f"ind must be a positive integer; got {ind=}")
@@ -2009,8 +2028,7 @@ def tensorsolve(a: ArrayLike, b: ArrayLike, axes: tuple[int, ...] | None = None)
     >>> jnp.allclose(b, b_reconstructed)
     Array(True, dtype=bool)
   """
-  check_arraylike("tensorsolve", a, b)
-  a_arr, b_arr = jnp.asarray(a), jnp.asarray(b)
+  a_arr, b_arr = ensure_arraylike("tensorsolve", a, b)
   if axes is not None:
     a_arr = jnp.moveaxis(a_arr, axes, len(axes) * (a_arr.ndim - 1,))
   out_shape = a_arr.shape[b_arr.ndim:]
@@ -2025,7 +2043,7 @@ def tensorsolve(a: ArrayLike, b: ArrayLike, axes: tuple[int, ...] | None = None)
 
 
 @export
-def multi_dot(arrays: Sequence[ArrayLike], *, precision: PrecisionLike = None) -> Array:
+def multi_dot(arrays: Sequence[ArrayLike], *, precision: lax.PrecisionLike = None) -> Array:
   """Efficiently compute matrix products between a sequence of arrays.
 
   JAX implementation of :func:`numpy.linalg.multi_dot`.
@@ -2096,8 +2114,7 @@ def multi_dot(arrays: Sequence[ArrayLike], *, precision: PrecisionLike = None) -
   >>> jax.jit(jnp.linalg.multi_dot).lower([x, y, z]).cost_analysis()['flops']
   30000.0
   """
-  check_arraylike('jnp.linalg.multi_dot', *arrays)
-  arrs: list[Array] = list(map(jnp.asarray, arrays))
+  arrs = list(ensure_arraylike('jnp.linalg.multi_dot', *arrays))
   if len(arrs) < 2:
     raise ValueError(f"multi_dot requires at least two arrays; got len(arrays)={len(arrs)}")
   if not (arrs[0].ndim in (1, 2) and arrs[-1].ndim in (1, 2) and
@@ -2113,12 +2130,12 @@ def multi_dot(arrays: Sequence[ArrayLike], *, precision: PrecisionLike = None) -
     einsum_axes[0] = einsum_axes[0][1:]
   if arrs[-1].ndim == 1:
     einsum_axes[-1] = einsum_axes[-1][:1]
-  return jnp.einsum(*itertools.chain(*zip(arrs, einsum_axes)),  # type: ignore[call-overload]
-                    optimize='auto', precision=precision)
+  return einsum.einsum(*itertools.chain(*zip(arrs, einsum_axes)),  # type: ignore[call-overload]
+                       optimize='auto', precision=precision)
 
 
 @export
-@partial(jit, static_argnames=['p'])
+@partial(api.jit, static_argnames=['p'])
 def cond(x: ArrayLike, p=None):
   """Compute the condition number of a matrix.
 
@@ -2157,8 +2174,7 @@ def cond(x: ArrayLike, p=None):
     >>> jnp.linalg.cond(x)
     Array(inf, dtype=float32)
   """
-  check_arraylike("cond", x)
-  arr = jnp.asarray(x)
+  arr = ensure_arraylike("cond", x)
   if arr.ndim < 2:
     raise ValueError(f"jnp.linalg.cond: input array must be at least 2D; got {arr.shape=}")
   if arr.shape[-1] == 0 or arr.shape[-2] == 0:
@@ -2174,7 +2190,7 @@ def cond(x: ArrayLike, p=None):
       raise ValueError(f"jnp.linalg.cond: for {p=}, array must be square; got {arr.shape=}")
     r = norm(x, ord=p, axis=(-2, -1)) * norm(inv(x), ord=p, axis=(-2, -1))
   # Convert NaNs to infs where original array has no NaNs.
-  return jnp.where(ufuncs.isnan(r) & ~ufuncs.isnan(x).any(axis=(-2, -1)), jnp.inf, r)
+  return jnp.where(ufuncs.isnan(r) & ~ufuncs.isnan(x).any(axis=(-2, -1)), np.inf, r)
 
 
 @export
@@ -2218,5 +2234,5 @@ def trace(x: ArrayLike, /, *,
     >>> jnp.linalg.trace(x)
     Array([15, 51], dtype=int32)
   """
-  check_arraylike('jnp.linalg.trace', x)
+  x = ensure_arraylike('jnp.linalg.trace', x)
   return jnp.trace(x, offset=offset, axis1=-2, axis2=-1, dtype=dtype)
