@@ -21,12 +21,14 @@ import dataclasses
 import functools
 import itertools as it
 import math
-from typing import Any
+from typing import Any, Protocol, TypeVar, cast
 
 import jax
+from jax import api_util
 from jax import lax
 from jax._src import core
 from jax._src import linear_util as lu
+from jax._src import state
 from jax._src import util
 from jax._src.interpreters import partial_eval as pe
 from jax._src.pallas import core as pallas_core
@@ -38,21 +40,36 @@ import jax.numpy as jnp
 
 map = util.safe_map
 zip = util.safe_zip
+T = TypeVar('T')
 
+def _get_block_size(
+    bd: pl.Blocked | pl.Element | pl.Squeezed | pl.BoundedSlice | int | None,
+) -> int:
+  match bd:
+    case int():
+      return bd
+    case pl.Blocked(block_size):
+      return block_size
+    case _:
+      raise NotImplementedError(f"Unsupported block size type: {type(bd)}")
+
+def _get_block_shape(spec: pallas_core.BlockSpec):
+  assert spec.block_shape is not None
+  return tuple(_get_block_size(bd) for bd in spec.block_shape)
 
 @jax.tree_util.register_dataclass
 @dataclasses.dataclass(frozen=True)
 class BufferedRef:
   spec: pallas_core.BlockSpec = dataclasses.field(metadata={"static": True})
   is_index_invariant: bool = dataclasses.field(metadata={"static": True})
-  gmem_ref: pallas_core.AbstractMemoryRef
+  gmem_ref: state.AbstractRef
   # ``None`` if the ref is pinned to GMEM; otherwise, has shape
   # [num_slots, *spec.block_shape].
-  smem_ref: pallas_core.AbstractMemoryRef | None
+  smem_ref: state.AbstractRef | None
 
   def get_ref_for_slot(
       self, slot: int | jax.Array
-  ) -> pallas_core.AbstractMemoryRef:
+  ) -> state.AbstractRef:
     if self.smem_ref is None:
       return self.gmem_ref
     return self.smem_ref.at[slot]
@@ -60,10 +77,14 @@ class BufferedRef:
   def compute_gmem_slice(self, grid_indices) -> tuple[pl.Slice, ...]:
     index_map = self.spec.index_map
     assert index_map is not None
+    # We don't allow Python scalars here, because they are interpreted
+    # differently depending on the x32/x64 mode.
+    assert all(i.dtype == jnp.dtype(jnp.int32) for i in grid_indices)
+    sizes = _get_block_shape(self.spec)
     return tuple(
         pl.Slice(idx * size, size)  # type: ignore[arg-type]
         for idx, size in zip(
-            index_map(*grid_indices), self.spec.block_shape  # type: ignore[arg-type]
+            index_map(*grid_indices), sizes  # type: ignore[arg-type]
         )
     )
 
@@ -74,7 +95,7 @@ class BufferedRef:
     gmem_slices = self.compute_gmem_slice(grid_indices)
     gpu_primitives.copy_gmem_to_smem(
         self.gmem_ref.at[gmem_slices],  # pytype: disable=unsupported-operands
-        self.smem_ref.at[slot],
+        self.smem_ref.at[slot],  # pytype: disable=unsupported-operands
         barrier_ref.at[slot],
     )
 
@@ -84,24 +105,33 @@ class BufferedRef:
     assert self.smem_ref is not None
     gmem_slices = self.compute_gmem_slice(grid_indices)
     gpu_primitives.copy_smem_to_gmem(
-        self.smem_ref.at[slot],
+        self.smem_ref.at[slot],  # pytype: disable=unsupported-operands
         self.gmem_ref.at[gmem_slices],  # pytype: disable=unsupported-operands
         predicate=predicate,
+        commit_group=False,
     )
 
 
 def _uses_arguments(
     index_map: Callable[..., Any], num_args: int
 ) -> Sequence[bool]:
-  jaxpr, _, _, () = pe.trace_to_jaxpr_dynamic(
-      lu.wrap_init(index_map), (core.ShapedArray((), jnp.int32),) * num_args
+  if not num_args:
+    return ()
+
+  jaxpr, _, _ = pe.trace_to_jaxpr_dynamic(
+      lu.wrap_init(
+          index_map,
+          debug_info=api_util.debug_info("pallas index_map",
+                                         index_map,
+                                         (0,) * num_args, {})),
+      (core.ShapedArray((), jnp.int32),) * num_args
   )
   _, used_inputs = pe.dce_jaxpr(jaxpr, used_outputs=[True] * len(jaxpr.outvars))
   return used_inputs
 
 
 def _is_index_invariant(
-    spec: pallas_core.BlockSpec, grid: pallas_core.StaticGrid
+    spec: pallas_core.BlockSpec, grid: pallas_core.TupleGrid
 ) -> bool:
   if (index_map := spec.index_map) is None:
     return True
@@ -109,14 +139,16 @@ def _is_index_invariant(
 
 
 def _inc_grid_by_1(
-    indices: tuple[jax.Array, ...], grid: Sequence[int]
+    indices: tuple[jax.Array, ...], grid: pallas_core.TupleGrid
 ) -> tuple[jax.Array, ...]:
   next_indices = []
   carry: bool | jax.Array = True
   for idx, size in reversed(list(zip(indices, grid))):
     next_idx = lax.select(carry, idx + 1, idx)
     carry = next_idx == size
-    next_indices.append(lax.select(carry, 0, next_idx).astype(idx.dtype))
+    next_indices.append(
+        lax.select(carry, jnp.asarray(0, dtype=idx.dtype), next_idx)
+    )
   return tuple(reversed(next_indices))
 
 
@@ -144,48 +176,65 @@ jax.tree_util.register_dataclass(
 
 
 def emit_pipeline(
-    body: Callable[..., None],
+    body: Callable[..., T],
     *,
-    grid: pallas_core.StaticGrid,
+    grid: pallas_core.TupleGrid,
     in_specs: Sequence[pallas_core.BlockSpec] = (),
     out_specs: Sequence[pallas_core.BlockSpec] = (),
     max_concurrent_steps: int = 1,
     delay_release: int = 0,
+    init_carry: T | None = None,
 ):
-  """Creates a function to emit a manual pipeline within a Pallas kernel.
+  r"""Creates a function to emit a manual pipeline within a Pallas kernel.
 
   Args:
-    body: The pipeline body.
-    grid: The grid to use for the pipeline.
-    in_specs: The block specs for the inputs.
-    out_specs: The block specs for the outputs.
-    max_concurrent_steps: The maximum number of sequential stages that are
-      active concurrently. Defaults to 1.
-    delay_release: The number of steps to wait before reusing the input/output
-      references. Defaults to 0, and must be strictly smaller than
-      ``max_concurrent_steps``. Generally, you'll want to set it to 1 if you
-      don't await the WGMMA in the body.
-  """
-  num_steps = math.prod(grid)
+    body: The pipeline body function, which is called with
 
+      - ``indices``: Tuple of current loop indices.
+      - ``*input_refs``: SMEM refs for inputs.
+      - ``*output_refs``: SMEM refs for outputs.
+
+      If ``init_carry`` is provided, ``body`` receives an additional argument
+      ``carry`` -- the carry from the previous iteration. It must then return
+      the next carry value.
+    grid: The grid dimensions for the pipeline.
+    in_specs: A sequence of :class:`~jax.experimental.pallas.BlockSpec`\s
+      for inputs.
+    out_specs: A sequence of :class:`~jax.experimental.pallas.BlockSpec`\s
+      for outputs.
+    max_concurrent_steps: Maximum concurrently active pipeline stages.
+    delay_release: Number of steps to delay before reusing input/output
+      references. Must be ``< max_concurrent_steps``. Useful for hiding WGMMA
+      latency (typically set to 1).
+    init_carry: Optional initial carry. If provided, ``body`` handles
+      carry-over state between iterations, and the pipeline returns the
+      final carry.
+
+  Returns:
+    A function that, when called with GMEM input and output refs, executes the
+    pipeline and returns the final carry value (if ``init_carry`` was used),
+    otherwise it returns None.
+  """
   if max_concurrent_steps <= delay_release:
     raise ValueError(
         "max_concurrent_steps must be greater than delay_release, but"
         f" {max_concurrent_steps=}, {delay_release=}"
     )
 
+  num_steps = math.prod(grid)
+  has_dynamic_grid = not isinstance(num_steps, int)
+
   # Shrink ``max_concurrent_steps`` if the total number of steps is lower to
   # reduce the size of the refs allocated in SMEM.
-  if max_concurrent_steps > num_steps:
-    max_concurrent_steps = num_steps
-    delay_release = 0  # No need to delay anything.
+  if not has_dynamic_grid and max_concurrent_steps > num_steps:
+    max_concurrent_steps = cast(int, num_steps)
 
-  def pipeline(*gmem_refs: pallas_core.AbstractMemoryRef):
+  def pipeline(*gmem_refs: state.AbstractRef):
     in_gmem_refs, out_gmem_refs = util.split_list(gmem_refs, [len(in_specs)])
     in_smem_refs, out_smem_refs = util.split_list(
         [
             gpu_core.SMEM(
-                (max_concurrent_steps, *spec.block_shape),  # type: ignore
+                (max_concurrent_steps, *_get_block_shape(spec)),  # type: ignore
                 ref.dtype,
                 transforms=tuple(
                     t.batch(1) for t in getattr(spec, "transforms", ())
@@ -197,6 +246,7 @@ def emit_pipeline(
         ],
         [len(in_specs)],
     )
+    num_arrivals = sum(map(_in_smem, in_specs))
     return pl.run_scoped(
         functools.partial(
             scoped_pipeline,
@@ -205,9 +255,11 @@ def emit_pipeline(
         ),
         in_smem_refs=in_smem_refs,
         out_smem_refs=out_smem_refs,
-        barrier_ref=gpu_core.Barrier(
+        barrier_ref=None
+        if num_arrivals == 0
+        else gpu_core.Barrier(
             # TODO(slebedev): Change this to arrive only once.
-            sum(map(_in_smem, in_specs)),
+            num_arrivals=num_arrivals,
             num_barriers=max_concurrent_steps,
         ),
     )
@@ -228,30 +280,42 @@ def emit_pipeline(
         )
     ]
 
-    for step, indices in enumerate(
-        it.islice(it.product(*map(range, grid)), max_concurrent_steps)
-    ):
-      map(lambda bref: bref.copy_in(step, indices, barrier_ref), in_brefs)
+    # Initialize the pipeline.
+    indices = (jnp.asarray(0, dtype=jnp.int32),) * len(grid)
+    fetch_indices = indices
+    for step in range(max_concurrent_steps):
+      for bref in in_brefs:
+        bref.copy_in(step, fetch_indices, barrier_ref)
+      fetch_indices = _inc_grid_by_1(fetch_indices, grid)
+    del fetch_indices
+
+    # This is true if any of the outputs need to be transferred inside the loop.
+    copies_out_in_loop = not all(bref.is_index_invariant for bref in out_brefs)
+    needs_epilogue = any(bref.is_index_invariant for bref in out_brefs)
 
     def loop_body(step, carry):
-      slot = step % max_concurrent_steps
-      indices, fetch_indices, last_store_slices = carry
+      slot = lax.rem(step, max_concurrent_steps)
+      indices, fetch_indices, last_store_slices, prev_body_carry = carry
 
-      if in_specs:
-        # Wait for the current GMEM->SMEM copy to complete.
+      if barrier_ref is not None:
+        # Wait for the current GMEM->SMEM copy to complete, if any.
         gpu_primitives.barrier_wait(barrier_ref.at[slot])
       # Wait for the previous output SMEM->GMEM copy to complete.
-      gpu_primitives.wait_smem_to_gmem(
-          max_concurrent_steps - (1 + delay_release), wait_read_only=True
+      if copies_out_in_loop:
+        gpu_primitives.wait_smem_to_gmem(
+            max_concurrent_steps - (1 + delay_release), wait_read_only=True
+        )
+
+      next_body_carry = body(
+          indices,
+          *(
+              bref.get_ref_for_slot(slot)
+              for bref in it.chain(in_brefs, out_brefs)
+          ),
+          *(prev_body_carry,) if init_carry is not None else (),
       )
 
-      with pallas_core.grid_env(map(pallas_core.GridAxis, indices, grid)):
-        body(*(
-            bref.get_ref_for_slot(slot)
-            for bref in it.chain(in_brefs, out_brefs)
-        ))
-
-      if not all(bref.is_index_invariant for bref in out_brefs):
+      if copies_out_in_loop:
         gpu_primitives.commit_smem()
 
       # Copy the output from SMEM to GMEM.
@@ -280,26 +344,33 @@ def emit_pipeline(
             predicate=lax.bitwise_or(slices_changed, is_last_step),
         )
 
+      if copies_out_in_loop:
+        gpu_primitives.commit_smem_to_gmem_group()
+
       fetch_step = step + (max_concurrent_steps - delay_release)
-      fetch_slot = slot  # (x + y) % y == x % y
+      fetch_slot = lax.rem(fetch_step, max_concurrent_steps)
+
+      def do_fetch():
+        for bref in in_brefs:
+          bref.copy_in(fetch_slot, fetch_indices, barrier_ref)
+
       jax.lax.cond(
-          lax.bitwise_and(fetch_step >= delay_release, fetch_step < num_steps),
-          lambda: map(
-              lambda bref: bref.copy_in(fetch_slot, fetch_indices, barrier_ref),
-              in_brefs,
-          ),
-          lambda: [None] * len(in_brefs),
+          lax.bitwise_and(step >= delay_release, fetch_step < num_steps),
+          do_fetch,
+          lambda: None,
       )
 
       return (
           _inc_grid_by_1(indices, grid),
           _inc_grid_by_1(fetch_indices, grid),
           new_store_slices,
+          next_body_carry if init_carry is not None else None,
       )
 
-    indices = (jnp.asarray(0, dtype=lax.dtype(0)),) * len(grid)
+    # Invariant: ``indices`` and ``fetch_indices`` are always
+    # ``max_concurrent_steps-delay_release`` apart.
     fetch_indices = indices
-    for _ in range(max_concurrent_steps):
+    for _ in range(max_concurrent_steps-delay_release):
       fetch_indices = _inc_grid_by_1(fetch_indices, grid)
     # TODO(justinfu): Only store base pointer instead of all indices.
     last_store_slices = [
@@ -308,39 +379,95 @@ def emit_pipeline(
         else (_Slice(-1, -1),) * len(bref.spec.block_shape)
         for bref in out_brefs
     ]
-    last_indices, _, _ = lax.fori_loop(
-        0, num_steps, loop_body, (indices, fetch_indices, last_store_slices)
+    last_indices, _, _, final_carry = lax.fori_loop(
+        0,
+        num_steps,
+        loop_body,
+        (indices, fetch_indices, last_store_slices, init_carry),
     )
 
     # Outputs invariant to the sequential axis are never written from inside the
     # loop. This is the only place where we store them.
-    if all(bref.is_index_invariant for bref in out_brefs):
+    if not copies_out_in_loop and needs_epilogue:
       gpu_primitives.commit_smem()
-    last_slot = (num_steps - 1) % max_concurrent_steps
-    for bref in out_brefs:
-      if bref.is_index_invariant:
-        bref.copy_out(last_slot, last_indices, predicate=None)
 
-    # Finalize the pipeline.
-    gpu_primitives.wait_smem_to_gmem(0)
+    if needs_epilogue:
+      last_slot = lax.rem(num_steps - 1, max_concurrent_steps)
+      for bref in out_brefs:
+        if bref.is_index_invariant:
+          bref.copy_out(last_slot, last_indices, predicate=None)
+
+      gpu_primitives.commit_smem_to_gmem_group()
+
+    if out_brefs:
+      # Finalize the pipeline.
+      gpu_primitives.wait_smem_to_gmem(0)
+    return final_carry if init_carry is not None else None
 
   return pipeline
 
+
+class ComputeContext(Protocol):
+  """Protocol for a compute context for the warp specialized pipeline.
+
+  The ComputeContext is run exclusively in the compute thread and allows
+  the user to set up a prologue to initialize a pipeline carry and an epilogue
+  to consume the final carry.
+
+  All values allocated in the ComputeContext will only be allocated in the
+  compute thread and not the memory thread. This can potentially reduce
+  register pressure if certain values are only consumed by the compute threads.
+
+  Usage will usually follow this structure:
+
+  ```
+  def compute_context(pipeline):
+    # Perform prologue work and compute the initial carry.
+    initial_carry = ...
+    # Run the pipeline.
+    final_carry = pipeline(*initial_carry)
+    # Perform epilogue work using the final carry.
+    do_work(final_carry)
+  ```
+
+  """
+  def __call__(self, pipeline: Callable[[T], T]) -> None:
+    ...
+
+
 def emit_pipeline_warp_specialized(
     body: Callable[..., None],
     *,
-    grid: pallas_core.StaticGrid,
+    grid: pallas_core.TupleGrid,
     memory_registers: int,
-    in_specs: Sequence[gpu_core.GPUBlockSpec] = (),
-    out_specs: Sequence[gpu_core.GPUBlockSpec] = (),
+    in_specs: Sequence[pl.BlockSpec] = (),
+    out_specs: Sequence[pl.BlockSpec] = (),
     max_concurrent_steps: int = 2,
     wg_axis: str,
     num_compute_wgs: int,
-    carry_coroutine: Any | None = None,
+    manual_consumed_barriers: bool = False,
+    compute_context: ComputeContext | None = None,
     memory_thread_idx: int | None = None,
 ):
   """Creates a function to emit a warp-specialized pipeline.
 
+  The ``body`` function should have the following signature (without carry).
+  ``consumed_barriers`` is an optional argument that is only passed if the
+  ``manual_consumed_barriers`` argument is True.
+
+  ```
+  def body(indices, *input_refs, *output_refs, [consumed_barriers]) -> None:
+  ```
+
+  or with a carries enabled (enabled via the ``compute_context`` argument),
+  where the body returns the next carry:
+
+  ```
+  def body(
+      indices, *input_refs, *output_refs, [consumed_barriers], carry
+  ) -> Carry:
+  ```
+
   Args:
     body: The pipeline body.
     grid: The grid to use for the pipeline.
@@ -352,17 +479,23 @@ def emit_pipeline_warp_specialized(
       active concurrently. Defaults to 2.
     wg_axis: The axis name for the warp group axis.
     num_compute_wgs: The number of compute warpgroups
-    carry_coroutine: If specified, enables carries in the pipeline.
-      The signature of the body function will be modified such that the last
-      argument will be the current carry and it must return the next carry.
-      The coroutine itself should yield the initial carry, and the
-      yield statement will return the final value of the carry.
+    manual_consumed_barriers: If True, consumed barriers will be
+      passed into the body function after the output refs. There will be one
+      barrier per input and will be passed in the same order.
+    compute_context: If specified, enables carries in the pipeline and allows
+      a user-specified prologue/epilogue that is only executed in the compute
+      thread. The signature of the pipeline body function will be modified
+      such that the last argument will be the current carry and it must
+      return the next carry.
+      The compute_context itself should follow the signature of `ComputeContext`
+      and take a pipeline function as its sole argument. Calling the
+      pipeline with the initial carry will run the pipeline and return the
+      final carry.
     memory_thread_idx: The index of the memory thread. If not specified,
       defaults to the last thread.
   """
   # TODO(justinfu): Factor out common code between warp-specialized and
   # normal pipelines.
-  # TODO(justinfu): Allow passing consumed_barrier into body.
 
   if memory_thread_idx is None:
     memory_thread_idx = num_compute_wgs
@@ -371,17 +504,18 @@ def emit_pipeline_warp_specialized(
     # thread is the last thread.
     raise NotImplementedError("Memory thread must be the last thread.")
 
-  has_carry = carry_coroutine is not None
+  has_carry = compute_context is not None
 
   # Trace the index maps to determine if they depend on the grid.
   # Grid-independent values will not be multiple-buffered.
   in_spec_has_seq_axis = [
-      ~_is_index_invariant(spec, grid) for spec in in_specs]
+      not _is_index_invariant(spec, grid) for spec in in_specs]
   out_spec_has_seq_axis = [
-      ~_is_index_invariant(spec, grid) for spec in out_specs]
+      not _is_index_invariant(spec, grid) for spec in out_specs]
   spec_has_seq_axis = [*in_spec_has_seq_axis, *out_spec_has_seq_axis]
 
-  num_pipeline_steps = math.prod(grid)
+  num_steps = math.prod(grid)
+  has_dynamic_grid = not isinstance(num_steps, int)
 
   def _get_slot(step, has_seq_dim):
     """Returns the buffer slot given the pipeline step."""
@@ -392,10 +526,10 @@ def emit_pipeline_warp_specialized(
 
   # Shrink ``max_concurrent_steps`` if the total number of steps is lower to
   # reduce the size of the refs allocated in SMEM.
-  if max_concurrent_steps > num_pipeline_steps:
-    max_concurrent_steps = num_pipeline_steps
+  if not has_dynamic_grid and max_concurrent_steps > num_steps:
+    max_concurrent_steps = cast(int, num_steps)
 
-  def pipeline(*gmem_refs: pallas_core.AbstractMemoryRef):
+  def pipeline(*gmem_refs: state.AbstractRef):
     in_gmem_refs, out_gmem_refs = util.split_list(gmem_refs, [len(in_specs)])
     if len(out_gmem_refs) != len(out_specs):
       raise ValueError(
@@ -411,19 +545,36 @@ def emit_pipeline_warp_specialized(
           gpu_core.SMEM(
               (slots, *spec.block_shape),   # type: ignore
               gmem_ref.dtype,
-              transforms=spec.transforms,
+              transforms=getattr(spec, "transforms", ()),
           )
       )
     in_smem_refs, out_smem_refs = util.split_list(
         smem_allocs, [len(in_specs)])
 
     in_smem_barriers = []
+    consumed_barriers = []
     for has_seq_dim in in_spec_has_seq_axis:
       num_barriers = max_concurrent_steps if has_seq_dim else 1
       in_smem_barriers.append(
           gpu_core.Barrier(
           num_arrivals=1,
           num_barriers=num_barriers))
+      if manual_consumed_barriers:
+        consumed_barriers.append(
+            gpu_core.Barrier(
+                num_arrivals=num_compute_wgs,
+                num_barriers=max_concurrent_steps,
+            )
+        )
+    if not manual_consumed_barriers:
+      # We only allocated one consumed barrier for all inputs when using
+      # automatic consumed barriers.
+      consumed_barriers = [
+          gpu_core.Barrier(
+              num_arrivals=num_compute_wgs,
+              num_barriers=max_concurrent_steps,
+          )
+      ]
     return pl.run_scoped(
         functools.partial(
             scoped_pipeline,
@@ -433,10 +584,8 @@ def emit_pipeline_warp_specialized(
         in_smem_refs=in_smem_refs,
         out_smem_refs=out_smem_refs,
         in_smem_barrier_refs=in_smem_barriers,
-        consumed_barrier_ref=gpu_core.Barrier(
-            num_arrivals=num_compute_wgs,
-            num_barriers=max_concurrent_steps,
-        ),
+        consumed_barrier_refs=consumed_barriers,
+        collective_axes=wg_axis,
     )
 
   def scoped_pipeline(
@@ -446,16 +595,16 @@ def emit_pipeline_warp_specialized(
       in_smem_refs,
       out_smem_refs,
       in_smem_barrier_refs,
-      consumed_barrier_ref,
+      consumed_barrier_refs,
   ):
     in_brefs: Sequence[BufferedRef] = [
-        BufferedRef(spec, ~has_seq_axis, gmem_ref, smem_ref)
+        BufferedRef(spec, not has_seq_axis, gmem_ref, smem_ref)
         for spec, has_seq_axis, gmem_ref, smem_ref in zip(
             in_specs, in_spec_has_seq_axis, in_gmem_refs, in_smem_refs
         )
     ]
     out_brefs: Sequence[BufferedRef] = [
-        BufferedRef(spec, ~has_seq_axis, gmem_ref, smem_ref)
+        BufferedRef(spec, not has_seq_axis, gmem_ref, smem_ref)
         for spec, has_seq_axis, gmem_ref, smem_ref in zip(
             out_specs, out_spec_has_seq_axis, out_gmem_refs, out_smem_refs
         )
@@ -466,9 +615,13 @@ def emit_pipeline_warp_specialized(
           _compute_registers(memory_registers, num_compute_wgs),
           action="increase")
 
+      # This is true if any of the outputs need to be transferred inside the loop.
+      copies_out_in_loop = not all(bref.is_index_invariant for bref in out_brefs)
+      needs_epilogue = any(bref.is_index_invariant for bref in out_brefs)
+
       def compute_loop_body(step, carry):
         indices, last_store_slices, prev_body_carry = carry
-        slot = step % max_concurrent_steps
+        slot = lax.rem(step, max_concurrent_steps)
         # Wait for the current GMEM->SMEM copies to complete.
         for in_barrier, has_seq_dim in zip(
             in_smem_barrier_refs, in_spec_has_seq_axis):
@@ -478,22 +631,27 @@ def emit_pipeline_warp_specialized(
               in_barrier.at[_get_slot(slot, has_seq_dim)])
 
         # Wait for the previous output SMEM->GMEM copy to complete.
-        gpu_primitives.wait_smem_to_gmem(max_concurrent_steps - 1)
+        if copies_out_in_loop:
+          gpu_primitives.wait_smem_to_gmem(max_concurrent_steps - 1)
 
-        with pallas_core.grid_env(map(pallas_core.GridAxis, indices, grid)):
-          body_refs = []
-          for bref in it.chain(in_brefs, out_brefs):
-            buf_slot = _get_slot(slot, ~bref.is_index_invariant)
-            body_refs.append(bref.get_ref_for_slot(buf_slot))
+        body_refs = []
+        for bref in it.chain(in_brefs, out_brefs):
+          buf_slot = _get_slot(slot, not bref.is_index_invariant)
+          body_refs.append(bref.get_ref_for_slot(buf_slot))
 
-          if has_carry:
-            next_body_carry = body(*body_refs, prev_body_carry)
-          else:
-            body(*body_refs)
-            next_body_carry = None
-        gpu_primitives.barrier_arrive(consumed_barrier_ref.at[slot])
+        body_args = body_refs
+        if manual_consumed_barriers:
+          body_args += [consumed_barrier_ref.at[slot] for consumed_barrier_ref in consumed_barrier_refs]
+        if has_carry:
+          body_args += [prev_body_carry]
+        next_body_carry = body(indices, *body_args)
+
+        if not manual_consumed_barriers:
+          [consumed_barrier_ref] = consumed_barrier_refs
+          gpu_primitives.barrier_arrive(consumed_barrier_ref.at[slot])
+        # TODO(justinfu,apaszke): This should probably be done by the memory WG.
         # Copy the output from SMEM to GMEM.
-        if not all(bref.is_index_invariant for bref in out_brefs):
+        if copies_out_in_loop:
           gpu_primitives.commit_smem()
 
         new_store_slices = last_store_slices[:]
@@ -511,12 +669,13 @@ def emit_pipeline_warp_specialized(
               new_store_slices[idx],
           )
           slices_changed = ~functools.reduce(lax.bitwise_and, are_same_slices)
-          bref.copy_out(_get_slot(slot, ~bref.is_index_invariant),
+          bref.copy_out(_get_slot(slot, not bref.is_index_invariant),
                         indices,
                         predicate=slices_changed)
+        gpu_primitives.commit_smem_to_gmem_group()
         next_indices = _inc_grid_by_1(indices, grid)
         return (next_indices, new_store_slices, next_body_carry)
-      init_indices = (jnp.asarray(0, dtype=lax.dtype(0)),) * len(grid)
+      init_indices = (jnp.asarray(0, dtype=jnp.int32),) * len(grid)
       # TODO(justinfu): Only store base pointer instead of all indices.
       last_store_slices = [
           None
@@ -526,62 +685,97 @@ def emit_pipeline_warp_specialized(
       ]
 
       if has_carry:
-        _carry = carry_coroutine()
-        try:
-          carry_init = next(_carry)
-        except StopIteration:
-          raise ValueError("carry_coroutine must yield the initial carry.")  # pylint: disable=raise-missing-from
+        last_indices = None
+        def pipeline_callback(user_init_carry):
+          nonlocal last_indices
+          if last_indices is not None:
+            raise ValueError(
+              "Cannot call pipeline more than once in `compute_context`")
+          init_loop_carry = (init_indices, last_store_slices, user_init_carry)
+          last_indices, _, final_body_carry = lax.fori_loop(0,
+                        num_steps,
+                        compute_loop_body,
+                        init_loop_carry)
+          return final_body_carry
+        compute_context(pipeline_callback)
+        if last_indices is None:
+          raise ValueError("Pipeline was not called in `compute_context`")
       else:
-        _carry = None
-        carry_init = None
-      init_loop_carry = (init_indices, last_store_slices, carry_init)
-      last_indices, _, final_body_carry = lax.fori_loop(0,
-                    num_pipeline_steps,
-                    compute_loop_body,
-                    init_loop_carry)
-      if has_carry:
-        try:
-          _carry.send(final_body_carry)  # pytype: disable=attribute-error
-          raise ValueError("carry_coroutine must only yield once.")
-        except StopIteration:
-          pass
+        assert compute_context is None
+        last_indices, _, _ = lax.fori_loop(
+            0, num_steps, compute_loop_body,
+            (init_indices, last_store_slices, None)
+        )
 
       # Handle index_invariant outputs after the loop. They are not
       # written in the main pipeline loop.
-      if all(bref.is_index_invariant for bref in out_brefs):
+      if not copies_out_in_loop and needs_epilogue:
         gpu_primitives.commit_smem()
-      last_slot = (num_pipeline_steps - 1) % max_concurrent_steps
-      for bref in out_brefs:
-        if bref.is_index_invariant:
-          bref.copy_out(last_slot, last_indices, predicate=None)
 
-      # Finalize the pipeline.
-      gpu_primitives.wait_smem_to_gmem(0)
+      if needs_epilogue:
+        last_slot = lax.rem(num_steps - 1, max_concurrent_steps)
+        for bref in out_brefs:
+          if bref.is_index_invariant:
+            bref.copy_out(_get_slot(last_slot, has_seq_dim=False),
+                          last_indices, predicate=None)
+
+        gpu_primitives.commit_smem_to_gmem_group()
+
+      if out_brefs:
+        # Finalize the pipeline.
+        gpu_primitives.wait_smem_to_gmem(0)
 
     # The memory thread executes this block which issues all pipelined DMAs.
     def memory_block():
       gpu_primitives.set_max_registers(memory_registers, action="decrease")
-      indices = (jnp.asarray(0, dtype=lax.dtype(0)),) * len(grid)
+      indices = (jnp.asarray(0, dtype=jnp.int32),) * len(grid)
+      if has_dynamic_grid:
+        prologue_steps = lax.min(max_concurrent_steps, num_steps)
+      else:
+        assert max_concurrent_steps <= num_steps
+        prologue_steps = max_concurrent_steps
 
       # Begin initial copies.
-      for step in range(max_concurrent_steps):
+      def _init_step(step, indices):
         for bref, barrier in zip(in_brefs, in_smem_barrier_refs):
-          buf_slot = _get_slot(step, ~bref.is_index_invariant)
+          buf_slot = _get_slot(step, not bref.is_index_invariant)
           bref.copy_in(buf_slot, indices, barrier)
-        indices = _inc_grid_by_1(indices, grid)
+        return _inc_grid_by_1(indices, grid)
+
+      indices = jax.lax.fori_loop(
+          0, prologue_steps, _init_step, indices, unroll=not has_dynamic_grid
+      )
 
       def memory_loop_body(step, carry):
         indices, = carry
-        slot = step % max_concurrent_steps
+        slot = lax.rem(step, max_concurrent_steps)
         fetch_slot = slot  # (x + y) % y == x % y
-        gpu_primitives.barrier_wait(consumed_barrier_ref.at[slot])
-        for bref, barrier in zip(in_brefs, in_smem_barrier_refs):
+
+        if not manual_consumed_barriers:
+          # We only have one consumed barrier when using automatic consumed
+          # barrier management.
+          [consumed_barrier_ref] = consumed_barrier_refs
+          gpu_primitives.barrier_wait(consumed_barrier_ref.at[slot])
+          consumed_barrier_it = [None] * len(in_brefs)
+        else:
+          consumed_barrier_it = consumed_barrier_refs
+
+        for bref, barrier, consumed_barrier in zip(
+            in_brefs, in_smem_barrier_refs, consumed_barrier_it):
+          if manual_consumed_barriers:
+            gpu_primitives.barrier_wait(consumed_barrier.at[slot])  # pytype: disable=attribute-error
           bref.copy_in(
-              _get_slot(fetch_slot, ~bref.is_index_invariant), indices, barrier)
+              _get_slot(fetch_slot, not bref.is_index_invariant), indices, barrier)
         next_indices = _inc_grid_by_1(indices, grid)
         return (next_indices,)
-      lax.fori_loop(0, num_pipeline_steps - max_concurrent_steps,
+      lax.fori_loop(0, num_steps - max_concurrent_steps,
                     memory_loop_body, (indices,))
+      # Await all the arrivals to not leave barriers in a bad state.
+      # We only need to account for the prologue steps.
+      @pl.loop(0, prologue_steps, unroll=not has_dynamic_grid)
+      def _epi_step(step):
+        for barrier in consumed_barrier_refs:
+          gpu_primitives.barrier_wait(barrier.at[step])
 
     wg_idx = lax.axis_index(wg_axis)
     lax.cond(
@@ -595,8 +789,16 @@ def _compute_registers(
     memory_registers: int,
     num_compute_wgs: int,
 ) -> int:
-  """Returns the number of registers to use for the compute thread."""
-  # TODO(justinfu): Configure this per-platform.
-  n_registers = (512 - memory_registers) / num_compute_wgs
+  """Returns the max number of registers to use in compute threads.
+
+  We start with the theoretical max registers per thread if one wargroup
+  (128 threads) used the entire SM's 64k register file (64k / 128 = 512).
+  Then reserve `memory_registers` for the producer warpgroup and distribute
+  the remaining registers evenly among the compute warpgroups.
+
+  Note: The maximum number of registers per thread is 255, so we clamp
+  the value.
+  """
+  n_registers = min(256, (512 - memory_registers) / num_compute_wgs)
   # Round down to the nearest multiple of 8.
   return int((n_registers // 8) * 8)
