@@ -16,16 +16,19 @@
 
 from __future__ import annotations
 
+from collections.abc import Callable, Sequence
+from collections.abc import Hashable
 import enum
 import functools
 import string
-from typing import Any, Callable
+from typing import Any
 
 import jax
 from jax import lax
 from jax import tree_util
 from jax._src import ad_util
 from jax._src import api_util
+from jax._src import callback
 from jax._src import core as jax_core
 from jax._src import dtypes
 from jax._src import effects
@@ -39,8 +42,8 @@ from jax._src.interpreters import partial_eval as pe
 from jax._src.pallas import core as pallas_core
 from jax._src.state import discharge as state_discharge
 from jax._src.state import indexing
-from jax._src.state import types as state_types
 from jax._src.state import primitives as sp
+from jax._src.state import types as state_types
 from jax.interpreters import mlir
 import jax.numpy as jnp
 
@@ -125,10 +128,9 @@ def _atomic_rmw_discharge_rule(
     in_avals, out_avals, *args_flat, args_tree, atomic_type: AtomicOpType
 ):
   del out_avals  # Unused.
-  ref, indexers, val, mask = args_tree.unflatten(args_flat)
-  if len(indexers) > 1:
-    raise NotImplementedError("Only one indexer is supported.")
-  idx = indexers[0]
+  ref, transforms, val, mask = args_tree.unflatten(args_flat)
+  *prev_transforms, idx = transforms
+  ref = state_discharge.transform_array(ref, prev_transforms)
 
   if mask is not None:
     raise NotImplementedError
@@ -144,7 +146,7 @@ def _atomic_rmw_discharge_rule(
 
   if all((isinstance(s, Slice) or not s.shape) for s in idx.indices):
     indices = idx.indices
-    scalar_dims = [not isinstance(s, Slice) and s.shape == () for s in indices]
+    scalar_dims = [not isinstance(s, Slice) and not s.shape for s in indices]
     slice_starts = [s.start if isinstance(s, Slice) else s for s in indices]
     slice_sizes = tuple(s.size if isinstance(s, Slice) else 1 for s in indices)
     out_ones = lax.dynamic_slice(ref, slice_starts, slice_sizes=slice_sizes)
@@ -344,9 +346,9 @@ max_contiguous_p.def_impl(lambda x, **_: x)
 mlir.register_lowering(max_contiguous_p, lambda _, x, **__: [x])
 
 def max_contiguous(x, values):
-  if not isinstance(values, list):
-    values = [values]
-  return max_contiguous_p.bind(x, values=values)
+  if not isinstance(values, (list, tuple)):
+    values = (values,)
+  return max_contiguous_p.bind(x, values=tuple(values))
 
 @max_contiguous_p.def_abstract_eval
 def _max_contiguous_abstract_eval(aval, **_):
@@ -357,9 +359,8 @@ multiple_of_p = jax_core.Primitive("multiple_of")
 multiple_of_p.def_impl(lambda x, **_: x)
 mlir.register_lowering(multiple_of_p, lambda _, x, **__: [x])
 
-def multiple_of(x: jax.Array, values: list[int] | int) -> jax.Array:
-  if not isinstance(values, list):
-    values = [values]
+def multiple_of(x: jax.Array, values: Sequence[int] | int) -> jax.Array:
+  values = (values,) if isinstance(values, int) else tuple(values)
   return multiple_of_p.bind(x, values=values)
 
 @multiple_of_p.def_abstract_eval
@@ -371,9 +372,11 @@ load_p = jax_core.Primitive('masked_load')
 
 @load_p.def_effectful_abstract_eval
 def _load_abstract_eval(*avals_flat, args_tree, **_):
-  ref, indexers, _, _ = args_tree.unflatten(avals_flat)
+  ref, transforms, _, _ = args_tree.unflatten(avals_flat)
+  assert transforms is not None and isinstance(transforms[-1], NDIndexer)
+  transformed_ref = pallas_core.TransformedRef(ref, transforms)
   return (
-      jax_core.ShapedArray(indexers[-1].get_indexer_shape(), ref.dtype),
+      jax_core.ShapedArray(transformed_ref.shape, transformed_ref.dtype),
       {state.ReadEffect(0)},
   )
 
@@ -381,15 +384,12 @@ def _load_abstract_eval(*avals_flat, args_tree, **_):
 def _load_pp_rule(eqn, context, settings):
   # Pretty prints `a = load x i` as `x[i] <- a`
   y, = eqn.outvars
-  x, indexers, mask, other  = tree_util.tree_unflatten(eqn.params["args_tree"],
-                                                       eqn.invars)
+  x, transforms, mask, other = tree_util.tree_unflatten(
+      eqn.params["args_tree"], eqn.invars
+  )
   # TODO(sharadmv): pretty print mask and other
   lhs = jax_core.pp_vars([y], context, print_shapes=settings.print_shapes)
-  result = [
-      lhs,
-      pp.text(' <- '),
-      sp.pp_ref_transforms(context, x, indexers)
-  ]
+  result = [lhs, pp.text(" <- "), sp.pp_ref_transforms(context, x, transforms)]
   if mask is not None:
     result += [
         pp.text(" "),
@@ -407,18 +407,20 @@ jax_core.pp_eqn_rules[load_p] = _load_pp_rule
 
 
 def _load_jvp(primals, tangents, args_tree, **params):
-  ref_primal, indexers, mask, other_primal = args_tree.unflatten(primals)
+  ref_primal, transforms, mask, other_primal = args_tree.unflatten(primals)
   ref_tangent, _, _, other_tangent = args_tree.unflatten(tangents)
   if other_tangent is not None:
     other_tangent = ad_util.instantiate(other_tangent)
   return (
       load_p.bind(
-          *tree_util.tree_leaves((ref_primal, indexers, mask, other_primal)),
+          *tree_util.tree_leaves((ref_primal, transforms, mask, other_primal)),
           args_tree=args_tree,
           **params,
       ),
       load_p.bind(
-          *tree_util.tree_leaves((ref_tangent, indexers, mask, other_tangent)),
+          *tree_util.tree_leaves(
+              (ref_tangent, transforms, mask, other_tangent)
+          ),
           args_tree=args_tree,
           **params,
       ),
@@ -474,11 +476,10 @@ _unpad_values_to_avoid_dynamic_slice_oob_shift = partial(
 @state_discharge.register_discharge_rule(load_p)
 def _load_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
   del out_avals  # Unused.
-  ref, indexers, mask, other = args_tree.unflatten(args_flat)
-  # TODO(sharadmv): add support for multiple indexers
-  if len(indexers) > 1:
-    raise NotImplementedError("Only one indexer supported in discharge rule.")
-  idx = indexers[0]
+  ref, transforms, mask, other = args_tree.unflatten(args_flat)
+  *prev_transforms, idx = transforms
+  assert isinstance(idx, NDIndexer)
+  ref = state_discharge.transform_array(ref, prev_transforms)
   if all((isinstance(s, Slice) or not s.shape) for s in idx.indices):
     # TODO(ayx): support strided load/store in interpret mode.
     for s in idx.indices:
@@ -488,15 +489,15 @@ def _load_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
     scalar_dims = [not isinstance(s, Slice) and not s.shape for s in indices]
     slice_starts = [s.start if isinstance(s, Slice) else s for s in indices]
     slice_sizes = tuple(s.size if isinstance(s, Slice) else 1 for s in indices)
-    # fixes an inconstency with lax.dynamic_slice where if the slice goes out
+    # fixes an inconsistency with lax.dynamic_slice where if the slice goes out
     # of bounds, it will instead move the start_index backwards so the slice
     # will fit in memory.
     ref = _pad_values_to_avoid_dynamic_slice_oob_shift(ref, slice_sizes)
     idx_dtype = dtypes.canonicalize_dtype(jnp.int64)
     out_ones = lax.dynamic_slice(
-      ref,
-      [jnp.astype(s, idx_dtype) for s in slice_starts],
-      slice_sizes=slice_sizes,
+        ref,
+        [jnp.astype(s, idx_dtype) for s in slice_starts],
+        slice_sizes=slice_sizes,
     )
     out_indexer = tuple(0 if scalar else slice(None) for scalar in scalar_dims)
     out = out_ones[out_indexer]
@@ -514,20 +515,23 @@ swap_p = jax_core.Primitive('masked_swap')
 
 @swap_p.def_effectful_abstract_eval
 def _swap_abstract_eval(*avals_flat, args_tree, **_):
-  ref, indexers, val, _ = args_tree.unflatten(avals_flat)
-  expected_output_shape = indexers[-1].get_indexer_shape()
+  ref, transforms, val, _ = args_tree.unflatten(avals_flat)
+  assert transforms is not None and isinstance(transforms[-1], NDIndexer)
+  transformed_ref = pallas_core.TransformedRef(ref, transforms)
+  expected_output_shape = transformed_ref.shape
+  expected_output_dtype = transformed_ref.dtype
   if expected_output_shape != val.shape:
     raise ValueError(
         f"Invalid shape for `swap`. Ref shape: {ref.shape}. "
-        f"Value shape: {val.shape}. Indices: {indexers}. "
+        f"Value shape: {val.shape}. Transforms: {transforms}. "
     )
-  if ref.dtype != val.dtype:
+  if expected_output_dtype != val.dtype:
     raise ValueError(
-        f"Invalid dtype for `swap`. Ref dtype: {ref.dtype}. "
+        f"Invalid dtype for `swap`. Ref dtype: {expected_output_dtype}. "
         f"Value dtype: {val.dtype}. "
     )
   return (
-      jax_core.ShapedArray(expected_output_shape, ref.dtype),
+      jax_core.ShapedArray(expected_output_shape, expected_output_dtype),
       {state.WriteEffect(0)},
   )
 
@@ -537,8 +541,8 @@ def _swap_pp_rule(eqn, context, settings):
   # or:
   # Pretty prints `_ = swap x v i` as `x[i] <- v`
   y, = eqn.outvars
-  x, indexers, val, mask = eqn.params["args_tree"].unflatten(eqn.invars)
-  x_i = sp.pp_ref_transforms(context, x, indexers)
+  x, transforms, val, mask = eqn.params["args_tree"].unflatten(eqn.invars)
+  x_i = sp.pp_ref_transforms(context, x, transforms)
   if isinstance(y, jax_core.DropVar):
     return pp.concat([
         x_i,
@@ -564,17 +568,17 @@ jax_core.pp_eqn_rules[swap_p] = _swap_pp_rule
 
 
 def _swap_jvp(primals, tangents, *, args_tree, **params):
-  ref_primal, indexers, val_primal, mask = args_tree.unflatten(primals)
+  ref_primal, transforms, val_primal, mask = args_tree.unflatten(primals)
   ref_tangent, _, val_tangent, _ = args_tree.unflatten(tangents)
   val_tangent = ad_util.instantiate(val_tangent)
   return (
       swap_p.bind(
-          *tree_util.tree_leaves((ref_primal, indexers, val_primal, mask)),
+          *tree_util.tree_leaves((ref_primal, transforms, val_primal, mask)),
           args_tree=args_tree,
           **params,
       ),
       swap_p.bind(
-          *tree_util.tree_leaves((ref_tangent, indexers, val_tangent, mask)),
+          *tree_util.tree_leaves((ref_tangent, transforms, val_tangent, mask)),
           args_tree=args_tree,
           **params,
       ),
@@ -587,10 +591,10 @@ ad.primitive_jvps[swap_p] = _swap_jvp
 @state_discharge.register_discharge_rule(swap_p)
 def _swap_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
   del out_avals  # Unused.
-  ref, indexers, val, mask = args_tree.unflatten(args_flat)
-  if len(indexers) > 1:
-    raise NotImplementedError("Only one indexer supported in discharge rule.")
-  idx = indexers[0]
+  ref, transforms, val, mask = args_tree.unflatten(args_flat)
+  *prev_transforms, idx = transforms
+  assert isinstance(idx, NDIndexer)
+  ref = state_discharge.transform_array(ref, prev_transforms)
   if all((isinstance(s, Slice) or not s.shape) for s in idx.indices):
     # TODO(ayx): support strided load/store in interpret mode.
     for s in idx.indices:
@@ -694,14 +698,51 @@ def dot(a, b, trans_a: bool = False, trans_b: bool = False,
     if precision is not None:
       raise ValueError("Only one of allow_tf32 and precision can be specified")
     precision = lax.Precision.HIGH if allow_tf32 else lax.Precision.HIGHEST
+
+  def _handle_f8(dtype: jax.typing.DTypeLike):
+    """Ugly workaround to support float8_e4m3b11fnuz in dot."""
+    if dtype == jnp.float8_e4m3b11fnuz:
+      return jnp.bfloat16
+    return dtype
+
+  dtype = jnp.promote_types(_handle_f8(a.dtype), _handle_f8(b.dtype))
+  out_dtype = jnp.int32 if jnp.issubdtype(dtype, jnp.integer) else jnp.float32
   return jax.lax.dot_general(
       a,
       b,
       dimension_numbers=(((lhs_contract_dim,), (rhs_contract_dim,)), ((), ())),
       precision=precision,
-      preferred_element_type=jnp.float32,
+      preferred_element_type=out_dtype,
   )
 
+reciprocal_p = jax_core.Primitive("reciprocal")
+
+
+def reciprocal(x, *, approx=False):
+  return reciprocal_p.bind(x, approx=approx)
+
+
+@reciprocal_p.def_abstract_eval
+def _reciprocal_abstract_eval(x, *, approx):
+  del approx
+  return x
+
+
+def _reciprocal_lowering_rule(
+    ctx: mlir.LoweringRuleContext, x, *, approx=False
+):
+  def _reciprocal(x, *, approx=False):
+    if approx:
+      return jnp.reciprocal(x.astype(jnp.bfloat16)).astype(jnp.float32)
+    return jnp.reciprocal(x)
+
+  return mlir.lower_fun(_reciprocal, multiple_results=False)(
+      ctx, x, approx=approx
+  )
+
+
+mlir.register_lowering(reciprocal_p, _reciprocal_lowering_rule)
+
 
 class PrintEffect(effects.Effect):
   __str__ = lambda self: "Print"
@@ -811,7 +852,7 @@ batching.primitive_batchers[debug_print_p] = functools.partial(
 
 @functools.partial(mlir.register_lowering, debug_print_p)
 def debug_print_lowering_rule(ctx, *args, **params):
-  result, _, _ = mlir.emit_python_callback(
+  result, _, _ = callback.emit_python_callback(
       ctx,
       functools.partial(debug_print_p.impl, **params),
       None,
@@ -840,15 +881,31 @@ run_scoped_p = jax_core.Primitive("run_scoped")
 run_scoped_p.multiple_results = True
 
 
-def run_scoped(f: Callable[..., Any], *types: Any, **kw_types: Any) -> Any:
+def run_scoped(
+    f: Callable[..., Any],
+    *types: Any,
+    collective_axes: Hashable | tuple[Hashable, ...] = (),
+    **kw_types: Any,
+) -> Any:
   """Calls the function with allocated references and returns the result.
 
   The positional and keyword arguments describe which reference types
   to allocate for each argument. Each backend has its own set of reference
   types in addition to :class:`jax.experimental.pallas.MemoryRef`.
+
+  When `collective_axes` is specified, the same allocation will be returned for
+  all programs that only differ in their program ids along the collective axes.
+  It is an error not to call the same `run_scoped` in all programs along that
+  axis.
   """
+  if not isinstance(collective_axes, tuple):
+    collective_axes = (collective_axes,)
   flat_types, in_tree = tree_util.tree_flatten((types, kw_types))
-  flat_fun, out_tree_thunk = api_util.flatten_fun(lu.wrap_init(f), in_tree)
+  flat_fun, out_tree_thunk = api_util.flatten_fun(
+      lu.wrap_init(f,
+                   debug_info=api_util.debug_info("pallas run_scoped",
+                                                  f, types, kw_types)),
+      in_tree)
   # We allow ref avals to be transformed references.
   ref_avals = [t.get_ref_aval() for t in flat_types]
   avals = [
@@ -865,14 +922,14 @@ def run_scoped(f: Callable[..., Any], *types: Any, **kw_types: Any) -> Any:
   # parent scope). Jax can't reason about effects to references that
   # are not in the invars of an operation so we just put them all
   # there.
-  jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(flat_fun, avals)
-  out = run_scoped_p.bind(*consts, jaxpr=jaxpr)
+  jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(flat_fun, avals)
+  out = run_scoped_p.bind(*consts, jaxpr=jaxpr, collective_axes=collective_axes)
   return tree_util.tree_unflatten(out_tree_thunk(), out)
 
 
 @run_scoped_p.def_effectful_abstract_eval
-def _run_scoped_abstract_eval(*args, jaxpr):
-  del args
+def _run_scoped_abstract_eval(*args, jaxpr, collective_axes):
+  del args, collective_axes
   # jaxpr will have effects for its inputs (Refs that are allocated) and for
   # constvars (closed over Refs). The effects for the allocated Refs are local
   # to the jaxpr and shouldn't propagate out.
@@ -893,25 +950,34 @@ def _run_scoped_discharge_rule(
     out_avals,
     *args_flat,
     jaxpr,
-    **_):
+    collective_axes):
   del out_avals
+  if collective_axes:
+    raise NotImplementedError(
+        "run_scoped discharge does not support collective_axes yet."
+    )
   num_consts = len(args_flat)
+  # discharge_state only discharges invars, not consts, so in order to
+  # discharge the requested refs we need to move them to the invar set.
   jaxpr_noconst = pe.convert_constvars_jaxpr(jaxpr)
   num_return_values = len(jaxpr_noconst.outvars)
-  should_discharge = should_discharge + [
-      isinstance(var.aval, state.AbstractRef) for var in jaxpr.invars
-  ]
   discharged_body, new_consts = state_discharge.discharge_state(
-      jaxpr_noconst, [], should_discharge=should_discharge)
+      jaxpr_noconst,
+      [],
+      should_discharge=should_discharge + [False] * len(jaxpr.invars),
+  )
   if new_consts:
     raise NotImplementedError(
         "Cannot handle new consts created by state discharge.")
-  # Create inputs filled with uninitialized values to the body.
-  body_avals = [v.aval for v in discharged_body.invars[num_consts:]]
-  init_vals = [uninitialized_value(
-      aval.shape, aval.dtype) for aval in body_avals]
-  init_vals_with_consts = args_flat + tuple(init_vals)
-  out = jax_core.eval_jaxpr(discharged_body, [], *init_vals_with_consts)
+
+  # Lowering expects that the jaxpr.consts to be the eqn.invals.
+  discharged_body = pe.convert_invars_to_constvars(discharged_body, num_consts)
+
+  # Run_scoped discharged the external variables but the scoped ones
+  # are not discharged.
+  out = run_scoped_p.bind(
+      *args_flat, jaxpr=discharged_body, collective_axes=collective_axes
+  )
   # Order of outputs:
   # (1) return values, (2) closed refs, (3) scoped refs.
   return_values = out[:num_return_values]
@@ -919,8 +985,8 @@ def _run_scoped_discharge_rule(
   # We update all ref values with their updated values from the discharged
   # body. For other values we leave them in place.
   updates = [
-      ref_outputs.pop(0) if isinstance(aval, pallas_core.AbstractMemoryRef)
-      else None for aval in in_avals]
+      ref_outputs.pop(0) if should and isinstance(aval, state.AbstractRef)
+      else None for should, aval in zip(should_discharge, in_avals)]
   assert len(updates) == len(in_avals), f'{len(updates)} != {len(in_avals)}'
   return updates, return_values
 
@@ -930,18 +996,282 @@ state_discharge.register_partial_discharge_rule(run_scoped_p)(
 
 
 @functools.partial(mlir.register_lowering, run_scoped_p)
-def _run_scoped_lowering_rule(ctx, *args, jaxpr):
-  # This lowering rule gets triggered when run_scoped is not discharged.
-  # In this case there are no stateful effects to handle.
-  should_discharge = [
-      isinstance(aval, state.AbstractRef) for aval in ctx.avals_in
-  ]
+def _run_scoped_lowering_rule(ctx, *args, jaxpr, collective_axes):
+  if collective_axes:
+    raise ValueError(
+        "run_scoped lowering outside of Pallas does not support"
+        " collective_axes."
+    )
+  jaxpr_noconst = pe.convert_constvars_jaxpr(jaxpr)
+  num_return_values = len(jaxpr_noconst.outvars)
+  discharged_body, new_consts = state_discharge.discharge_state(
+      jaxpr_noconst, [], should_discharge=True)
+  if new_consts:    raise NotImplementedError(
+        "Cannot handle new consts created by state discharge.")
 
   def _lower_fun(*lower_fun_args):
-    updates, out = _run_scoped_discharge_rule(
-        should_discharge,
-        [], [], *lower_fun_args,
-        jaxpr=jaxpr)
-    assert len(updates) == 0, 'Cannot lower run_scoped with effects.'
-    return out
+    # Create inputs filled with uninitialized values to the body.
+    num_consts = len(lower_fun_args)
+    body_avals = [v.aval for v in discharged_body.invars[num_consts:]]
+    init_vals = [uninitialized_value(
+        aval.shape, aval.dtype) for aval in body_avals]
+    out = jax_core.eval_jaxpr(discharged_body, [], *lower_fun_args, *init_vals)
+    return out[:num_return_values]
+
   return mlir.lower_fun(_lower_fun, multiple_results=True)(ctx, *args)
+
+
+def _get_ref_and_transforms(ref):
+  if isinstance(ref, state.TransformedRef):
+    return ref.ref, ref.transforms
+  return ref, ()
+
+
+class DeviceIdType(enum.Enum):
+  MESH = "mesh"
+  LOGICAL = "logical"
+
+
+def check_sem_avals(
+    sem_aval, sem_transforms_avals, name, allowed_semaphore_types=None
+):
+  if allowed_semaphore_types is None:
+    allowed_semaphore_types = {
+        pallas_core.semaphore,
+        pallas_core.barrier_semaphore,
+        # For interpret mode.
+        pallas_core.SEMAPHORE_INTERPRET_DTYPE,
+    }
+  if not isinstance(sem_aval, state.AbstractRef):
+    raise ValueError(f"Cannot {name} on a non-semaphore Ref: {sem_aval}")
+  sem_shape = sem_aval.shape
+  if sem_transforms_avals:
+    sem_shape = sem_transforms_avals[-1].get_indexer_shape()
+  if sem_shape:
+    raise ValueError(f"Cannot {name} on a non-()-shaped semaphore: {sem_shape}")
+  sem_dtype = sem_aval.dtype
+  if not any(
+      jnp.issubdtype(sem_dtype, sem_type)
+      for sem_type in allowed_semaphore_types
+  ):
+    raise ValueError(
+        f"Must {name} semaphores of the following types:"
+        f" {allowed_semaphore_types}."
+    )
+
+
+def _transform_semaphore(ref_value, transforms, ref_aval):
+  """Helper function for indexing into a semaphore during state_discharge."""
+  if ref_value.shape == ref_aval.shape:
+    return state_discharge.transform_array(ref_value, transforms)
+  elif len(ref_value.shape) == 0:
+    return ref_value
+  else:
+    raise ValueError(
+        f"Semaphore value shape {ref_value.shape} does not match aval shape"
+        f" {ref_aval.shape}"
+    )
+
+
+semaphore_read_p = jax_core.Primitive("semaphore_read")
+semaphore_read_p.multiple_results = False
+
+
+def semaphore_read(sem_or_view):
+  ref, transforms = _get_ref_and_transforms(sem_or_view)
+  args = [ref, transforms]
+  flat_args, args_tree = tree_util.tree_flatten(args)
+  return semaphore_read_p.bind(*flat_args, args_tree=args_tree)
+
+@semaphore_read_p.def_abstract_eval
+def _semaphore_read_abstract_eval(
+    *avals,
+    args_tree,
+):
+  del avals, args_tree
+  return jax_core.ShapedArray((), jnp.dtype("int32"))
+
+def _semaphore_read_discharge_rule(in_avals,
+                                   out_avals,
+                                   *flat_args,
+                                   args_tree):
+  del out_avals
+  [ref, transforms] = args_tree.unflatten(flat_args)
+  sem_value = _transform_semaphore(ref, transforms, in_avals[0])
+  sem_value = sem_value.astype(jnp.int32)
+  return (None,) * len(in_avals), sem_value
+state_discharge.register_discharge_rule(semaphore_read_p)(
+    _semaphore_read_discharge_rule
+)
+
+
+semaphore_signal_p = jax_core.Primitive('semaphore_signal')
+semaphore_signal_p.multiple_results = True
+
+
+def semaphore_signal(
+    sem_or_view,
+    inc: int | jax.Array = 1,
+    *,
+    device_id: int | jax.Array | None | tuple[int | jax.Array, ...] = None,
+    device_id_type: DeviceIdType = DeviceIdType.MESH,
+    core_index: int | jax.Array | None = None,
+):
+  ref, transforms = _get_ref_and_transforms(sem_or_view)
+  inc = jnp.asarray(inc, dtype=jnp.int32)
+  args = [ref, transforms, inc, device_id, core_index]
+  flat_args, args_tree = tree_util.tree_flatten(args)
+  semaphore_signal_p.bind(
+      *flat_args,
+      args_tree=args_tree,
+      device_id_type=device_id_type,
+  )
+
+
+# TODO(sharadmv,ivyzheng): remove this once we use axis dicts primarily
+class CommsEffect(effects.Effect):
+  pass
+_comms_effect = CommsEffect()
+effects.lowerable_effects.add_type(CommsEffect)
+effects.control_flow_allowed_effects.add_type(CommsEffect)
+effects.remat_allowed_effects.add_type(CommsEffect)
+effects.custom_derivatives_allowed_effects.add_type(CommsEffect)
+
+
+@semaphore_signal_p.def_effectful_abstract_eval
+def _semaphore_signal_abstract_eval(
+    *avals,
+    args_tree,
+    device_id_type: DeviceIdType,
+):
+  del device_id_type
+  (
+      sem_aval,
+      sem_transforms_avals,
+      value_aval,
+      device_id_avals,
+      core_index_aval,
+  ) = tree_util.tree_unflatten(args_tree, avals)
+  check_sem_avals(sem_aval, sem_transforms_avals, "signal")
+  if value_aval.dtype != jnp.dtype("int32"):
+    raise ValueError("Must signal an int32 value.")
+  effs : set[effects.Effect] = set()
+  if device_id_avals is not None:
+    device_id_flat_avals = tree_util.tree_leaves(device_id_avals)
+    for aval in device_id_flat_avals:
+      if aval.dtype != jnp.dtype("int32"):
+        raise ValueError("`device_id`s must be an int32 value.")
+    effs.add(_comms_effect)
+  return [], effs
+
+
+def _semaphore_signal_pp_eqn(eqn: jax_core.JaxprEqn,
+                             context: jax_core.JaxprPpContext,
+                             settings: jax_core.JaxprPpSettings):
+  del settings
+  invars = eqn.invars
+  tree = eqn.params["args_tree"]
+  (
+      sem,
+      sem_transforms,
+      value,
+      device_ids,
+      _,
+  ) = tree_util.tree_unflatten(tree, invars)
+  out = pp.concat([
+      pp.text("semaphore_signal"),
+      pp.text(" "),
+      sp.pp_ref_transforms(context, sem, sem_transforms),
+      pp.text(" "),
+      pp.text(jax_core.pp_var(value, context)),
+  ])
+  if device_ids is not None:
+    flat_device_ids = tree_util.tree_leaves(device_ids)
+    if not flat_device_ids:
+      return out
+    device_ids_pp = [pp.text(jax_core.pp_var(flat_device_ids[0], context))]
+    for device_id in flat_device_ids[1:]:
+      device_ids_pp.append(pp.text(" "))
+      device_ids_pp.append(pp.text(jax_core.pp_var(device_id, context)))
+    out = pp.concat([out, pp.concat(device_ids_pp)])
+  return out
+jax_core.pp_eqn_rules[semaphore_signal_p] = _semaphore_signal_pp_eqn
+
+
+def _semaphore_signal_discharge_rule(in_avals,
+                                     out_avals,
+                                     *flat_args,
+                                     args_tree,
+                                     device_id_type):
+  del out_avals, device_id_type
+  [ref, transforms, inc, device_id, core_index] = args_tree.unflatten(flat_args)
+  if device_id is not None:
+    raise NotImplementedError("Remote signal not implemented.")
+  if core_index is not None:
+    raise NotImplementedError("Multiple core support not implemented.")
+  sem_value = _transform_semaphore(ref, transforms, in_avals[0])
+  inc = inc.astype(pallas_core.SEMAPHORE_INTERPRET_DTYPE)
+  _, new_sem_value = state_discharge.transform_swap_array(
+      ref, transforms, sem_value + inc
+  )
+  return (new_sem_value,) + (None,) * (len(in_avals) - 1), ()
+state_discharge.register_discharge_rule(semaphore_signal_p)(
+    _semaphore_signal_discharge_rule
+)
+
+
+semaphore_wait_p = jax_core.Primitive('semaphore_wait')
+semaphore_wait_p.multiple_results = True
+
+def semaphore_wait(sem_or_view, dec: int | jax.Array = 1):
+  ref, transforms = _get_ref_and_transforms(sem_or_view)
+  dec = jnp.asarray(dec, dtype=jnp.int32)
+  args = [ref, transforms, dec]
+  flat_args, args_tree = tree_util.tree_flatten(args)
+  semaphore_wait_p.bind(*flat_args, args_tree=args_tree)
+
+@semaphore_wait_p.def_abstract_eval
+def _semaphore_wait_abstract_eval(*avals, args_tree):
+  sem_aval, sem_transforms_avals, value_aval = tree_util.tree_unflatten(
+      args_tree, avals
+  )
+  check_sem_avals(sem_aval, sem_transforms_avals, "wait")
+  if value_aval.dtype != jnp.dtype("int32"):
+    raise ValueError("Must wait an int32 value.")
+  return []
+
+def _semaphore_wait_pp_eqn(eqn: jax_core.JaxprEqn,
+                             context: jax_core.JaxprPpContext,
+                             settings: jax_core.JaxprPpSettings):
+  del settings
+  invars = eqn.invars
+  tree = eqn.params["args_tree"]
+  (
+      sem,
+      sem_transforms,
+      value,
+  ) = tree_util.tree_unflatten(tree, invars)
+  return pp.concat([
+      pp.text("semaphore_wait"),
+      pp.text(" "),
+      sp.pp_ref_transforms(context, sem, sem_transforms),
+      pp.text(" "),
+      pp.text(jax_core.pp_var(value, context)),
+  ])
+jax_core.pp_eqn_rules[semaphore_wait_p] = _semaphore_wait_pp_eqn
+
+def _semaphore_wait_discharge_rule(in_avals,
+                                     out_avals,
+                                     *flat_args,
+                                     args_tree):
+  del out_avals
+  [ref, transforms, dec] = args_tree.unflatten(flat_args)
+  sem_value = _transform_semaphore(ref, transforms, in_avals[0])
+  dec = dec.astype(pallas_core.SEMAPHORE_INTERPRET_DTYPE)
+  _, new_sem_value = state_discharge.transform_swap_array(
+      ref, transforms, sem_value - dec
+  )
+  return (new_sem_value,) + (None,) * (len(in_avals) - 1), ()
+state_discharge.register_discharge_rule(semaphore_wait_p)(
+    _semaphore_wait_discharge_rule
+)
