@@ -17,7 +17,7 @@ from __future__ import annotations
 from collections.abc import Mapping, Sequence
 import functools
 
-from jax._src.util import safe_zip, use_cpp_class
+from jax._src.util import safe_zip, use_cpp_class, cache
 from jax._src import xla_bridge as xb
 from jax._src.lib import xla_client as xc
 from jax._src.op_shardings import (
@@ -30,20 +30,18 @@ Index = tuple[slice, ...]
 XLADeviceAssignment = Sequence[Device]
 
 
-@functools.lru_cache(maxsize=4096)
+@cache(max_size=4096, trace_context_in_key=False)
 def _addressable_devices_indices_map(
     sharding: Sharding, global_shape: Shape) -> Mapping[Device, Index | None]:
   global_map = sharding.devices_indices_map(global_shape)
   if sharding.is_fully_addressable:
     return global_map
-  if hasattr(sharding, '_internal_device_list'):
-    return {d: global_map[d]
-            for d in sharding._internal_device_list.addressable_device_list}
-  return {d: ind for d, ind in global_map.items()
-          if d.process_index == d.client.process_index()}
+  return {d: global_map[d]
+          for d in sharding._internal_device_list.addressable_device_list}  # type: ignore
 
-@functools.lru_cache(maxsize=4096)
-def common_devices_indices_map(s, global_shape: Shape) -> Mapping[Device, Index]:
+@cache(max_size=4096, trace_context_in_key=False)
+def common_devices_indices_map(
+    s: Sharding, global_shape: Shape) -> Mapping[Device, Index]:
   s.shard_shape(global_shape)  # raises a good error message
   hlo_sharding = s._to_xla_hlo_sharding(len(global_shape))
   indices = op_sharding_to_indices(hlo_sharding, global_shape,
@@ -51,7 +49,7 @@ def common_devices_indices_map(s, global_shape: Shape) -> Mapping[Device, Index]
   return dict(safe_zip(s._device_assignment, indices))
 
 
-@functools.lru_cache(maxsize=4096)
+@cache(max_size=4096, trace_context_in_key=False)
 def _common_shard_shape(self, global_shape: Shape) -> Shape:
   hlo_sharding = self._to_xla_hlo_sharding(len(global_shape))
   if is_op_sharding_replicated(hlo_sharding):
@@ -110,6 +108,11 @@ class Sharding:
     """
     raise NotImplementedError('Subclasses should implement this method.')
 
+  @property
+  def num_devices(self) -> int:
+    """Number of devices that the sharding contains."""
+    raise NotImplementedError('Subclasses should implement this method.')
+
   @property
   def memory_kind(self) -> str | None:
     """Returns the memory kind of the sharding."""
@@ -123,13 +126,23 @@ class Sharding:
   def _device_assignment(self) -> XLADeviceAssignment:
     raise NotImplementedError('Subclasses should implement this method.')
 
+  @property
+  def _internal_device_list(self) -> xc.DeviceList:
+    raise NotImplementedError('Subclasses should implement this method.')
+
   def _to_xla_hlo_sharding(self, num_dimensions: int) -> xc.HloSharding:
     raise NotImplementedError('Subclasses should implement this method.')
 
+  def _to_sdy_sharding(self, num_dimensions: int):
+    raise NotImplementedError('Subclasses should implement this method.')
 
   #############################################################################
   # Default implementations below that all subclasses will inherit.
 
+  @property
+  def _is_concrete(self) -> bool:
+    return True
+
   @functools.cached_property
   def addressable_devices(self) -> set[Device]:
     """The set of devices in the :class:`Sharding` that are addressable by the
@@ -162,10 +175,7 @@ class Sharding:
   def _addressable_device_assignment(self) -> XLADeviceAssignment:
     if self.is_fully_addressable:
       return self._device_assignment
-    if hasattr(self, '_internal_device_list'):
-      return tuple(self._internal_device_list.addressable_device_list)
-    return tuple(d for d in self._device_assignment
-                 if d.process_index == d.client.process_index())
+    return tuple(self._internal_device_list.addressable_device_list)  # type: ignore
 
   def shard_shape(self, global_shape: Shape) -> Shape:
     """Returns the shape of the data on each device.
@@ -180,10 +190,6 @@ class Sharding:
 
     Two shardings are equivalent if they place the same logical array shards on
     the same devices.
-
-    For example, a :class:`NamedSharding` may be equivalent
-    to a :class:`PositionalSharding` if both place the same shards of the array
-    on the same devices.
     """
     try:
       return (are_op_shardings_equal(self._to_xla_hlo_sharding(ndim),
