@@ -30,24 +30,24 @@ executable protocols described above.
 """
 from __future__ import annotations
 
-import functools
+import dataclasses
+import enum
 from collections.abc import Sequence
 from dataclasses import dataclass
 from typing import Any, NamedTuple, Protocol, Union, runtime_checkable
 
-import jax
-
 from jax._src import core
 from jax._src import config
+from jax._src import sharding as sharding_lib
 from jax._src import source_info_util
 from jax._src import traceback_util
 from jax._src import tree_util
 from jax._src import util
-from jax._src import mesh as mesh_lib
 from jax._src.sharding_impls import UnspecifiedValue, AUTO
-from jax._src.layout import Layout
+from jax._src.layout import Format, Layout, AutoLayout
 from jax._src.interpreters import mlir
 from jax._src.lib.mlir import ir
+from jax._src.lib import _jax
 from jax._src.lib import xla_client as xc
 
 
@@ -55,44 +55,55 @@ source_info_util.register_exclusion(__file__)
 traceback_util.register_exclusion(__file__)
 
 
-xla_extension = xc._xla
 map, unsafe_map = util.safe_map, map
 zip, unsafe_zip = util.safe_zip, zip
 
 CompilerOptions = dict[str, Union[str, bool]]
 
 
-# -- Internal protocols
+# -- Internal types
 
-class Executable(Protocol):
-  """Protocol for executables, which a user-facing ``Compiled`` encapsulates."""
+
+class Executable:
+
+  def xla_extension_executable(self) -> xc.LoadedExecutable:
+    raise NotImplementedError(
+        "compiled executable carries no loaded XLA executable. It may be "
+        f"that {type(self)} defines an incomplete implementation.")
 
   def call(self, *args_flat) -> Sequence[Any]:
     """Execute on the flat list of arguments, returning flat outputs."""
-    # TODO(frostig): improve annotation (sequences of arrays/buffers)
-    raise NotImplementedError
+    raise NotImplementedError("compiled executable does not support invocation")
 
-  def input_shardings(self) -> Sequence[jax.sharding.Sharding]:
+  def create_cpp_call(self, no_kwargs, in_tree, out_tree) -> Any:
+    """Optionally constructs a fast c++ dispatcher."""
+    return None
+
+  def input_shardings(self) -> Sequence[sharding_lib.Sharding]:
     """Flat sequence of input shardings.
 
     May raise ``NotImplementedError`` if unavailable, e.g. based on backend,
     compiler, or runtime.
     """
-    raise NotImplementedError
+    raise NotImplementedError(
+        "compiled executable carries no input sharding information")
 
-  def output_shardings(self) -> Sequence[jax.sharding.Sharding]:
+  def output_shardings(self) -> Sequence[sharding_lib.Sharding]:
     """Flat sequence of output shardings.
 
     May raise ``NotImplementedError`` if unavailable, e.g. based on backend,
     compiler, or runtime.
     """
-    raise NotImplementedError
+    raise NotImplementedError(
+        "compiled executable carries no output sharding information")
 
-  def input_layouts(self):
-    raise NotImplementedError
+  def input_formats(self):
+    raise NotImplementedError(
+        "compiled executable carries no input layout information")
 
-  def output_layouts(self):
-    raise NotImplementedError
+  def output_formats(self):
+    raise NotImplementedError(
+        "compiled executable carries no output layout information")
 
   def as_text(self) -> str:
     """A human-readable text representation of this executable.
@@ -103,7 +114,19 @@ class Executable(Protocol):
     May raise ``NotImplementedError`` if unavailable, e.g. based on backend,
     compiler, or runtime.
     """
-    raise NotImplementedError
+    xla_ext_exe = self.xla_extension_executable()
+    err_msg = ("text view unsupported on current XLA backend: "
+               f"{type(xla_ext_exe)}")
+    if not hasattr(xla_ext_exe, "hlo_modules"):
+      raise NotImplementedError(err_msg)
+    try:
+      return "\n\n".join([m.to_string() for m in xla_ext_exe.hlo_modules()])
+    except _jax.XlaRuntimeError as e:
+      msg, *_ = e.args
+      if type(msg) is str and msg.startswith("UNIMPLEMENTED"):
+        raise NotImplementedError(err_msg) from e
+      else:
+        raise
 
   def cost_analysis(self) -> Any:
     """A summary of execution cost estimates.
@@ -118,8 +141,27 @@ class Executable(Protocol):
     May raise ``NotImplementedError`` if unavailable, e.g. based on backend,
     compiler, or runtime.
     """
-    # TODO(frostig): improve annotation (arbitrary pytree)
-    raise NotImplementedError
+    xla_ext_exe = self.xla_extension_executable()
+
+    if hasattr(xla_ext_exe, "cost_analysis"):
+      try:
+        return xla_ext_exe.cost_analysis()
+      except _jax.XlaRuntimeError as e:
+        msg, *_ = e.args
+        if not (type(msg) is str and msg.startswith("UNIMPLEMENTED")):
+          raise
+
+    if (
+        xla_ext_exe is None
+        and hasattr(self, "unsafe_call")
+        and hasattr(self.unsafe_call, "compiled")
+        and hasattr(self.unsafe_call.compiled, "cost_analysis")
+    ):
+      return self.unsafe_call.compiled.cost_analysis()
+
+    raise NotImplementedError(
+        f"cost analysis unsupported on current XLA backend: {type(xla_ext_exe)}"
+    )
 
   def memory_analysis(self) -> Any:
     """A summary of estimated memory requirements.
@@ -134,8 +176,19 @@ class Executable(Protocol):
     May raise ``NotImplementedError`` if unavailable, e.g. based on backend,
     compiler, or runtime.
     """
-    # TODO(frostig): improve annotation (arbitrary pytree)
-    raise NotImplementedError
+    xla_ext_exe = self.xla_extension_executable()
+    err_msg = ("memory analysis unsupported on current XLA backend: "
+               f"{type(xla_ext_exe)}")
+    if not hasattr(xla_ext_exe, "get_compiled_memory_stats"):
+      raise NotImplementedError(err_msg)
+    try:
+      return xla_ext_exe.get_compiled_memory_stats()
+    except _jax.XlaRuntimeError as e:
+      msg, *_ = e.args
+      if type(msg) is str and msg.startswith("UNIMPLEMENTED"):
+        raise NotImplementedError(err_msg) from e
+      else:
+        raise
 
   def runtime_executable(self) -> Any:
     """An arbitrary object representation of this executable.
@@ -147,29 +200,52 @@ class Executable(Protocol):
     May raise ``NotImplementedError`` if unavailable, e.g. based on backend or
     compiler.
     """
-    raise NotImplementedError
-
-  def create_cpp_call(self, no_kwargs, in_tree, out_tree) -> Any:
-    """Optionally constructs a fast c++ dispatcher."""
-    return None
+    return self.xla_extension_executable()
 
 
-class Lowering(Protocol):
-  """Protocol for lowerings, which a user-facing ``Lowered`` encapsulates."""
+class Lowering:
+
+  compile_args: dict[str, Any]
+
+  def hlo(self) -> xc.XlaComputation:
+    """Return an HLO representation of this computation."""
+    hlo = self.stablehlo()
+    m: str | bytes
+    m = mlir.module_to_bytecode(hlo)
+    return _jax.mlir.mlir_module_to_xla_computation(
+        m, use_tuple_args=self.compile_args["tuple_args"])
+
+  def stablehlo(self) -> ir.Module:
+    """Return a StableHLO representation of this computation."""
+    raise NotImplementedError(
+        f"cost analysis unsupported on XLA computation: {type(self)}")
 
   def compile(
-      self, compiler_options: CompilerOptions | None = None) -> Executable:
+      self, compiler_options: CompilerOptions | None = None, *,
+      device_assignment: tuple[xc.Device, ...] | None = None) -> Executable:
     """Compile and return a corresponding ``Executable``."""
-    raise NotImplementedError
+    raise NotImplementedError(
+        f"cost analysis unsupported on XLA computation: {type(self)}")
 
-  def as_text(self, dialect: str | None = None, *,
+  def as_text(self, dialect: str | None = None,
+              *,
               debug_info: bool = False) -> str:
     """A human-readable text representation of this lowering.
 
     Intended for visualization and debugging purposes. This need not be a valid
     nor reliable serialization. It is relayed directly to external callers.
     """
-    raise NotImplementedError
+    if dialect is None:
+      dialect = "stablehlo"
+    if dialect == "stablehlo":
+      return mlir.module_to_string(self.stablehlo(),
+                                   enable_debug_info=debug_info)
+    elif dialect == "hlo":
+      print_opts = xc._xla.HloPrintOptions.short_parsable()
+      print_opts.print_metadata = debug_info
+      return self.hlo().as_hlo_module().to_string(print_opts)
+    else:
+      raise ValueError(f"unknown dialect: {dialect}")
 
   def compiler_ir(self, dialect: str | None = None) -> Any:
     """An arbitrary object representation of this lowering.
@@ -185,7 +261,14 @@ class Lowering(Protocol):
       dialect: Optional string specifying a representation dialect
       (e.g. "stablehlo")
     """
-    raise NotImplementedError
+    if dialect is None:
+      dialect = "stablehlo"
+    if dialect == "stablehlo":
+      return self.stablehlo()
+    elif dialect == "hlo":
+      return self.hlo()
+    else:
+      raise ValueError(f"unknown dialect: {dialect}")
 
   def cost_analysis(self) -> Any:
     """A summary of execution cost estimates.
@@ -205,165 +288,8 @@ class Lowering(Protocol):
     May raise ``NotImplementedError`` if unavailable, e.g. based on backend,
     compiler, or runtime.
     """
-    # TODO(frostig): improve annotation (arbitrary pytree)
-    raise NotImplementedError
-
-
-# -- Internal adapters from XLA-related objects to the above protocols
-
-class XlaExecutable(Executable):
-
-  def xla_extension_executable(self) -> xc.LoadedExecutable:
-    raise NotImplementedError("must override")
-
-  def call(self, *args_flat) -> Sequence[Any]:
-    raise NotImplementedError("must override")
-
-  def input_shardings(self) -> Sequence[jax.sharding.Sharding]:
     raise NotImplementedError(
-        "compiled executable carries no input sharding information")
-
-  def output_shardings(self) -> Sequence[jax.sharding.Sharding]:
-    raise NotImplementedError(
-        "compiled executable carries no output sharding information")
-
-  def input_layouts(self):
-    raise NotImplementedError(
-        "compiled executable carries no input layout information")
-
-  def output_layouts(self):
-    raise NotImplementedError(
-        "compiled executable carries no input layout information")
-
-  def as_text(self) -> str:
-    xla_ext_exe = self.xla_extension_executable()
-    err_msg = ("text view unsupported on current XLA backend: "
-               f"{type(xla_ext_exe)}")
-    if not hasattr(xla_ext_exe, "hlo_modules"):
-      raise NotImplementedError(err_msg)
-    try:
-      return "\n\n".join([m.to_string() for m in xla_ext_exe.hlo_modules()])
-    except xla_extension.XlaRuntimeError as e:
-      msg, *_ = e.args
-      if type(msg) is str and msg.startswith("UNIMPLEMENTED"):
-        raise NotImplementedError(err_msg) from e
-      else:
-        raise
-
-  def cost_analysis(self) -> dict[str, float]:
-    xla_ext_exe = self.xla_extension_executable()
-
-    # TODO(b/259255524): Unify/merge the two cost_analysis calls below.
-    if hasattr(xla_ext_exe, "cost_analysis"):
-      try:
-        return xla_ext_exe.cost_analysis()
-      except xla_extension.XlaRuntimeError as e:
-        msg, *_ = e.args
-        if not (type(msg) is str and msg.startswith("UNIMPLEMENTED")):
-          raise
-
-    # Try client method if executable cost_analysis method is unimplemented
-    if hasattr(xla_ext_exe, "client"):
-      try:
-        # TODO(b/384741132): We expect that the executable has only one
-        # HloModule. We should be able to remove this check once we update the
-        # Executable class to have only a single HloModule (see bug).
-        hlo_modules = xla_ext_exe.hlo_modules()
-        assert len(hlo_modules) == 1, (
-            f"Exectuable should have only one HloModule ({len(hlo_modules)})"
-            " were found)."
-        )
-
-        return xla_extension.hlo_module_cost_analysis(
-            xla_ext_exe.client, hlo_modules[0]
-        )
-      except xla_extension.XlaRuntimeError as e:
-        msg, *_ = e.args
-        supported = not (type(msg) is str and
-                         (msg.startswith("UNIMPLEMENTED") or
-                          "PjRt-compatible backend only" in msg))
-        if supported:
-          raise
-
-    if (
-        xla_ext_exe is None
-        and hasattr(self, "unsafe_call")
-        and hasattr(self.unsafe_call, "compiled")
-        and hasattr(self.unsafe_call.compiled, "cost_analysis")
-    ):
-      return self.unsafe_call.compiled.cost_analysis()
-
-    raise NotImplementedError(
-        f"cost analysis unsupported on current XLA backend: {type(xla_ext_exe)}"
-    )
-
-  def memory_analysis(self) -> Any:
-    xla_ext_exe = self.xla_extension_executable()
-    err_msg = ("memory analysis unsupported on current XLA backend: "
-               f"{type(xla_ext_exe)}")
-    if not hasattr(xla_ext_exe, "get_compiled_memory_stats"):
-      raise NotImplementedError(err_msg)
-    try:
-      return xla_ext_exe.get_compiled_memory_stats()
-    except xla_extension.XlaRuntimeError as e:
-      msg, *_ = e.args
-      if type(msg) is str and msg.startswith("UNIMPLEMENTED"):
-        raise NotImplementedError(err_msg) from e
-      else:
-        raise
-
-  def runtime_executable(self) -> Any:
-    return self.xla_extension_executable()
-
-
-class XlaLowering(Lowering):
-  """Adapts our various internal XLA-backed computations into a ``Lowering``."""
-
-  compile_args: dict[str, Any]
-
-  def hlo(self) -> xc.XlaComputation:
-    """Return an HLO representation of this computation."""
-    hlo = self.stablehlo()
-    m: str | bytes
-    m = mlir.module_to_bytecode(hlo)
-    return xla_extension.mlir.mlir_module_to_xla_computation(
-        m, use_tuple_args=self.compile_args["tuple_args"])
-
-  def stablehlo(self) -> ir.Module:
-    """Return a StableHLO representation of this computation."""
-    raise NotImplementedError("must override")
-
-  def compile(
-      self, compiler_options: CompilerOptions | None = None) -> Executable:
-    raise NotImplementedError("must override")
-
-  def as_text(self, dialect: str | None = None,
-              *,
-              debug_info: bool = False) -> str:
-    if dialect is None:
-      dialect = "stablehlo"
-    if dialect == "stablehlo":
-      return mlir.module_to_string(self.stablehlo(),
-                                   enable_debug_info=debug_info)
-    elif dialect == "hlo":
-      print_opts = xc._xla.HloPrintOptions.short_parsable()
-      print_opts.print_metadata = debug_info
-      return self.hlo().as_hlo_module().to_string(print_opts)
-    else:
-      raise ValueError(f"unknown dialect: {dialect}")
-
-  def compiler_ir(self, dialect: str | None = None) -> Any:
-    if dialect is None:
-      dialect = "stablehlo"
-    if dialect == "stablehlo":
-      return self.stablehlo()
-    elif dialect == "hlo":
-      return self.hlo()
-    else:
-      raise ValueError(f"unknown dialect: {dialect}")
-
-  def cost_analysis(self) -> dict[str, float]:
-    raise NotImplementedError("must override")
+        f"cost analysis unsupported on XLA computation: {type(self)}")
 
 
 # -- Public-facing API, plus helpers
@@ -382,13 +308,6 @@ class ArgInfo:
     return self._aval.dtype  # pytype: disable=attribute-error
 
 
-@dataclass(frozen=True)
-class OutInfo:
-  shape: tuple[int, ...]
-  dtype: jax.typing.DTypeLike
-  sharding: jax.sharding.Sharding | None = None
-
-
 class Stage:
   args_info: Any  # PyTree of ArgInfo
 
@@ -499,6 +418,14 @@ class Compiled(Stage):
     except NotImplementedError:
       return None
 
+  @property
+  def out_info(self):  # PyTree of jax.ShapeDtypeStruct
+    out_avals = self._executable.out_avals
+    out_formats_flat = self._output_formats_flat
+    return self.out_tree.unflatten(
+        [core.ShapeDtypeStruct(o.shape, o.dtype, sharding=f)
+         for o, f in zip(out_avals, out_formats_flat)])
+
   def runtime_executable(self) -> Any | None:
     """An arbitrary object representation of this executable.
 
@@ -511,37 +438,52 @@ class Compiled(Stage):
     """
     return self._executable.runtime_executable()
 
-  @property
-  def input_shardings(self):  # PyTree[sharding.Sharding]
-    shardings_flat = self._executable.input_shardings()
+  def _input_shardings_flat(self):
+    shardings_flat = self._executable._in_shardings
     # Some input shardings got DCE'd
     if self.in_tree.num_leaves > len(shardings_flat):
       iter_shardings_flat = iter(shardings_flat)
       shardings_flat = [next(iter_shardings_flat) if i in self._executable._kept_var_idx
                         else None for i in range(self.in_tree.num_leaves)]
+    return shardings_flat
+
+  @property
+  def input_shardings(self):  # -> PyTree[sharding.Sharding]
+    shardings_flat = self._input_shardings_flat()
     return tree_util.tree_unflatten(self.in_tree, shardings_flat)  # pytype: disable=attribute-error
 
   @property
-  def output_shardings(self):  # PyTree[sharding.Sharding]
-    shardings_flat = self._executable.output_shardings()
+  def output_shardings(self):  # -> PyTree[sharding.Sharding]
+    shardings_flat = self._executable._out_shardings
     return tree_util.tree_unflatten(self.out_tree, shardings_flat)  # pytype: disable=attribute-error
 
-  @property
-  def input_layouts(self):
-    layouts_flat = self._executable.input_layouts()
-    assert all(isinstance(l, Layout) for l in layouts_flat)
+  def _input_layouts_flat(self):
+    layouts_flat = self._executable._xla_in_layouts
     # Some input layouts got DCE'd
     if self.in_tree.num_leaves > len(layouts_flat):
       iter_layouts_flat = iter(layouts_flat)
       layouts_flat = [next(iter_layouts_flat) if i in self._executable._kept_var_idx
-                      else Layout() for i in range(self.in_tree.num_leaves)]
-    return tree_util.tree_unflatten(self.in_tree, layouts_flat)  # pytype: disable=attribute-error
+                      else None for i in range(self.in_tree.num_leaves)]
+    return layouts_flat
 
   @property
-  def output_layouts(self):
-    layouts_flat = self._executable.output_layouts()
+  def input_formats(self):
+    layouts_flat = self._input_layouts_flat()
+    shardings_flat = self._input_shardings_flat()
+    formats_flat = [Format(l, s) for l, s in zip(layouts_flat, shardings_flat)]
+    return tree_util.tree_unflatten(self.in_tree, formats_flat)  # pytype: disable=attribute-error
+
+  @property
+  def _output_formats_flat(self):
+    layouts_flat = self._executable._xla_out_layouts
+    shardings_flat = self._executable._out_shardings
     assert all(isinstance(l, Layout) for l in layouts_flat)
-    return tree_util.tree_unflatten(self.out_tree, layouts_flat)  # pytype: disable=attribute-error
+    return [Format(l, s) for l, s in zip(layouts_flat, shardings_flat)]
+
+  @property
+  def output_formats(self):
+    formats_flat = self._output_formats_flat
+    return tree_util.tree_unflatten(self.out_tree, formats_flat)  # pytype: disable=attribute-error
 
   @staticmethod
   def call(*args, **kwargs):
@@ -618,14 +560,14 @@ class Lowered(Stage):
   lowering paths (:func:`~jax.jit`, :func:`~jax.pmap`, etc.).
   """
   __slots__ = ["_lowering", "args_info", "out_tree", "_no_kwargs"]
-  _lowering: XlaLowering
+  _lowering: Lowering
   args_info: Any                # PyTree of ArgInfo
   out_tree: tree_util.PyTreeDef
   _no_kwargs: bool
 
   def __init__(
       self,
-      lowering: XlaLowering,
+      lowering: Lowering,
       args_info,  # PyTree of ArgInfo
       out_tree: tree_util.PyTreeDef,
       no_kwargs: bool = False):
@@ -637,7 +579,7 @@ class Lowered(Stage):
 
   @classmethod
   def from_flat_info(cls,
-                     lowering: XlaLowering,
+                     lowering: Lowering,
                      in_tree: tree_util.PyTreeDef,
                      in_avals,
                      donate_argnums: tuple[int, ...],
@@ -662,14 +604,24 @@ class Lowered(Stage):
   def out_info(self):  # PyTree of OutInfo
     out_avals = self._lowering.compile_args["global_out_avals"]
     out_shardings = self._lowering.compile_args["out_shardings"]
-    return self.out_tree.unflatten(
-        [OutInfo(o.shape, o.dtype, None if isinstance(s, (UnspecifiedValue, AUTO)) else s)
-         for o, s in zip(out_avals, out_shardings)])
+    out_layouts = self._lowering.compile_args["out_layouts"]
+    outs = []
+    for o, l, s in zip(out_avals, out_layouts, out_shardings):
+      s = None if isinstance(s, (UnspecifiedValue, AUTO)) else s
+      l = None if isinstance(l, AutoLayout) else l
+      format = Format(l, s)
+      outs.append(core.ShapeDtypeStruct(o.shape, o.dtype, sharding=format))
+    return self.out_tree.unflatten(outs)
 
   def compile(
-      self, compiler_options: CompilerOptions | None = None) -> Compiled:
+      self, compiler_options: CompilerOptions | None = None, *,
+      device_assignment: tuple[xc.Device, ...] | None = None) -> Compiled:
     """Compile, returning a corresponding ``Compiled`` instance."""
-    kw: dict[str, Any] = {"compiler_options": compiler_options}
+
+    kw: dict[str, Any] = {
+        "compiler_options": compiler_options,
+        "device_assignment": device_assignment
+    }
     return Compiled(
         self._lowering.compile(**kw),  # pytype: disable=wrong-keyword-args
         self.args_info,
@@ -733,49 +685,62 @@ class Lowered(Stage):
 
 
 class Traced(Stage):
+  """Traced form of a function specialized to argument types and values.
+
+  A traced computation is ready for lowering. This class carries the
+  traced representation with the remaining information needed to later
+  lower, compile, and execute it.
+  """
   __slots__ = ["jaxpr", "args_info", "fun_name", "_out_tree", "_lower_callable",
                "_args_flat", "_arg_names", "_num_consts"]
 
   def __init__(self, jaxpr: core.ClosedJaxpr, args_info, fun_name, out_tree,
-               lower_callable, abstract_mesh=None,
-               args_flat=None, arg_names=None, num_consts: int = 0):
+               lower_callable, args_flat=None, arg_names=None,
+               num_consts: int = 0, params_out_shardings=None):
     self.jaxpr = jaxpr
     self.args_info = args_info
     self.fun_name = fun_name
     self._out_tree = out_tree
     self._lower_callable = lower_callable
-    self._abstract_mesh = abstract_mesh
     self._args_flat = args_flat
     self._arg_names = arg_names
     self._num_consts = num_consts
+    self._params_out_shardings = params_out_shardings
 
   @property
   def out_info(self):
-    return self._out_tree.unflatten(
-        [OutInfo(o.shape, o.dtype) for o in self.jaxpr.out_avals])
+    return self.eval_shape()
 
   def lower(self, *, lowering_platforms: tuple[str, ...] | None = None,
             _private_parameters: mlir.LoweringParameters | None = None):
-    from jax._src.interpreters import pxla
-    from jax._src import pjit
-
+    """Lower to compiler input, returning a ``Lowered`` instance."""
     if _private_parameters is None:
       _private_parameters = mlir.LoweringParameters()
-    new_callable = functools.partial(
-        self._lower_callable, lowering_platforms=lowering_platforms,
-        lowering_parameters=_private_parameters)
     try:
-      # TODO(yashkatariya): Maybe thread this into pjit params like resource_env
-      # and set the context manager down the stack?
-      with mesh_lib.set_abstract_mesh(self._abstract_mesh):
-        lowering = new_callable()
-    except pxla.DeviceAssignmentMismatchError as e:
+      lowering = self._lower_callable(
+          lowering_platforms=lowering_platforms,
+          lowering_parameters=_private_parameters)
+    except DeviceAssignmentMismatchError as e:
       fails, = e.args
-      msg = pjit._device_assignment_mismatch_error(
+      msg = _device_assignment_mismatch_error(
           self.fun_name, fails, self._args_flat, 'jit', self._arg_names)
       raise ValueError(msg) from None
     return Lowered(lowering, self.args_info, self._out_tree)
 
+  def eval_shape(self):
+    out_shardings = [None if isinstance(s, UnspecifiedValue) else s
+                     for s in self._params_out_shardings]
+    out = []
+    for a, out_s in zip(self.jaxpr.out_avals, out_shardings):
+      s = (a.sharding if a.sharding.mesh._are_all_axes_explicit else out_s
+           if out_s is None else out_s)
+      # TODO(yashkatariya): Add `Layout` to SDS.
+      out.append(
+          core.ShapeDtypeStruct(
+              a.shape, a.dtype, sharding=s, weak_type=a.weak_type,
+              vma=(a.vma if config._check_vma.value else None)))
+    return tree_util.tree_unflatten(self._out_tree, out)
+
 
 @runtime_checkable
 class Wrapped(Protocol):
@@ -805,6 +770,8 @@ class Wrapped(Protocol):
   def lower(self, *args, **kwargs) -> Lowered:
     """Lower this function explicitly for the given arguments.
 
+    This is a shortcut for ``self.trace(*args, **kwargs).lower()``.
+
     A lowered function is staged out of Python and translated to a
     compiler's input language, possibly in a backend-dependent
     manner. It is ready for compilation but not yet compiled.
@@ -813,3 +780,108 @@ class Wrapped(Protocol):
       A ``Lowered`` instance representing the lowering.
     """
     raise NotImplementedError
+
+
+class MismatchType(enum.Enum):
+  ARG_SHARDING = 0
+  OUT_SHARDING = 1
+  SHARDING_INSIDE_COMPUTATION = 2
+  CONTEXT_DEVICES = 3
+  IN_SHARDING = 4
+
+  def __str__(self):
+    if self.name == 'IN_SHARDING':
+      return 'explicit input sharding'
+    elif self.name == 'OUT_SHARDING':
+      return 'explicit output sharding'
+    elif self.name == 'CONTEXT_DEVICES':
+      return 'context mesh'
+    return f'{self.name}'
+
+
+class SourceInfo(NamedTuple):
+  source_info: source_info_util.SourceInfo
+  eqn_name: str
+
+
+@dataclasses.dataclass
+class DeviceAssignmentMismatch:
+  da: Sequence[xc.Device]
+  m_type: MismatchType
+  source_info: SourceInfo | None
+
+  @property
+  def device_ids(self) -> Sequence[int]:
+    return [d.id for d in self.da]
+
+  @property
+  def platform(self) -> str:
+    return self.da[0].platform.upper()
+
+  def _maybe_api_name(self, api_name) -> str:
+    return f" {api_name}'s" if self.m_type == MismatchType.CONTEXT_DEVICES else ""
+
+  @property
+  def source_info_str(self):
+    return (
+        "" if self.source_info is None
+        else f" at {source_info_util.summarize(self.source_info.source_info)}"
+    )
+
+  @property
+  def _dev_ids_plat_str(self):
+    return f"device ids {self.device_ids} on platform {self.platform}"
+
+  def m_type_str(self, api_name):
+    return (f'{self.source_info and self.source_info.eqn_name} inside {api_name}'
+            if self.m_type == MismatchType.SHARDING_INSIDE_COMPUTATION else self.m_type)
+
+  def _str(self, api_name):
+    return (f"{self._maybe_api_name(api_name)} {self.m_type_str(api_name)} with "
+            f"{self._dev_ids_plat_str}{self.source_info_str}")
+
+
+class DeviceAssignmentMismatchError(Exception):
+  pass
+
+
+def _find_arg_mismatch(arg_list, fails, fun_name):
+  mismatched_args_msg = []
+  def mismatch(err):
+    for name, inp_da, aval in arg_list:
+      if err.m_type == MismatchType.ARG_SHARDING and err.da == inp_da:
+        mismatched_args_msg.append(
+            f"argument {name} of {fun_name} with shape {aval.str_short()} and "
+            f"{err._dev_ids_plat_str}")
+        break
+  first_err, second_err = fails
+  mismatch(first_err)
+  mismatch(second_err)
+  return mismatched_args_msg
+
+
+def _device_assignment_mismatch_error(fun_name, fails, args_flat, api_name,
+                                      arg_names):
+  arg_list = []
+  if arg_names is None:
+    arg_names = [''] * len(args_flat)
+  for a, n in zip(args_flat, arg_names):
+    da = (a.sharding._device_assignment
+          if getattr(a, 'sharding', None) is not None else None)
+    arg_list.append((n, da, core.shaped_abstractify(a)))
+
+  mismatched_args_msg = _find_arg_mismatch(arg_list, fails, fun_name)
+
+  if len(mismatched_args_msg) == 2:
+    first, second = mismatched_args_msg  # pytype: disable=bad-unpacking
+    extra_msg = f" Got {first} and {second}"
+  elif len(mismatched_args_msg) == 1:
+    first, second  = fails
+    # Choose the failure left which is not already covered by ARG_SHARDING.
+    left = second if first.m_type == MismatchType.ARG_SHARDING else first
+    extra_msg = f" Got {mismatched_args_msg[0]} and{left._str(api_name)}"
+  else:
+    first, second = fails
+    extra_msg = f" Got{first._str(api_name)} and{second._str(api_name)}"
+  msg = (f"Received incompatible devices for {api_name}ted computation.{extra_msg}")
+  return msg
