@@ -25,6 +25,8 @@ from jax._src import ad_util
 from jax._src import api_util
 from jax._src import core
 from jax._src import linear_util as lu
+from jax._src import pjit
+from jax._src import sharding_impls
 from jax._src import source_info_util
 from jax._src import tree_util
 from jax._src.interpreters import ad
@@ -45,6 +47,7 @@ from jax._src.state.types import (
 from jax._src.state.utils import bitcast, hoist_consts_to_refs
 from jax._src.typing import Array
 from jax._src.util import (
+    foreach,
     merge_lists,
     partition_list,
     safe_map,
@@ -70,7 +73,7 @@ PyTreeDef = tree_util.PyTreeDef
 # `Read/Write/Accum` effects.
 
 def discharge_state(jaxpr: core.Jaxpr, consts: Sequence[Any], * ,
-                    should_discharge: bool | Sequence[bool] = True
+                    should_discharge: bool | Sequence[bool] = True,
                     ) -> tuple[core.Jaxpr, list[Any]]:
   """Converts a jaxpr that takes in `Ref`s into one that doesn't."""
   if isinstance(should_discharge, bool):
@@ -79,8 +82,9 @@ def discharge_state(jaxpr: core.Jaxpr, consts: Sequence[Any], * ,
               if isinstance(v.aval, AbstractRef) and d
               else v.aval for v, d in zip(jaxpr.invars, should_discharge)]
   eval_jaxpr = lu.wrap_init(partial(_eval_jaxpr_discharge_state, jaxpr,
-                                    should_discharge, consts))
-  new_jaxpr, _ , new_consts, () = pe.trace_to_jaxpr_dynamic(eval_jaxpr, in_avals)
+                                    should_discharge, consts),
+                            debug_info=jaxpr.debug_info)
+  new_jaxpr, _ , new_consts = pe.trace_to_jaxpr_dynamic(eval_jaxpr, in_avals)
   return new_jaxpr, new_consts
 
 @dataclasses.dataclass
@@ -139,18 +143,16 @@ def _eval_jaxpr_discharge_state(
     *args: Any):
   env = Environment({})
 
-  map(env.write, jaxpr.constvars, consts)
+  foreach(env.write, jaxpr.constvars, consts)
   # Here some args may correspond to `Ref` avals but they'll be treated like
   # regular values in this interpreter.
-  map(env.write, jaxpr.invars, args)
+  foreach(env.write, jaxpr.invars, args)
 
   refs_to_discharge = {id(v.aval) for v, d in zip(jaxpr.invars, should_discharge)
                        if d and isinstance(v.aval, AbstractRef)}
 
   for eqn in jaxpr.eqns:
-    name_stack = (
-        source_info_util.current_name_stack() + eqn.source_info.name_stack
-    )
+    name_stack = source_info_util.current_name_stack() + eqn.source_info.name_stack
     traceback = eqn.source_info.traceback
     with source_info_util.user_context(
         traceback, name_stack=name_stack), eqn.ctx.manager:
@@ -163,16 +165,14 @@ def _eval_jaxpr_discharge_state(
         [invar], [outvar] = eqn.invars, eqn.outvars
         ans = env.read(invar)
         refs_to_discharge.remove(id(invar.aval))
-      elif (any(should_discharge)
-            or core.internal_mutable_array_effect in eqn.effects
-        ):
+      elif any(should_discharge) or core.internal_mutable_array_effect in eqn.effects:
         if eqn.primitive in _partial_discharge_rules:
           rule: DischargeRule = partial(_partial_discharge_rules[eqn.primitive], should_discharge)
         elif eqn.primitive in _discharge_rules:
           rule = _discharge_rules[eqn.primitive]
         else:
-          raise NotImplementedError("No state discharge rule implemented for "
-              f"primitive: {eqn.primitive}")
+          raise NotImplementedError(
+              f"No state discharge rule implemented for primitive: {eqn.primitive}")
         invals = map(env.read, eqn.invars)
         in_avals = [v.aval for v in eqn.invars]
         out_avals = [v.aval for v in eqn.outvars]
@@ -194,7 +194,7 @@ def _eval_jaxpr_discharge_state(
         ans = eqn.primitive.bind(*subfuns, *map(env.read, eqn.invars),
                                 **bind_params)
     if eqn.primitive.multiple_results:
-      map(env.write, eqn.outvars, ans)
+      foreach(env.write, eqn.outvars, ans)
     else:
       env.write(eqn.outvars[0], ans)
   # By convention, we return the outputs of the jaxpr first and then the final
@@ -206,14 +206,13 @@ def _eval_jaxpr_discharge_state(
   return out_vals + ref_vals
 
 def _is_trivial_indexer(indexer: indexing.NDIndexer):
+  """Returns whether the indexer selects the entire shape."""
   for s, idx in zip(indexer.shape, indexer.indices):
     if not isinstance(idx, indexing.Slice):
       return False
-    if not isinstance(idx.start, int):
+    if idx.is_dynamic_start or idx.is_dynamic_size:
       return False
-    if idx.start:
-      return False
-    if idx.size != s:
+    if idx.start != 0 or idx.size != s:
       return False
   return True
 
@@ -273,33 +272,97 @@ def _maybe_convert_to_dynamic_slice(
   return starts, sizes, squeeze_dims
 
 
-def _convert_to_array_indexer(indexer: indexing.NDIndexer
-                              ) -> tuple[int | Array, ...]:
+# In this code, indexing is handled in three ways: `slice`, `dynamic_slice`, and
+# gather. For the gather case, the goal is to create a gather array, which means
+# that we need to convert all other types of indexers into integer array
+# indexers. This is done by looping over all indexers and checking if they are
+# not integer array indexers, and if not, performing the conversion. However,
+# during this process, the indexing semantics may change. Specifically,
+# according to the indexing rules of NumPy, when there are integer array
+# indexers separated by other indexers, the axes corresponding to the integer
+# array indexers need to be moved to the front. After we convert all other
+# indexers to integer array indexers, the distinction between integer array
+# indexers and other types of indexers is lost. As a result, it becomes
+# impossible to determine which axes should be moved to the front. In this case,
+# we need to transpose the target array before the gather operation. We also
+# need to transpose the target array back after the gather operation, if it is
+# used in subsequent computations.
+def _maybe_transpose_before_gather(
+    indexer: indexing.NDIndexer
+) -> tuple[int, ...] | None:
+  is_int_indexing, _, _ = indexing.unpack_ndindexer(indexer)
+
+  int_indexers_contiguous = bool(
+      np.all(np.diff(np.where(is_int_indexing)[0]) == 1)
+  )
+  if int_indexers_contiguous:
+    return None  # no transpose needed
+
+  int_indexer_idxs: list[int] = []
+  non_int_indexer_idxs: list[int] = []
+  for i, is_int_index in enumerate(is_int_indexing):
+    (int_indexer_idxs if is_int_index else non_int_indexer_idxs).append(i)
+  transpose_order = (*int_indexer_idxs, *non_int_indexer_idxs)
+  return transpose_order
+
+
+def _perform_transpose_before_gather(
+    target_arr: Array,
+    indexer: indexing.NDIndexer,
+    transpose_order: tuple[int, ...],
+) -> tuple[Array, indexing.NDIndexer]:
+  new_target_arr = target_arr.transpose(transpose_order)
+  reordered_indices = tuple(indexer.indices[i] for i in transpose_order)
+  new_indexer = indexing.NDIndexer(
+      indices=reordered_indices,
+      shape=indexer.shape,
+      int_indexer_shape=indexer.int_indexer_shape,
+  )
+  return new_target_arr, new_indexer
+
+
+def _convert_to_gather_arrays(indexer: indexing.NDIndexer) -> tuple[Array, ...]:
   # This is the general gather case. We need to create the gather arrays.
-  is_integer_indexer, _, integer_indexer = (
-      indexing.unpack_ndindexer(indexer)
-  )
   total_shape = indexer.get_indexer_shape()
-  int_indexer_shape = indexer.int_indexer_shape
-  slice_shape = total_shape[len(int_indexer_shape):]
-  slice_dims = tuple(
-      i + len(int_indexer_shape) for i in range(len(slice_shape))
-  )
-  slice_dim_iter = iter(slice_dims)
-  slice_indexer: list[Array] = []
-  for idx, is_int_index in zip(indexer.indices, is_integer_indexer):
-    if not is_int_index:
-      assert isinstance(idx, indexing.Slice)
-      slice_indices = lax.broadcasted_iota(
-          np.dtype("int32"), total_shape, next(slice_dim_iter)
-      ) * idx.stride + idx.start
-      slice_indexer.append(slice_indices)
-      integer_indexer = tuple(
-          lax.expand_dims(idx, (-1,)) for idx in integer_indexer
+  is_int_indexing, _, _ = indexing.unpack_ndindexer(indexer)
+
+  if any(is_int_indexing):
+    n_idxers = len(indexer.indices)
+    int_indexer_shape = indexer.int_indexer_shape
+    n_int_indexers = sum(1 for p in is_int_indexing if p)
+    last_int_index_idx = n_idxers - 1 - is_int_indexing[::-1].index(True)
+    n_slice_index_dims_after_int = n_idxers - last_int_index_idx - 1
+
+  def get_idx_in_shape_after_indexing(i):
+    if not any(is_int_indexing):
+      return i
+
+    if i < n_idxers - n_slice_index_dims_after_int - n_int_indexers:
+      return i
+    if i < n_idxers - n_slice_index_dims_after_int:
+      raise ValueError
+    return i - n_int_indexers + len(int_indexer_shape)
+
+  arrs = []
+  for i, idxer in enumerate(indexer.indices):
+    if isinstance(idxer, indexing.Slice):
+      idx_in_shape_after_indexing = get_idx_in_shape_after_indexing(i)
+      arr = (
+          lax.iota(np.int32, total_shape[idx_in_shape_after_indexing])
+          * idxer.stride
+          + idxer.start
       )
-      continue
-  assert next(slice_dim_iter, None) is None
-  return tuple(merge_lists(is_integer_indexer, slice_indexer, integer_indexer))
+      diff = len(total_shape) - idx_in_shape_after_indexing - 1
+      arr = arr.reshape(arr.shape + (1,) * diff)
+      arrs.append(arr)
+    elif isinstance(idxer, (np.ndarray, Array)):
+      diff = n_idxers - 1 - last_int_index_idx
+      arr = idxer.reshape(idxer.shape + (1,) * diff)
+      arrs.append(arr)
+    else:
+      raise ValueError(f"Invalid type of idxer: {type(idxer).__name__}")
+
+  return tuple(arrs)
 
 
 @register_discharge_rule(get_p)
@@ -311,20 +374,8 @@ def _get_discharge_rule(
   y = _get_discharge(x, idx, tree)
   return (None,) * (len(idx) + 1), y
 
-def _prepend_gather(x, indexer):
-  # NumPy advanced int indexing won't prepend w/ only one dim, so add dummy.
-  return x[None][(np.array(0, 'int32'), *indexer)]
 
-def _prepend_scatter(x, indexer, val, *, add=False):
-  # NumPy advanced int indexing won't prepend w/ only one dim, so add dummy.
-  # However, since this is scatter, we need to remove the 1-sized dimension
-  # we added at the front.
-  if add:
-    return x[None].at[(0, *indexer)].add(val)[0]
-  return x[None].at[(0, *indexer)].set(val)[0]
-
-
-def _index_array(x, indexer):
+def _index_array(x, indexer: indexing.NDIndexer):
   if _is_trivial_indexer(indexer):
     return x
   # Try the three APIs in the following order: `lax.slice`,
@@ -334,13 +385,16 @@ def _index_array(x, indexer):
   # If everything in the indexer is a slice or ()-shaped, we can also
   # use `lax.dynamic_slice` with 1-sized slices for ()-shaped indices.
   # We need to squeeze out the 1-sized slices at the end.
-  elif maybe_slice := _maybe_convert_to_dynamic_slice(indexer):
-    starts, sizes, squeeze_dims = maybe_slice
+  elif maybe_dynamic_slice := _maybe_convert_to_dynamic_slice(indexer):
+    starts, sizes, squeeze_dims = maybe_dynamic_slice
     y = lax_slicing.dynamic_slice(x, starts, sizes)
     x = lax.squeeze(y, squeeze_dims)
   else:
-    indexer = _convert_to_array_indexer(indexer)
-    x = x[None][(np.array(0, "int32"), *indexer)]
+    transpose_order = _maybe_transpose_before_gather(indexer)
+    if transpose_order is not None:
+      x, indexer = _perform_transpose_before_gather(x, indexer, transpose_order)
+    arrays = _convert_to_gather_arrays(indexer)
+    x = x[arrays]
   return x
 
 
@@ -365,53 +419,79 @@ def transform_array(x, transforms):
 def transform_swap_array(x, transforms, val):
   if transforms is None:
     transforms = []
-  result = x
-  result_val = val
-  # Compute updated "val" (result).
-  _results = [x]
+
+  # Will hold the value read from `x` before the swap, and will have the same
+  # shape as `val`.
+  new_val = x
+  # List of intermediate results by transforming `x`.
+  intermediates = [x]
+
+  # Read phase (forward loop)
   for transform in transforms:
     match transform:
       case indexing.NDIndexer():
         indexer = transform
         if _is_trivial_indexer(indexer):
-          _results.append(_results[-1])
+          intermediates.append(intermediates[-1])
           continue
         # If everything in the indexer is a slice or ()-shaped, we can also
         # use `lax.dynamic_slice` with 1-sized slices for ()-shaped indices.
         # We need to squeeze out the 1-sized slices at the end.
         if maybe_slice := _maybe_convert_to_dynamic_slice(indexer):
           starts, sizes, squeeze_dims = maybe_slice
-          result_old = lax_slicing.dynamic_slice(result, starts, sizes)
-          result = lax.squeeze(result_old, squeeze_dims)
+          new_val = lax.squeeze(
+              lax_slicing.dynamic_slice(new_val, starts, sizes), squeeze_dims
+          )
         else:
-          indexer = _convert_to_array_indexer(indexer)
-          result = _prepend_gather(result, indexer)
-        _results.append(result)
+          transpose_order = _maybe_transpose_before_gather(indexer)
+          if transpose_order is not None:
+            new_val, indexer = _perform_transpose_before_gather(
+                new_val, indexer, transpose_order
+            )
+          arrays = _convert_to_gather_arrays(indexer)
+          new_val = new_val[arrays]
+          # Here, we don't need to transpose `new_val` back because it now holds
+          # the result of the indexing, and is no longer the original array that
+          # was indexed into.
+        intermediates.append(new_val)
       case RefBitcaster():
-        _results.append(bitcast(result, transform.dtype))
+        intermediates.append(bitcast(new_val, transform.dtype))
       case RefReshaper():
-        _results.append(result.reshape(transform.shape))
+        intermediates.append(new_val.reshape(transform.shape))
       case _:
         raise NotImplementedError(f"Unsupported transform: {transform}")
 
-  # Compute updated "x" (result_val)
-  for i, transform in reversed(list(enumerate(transforms))):
+  # Will hold the final state of the `x` after `val` has been written to the
+  # transformed location, and will have the same shape as `x`.
+  new_x = val
+
+  # Write phase (reversed loop)
+  for intermediate, transform in reversed(zip(intermediates[:-1], transforms)):
     if isinstance(transform, indexing.NDIndexer):
       indexer = transform
       if _is_trivial_indexer(indexer):
         continue
       if maybe_slice := _maybe_convert_to_dynamic_slice(indexer):
         starts, _, squeeze_dims = maybe_slice
-        result_val = lax.expand_dims(result_val, squeeze_dims)
-        result_val = lax_slicing.dynamic_update_slice(
-            _results[i], result_val, starts
+        new_x = lax_slicing.dynamic_update_slice(
+            intermediate, lax.expand_dims(new_x, squeeze_dims), starts
         )
       else:
-        indexer = _convert_to_array_indexer(indexer)
-        result_val = _prepend_scatter(_results[i], indexer, result_val)
+        transpose_order = _maybe_transpose_before_gather(indexer)
+        if transpose_order is not None:
+          intermediate, indexer = _perform_transpose_before_gather(
+              intermediate, indexer, transpose_order
+          )
+        arrays = _convert_to_gather_arrays(indexer)
+        new_x = intermediate.at[arrays].set(new_x)  # pytype: disable=attribute-error
+        if transpose_order is not None:
+          transpose_order_inversed = np.argsort(transpose_order)
+          new_x = new_x.transpose(transpose_order_inversed)
     else:
       raise NotImplementedError(f"Unsupported transform: {transform}")
-  return result, result_val
+
+  return new_val, new_x
+
 
 def _get_discharge(x, idx, tree):
   transforms = tree_util.tree_unflatten(tree, idx)
@@ -444,8 +524,10 @@ def _addupdate_discharge(x, val, idx, tree):
   if len(transforms) > 1:
     raise NotImplementedError("Only single indexer is supported.")
   indexer = transforms[0]
+
   if _is_trivial_indexer(indexer):
     return x + val
+
   # If everything in the indexer is a slice or ()-shaped, we can also
   # use `lax.dynamic_slice` with 1-sized slices for ()-shaped indices.
   # We need to squeeze out the 1-sized slices at the end.
@@ -455,16 +537,26 @@ def _addupdate_discharge(x, val, idx, tree):
     val = lax.expand_dims(val, squeeze_dims)
     y = lax_slicing.dynamic_update_slice(x, x_old + val, starts)
     return y
-  indexer = _convert_to_array_indexer(indexer)
-  return _prepend_scatter(x, indexer, val, add=True)
+
+  transpose_order = _maybe_transpose_before_gather(indexer)
+  if transpose_order is not None:
+    x, indexer = _perform_transpose_before_gather(x, indexer, transpose_order)
+  arrays = _convert_to_gather_arrays(indexer)
+  x = x.at[arrays].add(val)
+  if transpose_order is not None:
+    transpose_order_inversed = np.argsort(transpose_order)
+    x = x.transpose(transpose_order_inversed)
+  return x
+
 
 @weakref_lru_cache
-def _cached_closed_jaxpr_discharge(closed_jaxpr):
+def _cached_closed_jaxpr_discharge(closed_jaxpr: core.ClosedJaxpr):
   jaxpr, consts = closed_jaxpr.jaxpr, closed_jaxpr.consts
   num_outs = len(jaxpr.outvars)
   discharged_jaxpr, discharged_consts = discharge_state(jaxpr, consts)
   discharged_closed_jaxpr = core.ClosedJaxpr(discharged_jaxpr, discharged_consts)
-  fun = lu.wrap_init(core.jaxpr_as_fun(discharged_closed_jaxpr))
+  fun = lu.wrap_init(core.jaxpr_as_fun(discharged_closed_jaxpr),
+                     debug_info=discharged_jaxpr.debug_info)
   return discharged_closed_jaxpr, num_outs, fun
 
 @register_discharge_rule(core.closed_call_p)
@@ -546,6 +638,9 @@ def _run_state_abstract_eval(*avals: core.AbstractValue, jaxpr: core.Jaxpr,
       nonlocal_effects.add(
           eff.replace(input_index=inner_to_outer_aval_mapping[eff.input_index])
       )
+  assert len(jaxpr.invars) == len(is_initialized)
+  if not all(is_initialized):
+    raise NotImplementedError  # Uninitialized refs are not in avals.
   return avals, nonlocal_effects
 run_state_p.def_effectful_abstract_eval(_run_state_abstract_eval)
 
@@ -594,7 +689,6 @@ def _convert_outputs_to_writes(
   assert not jaxpr.constvars, "Jaxpr shouldn't have constvars."
 
   in_avals = [v.aval for v in jaxpr.invars]
-  @lu.wrap_init
   def eval_jaxpr(*refs):
     # We split the refs into the original input refs and the dummy residual
     # refs.
@@ -605,15 +699,16 @@ def _convert_outputs_to_writes(
     return []
   res_ref_avals = [AbstractRef(v.aval) if not isinstance(v.aval, AbstractRef)
                    else v.aval for v in jaxpr.outvars]
-  jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(
-      eval_jaxpr, [*in_avals, *res_ref_avals])
+  jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(
+      lu.wrap_init(eval_jaxpr,
+                   debug_info=jaxpr.debug_info),
+      [*in_avals, *res_ref_avals])
   assert not consts
   return jaxpr, [core.ShapedArray(a.shape, a.dtype) for a in res_ref_avals]
 
 def _convert_inputs_to_reads(num_res: int, jaxpr: core.Jaxpr) -> core.Jaxpr:
   assert not jaxpr.constvars, "Jaxpr should not have constvars"
 
-  @lu.wrap_init
   def eval_jaxpr(*refs):
     residual_refs, orig_refs = split_list(refs, [num_res])
     residual_vals = [r[...] for r in residual_refs]
@@ -624,8 +719,10 @@ def _convert_inputs_to_reads(num_res: int, jaxpr: core.Jaxpr) -> core.Jaxpr:
       split_list([v.aval for v in jaxpr.invars], [num_res])
   res_ref_avals = [AbstractRef(aval) if not isinstance(aval, AbstractRef) else
                    aval for aval in res_val_avals]
-  jaxpr, _, (), () = pe.trace_to_jaxpr_dynamic(
-      eval_jaxpr, [*res_ref_avals, *orig_ref_avals])
+  jaxpr, _, () = pe.trace_to_jaxpr_dynamic(
+      lu.wrap_init(eval_jaxpr,
+                   debug_info=jaxpr.debug_info),
+      [*res_ref_avals, *orig_ref_avals])
   return jaxpr
 
 def _run_state_partial_eval(trace: pe.JaxprTrace, *tracers: pe.JaxprTracer,
@@ -729,7 +826,7 @@ def _run_state_partial_eval(trace: pe.JaxprTrace, *tracers: pe.JaxprTracer,
                    is_initialized=(True,) * len(jaxpr_unknown.invars))
   _, eqn_effects = run_state_p.abstract_eval(*[v.aval for v in unknown_inputs],
                                              **uk_params)
-  eqn = pe.new_eqn_recipe(unknown_inputs, res_ref_unknown_outputs,
+  eqn = pe.new_eqn_recipe(trace, unknown_inputs, res_ref_unknown_outputs,
                           run_state_p, uk_params,
                           eqn_effects, source)
   for t in res_ref_unknown_outputs: t.recipe = eqn
@@ -802,7 +899,7 @@ def _run_state_partial_eval_custom(
   jaxpr_known, res_avals = _convert_outputs_to_writes(jaxpr_known_resout)
 
   # In a stateful partial_eval, the residuals should be `Ref`s.
-  res_avals = map(AbstractRef, res_avals)  # type: ignore
+  res_avals = map(AbstractRef, res_avals)
 
   known_invars, staged_invars = partition_list(in_unknowns, eqn.invars)
   known_outvars, staged_outvars = partition_list(in_unknowns, eqn.outvars)
@@ -825,7 +922,7 @@ def _run_state_partial_eval_custom(
   eqn_known = pe.new_jaxpr_eqn(known_and_res_invars,
                                [*known_outvars, *known_out_resvars],
                                run_state_p, known_params,
-                               known_effects, eqn.source_info)
+                               known_effects, eqn.source_info, eqn.ctx)
 
   jaxpr_staged = _convert_inputs_to_reads(len(res_avals), jaxpr_staged_resin_)
 
@@ -841,17 +938,18 @@ def _run_state_partial_eval_custom(
       *[v.aval for v in res_staged_invars], **staged_params)
   _, staged_outvars = partition_list(in_unknowns, eqn.outvars)
   if num_res:
-    @lu.wrap_init
+
     def staged(*args):
       out = run_state_p.bind(*args, **staged_params)
       return out[num_res:]
-    staged_call_jaxpr, _, (), () = pe.trace_to_jaxpr_dynamic(staged,
-                                                             [v.aval for v in res_staged_invars])
+    staged_call_jaxpr, _, () = pe.trace_to_jaxpr_dynamic(
+        lu.wrap_init(staged, debug_info=jaxpr_staged.debug_info),
+        [v.aval for v in res_staged_invars])
     eqn_staged = pe.new_jaxpr_eqn(res_staged_invars,
                                   staged_outvars,
                                   core.closed_call_p,
                                   dict(call_jaxpr=pe.close_jaxpr(staged_call_jaxpr)),
-                                  staged_effects, eqn.source_info)
+                                  staged_effects, eqn.source_info, eqn.ctx)
     assert len(res_staged_invars) == len(staged_call_jaxpr.invars)
     assert len(staged_outvars) == len(staged_call_jaxpr.outvars)
   else:
@@ -859,7 +957,7 @@ def _run_state_partial_eval_custom(
                                   staged_outvars,
                                   run_state_p,
                                   staged_params,
-                                  staged_effects, eqn.source_info)
+                                  staged_effects, eqn.source_info, eqn.ctx)
   new_vars = [*new_inst, *nonref_resvars, *ref_resvars]
   return eqn_known, eqn_staged, in_unknowns, in_unknowns, new_vars
 pe.partial_eval_jaxpr_custom_rules[run_state_p] = _run_state_partial_eval_custom
@@ -913,8 +1011,10 @@ def _transpose_jaxpr(jaxpr: core.Jaxpr, which_linear: Sequence[bool],
     _, ct_args = partition_list(used_cts, tangent_args)
     ad.backward_pass(tangent_jaxpr, False, (), (*primals_args, *ct_args), ())
     return []
-  jaxpr_trans, _, consts, () = pe.trace_to_jaxpr_dynamic(
-      lu.wrap_init(trans), [v.aval for v in jaxpr.invars])
+  jaxpr_trans, _, consts = pe.trace_to_jaxpr_dynamic(
+      lu.wrap_init(trans,
+                   debug_info=jaxpr.debug_info),
+      [v.aval for v in jaxpr.invars])
   return jaxpr_trans, consts
 
 def _run_state_transpose(in_cts, *args, jaxpr: core.Jaxpr,
@@ -985,26 +1085,31 @@ def _run_state_discharge_rule(in_avals: Sequence[core.AbstractValue],
   return new_invals, out_vals
 
 def initial_style_jaxpr(
-    fun: Callable, in_tree: PyTreeDef, in_avals: Sequence[core.AbstractValue]
+    fun: Callable, in_tree: PyTreeDef, in_avals: Sequence[core.AbstractValue],
+    dbg: core.DebugInfo,
   ) -> tuple[core.Jaxpr, list[Any], PyTreeDef]:
-  return _initial_style_jaxpr(fun, in_tree, tuple(in_avals))
+  return _initial_style_jaxpr(fun, in_tree, tuple(in_avals), dbg)
 
 @weakref_lru_cache
-def _initial_style_jaxpr(fun, in_tree, in_avals):
-  fun_, out_tree_thunk = api_util.flatten_fun_nokwargs(lu.wrap_init(fun),
+def _initial_style_jaxpr(fun: Callable,
+                         in_tree: api_util.PyTreeDef,
+                         in_avals: Sequence[core.AbstractValue],
+                         debug: core.DebugInfo):
+  fun_, out_tree_thunk = api_util.flatten_fun_nokwargs(
+      lu.wrap_init(fun, debug_info=debug),
       tree_util.treedef_tuple((in_tree,)))
-  debug = pe.tracing_debug_info(fun_, in_tree, out_tree_thunk, False, 'run_state')
-  jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(fun_, in_avals, debug)
+  jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(fun_, in_avals)
   return jaxpr, consts, out_tree_thunk()
 
 
 T = TypeVar('T')
 def run_state(f: Callable[..., None]) -> Callable[[T], T]:
   def wrapped(args):
+    dbg = api_util.debug_info("run_state", f, (args,), {})
     flat_args, in_tree = tree_util.tree_flatten(args)
     ref_avals, ref_args = unzip2(map(get_ref_aval_from_value, flat_args))
     # There may be some uninitialized values here in ref_args.
-    jaxpr_, consts, _ = initial_style_jaxpr(f, in_tree, ref_avals)
+    jaxpr_, consts, _ = initial_style_jaxpr(f, in_tree, ref_avals, dbg)
     jaxpr = hoist_consts_to_refs(jaxpr_)
     which_linear = (False,) * (len(consts) + len(ref_args))
     refs_is_initialized = tuple(r is not uninitialized for r in ref_args)
@@ -1020,9 +1125,10 @@ def run_state(f: Callable[..., None]) -> Callable[[T], T]:
 
 def run_state_reference(f: Callable[..., None]):
   def wrapped(args):
+    dbg = api_util.debug_info("run_state", f, (args,), {})
     flat_args, in_tree = tree_util.tree_flatten(args)
     ref_avals, ref_args = unzip2(map(get_ref_aval_from_value, flat_args))
-    jaxpr_, consts, _ = initial_style_jaxpr(f, in_tree, ref_avals)
+    jaxpr_, consts, _ = initial_style_jaxpr(f, in_tree, ref_avals, dbg)
     jaxpr = hoist_consts_to_refs(jaxpr_)
     discharged_jaxpr, discharged_consts = discharge_state(jaxpr, ())
 
@@ -1037,3 +1143,33 @@ def run_state_reference(f: Callable[..., None]):
     _, out_flat = split_list(out_const_flat, [len(consts)])
     return in_tree.unflatten(out_flat)
   return wrapped
+
+@register_discharge_rule(pjit.jit_p)
+def _pjit_state_discharge_rule(
+    in_avals, out_avals, *args, jaxpr, in_shardings, out_shardings,
+    in_layouts, out_layouts, **params):
+  if not all(isinstance(s, sharding_impls.UnspecifiedValue) for s in (*in_shardings, *out_shardings)):
+    raise NotImplementedError
+
+  if not (all(l is None for l in in_layouts) and
+          all(l is None for l in out_layouts)):
+    raise NotImplementedError
+
+  jaxpr, consts = jaxpr.jaxpr, jaxpr.consts
+  discharged_jaxpr, discharged_consts = discharge_state(jaxpr, consts)
+  discharged_closed_jaxpr = core.ClosedJaxpr(discharged_jaxpr, discharged_consts)
+  new_in_shardings = (sharding_impls.UNSPECIFIED,) * len(discharged_jaxpr.invars)
+  new_out_shardings = (sharding_impls.UNSPECIFIED,) * len(discharged_jaxpr.outvars)
+  new_in_layouts = (None,) * len(discharged_jaxpr.invars)
+  new_out_layouts = (None,) * len(discharged_jaxpr.outvars)
+  out_and_ref_vals = pjit.jit_p.bind(
+      *args, jaxpr=discharged_closed_jaxpr, in_shardings=new_in_shardings,
+      out_shardings=new_out_shardings, in_layouts=new_in_layouts,
+      out_layouts=new_out_layouts, **params)
+  out_vals, ref_vals = split_list(out_and_ref_vals, [len(jaxpr.outvars)])
+  ref_vals_iter = iter(ref_vals)
+  new_invals = tuple(next(ref_vals_iter) if isinstance(aval, AbstractRef)
+                     else None for aval in in_avals)
+  sentinel = object()
+  assert next(ref_vals_iter, sentinel) is sentinel
+  return new_invals, out_vals
