@@ -18,7 +18,8 @@ from __future__ import annotations
 from collections.abc import Sequence
 import dataclasses
 import math
-from typing import Any, Callable, Protocol, Union
+from typing import Any, Protocol, Union
+from collections.abc import Callable
 
 from jax._src import core
 from jax._src import dtypes
@@ -75,6 +76,7 @@ class AccumEffect(RefEffect):
   name: str = "Accum"
 
 effects.control_flow_allowed_effects.add_type(RefEffect)
+effects.partial_eval_kept_effects.add_type(RefEffect)
 
 StateEffect = Union[ReadEffect, WriteEffect, AccumEffect]
 
@@ -119,6 +121,16 @@ class RefBitcaster:
     del dtype  # Unused
     return self.dtype
 
+  def transform_sharding(self, sharding):
+    # If there are no explicit axes, do nothing.
+    if all(p is None for p in sharding.spec):
+      return sharding
+    raise NotImplementedError
+
+  def pretty_print(self, context: core.JaxprPpContext) -> pp.Doc:
+    del context  # Unused.
+    return pp.text(f"{{bitcast({self.dtype}{list(self.shape)}])}}")
+
 
 @tree_util.register_pytree_node_class
 @dataclasses.dataclass(frozen=True)
@@ -166,6 +178,16 @@ class RefReshaper:
     del dtype  # Unused
     return self.dtype
 
+  def transform_sharding(self, sharding):
+    # If there are no explicit axes, do nothing.
+    if all(p is None for p in sharding.spec):
+      return sharding
+    raise NotImplementedError
+
+  def pretty_print(self, context: core.JaxprPpContext) -> pp.Doc:
+    del context  # Unused.
+    return pp.text(f"{{reshape({self.dtype}{list(self.shape)})}}")
+
 
 class Transform(Protocol):
 
@@ -189,6 +211,13 @@ class Transform(Protocol):
     """
     return dtype
 
+  def transform_sharding(self, sharding):
+    if all(p is None for p in sharding.spec): return sharding # no explicit axes
+    raise NotImplementedError
+
+  def pretty_print(self, context: core.JaxprPpContext) -> pp.Doc:
+    return pp.text(f"{{{self}}}")
+
 
 @dataclasses.dataclass
 class RefIndexer:
@@ -227,7 +256,7 @@ class TransformedRef:
     if not unprocessed:
       return shape
     # If there are any unprocessed transforms left, we apply them to the shape
-    # we've found previuously.
+    # we've found previously.
     for t in self.transforms[-unprocessed:]:
       shape = t.transform_shape(shape)
     assert shape is not None
@@ -250,6 +279,9 @@ class TransformedRef:
     assert dtype is not None
     return dtype
 
+  ndim = property(lambda self: len(self.shape))
+  size = property(lambda self: math.prod(self.shape))
+
   @property
   def at(self) -> RefIndexer:
     return RefIndexer(self)
@@ -266,6 +298,18 @@ class TransformedRef:
         (*self.transforms, RefReshaper.from_ref_new_shape(self, *shape)),
     )
 
+  def set(self, value, idx=()):
+    from jax._src.state.primitives import ref_set  # pytype: disable=import-error
+    return ref_set(self, idx, value)
+
+  def swap(self, value, idx=()):
+    from jax._src.state.primitives import ref_swap  # pytype: disable=import-error
+    return ref_swap(self, idx, value)
+
+  def get(self, idx=()):
+    from jax._src.state.primitives import ref_get  # pytype: disable=import-error
+    return ref_get(self, idx)
+
   def __getattr__(self, name):
     return getattr(self.ref, name)
 
@@ -280,10 +324,11 @@ class TransformedRef:
 
 # We need an aval for `Ref`s so we can represent `get` and `swap` in Jaxprs.
 class AbstractRef(core.AbstractValue):
-  __slots__ = ["inner_aval"]
+  __slots__ = ["inner_aval", "memory_space"]
 
-  def __init__(self, inner_aval: core.AbstractValue):
+  def __init__(self, inner_aval: core.AbstractValue, memory_space: Any = None):
     self.inner_aval = inner_aval
+    self.memory_space = memory_space
 
   @property
   def weak_type(self) -> bool:
@@ -292,23 +337,29 @@ class AbstractRef(core.AbstractValue):
     return self.inner_aval.weak_type
 
   def update_weak_type(self, weak_type):
-    return AbstractRef(self.inner_aval.update_weak_type(weak_type))
+    return self.update(inner_aval=self.inner_aval.update_weak_type(weak_type))
 
-  def update(self, inner_aval=None):
-    if inner_aval is None:
-      return AbstractRef(self.inner_aval)
-    return AbstractRef(inner_aval)
+  def update(self, inner_aval=None, memory_space=None):
+    inner_aval = self.inner_aval if inner_aval is None else inner_aval
+    memory_space = self.memory_space if memory_space is None else memory_space
+    return AbstractRef(inner_aval, memory_space)
 
   ndim = property(lambda self: len(self.shape))
   size = property(lambda self: math.prod(self.shape))
 
+  def _len(self, ignored_tracer) -> int:
+    try:
+      return self.shape[0]
+    except IndexError as err:
+      raise TypeError("len() of unsized object") from err  # same as numpy error
+
   @property
   def shape(self):
     try:
       return self.inner_aval.shape  # pytype: disable=attribute-error
     except AttributeError:
       raise AttributeError(
-          f"`Ref{{{self.inner_aval.str_short()}}} has no `shape`."
+          f"{self!r} has no `shape`."
       ) from None
 
   @property
@@ -317,7 +368,25 @@ class AbstractRef(core.AbstractValue):
       return self.inner_aval.dtype  # pytype: disable=attribute-error
     except AttributeError:
       raise AttributeError(
-          f"`Ref{{{self.inner_aval.str_short()}}} has no `dtype`."
+          f"{self!r} has no `dtype`."
+      ) from None
+
+  @property
+  def sharding(self):
+    try:
+      return self.inner_aval.sharding  # pytype: disable=attribute-error
+    except AttributeError:
+      raise AttributeError(
+          f"{self!r} has no `sharding`."
+      ) from None
+
+  @property
+  def vma(self):
+    try:
+      return self.inner_aval.vma  # pytype: disable=attribute-error
+    except AttributeError:
+      raise AttributeError(
+          f"{self!r} has no `vma`."
       ) from None
 
   @core.aval_property
@@ -338,6 +407,12 @@ class AbstractRef(core.AbstractValue):
     from jax._src.state.primitives import ref_get  # pytype: disable=import-error
     return ref_get(tracer, idx)
 
+  @core.aval_method
+  @staticmethod
+  def swap(tracer, value, idx=()):
+    from jax._src.state.primitives import ref_swap  # pytype: disable=import-error
+    return ref_swap(tracer, idx, value)
+
   @core.aval_method
   @staticmethod
   def set(tracer, value, idx=()):
@@ -353,23 +428,28 @@ class AbstractRef(core.AbstractValue):
     return ref_set(tracer, idx, value)
 
   def __repr__(self) -> str:
+    if self.memory_space is not None:
+      return f'Ref<{self.memory_space}>{{{self.inner_aval.str_short()}}}'
     return f'Ref{{{self.inner_aval.str_short()}}}'
 
   def to_tangent_aval(self):
-    return AbstractRef(self.inner_aval.to_tangent_aval())
+    return AbstractRef(self.inner_aval.to_tangent_aval(), self.memory_space)
 
   def __eq__(self, other):
-    return (type(self) is type(other) and self.inner_aval == other.inner_aval)
+    return (type(self) is type(other) and self.inner_aval == other.inner_aval
+            and self.memory_space == other.memory_space)
 
   def __hash__(self):
-    return hash((self.__class__, self.inner_aval))
+    return hash((self.__class__, self.inner_aval, self.memory_space))
 
 def _map_ref(size, axis, ref_aval):
-  return AbstractRef(core.mapped_aval(size, axis, ref_aval.inner_aval))
+  return AbstractRef(core.mapped_aval(size, axis, ref_aval.inner_aval),
+                     ref_aval.memory_space)
 
-def _unmap_ref(size, axis_name, axis, ref_aval):
-  return AbstractRef(core.unmapped_aval(size, axis_name, axis,
-                                        ref_aval.inner_aval))
+def _unmap_ref(size, axis, explicit_mesh_axis, ref_aval):
+  return AbstractRef(core.unmapped_aval(
+      size, axis, ref_aval.inner_aval, explicit_mesh_axis),
+                     ref_aval.memory_space)
 
 core.aval_mapping_handlers[AbstractRef] = (_map_ref, _unmap_ref)
 
@@ -384,20 +464,13 @@ def shaped_array_ref(
     shape: tuple[int, ...], dtype, weak_type: bool = False) -> AbstractRef:
   return AbstractRef(core.ShapedArray(shape, dtype, weak_type=weak_type))
 
-def _shard_ref(mesh, names, ref_aval: AbstractRef):
-  del mesh
-  if names:
-    # Can't actually shard a ref, can only close over it.
-    raise NotImplementedError("Can't shard a Ref.")
-  return ref_aval
+def _shard_ref(mesh, auto, check_rep, names, ref_aval: AbstractRef):
+  aval = core.shard_aval(mesh, auto, check_rep, names, ref_aval.inner_aval)
+  return AbstractRef(aval)
 core.shard_aval_handlers[AbstractRef] = _shard_ref
 
-def _unshard_ref(mesh, names, ref_aval: AbstractRef):
-  del mesh
-  if names:
-    # Can't actually shard a ref, can only close over it.
-    raise NotImplementedError("Can't unshard a Ref")
-  return ref_aval
+def _unshard_ref(mesh, check_rep, names, ref_aval: AbstractRef):
+  raise TypeError("can't unshard a ref")
 core.unshard_aval_handlers[AbstractRef] = _unshard_ref
 
 
