@@ -21,7 +21,7 @@ import functools
 from functools import partial
 import operator as op
 import textwrap
-from typing import Any, NamedTuple, TypeVar, overload
+from typing import Any, TypeVar, overload
 
 from jax._src import traceback_util
 from jax._src.lib import pytree
@@ -202,8 +202,11 @@ def all_leaves(iterable: Iterable[Any],
   if is_leaf is None:
     return pytree.all_leaves(default_registry, iterable)
   else:
-    lst = list(iterable)
-    return lst == tree_leaves(lst, is_leaf)
+    items = list(iterable)
+    leaves = tree_leaves(items, is_leaf)
+    return len(leaves) == len(items) and all(
+        item is leaf for item, leaf in zip(items, leaves, strict=True)
+    )
 
 
 _Children = TypeVar("_Children", bound=Iterable[Any])
@@ -287,14 +290,14 @@ def register_pytree_node(
     >>> jax.jit(f)(m)
     Array([1., 2., 3., 4., 5.], dtype=float32)
   """
-  default_registry.register_node(  # type: ignore[call-arg]
-      nodetype, flatten_func, unflatten_func, flatten_with_keys_func   # type: ignore[arg-type]
+  default_registry.register_node(
+      nodetype, flatten_func, unflatten_func, flatten_with_keys_func
   )
-  none_leaf_registry.register_node(  # type: ignore[call-arg]
-      nodetype, flatten_func, unflatten_func, flatten_with_keys_func   # type: ignore[arg-type]
+  none_leaf_registry.register_node(
+      nodetype, flatten_func, unflatten_func, flatten_with_keys_func
   )
-  dispatch_registry.register_node(  # type: ignore[call-arg]
-      nodetype, flatten_func, unflatten_func, flatten_with_keys_func   # type: ignore[arg-type]
+  dispatch_registry.register_node(
+      nodetype, flatten_func, unflatten_func, flatten_with_keys_func
   )
   _registry[nodetype] = _RegistryEntry(flatten_func, unflatten_func)
 
@@ -362,6 +365,8 @@ def tree_map(f: Callable[..., Any],
 def build_tree(treedef: PyTreeDef, xs: Any) -> Any:
   """Build a treedef from a nested iterable structure
 
+  DEPRECATED: Use :func:`jax.tree.unflatten` instead.
+
   Args:
     treedef: the PyTreeDef structure to build.
     xs: nested iterables matching the arity as the treedef
@@ -376,13 +381,6 @@ def build_tree(treedef: PyTreeDef, xs: Any) -> Any:
     >>> import jax
     >>> tree = [(1, 2), {'a': 3, 'b': 4}]
     >>> treedef = jax.tree.structure(tree)
-
-    Both ``build_tree`` and :func:`jax.tree_util.tree_unflatten` can reconstruct
-    the tree from new values, but ``build_tree`` takes these values in terms of
-    a nested rather than flat structure:
-
-    >>> jax.tree_util.build_tree(treedef, [[10, 11], [12, 13]])
-    [(10, 11), {'a': 12, 'b': 13}]
     >>> jax.tree_util.tree_unflatten(treedef, [10, 11, 12, 13])
     [(10, 11), {'a': 12, 'b': 13}]
   """
@@ -460,6 +458,42 @@ def tree_reduce(function: Callable[[T, Any], T],
     return functools.reduce(function, tree_leaves(tree, is_leaf=is_leaf), initializer)
 
 
+class Unspecified:
+  pass
+
+
+def _parallel_reduce(
+    sequence: list[T],
+    operation: Callable[[T, T], T],
+    identity: T | Unspecified = Unspecified(),
+) -> T:
+  length = len(sequence)
+  if length == 0:
+    if isinstance(identity, Unspecified):
+      raise TypeError("Must specify identity for parallel reduction of empty sequence.")
+    return identity
+  elif length == 1:
+    return sequence[0]
+  else:
+    index = length // 2
+    a = _parallel_reduce(sequence[:index], operation, identity)
+    b = _parallel_reduce(sequence[index:], operation, identity)
+    return operation(a, b)
+
+
+@export
+def tree_reduce_associative(
+    operation: Callable[[T, T], T],
+    tree: Any,
+    *,
+    identity: T | Unspecified = Unspecified(),
+    is_leaf: Callable[[Any], bool] | None = None,
+) -> T:
+  """Alias of :func:`jax.tree.reduce_associative`."""
+  sequence = tree_leaves(tree, is_leaf=is_leaf)
+  return _parallel_reduce(sequence, operation, identity)
+
+
 @export
 def tree_all(tree: Any, *, is_leaf: Callable[[Any], bool] | None = None) -> bool:
   """Alias of :func:`jax.tree.all`."""
@@ -534,7 +568,7 @@ class Partial(functools.partial):
   >>> print_zero()
   0
   >>> call_func(print_zero)  # doctest:+ELLIPSIS
-  Traced<ShapedArray(int32[], weak_type=True)>with<DynamicJaxprTrace...>
+  JitTracer<~int32[]>
   """
 
   def __new__(klass, func, *args, **kw):
@@ -562,17 +596,42 @@ register_pytree_node(
 )
 
 
-# broadcast_prefix is not exported.
+@export
+def tree_broadcast(prefix_tree: Any, full_tree: Any,
+                   is_leaf: Callable[[Any], bool] | None = None
+                  ) -> Any:
+  """Alias of :func:`jax.tree.broadcast`."""
+  broadcast_leaves = broadcast_prefix(prefix_tree, full_tree, is_leaf=is_leaf)
+  return tree_structure(full_tree).unflatten(broadcast_leaves)
+
+
+# broadcast_prefix is not exported
 def broadcast_prefix(prefix_tree: Any, full_tree: Any,
                      is_leaf: Callable[[Any], bool] | None = None
                      ) -> list[Any]:
-  # If prefix_tree is not a tree prefix of full_tree, this code can raise a
-  # ValueError; use prefix_errors to find disagreements and raise more precise
-  # error messages.
+  """Broadcasts tree prefix leaves into the full set of leaves for a given full tree.
+
+    Args:
+      prefix_tree: a pytree that is a tree prefix of full_tree.
+      full_tree: a pytree with the structure to broadcast the prefix leaves into.
+      is_leaf: an optionally specified function that will be called at each
+        flattening step. It should return a boolean, with true stopping the
+        traversal and the whole subtree being treated as a leaf, and false
+        indicating the flattening should traverse the current object.
+
+    Returns:
+      A list of leaves matching the expected count for the full tree,
+      with the leaf of each prefix tree being duplicated to match the count of
+      its corresponding subtree.
+  """
   result = []
   num_leaves = lambda t: tree_structure(t).num_leaves
   add_leaves = lambda x, subtree: result.extend([x] * num_leaves(subtree))
-  tree_map(add_leaves, prefix_tree, full_tree, is_leaf=is_leaf)
+  try:
+    tree_map(add_leaves, prefix_tree, full_tree, is_leaf=is_leaf)
+  except ValueError:
+      e, *_ = prefix_errors(prefix_tree, full_tree)
+      raise e('broadcast_prefix prefix_tree') from None
   return result
 
 
@@ -715,65 +774,56 @@ def _equality_errors(path, t1, t2, is_leaf):
     yield from _equality_errors((*path, k), c1, c2, is_leaf)
 
 
-SequenceKey: Any = pytree.SequenceKey  # type: ignore
-DictKey: Any = pytree.DictKey  # type: ignore
-GetAttrKey: Any = pytree.GetAttrKey  # type: ignore
-FlattenedIndexKey: Any = pytree.FlattenedIndexKey  # type: ignore
+SequenceKey: Any = pytree.SequenceKey
+DictKey: Any = pytree.DictKey
+GetAttrKey: Any = pytree.GetAttrKey
+FlattenedIndexKey: Any = pytree.FlattenedIndexKey
 
 
 @export
-def keystr(keys: KeyPath):
+def keystr(keys: KeyPath, *, simple: bool = False, separator: str = '') -> str:
   """Helper to pretty-print a tuple of keys.
 
   Args:
     keys: A tuple of ``KeyEntry`` or any class that can be converted to string.
+    simple: If True, use a simplified string representation for keys. The
+      simple representation of keys will be more compact than the default, but
+      is ambiguous in some cases (for example "0" might refer to the first item
+      in a list or a dictionary key for the integer 0 or string "0").
+    separator: The separator to use to join string representations of the keys.
 
   Returns:
     A string that joins all string representations of the keys.
 
   Examples:
     >>> import jax
-    >>> keys = (0, 1, 'a', 'b')
-    >>> jax.tree_util.keystr(keys)
-    '01ab'
+    >>> params = {'foo': {'bar': {'baz': 1, 'bat': [2, 3]}}}
+    >>> for path, _ in jax.tree_util.tree_leaves_with_path(params):
+    ...   print(jax.tree_util.keystr(path))
+    ['foo']['bar']['bat'][0]
+    ['foo']['bar']['bat'][1]
+    ['foo']['bar']['baz']
+    >>> for path, _ in jax.tree_util.tree_leaves_with_path(params):
+    ...   print(jax.tree_util.keystr(path, simple=True, separator='/'))
+    foo/bar/bat/0
+    foo/bar/bat/1
+    foo/bar/baz
   """
-  return ''.join(map(str, keys))
+  str_fn = _simple_entrystr if simple else str
+  return separator.join(map(str_fn, keys))
 
 
-# TODO(ivyzheng): remove this after another jaxlib release.
-class _RegistryWithKeypathsEntry(NamedTuple):
-  flatten_with_keys: Callable[..., Any]
-  unflatten_func: Callable[..., Any]
-
-
-def _register_keypaths(
-    ty: type[T], handler: Callable[[T], tuple[KeyEntry, ...]]
-) -> None:
-  def flatten_with_keys(xs):
-    children, treedef = _registry[ty].to_iter(xs)
-    return list(zip(handler(xs), children)), treedef
-  if ty in _registry:
-    _registry_with_keypaths[ty] = _RegistryWithKeypathsEntry(
-        flatten_with_keys, _registry[ty].from_iter
-    )
-
-_registry_with_keypaths: dict[type[Any], _RegistryWithKeypathsEntry] = {}
-
-_register_keypaths(
-    tuple, lambda xs: tuple(SequenceKey(i) for i in range(len(xs)))
-)
-_register_keypaths(
-    list, lambda xs: tuple(SequenceKey(i) for i in range(len(xs)))
-)
-_register_keypaths(dict, lambda xs: tuple(DictKey(k) for k in sorted(xs)))
-
-_register_keypaths(
-    collections.defaultdict, lambda x: tuple(DictKey(k) for k in x.keys())
-)
-
-_register_keypaths(
-    collections.OrderedDict, lambda x: tuple(DictKey(k) for k in x.keys())
-)
+def _simple_entrystr(key: KeyEntry) -> str:
+  match key:
+    case (
+        SequenceKey(idx=key)
+        | DictKey(key=key)
+        | GetAttrKey(name=key)
+        | FlattenedIndexKey(key=key)
+    ):
+      return str(key)
+    case _:
+      return str(key)
 
 
 @export
@@ -845,9 +895,6 @@ def register_pytree_with_keys(
   register_pytree_node(
       nodetype, flatten_func, unflatten_func, flatten_with_keys
   )
-  _registry_with_keypaths[nodetype] = _RegistryWithKeypathsEntry(
-      flatten_with_keys, unflatten_func
-  )
 
 
 @export
@@ -1040,11 +1087,6 @@ def register_dataclass(
         msg += f" Unexpected fields: {unexpected}."
       raise ValueError(msg)
 
-  def flatten_with_keys(x):
-    meta = tuple(getattr(x, name) for name in meta_fields)
-    data = tuple((GetAttrKey(name), getattr(x, name)) for name in data_fields)
-    return data, meta
-
   def unflatten_func(meta, data):
     meta_args = tuple(zip(meta_fields, meta))
     data_args = tuple(zip(data_fields, data))
@@ -1060,9 +1102,6 @@ def register_dataclass(
   none_leaf_registry.register_dataclass_node(nodetype, list(data_fields), list(meta_fields))
   dispatch_registry.register_dataclass_node(nodetype, list(data_fields), list(meta_fields))
   _registry[nodetype] = _RegistryEntry(flatten_func, unflatten_func)
-  _registry_with_keypaths[nodetype] = _RegistryWithKeypathsEntry(
-      flatten_with_keys, unflatten_func
-  )
   return nodetype
 
 
@@ -1123,34 +1162,38 @@ def register_static(cls: type[H]) -> type[H]:
 
 @export
 def tree_flatten_with_path(
-    tree: Any, is_leaf: Callable[[Any], bool] | None = None
+    tree: Any, is_leaf: Callable[..., bool] | None = None,
+    is_leaf_takes_path: bool = False,
 ) -> tuple[list[tuple[KeyPath, Any]], PyTreeDef]:
   """Alias of :func:`jax.tree.flatten_with_path`."""
-  return default_registry.flatten_with_path(tree, is_leaf)
+  is_leaf_with_kp: Callable[[Any, Any], bool] | None = is_leaf
+  if not is_leaf_takes_path and is_leaf is not None:
+    is_leaf_with_kp = lambda _, x: is_leaf(x)
+  return default_registry.flatten_with_path(tree, is_leaf_with_kp)
 
 
 @export
 def tree_leaves_with_path(
-    tree: Any, is_leaf: Callable[[Any], bool] | None = None
+    tree: Any, is_leaf: Callable[..., bool] | None = None,
+    is_leaf_takes_path: bool = False,
 ) -> list[tuple[KeyPath, Any]]:
   """Alias of :func:`jax.tree.leaves_with_path`."""
-  return tree_flatten_with_path(tree, is_leaf)[0]
-
-
-# generate_key_paths is not exported.
-def generate_key_paths(
-    tree: Any, is_leaf: Callable[[Any], bool] | None = None
-) -> list[tuple[KeyPath, Any]]:
-  return tree_leaves_with_path(tree, is_leaf)
-_generate_key_paths = generate_key_paths  # alias for backward compat
+  return tree_flatten_with_path(tree, is_leaf, is_leaf_takes_path)[0]
+generate_key_paths = tree_leaves_with_path
 
 
 @export
-def tree_map_with_path(f: Callable[..., Any],
-                       tree: Any, *rest: Any,
-                       is_leaf: Callable[[Any], bool] | None = None) -> Any:
+def tree_map_with_path(
+    f: Callable[..., Any],
+    tree: Any,
+    *rest: Any,
+    is_leaf: Callable[..., bool] | None = None,
+    is_leaf_takes_path: bool = False,
+) -> Any:
   """Alias of :func:`jax.tree.map_with_path`."""
-  keypath_leaves, treedef = tree_flatten_with_path(tree, is_leaf)
+  keypath_leaves, treedef = tree_flatten_with_path(
+      tree, is_leaf, is_leaf_takes_path
+  )
   keypath_leaves = list(zip(*keypath_leaves))
   all_keypath_leaves = keypath_leaves + [treedef.flatten_up_to(r) for r in rest]
   return treedef.unflatten(f(*xs) for xs in zip(*all_keypath_leaves))
