@@ -13,7 +13,7 @@
 # limitations under the License.
 
 # Note: import <name> as <name> is required for names to be exported.
-# See PEP 484 & https://github.com/google/jax/issues/7570
+# See PEP 484 & https://github.com/jax-ml/jax/issues/7570
 
 from jax._src.errors import (
   JAXTypeError as JAXTypeError,
@@ -26,4 +26,24 @@ from jax._src.errors import (
   UnexpectedTracerError as UnexpectedTracerError,
   KeyReuseError as KeyReuseError,
 )
-from jax._src.traceback_util import SimplifiedTraceback as SimplifiedTraceback
+
+from jax._src.lib import xla_client as _xc
+JaxRuntimeError = _xc.XlaRuntimeError
+del _xc
+
+import jax._src.traceback_util
+_deprecations = {
+  "SimplifiedTraceback": (
+    "jax.errors.SimplifiedTraceback is deprecated and will be removed in JAX v0.8.",
+    jax._src.traceback_util.SimplifiedTraceback
+  ),
+}
+
+import typing
+if typing.TYPE_CHECKING:
+  SimplifiedTraceback = jax._src.traceback_util.SimplifiedTraceback
+else:
+  from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
+  __getattr__ = _deprecation_getattr(__name__, _deprecations)
+  del _deprecation_getattr
+del typing
