@@ -15,9 +15,9 @@
 
 from __future__ import annotations
 
-from collections.abc import Sequence
+from collections.abc import Callable, Sequence
 import itertools
-from typing import Any, Callable, Optional, Union
+from typing import Any
 
 import jax
 from jax import lax
@@ -198,6 +198,12 @@ class Jax2TfLimitation(test_harnesses.Limitation):
             devices=("cpu", "gpu"),
             tol=1e-13,
             modes=("eager", "graph", "compiled")),
+        custom_numeric(
+            dtypes=[np.complex64],
+            devices=("tpu",),
+            tol=1e-3,
+            modes=("eager", "graph", "compiled"),
+            native_serialization=Jax2TfLimitation.FOR_NON_NATIVE),
     ]
 
   @classmethod
@@ -206,7 +212,17 @@ class Jax2TfLimitation(test_harnesses.Limitation):
         custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu", "tpu"),
                        tol=1e-3),
         custom_numeric(dtypes=[np.complex128], devices=("cpu", "gpu"), tol=1e-12),
-        cls.helper_get_trig_custom_limitation(np.cosh)
+        Jax2TfLimitation(
+            "TF2XLA impl for Acosh doesn't properly handle large complex types,"
+            " native serialization more closely matches numpy numerics.",
+            dtypes=[np.complex64, np.complex128],
+            devices=("cpu", "gpu", "tpu"),
+            modes="compiled",
+            expect_tf_error=False,
+            skip_comparison=True,
+            native_serialization=Jax2TfLimitation.FOR_NON_NATIVE,
+        ),
+        cls.helper_get_trig_custom_limitation(np.cosh),
     ]
 
   @classmethod
@@ -281,6 +297,11 @@ class Jax2TfLimitation(test_harnesses.Limitation):
         custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu", "tpu"),
                        tol=1e-3),
         custom_numeric(dtypes=[np.complex128], devices=("cpu", "gpu"), tol=1e-12),
+        custom_numeric(dtypes=[np.complex64, np.complex128],
+                       devices=("cpu", "gpu", "tpu"),
+                       modes=("compiled",),
+                       tol=1e-3,
+                       native_serialization=Jax2TfLimitation.FOR_NON_NATIVE),
         custom_numeric(dtypes=[np.complex128], devices=("cpu",),
                        modes=("eager", "compiled", "graph"),
                        tol=1e-13,
@@ -351,6 +372,13 @@ class Jax2TfLimitation(test_harnesses.Limitation):
             tol=5e-2,
             devices=("cpu", "gpu"),
             modes=("eager", "graph", "compiled")),
+        custom_numeric(
+            dtypes=[dtypes.bfloat16],
+            tol=5e-5,
+            # Error for GL
+            devices=("tpu",),
+            modes=("eager", "graph", "compiled"),
+            native_serialization=Jax2TfLimitation.FOR_NATIVE),
         custom_numeric(
             custom_assert=custom_assert,
             description=(
@@ -837,15 +865,15 @@ class Jax2TfLimitation(test_harnesses.Limitation):
     def custom_assert(tst, result_jax, result_tf, *, args, tol,
                       err_msg):  # noqa: F811
       arg1, arg2 = args
-      # lax.igammac returns 1. when arg1 <= 0; tf.math.igammac returns NaN
+      # lax.igammac returns nan. when arg1 <= 0; tf.math.igammac returns 1
       special_cases = (arg1 <= 0.) | (arg2 <= 0)
       nr_special_cases = np.count_nonzero(special_cases)
       tst.assertAllClose(
-          np.full((nr_special_cases,), 1., dtype=dtype),
+          np.full((nr_special_cases,), np.nan, dtype=dtype),
           result_jax[special_cases],
           err_msg=err_msg)
       tst.assertAllClose(
-          np.full((nr_special_cases,), np.nan, dtype=dtype),
+          np.full((nr_special_cases,), 1, dtype=dtype),
           result_tf[special_cases],
           err_msg=err_msg)
       # non-special cases are equal
@@ -864,12 +892,12 @@ class Jax2TfLimitation(test_harnesses.Limitation):
         custom_numeric(dtypes=[np.float64], tol=1e-9),
         custom_numeric(devices="gpu", tol=1e-3),
         custom_numeric(
+            modes=("compiled",),
             custom_assert=custom_assert,
-            devices=("cpu", "gpu"),
+            devices=("cpu", "gpu", "tpu"),
             description=(
                 "May return different results at undefined points "
-                "(both arguments less or equal 0). JAX returns `NaN` and TF returns 0 or "
-                "JAX returns 1 and TF returns `NaN`")),
+                "(both arguments less or equal 0). JAX returns `NaN` and TF returns 1")),
     ]
 
   @classmethod
@@ -1038,9 +1066,9 @@ class Jax2TfLimitation(test_harnesses.Limitation):
 
   @classmethod
   def qr(cls, harness: test_harnesses.Harness):
-    # See https://github.com/google/jax/pull/3775#issuecomment-659407824;
+    # See https://github.com/jax-ml/jax/pull/3775#issuecomment-659407824;
     #     # jit_compile=True breaks for complex types.
-    # TODO: see https://github.com/google/jax/pull/3775#issuecomment-659407824.
+    # TODO: see https://github.com/jax-ml/jax/pull/3775#issuecomment-659407824.
     # - for now, the performance of the HLO QR implementation called when
     #   compiling with TF is expected to have worse performance than the
     #   custom calls made in JAX.
