@@ -38,7 +38,7 @@ from jax._src import test_util as jtu
 from jax._src import xla_bridge as xb
 from jax.experimental import jax2tf
 from jax.experimental.jax2tf.tests import tf_test_util
-from jax.experimental.shard_map import shard_map
+from jax._src.shard_map import shard_map
 from jax.experimental import pjit
 from jax.sharding import PartitionSpec as P
 
@@ -52,6 +52,13 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
   def setUp(self):
     super().setUp()
+    versions = tf.version.VERSION.split(".")
+    if versions < ["2", "19", "1"]:
+      # StableHLO changed on March 18th, 2025 ,to version 1.10.0, and this
+      # introduces ops like vhlo_sine_v2. These ops require a TF version
+      # released after this date.
+      self.skipTest("Need version of TensorFlow at least 2.19.1")
+
     # One TF device of each device_type
     self.tf_devices = []
     for tf_device in (tf.config.list_logical_devices("TPU") +
@@ -832,11 +839,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     arg = np.array(3.)
     f_tf = jax2tf.convert(jax.grad(remat_f))
     f_tf_hlo = self.TfToHlo(f_tf, arg)
-    if config.remat_opt_barrier.value:
-      self.assertRegex(f_tf_hlo, r"opt-barrier")
-    else:
-      self.assertRegex(f_tf_hlo,
-                       r'transpose/jax2tf_f_/jvp/checkpoint/cond/branch_1_fun/Sin')
+    self.assertRegex(f_tf_hlo, r"opt-barrier")
 
   def test_remat_free_var(self):
     def f(x):
@@ -1017,7 +1020,9 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
   def test_shared_constants_under_scan(self):
     # See https://github.com/jax-ml/jax/issues/7992.
     if config.jax2tf_default_native_serialization.value:
-      raise unittest.SkipTest("shared constants tests not interesting for native serialization")
+      self.skipTest("shared constants tests not interesting for native serialization")
+    if config.use_simplified_jaxpr_constants.value:
+      self.skipTest("shared constants tests broken with new handling of constants")
     const_size = 512
     const = np.random.uniform(size=const_size).astype(np.float32)  # A shared constant
     xs = np.ones((8, const_size), dtype=np.float32)
@@ -1131,7 +1136,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.skipTest("include_xla_op_metadata not yet enabled")
     # A simple example
     # The user_frame is used to compute line numbers for ops in the test.
-    user_frame = source_info_util.user_frame(source_info_util.current())
+    user_frame = source_info_util.user_frame(
+        source_info_util.current().traceback)
     def f_simple(x):
       return jnp.sin(x)
 
@@ -1150,7 +1156,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.skipTest("include_xla_op_metadata not yet enabled")
     # Calling a jitted-function
     # The user_frame is used to compute line numbers for ops in the test.
-    user_frame = source_info_util.user_frame(source_info_util.current())
+    user_frame = source_info_util.user_frame(
+        source_info_util.current().traceback)
     def f_callee(x):
       return jnp.cos(x)
     def f_caller(x):
@@ -1184,7 +1191,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.skipTest("include_xla_op_metadata not yet enabled")
     # Calling a jax.named_call
     # The user_frame is used to compute line numbers for ops in the test.
-    user_frame = source_info_util.user_frame(source_info_util.current())
+    user_frame = source_info_util.user_frame(
+        source_info_util.current().traceback)
     def f_callee(x):
       return jnp.cos(x)
     def f_caller(x):
@@ -1218,7 +1226,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.skipTest("include_xla_op_metadata not yet enabled")
     # An example with while and cond
     # The user_frame is used to compute line numbers for ops in the test.
-    user_frame = source_info_util.user_frame(source_info_util.current())
+    user_frame = source_info_util.user_frame(
+        source_info_util.current().traceback)
     def f_while_cond(x):
       def body_fun(i_acc):
         i, acc = i_acc
@@ -1259,7 +1268,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.skipTest("include_xla_op_metadata not yet enabled")
     # An example with while and cond
     # The user_frame is used to compute line numbers for ops in the test.
-    user_frame = source_info_util.user_frame(source_info_util.current())
+    user_frame = source_info_util.user_frame(
+        source_info_util.current().traceback)
     @jax.vmap
     def f_while(x):
       def body_fun(carry):
@@ -1476,7 +1486,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
               in_shardings=(sharding.NamedSharding(mesh, P("a")),),
               out_shardings=sharding.NamedSharding(mesh, P("a"))),
           shard_map=(
-              shard_map(func, mesh, in_specs=(P("a", None),),
+              shard_map(func, mesh=mesh, in_specs=(P("a", None),),
                         out_specs=P("a", None))),
           pmap=jax.pmap(func, in_axes=0, out_axes=0),
       )[transform]
@@ -1698,6 +1708,54 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
                                 "Unsupported precision in dot_general"):
       jax2tf.convert(f_jax, native_serialization=False)(x)
 
+  def test_jvp_through_loop(self):
+    # Context: b/388929258
+
+    num_actions = 512
+
+    def tf_preprocessor(features):
+      features["num_c_actions"] = tf.constant(256, tf.int32)
+      return features
+
+    def postprocessor(prob, features):
+      actions = jnp.arange(num_actions, dtype=jnp.int32)
+      r = actions // features["num_c_actions"]
+      c = actions - r * features["num_c_actions"]
+      rr = jnp.array([0.12, 0.3])[r] * prob
+      rc = (jnp.arange(256) * 0.7)[c] * prob
+      return rr, rc
+
+    def loop_step(features, params):
+      features = jax2tf.call_tf(tf_preprocessor)(features)
+      odds = features["f1"] @ params["w1"] + features["f2"] @ params["w2"]
+      prob = jax.nn.sigmoid(odds)
+      rr, rc = postprocessor(prob, features)
+      new_f1 = jnp.mean(rr, keepdims=True)
+      new_f2 = jnp.mean(rc, keepdims=True)
+      return new_f1, new_f2
+
+    def loop(init_features, params):
+      def body(carry, unused_x):
+        f1, f2 = carry
+        return loop_step({"f1": f1, "f2": f2}, params), None
+
+      (rr, rc), _ = jax.lax.scan(
+          body, (init_features["f1"], init_features["f2"]), length=10
+      )
+      return rr, rc
+
+    def loss(features, params):
+      rr, rc = loop(features, params)
+      return jnp.mean((rr - rc) ** 2)
+
+    jax.grad(loss, argnums=(1,))(
+        {"f1": jnp.array([0.5]), "f2": jnp.array([0.7])},
+        {
+            "w1": jnp.ones((1, num_actions)) * 0.01,
+            "w2": jnp.ones((1, num_actions)) * 0.01,
+        },
+    )
+
 
 @jtu.with_config(jax_enable_custom_prng=True)
 class Jax2tfWithCustomPRNGTest(tf_test_util.JaxToTfTestCase):
@@ -1738,10 +1796,17 @@ class Jax2tfWithCustomPRNGTest(tf_test_util.JaxToTfTestCase):
     jax_result = func()
     self.assertEqual(tf_result, jax_result)
 
+
 class Jax2TfVersioningTest(tf_test_util.JaxToTfTestCase):
   # Use a separate test case with the default jax_serialization_version
   def setUp(self):
     self.use_max_serialization_version = False
+    versions = tf.version.VERSION.split(".")
+    if versions < ["2", "19", "1"]:
+      # StableHLO changed on March 18th, 2025 ,to version 1.10.0, and this
+      # introduces ops like vhlo_sine_v2. These ops require a TF version
+      # released after this date.
+      self.skipTest("Need version of TensorFlow at least 2.19.1")
     super().setUp()
 
   @jtu.ignore_warning(
