@@ -33,10 +33,11 @@ from jax._src import compiler
 from jax._src import config
 from jax._src import test_util as jtu
 from jax._src import xla_bridge
+from jax._src.lib import xla_client as xc
 from jax import lax
 from jax.experimental import jax2tf
 from jax.experimental import pjit
-from jax.experimental.shard_map import shard_map
+from jax._src.shard_map import shard_map
 from jax.sharding import NamedSharding
 from jax.sharding import Mesh
 from jax.sharding import PartitionSpec as P
@@ -86,6 +87,11 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
     )
     self.warning_ctx.__enter__()
 
+  def get_xla_options(self):
+    return tf.tpu.XLAOptions(
+        use_shardy_partitioner=jax.config.jax_use_shardy_partitioner
+    )
+
   def tearDown(self):
     self.warning_ctx.__exit__(None, None, None)
     super().tearDown()
@@ -109,8 +115,9 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
           device_assignment=device_assignment,
           use_spmd_partitioning=use_spmd_partitioning,
       )
-      jax_optimized_hlo = backend.compile(
-          jax_hlo, compile_options).hlo_modules()[0].to_string()
+      executable = backend.compile_and_load(
+          jax_hlo, xc.DeviceList(tuple(self.devices.flat)), compile_options)  # type: ignore
+      jax_optimized_hlo = executable.hlo_modules()[0].to_string()
       logging.info("[%s] got JAX optimized HLO for platform %s %s",
                    self._testMethodName, backend.platform, jax_optimized_hlo)
 
@@ -209,10 +216,13 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
       f_converted = jax2tf.convert(f_jax)
       if jtu.test_device_matches(["tpu"]):
         return tf.compat.v1.tpu.rewrite(
-            f_converted, [tf.convert_to_tensor(x)],
+            f_converted,
+            [tf.convert_to_tensor(x)],
             device_assignment=self.device_assignment(
                 computation_shape=[1, 1, 1, 2],
-            ))[0]
+            ),
+            xla_options=self.get_xla_options(),
+        )[0]
       else:
         return f_converted(x)
 
@@ -231,10 +241,10 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
         jax2tf.convert(f_jax), [x],
         checks=[
             # The argument
-            (r"f32\[10,20\].*custom_call_target.*Sharding.*sharding.*devices=\[1,2\]",
+            (r"f32\[10,20\].*custom_call_target.*\"Sharding.*sharding.*devices=\[1,2\]",
              count_in_P),
             # The result
-            (r"f32\[20,10\].*custom_call_target.*Sharding.*sharding.*devices=\[2,1\]",
+            (r"f32\[20,10\].*custom_call_target.*\"Sharding.*sharding.*devices=\[2,1\]",
              count_out_P),
         ])
     # TODO(b/326476605): Change the condition below if required.
@@ -242,11 +252,11 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
       self.check_sharding(
         jax2tf.convert(f_jax), [x],
         checks=[
-            (r"f32\[10,20\].*custom_call_target.*Sharding.*sharding.*replicated",
+            (r"f32\[10,20\].*custom_call_target.*\"Sharding.*sharding.*replicated",
              count_in_replicated),
-            (r"f32\[20,10\].*custom_call_target.*Sharding.*sharding.*replicated",
+            (r"f32\[20,10\].*custom_call_target.*\"Sharding.*sharding.*replicated",
              count_out_replicated),
-            (r"custom_call_target.*Sharding",
+            (r"custom_call_target.*\"Sharding",
              count_in_P + count_in_replicated + count_out_P + count_out_replicated),
         ])
 
@@ -276,13 +286,13 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
         f_tf, [y],
         checks=[
             # The variable argument
-            (r"f32\[10,20\].*custom_call_target.*Sharding.*sharding.*devices=\[1,2\]", 1),
+            (r"f32\[10,20\].*custom_call_target.*\"Sharding.*sharding.*devices=\[1,2\]", 1),
             # The y argument
-            (r"f32\[20,30\].*custom_call_target.*Sharding.*sharding.*devices=\[2,1\]", 1),
+            (r"f32\[20,30\].*custom_call_target.*\"Sharding.*sharding.*devices=\[2,1\]", 1),
             # The output sharding
-            (r"f32\[10,30\].*custom_call_target.*Sharding.*sharding.*replicated", 1),
+            (r"f32\[10,30\].*custom_call_target.*\"Sharding.*sharding.*replicated", 1),
             # No other annotations
-            (r"custom_call_target.*Sharding", 3)
+            (r"custom_call_target.*\"Sharding", 3)
         ])
 
   @jtu.with_mesh([("x", 2)])
@@ -299,9 +309,12 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
       f_converted = jax2tf.convert(f_jax)
       if jtu.test_device_matches(["tpu"]):
         return tf.compat.v1.tpu.rewrite(
-            f_converted, [tf.convert_to_tensor(x)],
+            f_converted,
+            [tf.convert_to_tensor(x)],
             device_assignment=self.device_assignment(
-                computation_shape=[1, 1, 1, 2])
+                computation_shape=[1, 1, 1, 2]
+            ),
+            xla_options=self.get_xla_options(),
         )[0]
       else:
         return f_converted(x)
@@ -310,10 +323,10 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
         jax2tf.convert(f_jax), [x],
         checks=[
             # x
-            (r"f32\[10,20\].*custom_call_target.*Sharding.*sharding.*devices=\[2,1\]",
+            (r"f32\[10,20\].*custom_call_target.*\"Sharding.*sharding.*devices=\[2,1\]",
              1),
             # The result
-            (r"f32\[20,10\].*custom_call_target.*Sharding.*sharding.*replicated",
+            (r"f32\[20,10\].*custom_call_target.*\"Sharding.*sharding.*replicated",
              self.GEQ(1)),
         ])
 
@@ -357,16 +370,16 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
         f_tf, [x],
         checks=[
             # The input argument
-            (r"f32\[10,20\].*custom_call_target.*Sharding.*sharding.*replicated", 1),
+            (r"f32\[10,20\].*custom_call_target.*\"Sharding.*sharding.*replicated", 1),
             # The y argument
-            (r"f32\[10,40\].*custom_call_target.*Sharding.*sharding.*devices=\[2,1\]",
+            (r"f32\[10,40\].*custom_call_target.*\"Sharding.*sharding.*devices=\[2,1\]",
              count_inner_sharding),
-            (r"f32\[10,40\].*custom_call_target.*Sharding.*sharding.*replicated",
+            (r"f32\[10,40\].*custom_call_target.*\"Sharding.*sharding.*replicated",
              count_inner_replicated),
             # The output sharding
-            (r"f32\[10,80\].*custom_call_target.*Sharding.*sharding.*replicated", 1),
+            (r"f32\[10,80\].*custom_call_target.*\"Sharding.*sharding.*replicated", 1),
             # No other annotations
-            (r"custom_call_target.*Sharding", 2 + count_inner_sharding + count_inner_replicated)
+            (r"custom_call_target.*\"Sharding", 2 + count_inner_sharding + count_inner_replicated)
         ])
 
   @jtu.parameterized_filterable(
@@ -427,17 +440,17 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
     self.check_sharding(f_grad_tf, [x, x.T],
         checks=[
             # The input primal argument, and the output grad
-            (r"f32\[10,20\].*custom_call_target.*Sharding.*sharding.*devices=\[1,2\]", count_in_P),
+            (r"f32\[10,20\].*custom_call_target.*\"Sharding.*sharding.*devices=\[1,2\]", count_in_P),
             # The primal result, and the input cotangent
-            (r"f32\[20,10\].*custom_call_target.*Sharding.*sharding.*devices=\[2,1\]", count_out_P),
+            (r"f32\[20,10\].*custom_call_target.*\"Sharding.*sharding.*devices=\[2,1\]", count_out_P),
         ])
     # TODO(b/326476605): Change the condition below if required.
     if out_shardings not in [None, "missing"] and in_shardings not in [None, "missing"]:
       self.check_sharding(f_grad_tf, [x, x.T],
         checks=[
-            (r"f32\[10,20\].*custom_call_target.*Sharding.*sharding.*replicated", count_in_replicated),
+            (r"f32\[10,20\].*custom_call_target.*\"Sharding.*sharding.*replicated", count_in_replicated),
             # The primal result, and the input cotangent
-            (r"f32\[20,10\].*custom_call_target.*Sharding.*sharding.*devices=\[2,1\]", count_out_P),
+            (r"f32\[20,10\].*custom_call_target.*\"Sharding.*sharding.*devices=\[2,1\]", count_out_P),
         ])
 
   def test_grad_sharding_different_mesh(self):
@@ -543,9 +556,12 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
       f_converted = jax2tf.convert(f_jax, native_serialization=True)
       if jtu.test_device_matches(["tpu"]):
         return tf.compat.v1.tpu.rewrite(
-            f_converted, [tf.convert_to_tensor(a)],
+            f_converted,
+            [tf.convert_to_tensor(a)],
             device_assignment=self.device_assignment(
-                computation_shape=[1, 1, 1, 2])
+                computation_shape=[1, 1, 1, 2]
+            ),
+            xla_options=self.get_xla_options(),
         )[0]
       else:
         return f_converted(a)
@@ -576,7 +592,7 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
     @partial(shard_map, mesh=mesh,
              in_specs=(P('x', None),), out_specs=P('x', None))
     def f_jax(b):  # b: f32[2, 4]
-      axis_size = lax.psum(1, 'x')
+      axis_size = lax.axis_size('x')
       perm = [(j, (j + 1) % axis_size) for j in range(axis_size)]
       return lax.ppermute(b, 'x', perm=perm)
 
@@ -612,7 +628,7 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
     @partial(shard_map, mesh=mesh,
              in_specs=(P('x', None),), out_specs=P('x', None))
     def f_jax(b):  # b: f32[2, 4]
-      axis_size = lax.psum(1, 'x')
+      axis_size = lax.axis_size('x')
       perm = [(j, (j + 1) % axis_size) for j in range(axis_size)]
       return lax.ppermute(b, 'x', perm=perm)
 
@@ -622,9 +638,12 @@ class ShardingTest(tf_test_util.JaxToTfTestCase):
                                    polymorphic_shapes=poly)
       if jtu.test_device_matches(["tpu"]):
         res = tf.compat.v1.tpu.rewrite(
-            f_converted, [tf.convert_to_tensor(a)],
+            f_converted,
+            [tf.convert_to_tensor(a)],
             device_assignment=self.device_assignment(
-                computation_shape=[1, 1, 1, 2])
+                computation_shape=[1, 1, 1, 2]
+            ),
+            xla_options=self.get_xla_options(),
         )[0]
       else:
         res = f_converted(a)
