@@ -24,13 +24,18 @@ import os
 import pathlib
 import time
 from typing import Any, Generic, TypeVar
+from collections.abc import Callable
 import weakref
 
+import itertools
 import jax
+from jax._src import lib
+from jax._src import sharding_impls
 from jax._src.interpreters import mlir
 from jax._src.lib import mosaic_gpu_dialect as dialect
 from jaxlib.mlir import ir
 from jaxlib.mlir import passmanager
+from jaxlib.mlir.dialects import arith
 from jaxlib.mlir.dialects import builtin
 from jaxlib.mlir.dialects import func
 from jaxlib.mlir.dialects import gpu
@@ -39,29 +44,21 @@ from jaxlib.mlir.dialects import memref
 from jaxlib.mlir.dialects import nvvm
 import numpy as np
 
-if dialect is not None:
-  from . import dialect_lowering
-  from . import layout_inference
-else:
-  dialect_lowering = None
-  layout_inference = None
-
-from . import profiler
-from . import utils
-from . import launch_context
 
 # mypy: ignore-errors
 
-# MLIR can't find libdevice unless we point it to the CUDA path
-# TODO(apaszke): Unify with jax._src.lib.cuda_path
-CUDA_ROOT = "/usr/local/cuda"
-if os.environ.get("CUDA_ROOT") is None:
-  os.environ["CUDA_ROOT"] = CUDA_ROOT
-else:
-  CUDA_ROOT = os.environ["CUDA_ROOT"]
+from . import dialect_lowering
+from . import launch_context
+from . import layout_inference
+from . import profiler
+from . import tcgen05
+from . import transform_inference
+from . import utils
 
-PTXAS_PATH = os.path.join(CUDA_ROOT, "bin/ptxas")
-NVDISASM_PATH = os.path.join(CUDA_ROOT, "bin/nvdisasm")
+# MLIR can't find libdevice unless we point it to the CUDA path
+cuda_root = lib.cuda_path or "/usr/local/cuda"
+os.environ["CUDA_ROOT"] = cuda_root
+PYTHON_RUNFILES = os.environ.get("PYTHON_RUNFILES")
 
 # This tracks the latest Mosaic GPU IR version with a monthly delay.
 FWD_COMPAT_IR_VERSION = 1
@@ -85,14 +82,74 @@ if RUNTIME_PATH and RUNTIME_PATH.exists():
   os.environ["MOSAIC_GPU_RUNTIME_LIB_PATH"] = str(RUNTIME_PATH)
 
 
+try:
+  from nvidia import nvshmem  # pytype: disable=import-error
+except ImportError:
+  # Try to find the nvshmem library in Bazel runfiles.
+  if PYTHON_RUNFILES:
+    libdevice_path = os.path.join(
+        PYTHON_RUNFILES, "nvidia_nvshmem", "lib", "libnvshmem_device.bc"
+    )
+    if os.path.exists(libdevice_path):
+      os.environ["MOSAIC_GPU_NVSHMEM_BC_PATH"] = libdevice_path
+    for root, _, files in os.walk(os.path.join(os.getcwd(), "_solib_local")):
+      if "libnvshmem_host.so.3" in files:
+        os.environ["MOSAIC_GPU_NVSHMEM_SO_PATH"] = os.path.join(
+            root, "libnvshmem_host.so.3"
+        )
+        break
+  else:
+    pass
+else:
+  if os.environ.get("MOSAIC_GPU_NVSHMEM_BC_PATH") is None:
+    os.environ["MOSAIC_GPU_NVSHMEM_BC_PATH"] = os.path.join(
+        nvshmem.__path__[0], "lib/libnvshmem_device.bc"
+    )
+  if os.environ.get("MOSAIC_GPU_NVSHMEM_SO_PATH") is None:
+    os.environ["MOSAIC_GPU_NVSHMEM_SO_PATH"] = os.path.join(
+        nvshmem.__path__[0], "lib/libnvshmem_host.so.3"
+    )
+
+
+def supports_cross_device_collectives():
+  try:
+    nvshmem_bc_path = os.environ["MOSAIC_GPU_NVSHMEM_BC_PATH"]
+  except KeyError:
+    return False
+  if nvshmem_so_path := os.environ.get("MOSAIC_GPU_NVSHMEM_SO_PATH", ""):
+    try:
+      # This both ensures that the file exists, and it populates the dlopen
+      # cache, helping XLA find the library even if the RPATH is not right...
+      ctypes.CDLL(nvshmem_so_path)
+    except OSError:
+      return False
+  xla_flags = os.environ.get("XLA_FLAGS", "")
+  return (
+      os.path.exists(nvshmem_bc_path)
+      and "--xla_gpu_experimental_enable_nvshmem" in xla_flags
+  )
+
+
 mosaic_gpu_p = jax._src.core.Primitive("mosaic_gpu_p")
 mosaic_gpu_p.multiple_results = True
 
 
 @mosaic_gpu_p.def_abstract_eval
-def _mosaic_gpu_abstract_eval(*_, module, out_types):
+def _mosaic_gpu_abstract_eval(*_, module, out_types, inout_types):
   del module  # Unused.
-  return [jax._src.core.ShapedArray(t.shape, t.dtype) for t in out_types]
+  return [
+      jax._src.core.ShapedArray(t.shape, t.dtype)
+      for t in itertools.chain(out_types, inout_types)
+  ]
+
+
+def _has_communication(module, **_):
+  empty_str_attr = ir.StringAttr.get("")
+  for op in module.body:
+    if "nvshmem" in getattr(op, "sym_name", empty_str_attr).value:
+      return True
+  return False
+
 
 # TODO(apaszke): Implement a proper system for managing kernel lifetimes
 KNOWN_KERNELS = {}
@@ -103,9 +160,44 @@ def _mosaic_gpu_lowering_rule(
     *args,
     module,
     out_types,
+    inout_types,
     input_output_aliases: tuple[tuple[int, int], ...] = (),
+    use_custom_barrier: bool = False,
 ):
-  assert len(out_types) == len(ctx.avals_out)
+  axis_context = ctx.module_context.axis_context
+  if _has_communication(module):
+    # Those checks are trying to ensure that the logical device ids are
+    # consistent with the NVSHMEM PE ids that Mosaic will be using for
+    # communication. Any divergence here would require us to implement a logical
+    # to physical translation, which is currently not implemented.
+    if isinstance(axis_context, sharding_impls.SPMDAxisContext):
+      mesh = axis_context.mesh
+      if not np.array_equal(mesh.device_ids.ravel(), np.arange(mesh.size)):
+        raise NotImplementedError(
+            "Mosaic GPU only supports meshes with device ordering that follows"
+            " row-major device ids."
+        )
+    elif isinstance(axis_context, sharding_impls.ShardingContext):
+      if axis_context.num_devices != 1:
+        raise NotImplementedError(
+            "Mosaic GPU only supports single-device meshes in ShardingContext."
+        )
+    else:
+      raise NotImplementedError(f"Unsupported sharding context: {axis_context}")
+
+  if inout_types:
+    if input_output_aliases:
+      raise ValueError(
+          "input_output_aliases and inout_types are mutually exclusive"
+      )
+    num_inputs = len(ctx.avals_in)
+    num_outputs = len(ctx.avals_out)
+    input_output_aliases = tuple(
+        (num_inputs - 1 - i, num_outputs - 1 - i)
+        for i in range(len(inout_types))
+    )
+  assert len(ctx.avals_in) == len(args)
+  assert len(ctx.avals_out) == len(out_types) + len(inout_types)
   module = _run_serde_pass(
       module,
       serialize=True,
@@ -121,15 +213,35 @@ def _mosaic_gpu_lowering_rule(
       raise RuntimeError("Hash collision!")
   else:
     KNOWN_KERNELS[kernel_id] = module_asm
-  op = mlir.custom_call(
-      "mosaic_gpu",
-      result_types=[mlir.aval_to_ir_type(aval) for aval in ctx.avals_out],
-      operands=args,
-      operand_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_in],
-      result_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_out],
-      backend_config=kernel_id + module_asm,
-      operand_output_aliases=dict(input_output_aliases),
-  )
+
+  if ctx.is_forward_compat():
+    if use_custom_barrier:
+      raise ValueError("Barrier semaphore is not supported in forward compatibility mode. "
+                       "Please, use 'export_ignore_forward_compatibility=True'.")
+    op = mlir.custom_call(
+        "mosaic_gpu",
+        result_types=[mlir.aval_to_ir_type(aval) for aval in ctx.avals_out],
+        operands=args,
+        operand_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_in],
+        result_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_out],
+        backend_config=kernel_id + module_asm,
+        operand_output_aliases=dict(input_output_aliases),
+    )
+  else:
+    op = mlir.custom_call(
+        "mosaic_gpu_v2",
+        result_types=[mlir.aval_to_ir_type(aval) for aval in ctx.avals_out],
+        operands=args,
+        operand_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_in],
+        result_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_out],
+        backend_config=dict(
+            kernel_hash=ir.StringAttr.get(kernel_id),
+            module=ir.StringAttr.get(module_asm),
+            use_custom_barrier=ir.BoolAttr.get(use_custom_barrier),
+        ),
+        operand_output_aliases=dict(input_output_aliases),
+        api_version=4,
+    )
   return op.results
 
 
@@ -158,87 +270,188 @@ class Barrier:
   arrival_count: int
   num_barriers: int = 1
 
+  def __post_init__(self):
+    if self.arrival_count < 1:
+      raise ValueError(
+          f"Arrival count must be at least 1, but got {self.arrival_count}"
+      )
+
 @dataclasses.dataclass(frozen=True)
 class ClusterBarrier:
   collective_dims: Sequence[gpu.Dimension]
   num_barriers: int = 1
 
+@dataclasses.dataclass(frozen=True)
+class TMEM:
+  shape: tuple[int, int]
+  dtype: Any
+  _: dataclasses.KW_ONLY
+  layout: tcgen05.TMEMLayout | None = None
+  collective: bool = False
+  packing: int | None = None
+
+  def __post_init__(self):
+    if self.layout is not None:
+      self.layout.check_type(
+          self.shape, utils.bitwidth(utils.dtype_to_ir_type(self.dtype))
+      )
+      if self.packing is not None:
+        raise ValueError("Cannot specify both layout and packing")
+
 
 def _count_buffer_bytes(shape_dtype: jax.ShapeDtypeStruct) -> int:
-  return np.prod(shape_dtype.shape) * np.dtype(shape_dtype.dtype).itemsize
+  return math.prod(shape_dtype.shape) * np.dtype(shape_dtype.dtype).itemsize
 
 
-class ThreadSemantics(enum.Enum):
+class LoweringSemantics(enum.Enum):
   """Semantics for the kernel's instruction stream."""
 
   Lane = enum.auto()
   Warpgroup = enum.auto()
 
 
+@dataclasses.dataclass(frozen=True)
+class _TMEMAlloc:
+  addr_ref: ir.Value
+  num_cols: int
+  collective: bool
+
+  def alloc(self) -> int:
+    """Allocates TMEM and returns the number of columns allocated."""
+    _, cols = tcgen05.tmem_alloc(
+        self.addr_ref, self.num_cols, collective=self.collective, exact=False
+    )
+    return cols
+
+  def dealloc(self):
+    addr = memref.load(self.addr_ref, [])
+    tcgen05.tmem_dealloc(
+        addr, self.num_cols, collective=self.collective, exact=False
+    )
+
+
+def _slice_smem(
+    result: ir.Type,
+    smem_base: ir.Value,
+    offset: ir.Value,  # This should be an ir.IndexType.
+    lowering_semantics: LoweringSemantics,
+) -> ir.Value:
+  if lowering_semantics == LoweringSemantics.Warpgroup:
+    offset = arith.index_cast(ir.IntegerType.get_signless(32), offset)
+    return dialect.slice_smem(result, offset)
+  else:
+    return memref.view(result, smem_base, offset, [])
+
+
 def _construct_smem_reftree(
     cluster_shape: tuple[int, int, int],
     dynamic_smem: ir.Value,
     smem_buffers: ShapeTree,
+    tmem_allocs: list[_TMEMAlloc],  # Mutated by this function!
+    lowering_semantics: LoweringSemantics,
     dynamic_smem_offset: int = 0,
-) -> RefTree:
+) -> Callable[[], RefTree]:
   index = ir.IndexType.get()
-  i8 = ir.IntegerType.get_signless(8)
-  smem = ir.Attribute.parse("#gpu.address_space<workgroup>")
+  i32 = ir.IntegerType.get_signless(32)
+  i64 = ir.IntegerType.get_signless(64)
   flat_ref_tys, smem_buffer_tree = jax.tree.flatten(
       smem_buffers, is_leaf=lambda x: isinstance(x, Union)
   )
   smem_refs = []
+
   for ref_ty in flat_ref_tys:
-    def get_barrier_ptr(num_barriers: int) -> ir.Value:
+    def barrier_memref(num_barriers: int) -> ir.Value:
       nonlocal dynamic_smem_offset
-      workgroup_nvptx_address_space = (
-          utils.gpu_address_space_to_nvptx(gpu.AddressSpace.Workgroup)
-      )
-      smem_base_ptr = utils.memref_ptr(
-          dynamic_smem, memory_space=workgroup_nvptx_address_space
-      )
-      smem_ptr_ty = ir.Type.parse(f"!llvm.ptr<{workgroup_nvptx_address_space}>")
-      barrier_base_ptr = llvm.getelementptr(
-          smem_ptr_ty, smem_base_ptr, [], [dynamic_smem_offset], i8
+      barrier_ty = ir.MemRefType.get(
+          (num_barriers,),
+          ir.Type.parse("!mosaic_gpu.barrier")
+          if lowering_semantics == LoweringSemantics.Warpgroup
+          else i64,
+          memory_space=utils.smem(),
       )
+      barrier_memref = _slice_smem(
+            barrier_ty,
+            dynamic_smem,
+            c(dynamic_smem_offset, index),
+            lowering_semantics,
+        )
       dynamic_smem_offset += num_barriers * utils.MBARRIER_BYTES
-      return barrier_base_ptr
+      return barrier_memref
     match ref_ty:
       case Union(members):
-        member_trees = [
-            _construct_smem_reftree(cluster_shape, dynamic_smem, m, dynamic_smem_offset)
+        member_thunks = [
+            _construct_smem_reftree(
+                cluster_shape,
+                dynamic_smem,
+                m,
+                tmem_allocs,
+                lowering_semantics,
+                dynamic_smem_offset,
+            )
             for m in members
         ]
         # TODO(apaszke): This is quadratic, but it shouldn't matter for now...
         dynamic_smem_offset += _smem_tree_size(ref_ty)
-        ref = Union(member_trees)
+
+        def ref(member_thunks=member_thunks):
+          return Union([t() for t in member_thunks])
+
       case TMABarrier(num_barriers):
-        ref = utils.BarrierRef.initialize(
-            get_barrier_ptr(num_barriers), num_barriers, arrival_count=1
-        )
+        init_fn = utils.DialectBarrierRef.initialize if (
+            lowering_semantics == LoweringSemantics.Warpgroup
+        ) else utils.BarrierRef.initialize
+        ref = init_fn(barrier_memref(num_barriers), arrival_count=1)
       case Barrier(arrival_count, num_barriers):
-        ref = utils.BarrierRef.initialize(
-            get_barrier_ptr(num_barriers),
-            num_barriers,
-            arrival_count=arrival_count,
-        )
+        init_fn = utils.DialectBarrierRef.initialize if (
+            lowering_semantics == LoweringSemantics.Warpgroup
+        ) else utils.BarrierRef.initialize
+        ref = init_fn(barrier_memref(num_barriers), arrival_count=arrival_count)
       case ClusterBarrier(collective_dims, num_barriers):
         ref = utils.CollectiveBarrierRef.initialize(
-            get_barrier_ptr(num_barriers),
-            num_barriers,
+            barrier_memref(num_barriers),
             collective_dims,
             cluster_shape,
         )
+      case TMEM(shape, dtype, layout=layout, collective=collective, packing=packing):
+        addr_ref = _slice_smem(
+            ir.MemRefType.get([], i32, memory_space=utils.smem()),
+            dynamic_smem,
+            c(dynamic_smem_offset, index),
+            lowering_semantics,
+        )
+        if layout is None:
+          layout = tcgen05._infer_tmem_layout(
+              shape, collective, 1 if packing is None else packing
+          )
+        num_cols = layout.cols_in_shape(
+            shape, utils.bitwidth(utils.dtype_to_ir_type(dtype))
+        )
+        tmem_allocs.append(_TMEMAlloc(addr_ref, num_cols, collective))
+        def ref(addr_ref=addr_ref, shape=shape, dtype=dtype, layout=layout):
+          addr = memref.load(addr_ref, [])
+          return tcgen05.TMEMRef(
+              addr, shape, utils.dtype_to_ir_type(dtype), layout
+          )
+        dynamic_smem_offset += 4  # i32 takes up 4 bytes
       case _:
         mlir_dtype = utils.dtype_to_ir_type(ref_ty.dtype)
-        tile_smem = memref.view(
-            ir.MemRefType.get(ref_ty.shape, mlir_dtype, memory_space=smem),
-            dynamic_smem, c(dynamic_smem_offset, index), [],
+        tile_smem = _slice_smem(
+            ir.MemRefType.get(ref_ty.shape, mlir_dtype, memory_space=utils.smem()),
+            dynamic_smem,
+            c(dynamic_smem_offset, index),
+            lowering_semantics,
         )
         dynamic_smem_offset += _count_buffer_bytes(ref_ty)
         ref = tile_smem
     smem_refs.append(ref)
-  return jax.tree.unflatten(smem_buffer_tree, smem_refs)
+  def ref_tree_thunk():
+    refs = []
+    for ref in smem_refs:
+      if callable(ref):
+        ref = ref()
+      refs.append(ref)
+    return jax.tree.unflatten(smem_buffer_tree, refs)
+  return ref_tree_thunk
 
 
 def _smem_tree_size(smem_buffers: ShapeTree) -> int:
@@ -258,6 +471,10 @@ def _smem_tree_size(smem_buffers: ShapeTree) -> int:
         if size % utils.MBARRIER_BYTES:
           raise NotImplementedError("Misaligned barrier allocation")
         size += num_barriers * utils.MBARRIER_BYTES
+      case TMEM(_):
+        # TODO(justinfu): This can trigger misaligned barrier allocations
+        # if TMEM is requested before barriers b/c it's not divisible by 8.
+        size += 4  # i32 takes up 4 bytes
       case _:
         size += _count_buffer_bytes(l)
   return size
@@ -270,8 +487,9 @@ def _launch(
     grid: tuple[int, int, int],
     cluster: tuple[int, int, int],
     block: tuple[int, int, int],
-    scratch_arr,
     smem_buffers: ShapeTree | Union[ShapeTree],
+    lowering_semantics: LoweringSemantics,
+    module: ir.Module,
     profiler_spec: profiler.ProfilerSpec | None = None,
     maybe_prof_buffer: ir.Value | None = None,
 ):
@@ -287,7 +505,13 @@ def _launch(
 
   smem_bytes = user_smem_bytes
   if profiler_spec is not None:
-    smem_bytes += profiler_spec.smem_bytes(block=block)
+    # Profiler array stores values in 64 bit chunks (vectors of size 2
+    # of 32-bit elements), and so the starting address needs to be 64
+    # bit = 8 byte aligned.
+    # https://docs.nvidia.com/cuda/parallel-thread-execution/#addresses-as-operands:~:text=The%20address%20must%20be%20naturally%20aligned%20to%20a%20multiple%20of%20the%20access%20size.
+    align = 8
+    profiler_start = (smem_bytes + align - 1) & ~(align - 1)
+    smem_bytes = profiler_start + profiler_spec.smem_bytes(block=block)
 
   # TODO(cperivol): Query the shared memory size programmatically.
   if smem_bytes > 228 * 1024:
@@ -310,21 +534,21 @@ def _launch(
       token.type, [token], *grid_vals, *block_vals,
       dynamicSharedMemorySize=c(smem_bytes, i32), **cluster_kwargs)
   launch_op.body.blocks.append(*([index] * (12 + 2 * len(cluster_kwargs))))  # Append an empty block
-  smem = ir.Attribute.parse("#gpu.address_space<workgroup>")
   with ir.InsertionPoint(launch_op.body.blocks[0]):
     dynamic_smem = gpu.dynamic_shared_memory(
-        ir.MemRefType.get(
-            (ir.ShapedType.get_dynamic_size(),), i8, memory_space=smem
-        )
+        ir.MemRefType.get((utils.DYNAMIC,), i8, memory_space=utils.smem())
     )
 
     if profiler_spec:
-      prof_smem = memref.view(
+      prof_smem = _slice_smem(
           ir.MemRefType.get(
               (profiler_spec.smem_i32_elements(block=block),),
-              i32, memory_space=smem,
+              i32,
+              memory_space=utils.smem(),
           ),
-          dynamic_smem, c(user_smem_bytes, index), [],
+          dynamic_smem,
+          c(profiler_start, index),
+          lowering_semantics,
       )
       prof = profiler.OnDeviceProfiler(
           profiler_spec, prof_smem, maybe_prof_buffer
@@ -332,21 +556,50 @@ def _launch(
     else:
       prof = None
 
-    ptr_ty = ir.Type.parse("!llvm.ptr")
-    scratch_ptr = builtin.unrealized_conversion_cast([ptr_ty], [scratch_arr])
-    ctx = launch_context.LaunchContext(launch_op, scratch_ptr, cluster, prof)
+    ctx = launch_context.LaunchContext(
+        module, launch_context.Scratch(launch_op), cluster, prof
+    )
     with ctx.named_region("Init"):
-      smem_ref_tree = _construct_smem_reftree(
-          cluster, dynamic_smem, smem_buffers
+      tmem_allocs: list[_TMEMAlloc] = []
+      smem_ref_tree_thunk = _construct_smem_reftree(
+          cluster, dynamic_smem, smem_buffers, tmem_allocs, lowering_semantics
       )
-      # TODO(apaszke): Skip the following if no barriers were initialized.
+      # TODO(apaszke): Skip fences if no barriers or TMEM is initialized.
+      # TODO(apaszke): Only initialize cluster barriers before the cluster wait.
       nvvm.fence_mbarrier_init()
       if math.prod(cluster) != 1:
         nvvm.cluster_arrive_relaxed(aligned=ir.UnitAttr.get())
         nvvm.cluster_wait(aligned=ir.UnitAttr.get())
-      gpu.barrier()
+      if tmem_allocs:
+        eq = arith.CmpIPredicate.eq
+        is_init_warp = arith.cmpi(eq, utils.warp_idx(sync=False), c(0, i32))
+        with utils.when(is_init_warp):
+          cols_used = 0
+          for alloc in tmem_allocs:
+            cols_used += alloc.alloc()
+          if cols_used > tcgen05.TMEM_MAX_COLS:
+            raise ValueError(
+                "Total TMEM allocation exceeds memory limit. "
+                f"Requested {cols_used} columns which exceeds limit of "
+                f"{tcgen05.TMEM_MAX_COLS}."
+            )
+          if any(alloc.collective for alloc in tmem_allocs):
+            tcgen05.tmem_relinquish_alloc_permit(collective=True)
+          if any(not alloc.collective for alloc in tmem_allocs):
+            tcgen05.tmem_relinquish_alloc_permit(collective=False)
+      gpu.barrier()  # Make sure the init is visible to all threads.
+      smem_ref_tree = smem_ref_tree_thunk()
 
     yield ctx, smem_ref_tree
+
+    if tmem_allocs:
+      gpu.barrier()  # Make sure everyone is done before we release TMEM.
+      if any(alloc.collective for alloc in tmem_allocs):
+        nvvm.cluster_arrive_relaxed(aligned=ir.UnitAttr.get())
+        nvvm.cluster_wait(aligned=ir.UnitAttr.get())
+      with utils.when(is_init_warp):
+        for alloc in tmem_allocs:
+          alloc.dealloc()
     if prof is not None:
       prof.finalize(grid=grid, block=block)
     gpu.terminator()
@@ -359,7 +612,9 @@ def _lower_as_gpu_kernel(
     block: tuple[int, int, int],
     in_shapes: tuple[Any, ...],
     out_shape,
+    inout_shape,
     smem_scratch_shape: ShapeTree | Union[ShapeTree],
+    lowering_semantics: LoweringSemantics,
     module_name: str,
     kernel_name: str | None = None,
     prof_spec: profiler.ProfilerSpec | None = None,
@@ -367,32 +622,33 @@ def _lower_as_gpu_kernel(
   ptr_ty = ir.Type.parse("!llvm.ptr")
   token_ty = ir.Type.parse("!gpu.async.token")
   i32 = ir.IntegerType.get_signless(32)
-  i64 = ir.IntegerType.get_signless(64)
 
   def _shape_to_ref_ty(shape: jax.ShapeDtypeStruct) -> ir.MemRefType:
     return ir.MemRefType.get(shape.shape, utils.dtype_to_ir_type(shape.dtype))
 
   in_ref_tys = [_shape_to_ref_ty(t) for t in in_shapes]
+  inout_ref_tys = [_shape_to_ref_ty(t) for t in inout_shape]
 
   unwrap_output_tuple = False
   if isinstance(out_shape, list):
     out_shape = tuple(out_shape)
   elif not isinstance(out_shape, tuple):
     out_shape = (out_shape,)
-    unwrap_output_tuple = True
+    unwrap_output_tuple = not inout_shape
   out_ref_tys = [_shape_to_ref_ty(t) for t in out_shape]
   if prof_spec is not None:
     out_shape = (*out_shape, prof_spec.jax_buffer_type(grid, block))
     out_ref_tys.append(prof_spec.mlir_buffer_type(grid, block))
 
   module = ir.Module.create()
+  dialect.register_dialect(module.context)
   attrs = module.operation.attributes
   attrs["sym_name"] = ir.StringAttr.get(module_name)
   if kernel_name is None:
-    kernel_name = getattr(body, "__name__", "anonymous")
+    kernel_name = module_name
 
   # These are needed as nonlocal below.
-  launch_ctx, scratch_arr = None, None
+  launch_ctx = None
   with ir.InsertionPoint(module.body):
     _declare_runtime_functions()
     global_scratch = llvm.GlobalOp(
@@ -401,37 +657,31 @@ def _lower_as_gpu_kernel(
         ir.Attribute.parse("#llvm.linkage<external>"),
         addr_space=ir.IntegerAttr.get(i32, 4),  # GPU constant memory.
     )
-    @func.FuncOp.from_py_func(ptr_ty, ptr_ty, name=f"mosaic_gpu_{kernel_name}")
+    @func.FuncOp.from_py_func(ptr_ty, ptr_ty, name=f"{kernel_name}_mosaic_gpu")
     def main(token_ptr, buffers):
-      nonlocal launch_ctx, scratch_arr
+      nonlocal launch_ctx
       token = builtin.unrealized_conversion_cast([token_ty], [token_ptr])
       arg_refs = []
-      for i, ref_ty in enumerate([*in_ref_tys, *out_ref_tys]):
-        ptr = llvm.LoadOp(ptr_ty, llvm.GEPOp(ptr_ty, buffers, [], [i], ptr_ty))
+      # XLA will pass in inout refs again as outputs, but we ignore them.
+      for i, ref_ty in enumerate([*in_ref_tys, *inout_ref_tys, *out_ref_tys]):
+        ptr = llvm.LoadOp(ptr_ty, llvm.GEPOp(ptr_ty, buffers, [], [i], ptr_ty, llvm.GEPNoWrapFlags.none))
         arg_refs.append(utils.ptr_as_memref(ptr, ir.MemRefType(ref_ty)))
-      in_refs = arg_refs[:len(in_ref_tys)]
-      out_refs = arg_refs[len(in_ref_tys):]
-      prof_buffer = out_refs.pop() if prof_spec is not None else None
-      empty_arr_ty = ir.Type.parse("!llvm.array<0 x i8>")
-      scratch_alloc = llvm.AllocaOp(
-          ptr_ty, c(1, i64), empty_arr_ty,
-          alignment=launch_context.TMA_DESCRIPTOR_ALIGNMENT
-      )
-      scratch_arr = llvm.load(empty_arr_ty, scratch_alloc.result)
+      prof_buffer = arg_refs.pop() if prof_spec is not None else None
       with _launch(
-          token, grid, cluster, block, scratch_arr, smem_scratch_shape,
-          prof_spec, prof_buffer
+          token, grid, cluster, block, smem_scratch_shape,
+          lowering_semantics, module, prof_spec, prof_buffer
       ) as (_launch_ctx, smem_refs):
         nonlocal launch_ctx
         launch_ctx = _launch_ctx
-        body(launch_ctx, *in_refs, *out_refs, smem_refs)
+        body(launch_ctx, *arg_refs, smem_refs)
     main.func_op.attributes["llvm.emit_c_interface"] = ir.UnitAttr.get()
   sym_tab = ir.SymbolTable(module.operation)
   sym_tab.insert(main.func_op)
   sym_tab.insert(global_scratch)
   module.operation.verify()
 
-  return module, out_shape, unwrap_output_tuple, launch_ctx, scratch_arr
+  assert launch_ctx is not None
+  return module, out_shape, unwrap_output_tuple, launch_ctx
 
 
 def _run_serde_pass(
@@ -458,27 +708,6 @@ def _run_serde_pass(
   return module
 
 
-def _initialize_scratch(
-    launch_ctx : launch_context.LaunchContext,
-    scratch_arr: ir.Value,
-    ):
-  """
-  Allocates and initializes the host buffer right before the launch. This needs
-  to be done after all TMA descriptors have been recorded by the launch context.
-  Only then we know what the scratch contains.
-
-  When using the Mosaic GPU dialect, the necessary information is known only
-  after the lowering passes have run.
-  """
-  with ir.InsertionPoint(scratch_arr.owner):
-    gmem_scratch_bytes = launch_ctx.next_scratch_offset
-    scratch_alloc_op = scratch_arr.owner.opview.addr.owner.opview
-    scratch_arr_ty = ir.Type.parse(f"!llvm.array<{gmem_scratch_bytes} x i8>")
-    scratch_alloc_op.elem_type = ir.TypeAttr.get(scratch_arr_ty)
-    scratch_arr.set_type(scratch_arr_ty)
-    for init_callback in launch_ctx.host_scratch_init:
-      init_callback(scratch_alloc_op.result)
-
 def _declare_runtime_functions():
   """Declares the runtime functions that can be used by the generated code."""
   ptr_ty = ir.Type.parse("!llvm.ptr")
@@ -502,30 +731,49 @@ def as_gpu_kernel(
     module_name: str = "unknown",
     kernel_name: str | None = None,
     ir_version: int | None = None,
-    thread_semantics: ThreadSemantics = ThreadSemantics.Lane,
+    thread_semantics: LoweringSemantics = LoweringSemantics.Lane,
+    inout_shape = (),
 ):
   if isinstance(in_shape, list):
     in_shape = tuple(in_shape)
   elif not isinstance(in_shape, tuple):
     in_shape = (in_shape,)
+  if isinstance(inout_shape, list):
+    inout_shape = tuple(inout_shape)
+  elif not isinstance(inout_shape, tuple):
+    inout_shape = (inout_shape,)
 
-  module, out_shape, unwrap_output_tuple, launch_ctx, scratch_arr = (
+  inout_shape = jax.tree.map(lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype),
+                             inout_shape)
+  out_shape = jax.tree.map(lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype),
+                           out_shape)
+  module, out_shape, unwrap_output_tuple, launch_ctx = (
       _lower_as_gpu_kernel(
-          body, grid, cluster, block, in_shape, out_shape, smem_scratch_shape,
-          module_name, kernel_name, prof_spec
+          body, grid, cluster, block, in_shape, out_shape, inout_shape,
+          smem_scratch_shape, thread_semantics, module_name, kernel_name,
+          prof_spec
       )
   )
 
-  if thread_semantics == ThreadSemantics.Warpgroup and dialect is not None:
+  if thread_semantics == LoweringSemantics.Warpgroup and dialect is not None:
+    # We need to run a pass that removes dead-code for which layout inference
+    # does not work.
+    pm = mlir.passmanager.PassManager.parse("builtin.module(canonicalize)", module.context)
+    pm.run(module.operation)
+
     # Run Python lowering passes. The remaining passes will be run in C++ in
     # jax/jaxlib/mosaic/gpu/custom_call.cc
     layout_inference.infer_layout(module)  # pytype: disable=attribute-error
+    transform_inference.infer_transforms(module)  # pytype: disable=attribute-error
     dialect_lowering.lower_mgpu_dialect(module, launch_ctx)  # pytype: disable=attribute-error
 
-  _initialize_scratch(launch_ctx, scratch_arr)
+  launch_ctx.scratch.finalize_size()
   module.operation.verify()
 
-  expected_arg_treedef = jax.tree.structure(in_shape)
+  if launch_ctx.is_device_collective and not supports_cross_device_collectives():
+    raise RuntimeError("Kernel is a cross-device collective but no support is available.")
+
+  expected_arg_tys, expected_arg_treedef = jax.tree.flatten((*in_shape, *inout_shape))
   def _check_args(*args):
     arg_treedef = jax.tree.structure(args)
     if arg_treedef != expected_arg_treedef:
@@ -533,9 +781,23 @@ def as_gpu_kernel(
           f"Invalid argument structure: expected {expected_arg_treedef}, got"
           f" {arg_treedef}, ({args=})"
       )
+    for arg, expected_ty in zip(args, expected_arg_tys):
+      if arg.shape != expected_ty.shape:
+        raise ValueError(
+            f"Argument shape mismatch: expected {expected_ty.shape}, got"
+            f" {arg.shape}"
+        )
+      if arg.dtype != expected_ty.dtype:
+        hint = ""
+        if not arg.shape:
+          hint = f". Hint: cast the scalar to {expected_ty.dtype} explicitly."
+        raise ValueError(
+            f"Argument dtype mismatch: expected {expected_ty.dtype}, got"
+            f" {arg.dtype}{hint}"
+        )
 
   def bind(*args) -> Any:
-    return mosaic_gpu_p.bind(*args, module=module, out_types=out_shape)
+    return mosaic_gpu_p.bind(*args, module=module, out_types=out_shape, inout_types=inout_shape)
 
   if prof_spec is not None:
     @jax.jit
@@ -544,7 +806,7 @@ def as_gpu_kernel(
       *results, prof_buffer = bind(*args)
       def dump_profile(prof_buffer):
         out_file = os.path.join(
-            os.getenv("TEST_UNDECLARED_OUTPUTS_DIR"),
+            os.getenv("TEST_UNDECLARED_OUTPUTS_DIR", "/tmp"),
             f"{time.time_ns()}-trace.json",
         )
         try:
@@ -575,10 +837,10 @@ def as_torch_gpu_kernel(
     cluster: tuple[int, int, int] = (1, 1, 1),
     module_name: str = "unknown",
     kernel_name: str | None = None,
-    thread_semantics: ThreadSemantics = ThreadSemantics.Lane,
+    lowering_semantics: LoweringSemantics = LoweringSemantics.Lane,
 ):
   try:
-    import torch
+    import torch  # pytype: disable=import-error
   except ImportError:
     raise RuntimeError("as_torch_gpu_kernel requires PyTorch")
   torch.cuda.init()  # Make sure CUDA context is set up.
@@ -588,28 +850,44 @@ def as_torch_gpu_kernel(
   elif not isinstance(in_shape, tuple):
     in_shape = (in_shape,)
 
+  # TODO(slebedev): Make this a parameter.
+  inout_shape = ()
+
   flat_out_types, out_treedef = jax.tree.flatten(out_shape)
   expected_arg_treedef = jax.tree.structure(in_shape)
 
-  module, out_shape, unwrap_output_tuple, launch_ctx, scratch_arr = (
+  module, out_shape, unwrap_output_tuple, launch_ctx = (
       _lower_as_gpu_kernel(
-          body, grid, cluster, block, in_shape, out_shape, smem_scratch_shape,
-          module_name, kernel_name, prof_spec
+          body, grid, cluster, block, in_shape, out_shape, inout_shape,
+          smem_scratch_shape, lowering_semantics, module_name, kernel_name,
+          prof_spec
       )
   )
 
-  if thread_semantics == ThreadSemantics.Warpgroup and dialect is not None:
+  if lowering_semantics == LoweringSemantics.Warpgroup and dialect is not None:
+    # We need to run a pass that removes dead-code for which layout inference
+    # does not work.
+    pm = mlir.passmanager.PassManager.parse("builtin.module(canonicalize)", module.context)
+    pm.run(module.operation)
+
     # Run Python lowering passes. The remaining passes will be run in C++ in
     # jax/jaxlib/mosaic/gpu/custom_call.cc
     layout_inference.infer_layout(module)  # pytype: disable=attribute-error
+    transform_inference.infer_transforms(module)  # pytype: disable=attribute-error
     dialect_lowering.lower_mgpu_dialect(module, launch_ctx)  # pytype: disable=attribute-error
 
-  _initialize_scratch(launch_ctx, scratch_arr)
+  launch_ctx.scratch.finalize_size()
   module.operation.verify()
 
+  if launch_ctx.is_device_collective:
+    raise RuntimeError("Kernel is a cross-device collective but no support is available.")
+
   # Get our hands on the compilation and unload functions
   try:
-    import jax_plugins.xla_cuda12 as cuda_plugin
+    try:
+      import jax_plugins.xla_cuda13 as cuda_plugin  # pytype: disable=import-error
+    except ImportError:
+      import jax_plugins.xla_cuda12 as cuda_plugin  # pytype: disable=import-error
   except ImportError:
     raise RuntimeError("as_torch_gpu_kernel only works with recent jaxlib builds "
                        "that use backend plugins")
