@@ -27,11 +27,11 @@ from jax.interpreters import mlir
 from jax.interpreters import xla
 
 from jax._src import core
+from jax._src import ffi
 from jax._src.interpreters import ad
-from jax._src.lib import gpu_solver
 
 import numpy as np
-from scipy.sparse import csr_matrix, linalg
+import scipy.sparse
 
 
 def lobpcg_standard(
@@ -533,18 +533,17 @@ def _spsolve_abstract_eval(data, indices, indptr, b, *, tol, reorder):
 
 
 def _spsolve_gpu_lowering(ctx, data, indices, indptr, b, *, tol, reorder):
-  data_aval, _, _, _, = ctx.avals_in
-  return gpu_solver.cuda_csrlsvqr(data_aval.dtype, data, indices,
-                                  indptr, b, tol, reorder)
-
+  return ffi.ffi_lowering("cusolver_csrlsvqr_ffi")(
+      ctx, data, indices, indptr, b, tol=np.float64(tol),
+      reorder=np.int32(reorder))
 
 def _spsolve_cpu_lowering(ctx, data, indices, indptr, b, tol, reorder):
   del tol, reorder
   args = [data, indices, indptr, b]
 
   def _callback(data, indices, indptr, b, **kwargs):
-    A = csr_matrix((data, indices, indptr), shape=(b.size, b.size))
-    return (linalg.spsolve(A, b).astype(b.dtype),)
+    A = scipy.sparse.csr_matrix((data, indices, indptr), shape=(b.size, b.size))
+    return (scipy.sparse.linalg.spsolve(A, b).astype(b.dtype),)
 
   result, _, _ = mlir.emit_python_callback(
       ctx, _callback, None, args, ctx.avals_in, ctx.avals_out,
