@@ -55,23 +55,24 @@ import numpy as np
 
 import jax
 from jax import lax
+from jax._src import api_util
 from jax._src import config
 from jax._src import core
 from jax._src.custom_derivatives import lift_jvp
 from jax._src import linear_util as lu
 from jax._src import pjit
 from jax._src import sharding_impls
-from jax.experimental.sparse.bcoo import bcoo_multiply_dense, bcoo_multiply_sparse
+from jax.experimental.sparse.bcoo import bcoo_multiply_dense, bcoo_multiply_sparse, BCOO
+from jax.experimental.sparse.bcsr import BCSR
 import jax.numpy as jnp
 from jax._src.api_util import flatten_fun_nokwargs
 from jax._src.lib import pytree
 from jax._src.interpreters import partial_eval as pe
 from jax.tree_util import tree_flatten, tree_map, tree_unflatten
-from jax.util import safe_map, safe_zip, split_list
+from jax._src.util import safe_map, safe_zip, split_list
 from jax._src.lax.control_flow import _check_tree_and_avals
-from jax._src.numpy import lax_numpy
+from jax._src.numpy import indexing as jnp_indexing
 from jax.experimental import sparse
-from jax.experimental.sparse import BCOO, BCSR
 
 sparse_rules_bcoo : dict[core.Primitive, Callable] = {}
 sparse_rules_bcsr : dict[core.Primitive, Callable] = {}
@@ -297,7 +298,10 @@ class SparseTracer(core.Tracer):
 
 class SparseTrace(core.Trace):
 
+  __slots__ = ("parent_trace", "tag", "spenv")
+
   def __init__(self, parent_trace, tag, spenv):
+    super().__init__()
     self.parent_trace = parent_trace
     self.tag = tag
     self.spenv = spenv
@@ -363,12 +367,15 @@ def sparsify_fun(wrapped_fun, args: list[ArrayOrSparse]):
   spenv = SparsifyEnv(out_bufs)
   return spvalues_to_arrays(spenv, out_spvalues())
 
-def _sparsify_with_tracer(fun):
+def _sparsify_with_tracer(fun: Callable):
   """Implementation of sparsify() using tracers."""
   @functools.wraps(fun)
   def _wrapped(*args):
     args_flat, in_tree = tree_flatten(args, is_leaf=_is_sparse_obj)
-    wrapped_fun, out_tree = flatten_fun_nokwargs(lu.wrap_init(fun), in_tree)
+    wrapped_fun, out_tree = flatten_fun_nokwargs(
+        lu.wrap_init(fun,
+                     debug_info=api_util.debug_info("sparsify", fun, args, {})),
+        in_tree)
     out = sparsify_fun(wrapped_fun, args_flat)
     return tree_unflatten(out_tree(), out)
   return _wrapped
@@ -437,8 +444,13 @@ def sparsify_raw(f):
   ) -> tuple[Sequence[SparsifyValue], pytree.PyTreeDef]:
     spvalues_flat, in_tree = tree_flatten(spvalues, is_leaf=_is_spvalue)
     in_avals_flat = spvalues_to_avals(spenv, spvalues_flat)
-    wrapped_fun, out_tree = flatten_fun_nokwargs(lu.wrap_init(f, params), in_tree)
-    jaxpr, out_avals_flat, consts, () = pe.trace_to_jaxpr_dynamic(wrapped_fun, in_avals_flat)
+    wrapped_fun, out_tree = flatten_fun_nokwargs(
+        lu.wrap_init(
+            f, params,
+            debug_info=api_util.debug_info("sparsify", f,
+                                           spvalues_to_arrays(spenv, spvalues), {})),
+        in_tree)
+    jaxpr, out_avals_flat, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, in_avals_flat)
     result = eval_sparse(jaxpr, consts, spvalues_flat, spenv)
     if len(out_avals_flat) != len(result):
       raise Exception("Internal: eval_sparse does not return expected number of arguments. "
@@ -714,14 +726,14 @@ def _gather_sparse_rule(spenv, *args, dimension_numbers, slice_sizes, unique_ind
 
 sparse_rules_bcoo[lax.gather_p] = _gather_sparse_rule
 
-def _sparsify_jaxpr(spenv, jaxpr, *spvalues):
+def _sparsify_jaxpr(spenv: SparsifyEnv,
+                    jaxpr: core.ClosedJaxpr, *spvalues):
   # TODO(jakevdp): currently this approach discards all information about
   #   shared data & indices when generating the sparsified jaxpr. The
   #   current approach produces valid sparsified while loops, but they
   #   don't work in corner cases (see associated TODO in sparsify_test.py)
   out_tree: pytree.PyTreeDef | None = None
 
-  @lu.wrap_init
   def wrapped(*args_flat):
     # TODO(frostig,jakevdp): This closes over `spenv`, which can bring
     # in buffers from the "outer scope" as constants. Is this a
@@ -738,7 +750,8 @@ def _sparsify_jaxpr(spenv, jaxpr, *spvalues):
   args = spvalues_to_arrays(spenv, spvalues)
   args_flat, in_tree = tree_flatten(args)
   avals_flat = [core.get_aval(arg) for arg in args_flat]
-  sp_jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(wrapped, avals_flat)
+  sp_jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(
+      lu.wrap_init(wrapped, debug_info=jaxpr.jaxpr.debug_info), avals_flat)
   sp_jaxpr = pe.ClosedJaxpr(sp_jaxpr, consts)
   assert out_tree is not None
   return sp_jaxpr, out_tree
@@ -763,7 +776,7 @@ sparse_rules_bcoo[lax.while_p] = _while_sparse
 
 
 def _pjit_sparse(spenv, *spvalues, jaxpr, in_shardings, out_shardings,
-                 in_layouts, out_layouts, resource_env, donated_invars, name,
+                 in_layouts, out_layouts, donated_invars, ctx_mesh, name,
                  keep_unused, inline, compiler_options_kvs):
   if any(donated_invars):
     raise NotImplementedError("sparse xla_call with donated_invars")
@@ -789,22 +802,22 @@ def _pjit_sparse(spenv, *spvalues, jaxpr, in_shardings, out_shardings,
       None for _ in range(len(sp_call_jaxpr.out_avals) - len(out_layouts))
   )
 
-  out_flat = pjit.pjit_p.bind(
+  out_flat = pjit.jit_p.bind(
       *args_flat,
       jaxpr=sp_call_jaxpr,
       in_shardings=in_shardings,
       out_shardings=out_shardings,
       in_layouts=in_layouts,
       out_layouts=out_layouts,
-      resource_env=resource_env,
       donated_invars=donated_invars,
+      ctx_mesh=ctx_mesh,
       name=name,
       keep_unused=keep_unused,
       inline=inline,
       compiler_options_kvs=compiler_options_kvs)
   return arrays_to_spvalues(spenv, tree_unflatten(out_tree, out_flat))
 
-sparse_rules_bcoo[pjit.pjit_p] = _pjit_sparse
+sparse_rules_bcoo[pjit.jit_p] = _pjit_sparse
 
 
 def _duplicate_for_sparse_spvalues(spvalues, params):
@@ -848,7 +861,7 @@ def _cond_sparse(spenv, pred, *operands, branches, **params):
                         "sparsified false_fun output",
                         treedefs[1], sp_branches[1].out_avals)
   args, _ = tree_flatten(spvalues_to_arrays(spenv, (pred, *operands)))
-  out_flat = lax.cond_p.bind(*args, branches=sp_branches, **params)
+  out_flat = lax.cond_p.bind(*args, branches=tuple(sp_branches), **params)
   out = tree_unflatten(treedefs[0], out_flat)
   return arrays_to_spvalues(spenv, out)
 
@@ -863,18 +876,19 @@ sparse_rules_bcoo[sparse.todense_p] = _todense_sparse_rule
 sparse_rules_bcsr[sparse.todense_p] = _todense_sparse_rule
 
 def _custom_jvp_sparse_rule(spenv, *spvalues, **params):
-  call_jaxpr = params.pop('call_jaxpr')
-  jvp_jaxpr_thunk = params.pop('jvp_jaxpr_thunk')
-  num_consts = params.pop('num_consts')
+  call_jaxpr: core.ClosedJaxpr = params.pop('call_jaxpr')
+  jvp_jaxpr_fun: lu.WrappedFun = params.pop('jvp_jaxpr_fun')
+  num_consts: int = params.pop('num_consts')
   sp_call_jaxpr, out_tree = _sparsify_jaxpr(spenv, call_jaxpr, *spvalues)
-  @lu.wrap_init
   def fun(*arrs):
     sparrs = arrays_to_spvalues(spenv, arrs)
     out = eval_sparse(call_jaxpr.jaxpr, call_jaxpr.consts, sparrs, spenv)
     return spvalues_to_arrays(spenv, out)
-  jvp = lift_jvp(num_consts, jvp_jaxpr_thunk)
+  jvp = lift_jvp(num_consts, jvp_jaxpr_fun)
   invals = spvalues_to_arrays(spenv, spvalues)
-  outvals = jax.custom_derivatives.custom_jvp_call_p.bind(fun, jvp, *invals, **params)
+  outvals = jax.custom_derivatives.custom_jvp_call_p.bind(
+      lu.wrap_init(fun, debug_info=call_jaxpr.jaxpr.debug_info),
+      jvp, *invals, **params)
   return arrays_to_spvalues(spenv, outvals)
 
 sparse_rules_bcoo[jax.custom_derivatives.custom_jvp_call_p] = _custom_jvp_sparse_rule
@@ -901,7 +915,7 @@ def _bcoo_rewriting_take(arr, idx, indices_are_sorted=False, unique_indices=Fals
                            mode=None, fill_value=None):
   # Only sparsify the array argument; sparse indices not yet supported
   result = sparsify(functools.partial(
-    lax_numpy._rewriting_take, idx=idx, indices_are_sorted=indices_are_sorted,
+    jnp_indexing.rewriting_take, idx=idx, indices_are_sorted=indices_are_sorted,
     mode=mode, unique_indices=unique_indices, fill_value=fill_value))(arr)
   # Account for a corner case in the rewriting_take implementation.
   if not isinstance(result, BCOO) and np.size(result) == 0:
@@ -953,7 +967,7 @@ def _bcsr_rewriting_take(arr, idx, indices_are_sorted=False, unique_indices=Fals
                            mode=None, fill_value=None):
   # Only sparsify the array argument; sparse indices not yet supported
   result = sparsify(functools.partial(
-    lax_numpy._rewriting_take, idx=idx, indices_are_sorted=indices_are_sorted,
+    jnp_indexing.rewriting_take, idx=idx, indices_are_sorted=indices_are_sorted,
     mode=mode, unique_indices=unique_indices, fill_value=fill_value))(arr)
   return result
 
