@@ -13,8 +13,25 @@
 # limitations under the License.
 
 # Note: import <name> as <name> is required for names to be exported.
-# See PEP 484 & https://github.com/google/jax/issues/7570
+# See PEP 484 & https://github.com/jax-ml/jax/issues/7570
 
 from jax._src.api import (
   clear_backends as clear_backends,
 )
+from jax._src.compiler import (
+    get_compile_options as get_compile_options,
+)
+from jax._src.xla_bridge import (
+  backends as backends,
+  backend_xla_version as backend_xla_version,
+  get_backend as get_backend,
+  register_backend_factory as register_backend_factory,
+)
+from jax._src.interpreters.pxla import (
+  get_default_device as get_default_device
+)
+from jax._src import (
+    util as _util
+)
+register_backend_cache = _util.register_cache  # type: ignore
+del _util
