@@ -27,27 +27,25 @@ from jax._src.interpreters.mlir import (
   Token as Token,
   TokenSet as TokenSet,
   Value as Value,
-  call_lowering as _call_lowering,
+  call_lowering as _call_lowering,  # noqa: F401
   _lowerings as _lowerings,
   _platform_specific_lowerings as _platform_specific_lowerings,
   aval_to_ir_type as aval_to_ir_type,
   aval_to_ir_types as aval_to_ir_types,
   core_call_lowering as core_call_lowering,
-  custom_call as custom_call,
+  custom_call as _custom_call,
   dense_bool_elements as dense_bool_elements,
   dense_bool_array as dense_bool_array,
   dense_int_array as dense_int_array,
   dense_int_elements as dense_int_elements,
   dtype_to_ir_type as dtype_to_ir_type,
-  emit_python_callback as emit_python_callback,
-  flatten_lowering_ir_args as flatten_lowering_ir_args,
-  func_dialect as func_dialect,
-  hlo as hlo,
+  flatten_ir_types as flatten_ir_types,
+  flatten_ir_values as flatten_ir_values,
+  unflatten_ir_values_like_types as unflatten_ir_values_like_types,
   i32_attr as i32_attr,
   i64_attr as i64_attr,
   ir as ir,
   ir_constant as ir_constant,
-  ir_constants as ir_constants,
   ir_type_handlers as ir_type_handlers,
   jaxpr_subcomp as jaxpr_subcomp,
   lower_fun as lower_fun,
@@ -62,7 +60,6 @@ from jax._src.interpreters.mlir import (
   register_lowering as register_lowering,
   shape_tensor as shape_tensor,
   token_type as token_type,
-  xla_computation_to_mlir_module as xla_computation_to_mlir_module,
 )
 
 from jax._src.mesh import Mesh as Mesh
@@ -72,3 +69,30 @@ from jax._src.sharding_impls import (
   SPMDAxisContext as SPMDAxisContext,
   ShardingContext as ShardingContext,
 )
+
+
+# TODO(dsuo): Temporarily maintain symbols related to callback lowering for sake
+# of public APIs.
+from jax._src.callback import (
+  emit_python_callback as emit_python_callback,
+)
+
+_deprecations = {
+    # Added Apr 7 2025
+    "custom_call": (
+        "mlir.custom_call is deprecated; use the APIs provided by jax.ffi instead.",
+        _custom_call,
+    )
+}
+
+import typing as _typing
+
+if _typing.TYPE_CHECKING:
+  custom_call = _custom_call
+else:
+  from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
+
+  __getattr__ = _deprecation_getattr(__name__, _deprecations)
+  del _deprecation_getattr
+del _typing
+del _custom_call
