@@ -20,30 +20,45 @@ from jax._src.sharding_impls import (
     NamedSharding as NamedSharding,
     SingleDeviceSharding as SingleDeviceSharding,
     PmapSharding as PmapSharding,
-    GSPMDSharding as GSPMDSharding,
-    PositionalSharding as PositionalSharding,
+    use_mesh as use_mesh,
+    set_mesh as set_mesh,
 )
 from jax._src.partition_spec import (
     PartitionSpec as PartitionSpec,
 )
-from jax._src.interpreters.pxla import Mesh as Mesh
 from jax._src.mesh import (
+    Mesh as Mesh,
     AbstractMesh as AbstractMesh,
-    AxisTypes as AxisTypes,
-    use_mesh as use_mesh
+    AxisType as AxisType,
+    get_abstract_mesh as get_abstract_mesh,
 )
 
+from jax._src.pjit import (
+    reshard as reshard,
+    auto_axes as auto_axes,
+    explicit_axes as explicit_axes,
+)
+
+
 _deprecations = {
-    # Finalized 2024-10-01; remove after 2025-01-01.
-    "XLACompatibleSharding": (
+    # Added April 11, 2025.
+    "PositionalSharding": (
         (
-            "jax.sharding.XLACompatibleSharding was removed in JAX v0.4.34. "
-            "Use jax.sharding.Sharding instead."
+            "jax.sharding.PositionalSharding was deprecated in JAX v0.6.0 and"
+            " removed in JAX v0.7.0"
         ),
         None,
-    )
+    ),
+    "GSPMDSharding": (
+        (
+            "jax.sharding.GSPMDSharding was deprecated in JAX v0.6.0 and"
+            " removed in JAX v0.7.0"
+        ),
+        None,
+    ),
 }
 
+
 from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
 __getattr__ = _deprecation_getattr(__name__, _deprecations)
 del _deprecation_getattr
